#include "bus.h"
#include "backend-vdcs.h"
#include "model/apartment.h"
#include "model/dsmeter.h"

static ds::LogChannel dsLogChannel("bus");

namespace dss {

struct Bus::Impl : ds::asio::Resource<> {
  Impl(ds::asio::ResourceBase& parent, Bus& bus, BusInterface& ds485, BusHooks&& busHooks)
      : ds::asio::Resource<>{parent, "dssBus"}, m_bus(bus), m_ds485(ds485), m_busHooks(busHooks) {}

  Bus& m_bus;
  BusInterface& m_ds485;
  BusHooks m_busHooks;

  ds::Slot<> DS_UNIQUE_NAME(slot){context(), m_ds485.linkPresent, [&]() {
                                    m_busHooks.linkPresent();
                                    m_bus.linkPresent.emit();
                                  }};

  ds::Slot<> DS_UNIQUE_NAME(slot){context(), m_ds485.linkLost, [&]() {
                                    m_busHooks.linkLost();
                                    m_bus.linkLost.emit();
                                  }};
  ds::Slot<dsuid_t> DS_UNIQUE_NAME(slot){context(), m_ds485.ds485MemberLost, [&](auto arg) {
                                           m_busHooks.ds485MemberLost(arg);
                                           m_bus.ds485MemberLost.emit(std::move(arg));
                                         }};
  ds::Slot<dsuid_t> DS_UNIQUE_NAME(slot){context(), m_ds485.ds485MemberJoin, [&](auto arg) {
                                           m_busHooks.ds485MemberJoin(arg);
                                           m_bus.ds485MemberJoin.emit(std::move(arg));
                                         }};
};

Bus::Bus(Security& security, BusInterface& ds485Bus)
    : m_ds485Bus(ds485Bus),
      m_ds485Query(*m_ds485Bus.getStructureQueryBusInterface()),
      m_ds485Modify(*m_ds485Bus.getStructureModifyingBusInterface()),
      m_queries{std::ref(m_ds485Query)},
      m_modifies{std::ref(m_ds485Modify)} {}

Bus::~Bus() = default;

int Bus::makeImpl(ds::asio::ResourceBase& parent, BusHooks&& hooks) {
  m_impl = std::make_unique<Impl>(parent, *this, m_ds485Bus, std::move(hooks));
  return 0;
}

StructureQueryBusInterface& Bus::dsmQueryBus(const dsuid_t& dsmId) {
  if (dsmId == DSUID_BROADCAST) {
    return m_ds485Query;
  }
  if (m_backendVdcs->isBackendVdc(dsmId)) {
    return m_backendVdcs->structureQueryBusInterface();
  }
  return m_ds485Query;
}

StructureModifyingBusInterface& Bus::dsmModifyBus(const dsuid_t& dsmId) {
  if (dsmId == DSUID_BROADCAST) {
    return m_ds485Modify;
  }
  if (m_backendVdcs->isBackendVdc(dsmId)) {
    return m_backendVdcs->structureModifyingBusInterface();
  }
  return m_ds485Modify;
}

std::string Bus::name() const { return dsLogChannel.name(); }

void Bus::removeDeviceFromDSMeters(const dsuid_t& device) {
  for (const auto& modify : m_modifies) {
    try {
      modify.get().removeDeviceFromDSMeters(device);
    } catch (const std::exception& e) {
      DS_ERROR("Failed to remove device.", device, modify);
    }
  }
}

void Bus::sceneSetName(uint16_t zone, uint8_t group, uint8_t scene, const std::string& name) {
  for (const auto& modify : m_modifies) {
    try {
      modify.get().sceneSetName(zone, group, scene, name);
    } catch (const std::exception& e) {
      DS_ERROR("Failed to set scene name.", zone, scene, name, modify);
    }
  }
}

} // namespace dss
