/*
    Copyright (c) 2018 digitalSTROM AG, Zurich, Switzerland

    This file is part of dSS.

    vdSM is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 2 of the License.

    vdSM is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with dSS. If not, see <http://www.gnu.org/licenses/>.
*/

#pragma once

#include <memory>
#include <optional>
#include <unordered_map>

#include <cxxabi.h>
#include <mutex>
#include "ds/json.h"
#include "ds/log.h"
#include "ds/sqlite3/ds-sqlite3.h"
#include "ds485types.h"

namespace dss {

extern ds::LogChannel cacheDbLogChannel;

class CacheDbBase {
public:
  virtual ~CacheDbBase() = default;

  std::vector<Dsuid> devices() const;
  std::string getTableName() { return m_tableName; }

protected:
  CacheDbBase(ds::sqlite3::Database* db) : m_db(db) {}

  void setTableName(const std::string& name) {
    m_tableName = name;
    initializeDbTable();
  }

private:
  void initializeDbTable() {
    std::lock_guard<std::mutex> lock(m_mutex);
    m_db->exec("CREATE TABLE IF NOT EXISTS " + m_tableName + " (dsuid TEXT(40) UNIQUE, data);");

    // Activate WAL journal.
    // Writes are several times faster even on SSD, but benchmark showed 2x speedup also for reads.
    m_db->exec("PRAGMA journal_mode=WAL;");

    // Reduce size of WAL journal from default 4MB. 4MB is quite a bit on embedded device.
    m_db->exec("PRAGMA journal_size_limit=131072;");

    // Benchmarks show more than 10x speed up on PC + SSD in NORMAL mode compared to default FULL mode.
    // As we are using WAL mode, the database is always consistent.
    // But the write transaction finishes without waiting for date being written on disc.
    m_db->exec("PRAGMA synchronous=NORMAL;");
  }

protected:
  ds::sqlite3::Database* m_db;
  std::string m_tableName;
  inline static std::mutex m_mutex;
};

template <typename T>
class CacheDb : public CacheDbBase {
public:
  CacheDb(ds::sqlite3::Database* db, bool cached = true) : CacheDbBase(db), m_cached(cached) {
    char* name = abi::__cxa_demangle(typeid(T).name(), 0, 0, NULL);
    std::string typeName(name);
    setTableName(typeName.substr(typeName.find_last_of(":") + 1) + "_table");
    free(name);
  }

  T tryLookup(const Dsuid& dsuid) {
    DS_CONTEXT(m_tableName, dsuid);
    std::lock_guard<std::mutex> lock(m_mutex);
    auto dsuidStr = ds::str(dsuid);
    if (m_cached) {
      if (auto cacheHit = m_cache.find(dsuidStr); cacheHit != m_cache.end()) {
        return cacheHit->second;
      }
    }
    try {
      std::string entryStr;
      std::tie(entryStr) =
          m_db->exec("SELECT data FROM " + m_tableName + " WHERE dsuid = ?;", dsuidStr).template row<std::string>();

      DS_CHANNEL_DEBUG(cacheDbLogChannel, entryStr);
      DS_CHANNEL_DEBUG(cacheDbLogChannel, "Cache lookup before json");
      auto entry = ds::fromJson<T>(entryStr);
      DS_CHANNEL_DEBUG(cacheDbLogChannel, "Cache lookup after json");
      m_cache[dsuidStr] = entry;

      return entry;
    } catch (const std::exception& e) {
      DS_CHANNEL_DEBUG(cacheDbLogChannel, "Cache lookup failed.", e.what());
    }
    m_cache[dsuidStr] = T();

    return T();
  }

  void update(const Dsuid& dsuid, const T& entry) {
    DS_CONTEXT(m_tableName, dsuid);
    std::lock_guard<std::mutex> lock(m_mutex);
    auto dsuidStr = ds::str(dsuid);
    m_cache[dsuidStr] = entry;

    DS_CHANNEL_DEBUG(cacheDbLogChannel, "Cache update before json");
    auto value = ds::toJson(entry);
    DS_CHANNEL_DEBUG(cacheDbLogChannel, "Cache update after json");
    DS_CHANNEL_DEBUG(cacheDbLogChannel, ds::toJson(entry));
    m_db->exec("INSERT OR REPLACE INTO " + m_tableName + " (dsuid, data) VALUES(?, ?);", dsuidStr, value);
  }

  void remove(const Dsuid& dsuid) {
    std::lock_guard<std::mutex> lock(m_mutex);
    DS_CONTEXT(m_tableName, dsuid);
    auto dsuidStr = ds::str(dsuid);
    m_cache.erase(dsuidStr);
    m_db->exec("DELETE FROM " + m_tableName + " WHERE dsuid = ?;", dsuidStr);
  }

  void empty() {
    std::lock_guard<std::mutex> lock(m_mutex);
    DS_CONTEXT(m_tableName);
    m_cache.clear();
    m_db->exec("DELETE FROM " + m_tableName + ";");
  }

private:
  std::unordered_map<std::string, T> m_cache{};
  bool m_cached;
};

} /* namespace dss */
