// This program is a code generator plugin for `capnp compile` which generates html code.

#include <capnp/schema.capnp.h>
#include <capnp/serialize.h>
#include <capnp/pretty-print.h>
#include <capnp/compat/json.h>
#include <kj/debug.h>
#include <kj/io.h>
#include <kj/string-tree.h>
#include <kj/tuple.h>
#include <kj/vector.h>
#include <capnp/schema-loader.h>
#include <capnp/dynamic.h>
#include <kj/miniposix.h>
#include <unordered_map>
#include <unordered_set>
#include <map>
#include <set>
#include <kj/main.h>
#include <algorithm>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <iostream>
#include "form.capnp.h"

#ifndef VERSION
#define VERSION "(unknown)"
#endif

static constexpr uint64_t OPTION_ANNOTATION_ID = 10393266975298982020ull;
static constexpr uint64_t FIELD_ANNOTATION_ID = 17464266750853295555ull;

namespace capnp {
namespace {

kj::StringPtr baseName(kj::StringPtr path) {
  KJ_IF_MAYBE(slashPos, path.findLast('/')) {
    return path.slice(*slashPos + 1);
  } else {
    return path;
  }
}

// =======================================================================================

class CapnpcDssFromMain {
public:
  CapnpcDssFromMain(kj::ProcessContext& context): context(context) {}

  kj::MainFunc getMain() {
    return kj::MainBuilder(context, "Cap'n Proto plugin version " VERSION,
          "This is a Cap'n Proto compiler plugin which generates dss view code. "
          "It is meant to be run using the Cap'n Proto compiler, e.g.:\n"
          "    capnp compile -oc++ foo.capnp")
        .callAfterParsing(KJ_BIND_METHOD(*this, run))
        .build();
  }

private:
  kj::ProcessContext& context;
  SchemaLoader schemaLoader;

  kj::String toTitleCase(kj::StringPtr name) {
    kj::String result = kj::heapString(name);
    if ('a' <= result[0] && result[0] <= 'z') {
      result[0] = result[0] - 'a' + 'A';
    }
    return kj::mv(result);
  }

  kj::String toFormCase(kj::StringPtr name) {
    kj::Vector<char> result(name.size() + 4);

    char last = 0;
    for (char c: name) {
      if (result.size() == 0 && 'a' <= c && c <= 'z') {
        result.add(c - 'a' + 'A');
      } else if (result.size() > 0 && 'A' <= c && c <= 'Z') {
        result.add(' ');
        result.add(c - 'A' + 'a');
      } else if (result.size() > 0 && '0' <= c && c <= '9' && !('0' <= last && last < '9')) {
        result.add(' ');
        result.add(c);
      } else {
        result.add(c);
      }
      last = result[result.size() - 1];
    }

    result.add('\0');
    return kj::String(result.releaseAsArray());
  }

  template <typename P>
  kj::Maybe<schema::Value::Reader> annotationValue(P proto, uint64_t annotationId) {
    kj::Maybe<schema::Value::Reader> out;
    for (auto annotation: proto.getAnnotations()) {
      if (annotation.getId() == annotationId) {
        KJ_REQUIRE(out == nullptr, "Duplicate annotation", annotation);
        out = annotation.getValue();
      }
    }
    return out;
  }

  template <typename P>
  dss::view::Option::Reader optionAnnotation(P proto) {
    return annotationValue(proto, OPTION_ANNOTATION_ID).map(
      [&](schema::Value::Reader value) {
        return value.getStruct().getAs<dss::view::Option>();
      }).orDefault(dss::view::Option::Reader());
  }

  template <typename P>
  dss::view::Field::Reader fieldAnnotation(P proto) {
    return annotationValue(proto, FIELD_ANNOTATION_ID).map(
      [&](schema::Value::Reader value) {
        return value.getStruct().getAs<dss::view::Field>();
      }).orDefault(dss::view::Field::Reader());
  }

  kj::String formTitle(kj::StringPtr annotatedTitle, kj::StringPtr name) {
    return annotatedTitle.size() ? kj::heapString(annotatedTitle) : toFormCase(name);
  }

  kj::String formTitle(dss::view::Field::Reader annotation, kj::StringPtr name) {
    return formTitle(annotation.getTitle(), name);
  }

  kj::String formTitle(dss::view::Option::Reader annotation, kj::StringPtr name) {
    return formTitle(annotation.getTitle(), name);
  }

  kj::String evalString(kj::StringPtr s) {
    return s.startsWith("{{")
      ? kj::str("(", s.slice(2, s.size() - 2), ")")
      : kj::str("_('", s, "')");
  }

  template <typename P>
  kj::StringPtr protoName(P proto) {
    return proto.getName();
  }

  struct ValueTypeText {
    kj::StringTree script;
    kj::StringTree html;
  };

  ValueTypeText makeValueTypeText(kj::StringPtr indent, kj::StringPtr valueName, kj::StringPtr fullName_, StructSchema schema) {
    auto subIndent = kj::str(indent, "");
    auto valueSubScope = kj::str(valueName, ".");
    auto subScope_ = fullName_.size() ? kj::str(fullName_, "_") : kj::str();

    auto fieldTexts = KJ_MAP(f, schema.getFields()) {
      return makeFieldText(subIndent, valueSubScope, subScope_, f);
    };

    return ValueTypeText {
      // script
      kj::strTree(
        KJ_MAP(x, fieldTexts) { return kj::mv(x.script); }
      ),
      // html
      kj::strTree(
        KJ_MAP(x, fieldTexts) { return kj::mv(x.html); }
      )
    };
  }

  struct FieldText {
    kj::StringTree script;
    kj::StringTree html;
  };

  FieldText makeFieldText(kj::StringPtr indent, kj::StringPtr valueScope, kj::StringPtr scope_, StructSchema::Field capnpField) {
    auto proto = capnpField.getProto();
    auto name = protoName(proto);
    auto valueName = kj::str(valueScope, name);
    auto fullName_ = kj::str(scope_, name);
    auto subIndent = kj::str(indent, "  ");
    auto showFuntion= kj::str(fullName_, "__show");
    auto disabledFunction = kj::str(fullName_, "__disabled");
    auto fieldName = kj::str("fields.", fullName_);
    auto titleName = kj::str(fieldName, ".title");
    auto showName = kj::str(fieldName, ".show");
    auto disabledName = kj::str(fieldName, ".disabled");
    auto optionsName = kj::str(fieldName, ".options");
    auto minName = kj::str(fieldName, ".min");
    auto maxName = kj::str(fieldName, ".max");
    auto unitName = kj::str(fieldName, ".unit");
    auto typeSchema = capnpField.getType();
    auto field = fieldAnnotation(proto);
    kj::String type;
    auto jsIndent = "      ";
    // std::cout << kj::str(field).cStr() << "\n";

    auto common = kj::strTree(
      (field.getEnableIf().size() ? kj::strTree(" ng-disabled=\"", disabledName, "\"") : kj::strTree())
    );

    ValueTypeText structText;
    if (typeSchema.which() == schema::Type::STRUCT) {
      structText = makeValueTypeText(subIndent, valueName, fullName_, typeSchema.asStruct());
    }
    kj::StringTree enumScript;
    if (typeSchema.which() == schema::Type::ENUM) {
      enumScript = kj::strTree(
        jsIndent, "$scope.$watch(function() {\n",
        jsIndent, "  var options = [];\n",
        KJ_MAP(e, typeSchema.asEnum().getEnumerants()) {
          auto&& eProto = e.getProto();
          auto&& eOption = optionAnnotation(eProto);
          auto&& eName = protoName(e.getProto());
          auto&& eTitle = formTitle(eOption, eName);
          return kj::strTree(
            jsIndent, "  if (", (eOption.getShowIf().size() ? kj::strTree(eOption.getShowIf()) : kj::strTree("true")), ")",
            " { options.push({ id:'", eName, "', title:", evalString(eTitle),
            (field.getIconInOptions() ? kj::strTree( ", icon:'", eOption.getIcon(), "'") : kj::strTree()), "}); }\n"
          );
        },
        jsIndent, "  return { show : ", showFuntion, "(), options : options };\n",
        jsIndent, "}, function(x)  { \n",
        jsIndent, "  if (x.show) { ", optionsName, " = x.options; }\n",
        jsIndent, "  if (x.show && x.options.find(function (x) { return ", valueName, " === x.id;}) === undefined "
            "&& x.options.length > 0) { ", valueName, " = x.options[0].id; }\n",
        jsIndent, "}, true);\n",
        "\n"
      );
      type = kj::heapString("options");
    }
    if (field.getOptionsQuery().getId().size()) {
      auto optionsQuery = field.getOptionsQuery();
      enumScript = kj::strTree(
        jsIndent, "$scope.$watch(function() {\n",
        jsIndent, "  var options = [];\n",
        jsIndent, "  var a = ", optionsQuery.getForXIn(), ";\n",
        jsIndent, "  if (a === undefined) { a = []; }\n",
        jsIndent, "  for (i = 0, l = a.length; i < l; i++) {\n",
        jsIndent, "    x = a[i];\n",
        jsIndent, "    options.push({ id:(", optionsQuery.getId(), "), title:(", optionsQuery.getTitle(), ")});",
        jsIndent, "  }\n",
        jsIndent, "  return { show : ", showFuntion, "(), options : options };\n",
        jsIndent, "}, function(x)  { \n",
        jsIndent, "  if (x.show) { ", optionsName, " = x.options; }\n",
        jsIndent, "  if (x.show && x.options.find(function (x) { return ", valueName, " === x.id;}) === undefined "
            "&& x.options.length > 0) { ", valueName, " = x.options[0].id; }\n",
        jsIndent, "}, true);\n",
        "\n"
      );
      type = kj::heapString("options");
    }

    if (type.size() == 0) {
        type = kj::heapString([&] {
          switch (typeSchema.which()) {
            case schema::Type::VOID: return "";
            case schema::Type::BOOL: return "bool";
            case schema::Type::INT8:
            case schema::Type::INT16:
            case schema::Type::INT32:
            case schema::Type::INT64:
            case schema::Type::UINT8:
            case schema::Type::UINT16:
            case schema::Type::UINT32:
            case schema::Type::UINT64:
            case schema::Type::FLOAT32:
            case schema::Type::FLOAT64: return "number";
            case schema::Type::TEXT: return "text";
            case schema::Type::DATA: return "";
            case schema::Type::ENUM: KJ_FAIL_ASSERT("Handled sooner.");
            case schema::Type::STRUCT: return "struct";
            case schema::Type::LIST: return "list";
            case schema::Type::INTERFACE: return "";
            case schema::Type::ANY_POINTER: return "";
          };
          KJ_FAIL_REQUIRE("Unrecognized typeSchema", typeSchema.which());
        }());
    }

    return FieldText {
      // script
      kj::strTree(
        jsIndent, "var ", showFuntion, " = function() { return ", (field.getShowIf().size() ? kj::strTree("(", field.getShowIf(), ")") : kj::strTree("true")), "; };\n",
        jsIndent, "var ", disabledFunction, " = function() { return ", (field.getEnableIf().size() ? kj::strTree("!(", field.getEnableIf(), ")") : kj::strTree("false")), "; };\n",
        jsIndent, fieldName, " = {}\n",
        jsIndent, "fieldsArray.push(", fieldName, ");\n",
        jsIndent, fieldName, ".id = \"", name, "\"\n",
        jsIndent, fieldName, ".type = \"", type, "\";\n",
        (type == "select" ? kj::strTree(jsIndent, fieldName, ".iconInOptions = ", field.getIconInOptions(), ";\n") : kj::strTree()),
        jsIndent, "$scope.$watch(function() { return ", evalString(formTitle(field, name)), "; }, function(x) { ", titleName, " = x;}, true);\n",
        jsIndent, "$scope.$watch(", showFuntion, ", function(x) { ", showName, " = x;}, true);\n",
        jsIndent, "$scope.$watch(", disabledFunction, ", function(x) { ", disabledName, " = x;}, true);\n",
        [&] {
          auto number = [&](kj::StringPtr min, kj::StringPtr max) {
            return kj::strTree(
              jsIndent, "$scope.$watch(function() { return ", evalString(field.getMin().size() ? field.getMin() : min), "; }, function(x) { ", minName, " = Number(x); }, true);\n",
              jsIndent, "$scope.$watch(function() { return ", evalString(field.getMax().size() ? field.getMax() : max), "; }, function(x) { ", maxName, " = Number(x); }, true);\n"
            );
          };
          switch (typeSchema.which()) {
            case schema::Type::INT8: return number("-128", "127");
            case schema::Type::INT16: return number("-32768", "32767");
            case schema::Type::INT32: return number("-2147483648", "2147483647");
            case schema::Type::INT64: return number("-9223372036854775808", "9223372036854775807");
            case schema::Type::UINT8: return number("0", "255");
            case schema::Type::UINT16: return number("0", "65535");
            case schema::Type::UINT32: return number("0", "4294967295");
            case schema::Type::UINT64: return number("0", "9223372036854775807");
            case schema::Type::FLOAT32: return number("Number.NEGATIVE_INFINITY", "Number.POSITIVE_INFINITY");
            case schema::Type::FLOAT64: return number("Number.NEGATIVE_INFINITY", "Number.POSITIVE_INFINITY");
            case schema::Type::VOID:
            case schema::Type::BOOL:
            case schema::Type::TEXT:
            case schema::Type::DATA:
            case schema::Type::ENUM:
            case schema::Type::STRUCT:
            case schema::Type::LIST:
            case schema::Type::INTERFACE:
            case schema::Type::ANY_POINTER:
              break;
          };
          return kj::strTree();
        }(),
      jsIndent, unitName, " = \"",
        [&] {
          switch (field.getUnit()) {
            case dss::view::Field::Unit::NONE: return "";
            case dss::view::Field::Unit::SECOND: return "s";
            case dss::view::Field::Unit::PERCENT: return "%";
          }
          return "";
        }(), "\";\n",
        kj::mv(enumScript),
        KJ_MAP(watch, field.getWatches()) {
          return kj::strTree(
            jsIndent, "$scope.$watch(\"", watch.getWatch(), "\", function(newWatchValue, oldWatchValue) {\n",
            jsIndent, "  console.log(\"watch ", valueName, "\", newWatchValue, oldWatchValue);\n",
            jsIndent, "  (function(newWatchValue, oldWatchValue) {\n",
            ( watch.getDoOnInit() == false ? kj::strTree(jsIndent, "      if (newWatchValue === oldWatchValue) { return; }\n") : kj::strTree()),
            jsIndent, "    ", watch.getDo(), ";\n",
            jsIndent, "  })(newWatchValue, oldWatchValue);\n",
            jsIndent, "}, true);\n"
          );
        },
        kj::mv(structText.script)
      ),
      // html
      kj::strTree(
        indent, "<div class=\"dssFormField\" ng-show=\"", showName, "\">\n",
        indent, "  <span class=\"dssFormFieldTitle\">{{", titleName, "}}:</span>\n",
        [&] {
          if (typeSchema.which() == schema::Type::ENUM || field.getOptionsQuery().getId().size()) {
            return kj::strTree(
                indent, "  <select class=\"dssFormFieldInput\" ng-model=\"", valueName, "\"", kj::mv(common), "\n",
                indent, "    ng-options=\"x.id as x.title for x in ", optionsName, "\">\n",
                indent, "  </select>\n"
              );
          }
          switch (typeSchema.which()) {
            case schema::Type::VOID: return kj::strTree();
            case schema::Type::BOOL: return kj::strTree(
                indent, "  <input class=\"dssFormFieldInput\" type=\"checkbox\" ng-model=\"", valueName,"\"", kj::mv(common), "/>\n"
              );
            case schema::Type::INT8:
            case schema::Type::INT16:
            case schema::Type::INT32:
            case schema::Type::INT64:
            case schema::Type::UINT8:
            case schema::Type::UINT16:
            case schema::Type::UINT32:
            case schema::Type::UINT64:
            case schema::Type::FLOAT32:
            case schema::Type::FLOAT64:
              return kj::strTree(
                indent, "  <input class=\"dssFormFieldInput\" type=\"number\"",
                " ng-model=\"", valueName, "\"",
                " min=\"{{", minName, "}}\"",
                " max=\"{{", maxName, "}}\"",
                (field.getStep().size() ? kj::strTree(" step=\"", field.getStep(), "\"") : kj::strTree()),
                kj::mv(common),
                "/>\n");
            case schema::Type::TEXT: return kj::strTree(
                indent, "  <input class=\"dssFormFieldInput\" type=\"text\" ng-model=\"", valueName, "\"", kj::mv(common), "/>\n"
              );
            case schema::Type::DATA: return kj::strTree();
            case schema::Type::ENUM: KJ_FAIL_ASSERT("Handled sooner.");
            case schema::Type::STRUCT: return kj::mv(structText.html);
            case schema::Type::LIST: return kj::strTree();
            case schema::Type::INTERFACE: return kj::strTree();
            case schema::Type::ANY_POINTER: return kj::strTree();
          };
          return kj::strTree();
        }(),
        indent, " <span class=\"dssFormFieldUnit\"> {{", unitName, "}}</span>\n",
        indent, "</div>\n"
      )
    };
  }

  // -----------------------------------------------------------------

  kj::StringTree toJson(ConstSchema value) {
    JsonCodec codec;
    return codec.encode(value.as<DynamicStruct>(), value.as<DynamicStruct>().getSchema());
  }

  struct FileText {
    kj::StringTree html;
  };

  FileText makeFileText(Schema schema,
                        schema::CodeGeneratorRequest::RequestedFile::Reader request) {

    auto node = schema.getProto();
    auto displayName = node.getDisplayName();

    auto getNodeByName = [&](kj::StringPtr name) {
      for (auto&& nestedNode : node.getNestedNodes()) {
        if (nestedNode.getName() == name) {
          return schemaLoader.getUnbound(nestedNode.getId());
        }
      }
      KJ_FAIL_REQUIRE("Top level node with", name, " not found.");
    };


    auto valueTypeText = makeValueTypeText("      ", "value", "", getNodeByName("Value").asStruct());
    auto value = getNodeByName("value").asConst();
    auto args = getNodeByName("args").asConst();
    auto view = getNodeByName("view").asConst();
    return FileText {
      kj::strTree(
          "<!-- Generated by Cap'n Proto compiler, DO NOT EDIT-->\n"
          "<!-- source: ", baseName(displayName), "-->\n",
          "\n",
R"(
<!DOCTYPE html>
<style media="screen" type="text/css">
.dssForm { display : table; table-layout: fixed; }
.dssFormField { display : table-row; }
.dssFormFieldTitle { display : table-cell; width : 100px; padding: 5px; }
.dssFormFieldInput { display : table-cell; width : 150px; }
</style>
<html>
  <script src="https://ajax.googleapis.com/ajax/libs/angularjs/1.6.4/angular.min.js"></script>
  <script>
    var dssDeviceConfigModule = angular.module('dssDeviceConfig', [])
    dssDeviceConfigModule.controller('dssDeviceConfigController', function($scope) {
      var _ = function(s) { return s; };
      $scope._ = _;
)",
          "      var initValue = ", toJson(value), ";\n",
          "      $scope.initValue = initValue;\n",
          "\n",
          "      var value = ", toJson(value), ";\n",
          "      $scope.value = value;\n",
          "\n",
          "      var args = ", toJson(args), ";\n",
          "      $scope.args = args;\n",
          "\n",
          "      var view = ", toJson(view), ";\n",
          "      $scope.view = view;\n",
          "\n",
          "      var fields = {}\n",
          "      $scope.fields = fields;\n",
          "      var fieldsArray = [];\n",
          "      $scope.fieldsArray = fieldsArray;\n",
          "\n",
          kj::mv(valueTypeText.script),
R"(
    });
  </script>
)",

          "  <body ng-app=\"dssDeviceConfig\">\n"
          "    <h1>", baseName(displayName),"</h1>\n",
          "    <form class=\"dssForm\" ng-controller=\"dssDeviceConfigController\">\n",
          kj::mv(valueTypeText.html),
R"(
      <div style="position: absolute; bottom: 0; right: 0; overflow-y: scroll; height:100%">
        <b>Value:</b>
        <pre>{{value | json}}</pre>
        <b>Args:</b>
        <pre>{{args | json}}</pre>
        <b>View:</b>
        <pre>{{view | json}}</pre>
        <b>Fields:</b>
        <pre>{{fields | json}}</pre>
      </div>
    </form>
  </body>
</html>
)"
      )
    };
  }

  // -----------------------------------------------------------------

  void makeDirectory(kj::StringPtr path) {
    KJ_IF_MAYBE(slashpos, path.findLast('/')) {
      // Make the parent dir.
      makeDirectory(kj::str(path.slice(0, *slashpos)));
    }

    if (kj::miniposix::mkdir(path.cStr(), 0777) < 0) {
      int error = errno;
      if (error != EEXIST) {
        KJ_FAIL_SYSCALL("mkdir(path)", error, path);
      }
    }
  }

  void writeFile(kj::StringPtr filename, const kj::StringTree& text) {
    if (!filename.startsWith("/")) {
      KJ_IF_MAYBE(slashpos, filename.findLast('/')) {
        // Make the parent dir.
        makeDirectory(kj::str(filename.slice(0, *slashpos)));
      }
    }

    int fd;
    KJ_SYSCALL(fd = open(filename.cStr(), O_CREAT | O_WRONLY | O_TRUNC, 0666), filename);
    kj::FdOutputStream out((kj::AutoCloseFd(fd)));

    text.visit(
        [&](kj::ArrayPtr<const char> text) {
          out.write(text.begin(), text.size());
        });
  }

  kj::MainBuilder::Validity run() {
    ReaderOptions options;
    options.traversalLimitInWords = 1 << 30;  // Don't limit.
    StreamFdMessageReader reader(STDIN_FILENO, options);
    auto request = reader.getRoot<schema::CodeGeneratorRequest>();

    for (auto node: request.getNodes()) {
      schemaLoader.load(node);
    }

    kj::FdOutputStream rawOut(STDOUT_FILENO);
    kj::BufferedOutputStreamWrapper out(rawOut);

    for (auto requestedFile: request.getRequestedFiles()) {
      auto schema = schemaLoader.get(requestedFile.getId());
      auto fileText = makeFileText(schema, requestedFile);

      writeFile(kj::str(schema.getProto().getDisplayName(), ".html"), fileText.html);
    }
    return true;
  }
};

}  // namespace
}  // namespace capnp

KJ_MAIN(capnp::CapnpcDssFromMain)
