/*
    Copyright (c) 2020 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/
#include "api-revision.h"

static ds::LogChannel dsLogChannel("dsApartmentApiRevision");

namespace dsApartment {

static const ApiRevisionConfig s_default = {.additions = {.default_ = 4, .min = 2, .max = 4},
                                            .removals = {.default_ = 0, .min = 0, .max = 1}};

struct ApiRevisionImpl : ApiRevision, ds::asio::Resource<ApiRevisionConfig> {
  ApiRevisionImpl(ds::asio::ResourceBase &parent)
      : ds::asio::Resource<ApiRevisionConfig>{parent, "apiRevision", ApiRevisionConfig{s_default}} {}
  int m_additions = config().additions.default_;
  int m_removals = config().removals.default_;

  const ds::asio::ResourceInside &inside() override {
    using This = ApiRevisionImpl;
    static constexpr ds::asio::ResourceInside instance{
        .description = "Apartment api revision.",
        .properties = {
            {.id = "additions",
             .description = "Additions revision.",
             .value =
                 [](const ResourceBase &x) { return ds::toJsonFragment(static_cast<const This &>(x).m_additions); },
             .test = false,
             .defaultValue =
                 [](const ResourceBase &x) {
                   return ds::toJsonFragment(static_cast<const This &>(x).config().additions);
                 },
             .setValue = [](ResourceBase &x,
                            ds::JsonFragment &&data) { static_cast<This &>(x).setAdditions(ds::fromJson<int>(data)); }},
            {.id = "removals",
             .description = "Removals revision.",
             .value = [](const ResourceBase &x) { return ds::toJsonFragment(static_cast<const This &>(x).m_removals); },
             .test = false,
             .defaultValue =
                 [](const ResourceBase &x) {
                   return ds::toJsonFragment(static_cast<const This &>(x).config().removals);
                 },
             .setValue = [](ResourceBase &x,
                            ds::JsonFragment &&data) { static_cast<This &>(x).setRemovals(ds::fromJson<int>(data)); }},
        }};
    return instance;
  }

  void setAdditions(int value) {
    DS_REQUIRE(config().additions.min <= value && value <= config().additions.max, "Invalid additions revision", value,
               config().additions);
    DS_NOTICE("Set additions", value)
    m_additions = value;
  }

  void setRemovals(int value) {
    DS_REQUIRE(config().removals.min <= value && value <= config().removals.max, "Invalid removals revision", value,
               config().removals);
    DS_NOTICE("Set removals", value)
    m_removals = value;
  }

  int additions() const override { return m_additions; }
  int removals() const override { return m_removals; }
};

std::unique_ptr<ApiRevision> makeApiRevision(ds::asio::ResourceBase &parent) {
  return std::make_unique<ApiRevisionImpl>(parent);
}

} // namespace dsApartment

DS_JSON_DEFINE_STRUCT((dsApartment), ApiRevisionConfigDomain);
DS_JSON_DEFINE_STRUCT((dsApartment), ApiRevisionConfig);
