/*
    Copyright (c) 2020 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/
#pragma once

#include <ds/box.h>
#include <ds/log.h>

namespace dsApi {

// unbox array
template <class T, class Enable = void>
struct IdOf {
  const auto& operator()(const T& x) { return x.id; }
};
template <class T>
struct IdOf<ds::Box<T>> {
  const auto& operator()(const ds::Box<T>& x) { return x->id; }
};
template <class T>
struct IdOf<T, typename std::enable_if_t<std::is_convertible_v<T, std::string>>> {
  const std::string& operator()(const T& x) { return x; }
};
template <class T>
const auto& idOf(const T& x) {
  return IdOf<T>{}(x);
}

template <class T>
const std::string& stringIdOf(const T& value) {
  return static_cast<const std::string&>(idOf(value));
}

template <class T>
inline bool idLess(const T& a, const T& b) {
  return idOf(a) < idOf(b);
}
template <class T>
inline bool idGreaterEq(const T& a, const T& b) {
  return !idLess(a, b);
}
template <class T>
inline bool idEqual(const T& a, const T& b) {
  return idOf(a) == idOf(b);
}

/// Sort, deduplicate, remove empty ids
template <class T>
void normalizeIdArray(std::vector<T>& values) {
  auto begin = [&] { return values.begin(); };
  auto end = [&] { return values.end(); };
  std::stable_sort(begin(), end(), idLess<T>);
  values.erase(std::unique(begin(), end(), idEqual<T>), end());
  if (begin() != end() && stringIdOf(*begin()).empty()) {
    values.erase(begin());
  }
}

template <class T>
struct ArrayBox;

/// Vector of boxed objects mapped by id field.
template <class T>
struct Array : std::vector<ds::Box<T>> {
  using Super = std::vector<ds::Box<T>>;
  using Super::back;
  using Super::begin;
  using Super::end;
  using Super::erase;
  using Super::front;
  using Super::insert;
  using Super::vector::vector;

  using Id = std::decay_t<decltype(std::declval<T>().id)>;
  const std::string& stringId(const T& value) const { return static_cast<const std::string&>(value.id); }

  std::optional<T> tryAt(const Id& id) const {
    auto it = std::find_if(begin(), end(), [&](const auto& x) { return x->id == id; });
    if (it != end()) {
      return *it;
    }
    return std::nullopt;
  }

  /// Query one value by id.
  T at(const Id& id) const {
    if (auto x = tryAt(id)) {
      return std::move(*x);
    }
    DS_FAIL_REQUIRE("Not found", id, *this);
  }

  /// Insert or replace value by id.
  void replace(ds::Box<T> value) {
    if (auto it = std::lower_bound(begin(), end(), value, less); it != end() && (*it)->id == value->id) {
      *it = std::move(value);
    } else {
      insert(it, std::move(value));
    }
  }

  /// Array is normalized when it is
  /// * sorted by id
  /// * ids are unique
  /// * ids are non-empty
  bool isNormalized() const {
    if (std::adjacent_find(begin(), end(), greaterEq) != end()) {
      return false;
    }
    if (begin() != end() && stringId(*begin()).empty()) {
      return false;
    }
    return true;
  }

  void requireNormalized() const { DS_REQUIRE(isNormalized(), "Array is not normalized.", *this); }

  /// Sort, deduplicate, remove empty ids
  void normalize() { normalizeIdArray(*this); }

  /// Normalize and move to box
  ArrayBox<T> boxed() &&;

  static bool less(const ds::Box<T>& a, const ds::Box<T>& b) { return a->id < b->id; }
  static bool greaterEq(const ds::Box<T>& a, const ds::Box<T>& b) { return !(a->id < b->id); }
  static bool equal(const ds::Box<T>& a, const ds::Box<T>& b) { return a->id == b->id; }
};

/// Normalized array in ds::Box.
template <class T>
struct ArrayBox : ds::Box<Array<T>> {
  using ds::Box<Array<T>>::Box;
  ArrayBox(Array<T>&& x) : ds::Box<Array<T>>{std::move(x)} { (*this)->requireNormalized(); }
};

// Makes normalized array box from a range.
template <class Range>
auto makeArrayBox(Range&& range) {
  auto array = dsApi::Array<std::decay_t<decltype(**std::begin(range))>>{};
  array.reserve(std::size(range));
  for (auto&& item : std::forward<Range>(range)) {
    array.push_back(std::forward<decltype(item)>(item));
  }
  return std::move(array).boxed();
};

template <class T>
ArrayBox<T> Array<T>::boxed() && {
  normalize();
  return ArrayBox<T>{std::move(*this)};
}

} // namespace dsApi
