#include "inside.h"

#include <ds-api/service.h>
#include <ds/asio/resource.h>
#include <ds/json.h>
#include <ds/vector.h>

#include <boost/algorithm/string.hpp>
#include <boost/range/adaptor/filtered.hpp>
#include <boost/range/adaptor/transformed.hpp>
#include <boost/range/algorithm_ext.hpp>

static ds::LogChannel dsLogChannel("dsApiInside");

DS_REFLECTED_ENUM((dsApi), InsideMode, (none)(inside)(test));

namespace dsApi {

struct InsideArgs {
  std::string path{};
};

struct InsidePropertyAttributes {
  std::string description{};
  ds::JsonFragment value;
  ds::JsonFragment defaultValue;
  bool mutable_{}; // TODO(someday): drop _ suffix in json. `mutable` is c++ keyword.
};
struct InsideProperty {
  std::string id;
  InsidePropertyAttributes attributes;
};

struct InsideMethodAttributes {
  std::string description{};
};
struct InsideMethod {
  std::string id;
  InsideMethodAttributes attributes;
};

struct InsideResourceAttributes {
  std::string description{};
  ds::JsonFragment config{};
  std::vector<std::string> children{};
  std::vector<std::string> properties{};
  std::vector<std::string> methods{};
};
struct InsideResourceIncluded {
  std::vector<InsideProperty> properties{};
  std::vector<InsideMethod> methods{};
};
struct InsideResource {
  std::string id;
  InsideResourceAttributes attributes;
  InsideResourceIncluded included;
};

namespace {

struct ServiceImpl : public dsApi::Service {
  InsideMode m_mode;

  ServiceImpl(InsideMode mode) : m_mode{mode} {
    routes.push_back(
        dsApi::Route{.pattern = "/{*path}", .invoke = [this](HttpRequest&& x) { return invoke(std::move(x)); }});
  }

  ds::asio::ResourceBase& findResource(ds::asio::ResourceContext& context, std::string_view path) {
    // already synchronized to context
    auto result = std::reference_wrapper<ds::asio::ResourceBase>{context.root()};
    for (const auto& resource : context.resources()) {
      if (boost::starts_with(path, resource.get().path()) &&
          result.get().path().size() < resource.get().path().size()) {
        result = resource;
      }
    }
    if (!result.get().path().empty()) {
      return result.get();
    }
    DS_FAIL_REQUIRE("Resource not found.", path);
  }

  template <class T>
  bool present(const T& x) {
    return m_mode == InsideMode::inside || (m_mode == InsideMode::test && x.test);
  }

  template <class T>
  void requirePresent(const T& x) {
    DS_REQUIRE(present(x), "Property / method not available in api mode.", m_mode);
  }

  std::optional<std::string> startsWithSegment(const std::string_view path, const std::string_view token) {
    if (!boost::starts_with(path, token)) {
      return {};
    }
    const auto remainingPath = path.substr(token.size());
    if (remainingPath.empty() || remainingPath.at(0) == '/') {
      return std::string{remainingPath};
    }
    return {};
  }

  // routes the request to root, resource, property, method etc
  HttpResult invoke(HttpRequest&& request) {
    DS_INFO_ENTER(request);
    auto& path = request.path; // mutate inside request
    path = request.routeArgs.at("path");

    if (path.empty()) {
      DS_DEBUG("root")
      // matched root
      return dsApi::invoke(
          std::vector{dsApi::makeDocumentRoute("", {.methods = {dsApi::makeGet([&]() { return insideRoot(); })}})},
          std::move(request));
    }

    auto& context = ds::asio::resourceContext(path);
    DS_DEBUG("context")
    return context.invoke([&] {
      auto& resource = findResource(context, path);

      // resource match?
      path.erase(0, resource.path().size());
      if (path.empty()) {
        DS_DEBUG(resource)
        return dsApi::invoke(std::vector{dsApi::makeDocumentRoute(
                                 "", {.methods = {dsApi::makeGet([&]() { return insideResource(resource); })}})},
                             std::move(request));
      }
      DS_REQUIRE(boost::starts_with(path, "/"), path);
      path.erase(0, 1);

      // property match?
      const auto& inside = resource.inside();
      for (const auto& property : inside.properties) {
        if (auto remainingPath = startsWithSegment(path, std::string_view{property.id})) {
          DS_DEBUG(resource, property.id);
          requirePresent(property);
          path = std::move(*remainingPath);
          return dsApi::invoke(std::vector{
              dsApi::makeDocumentRoute("", {.methods = {dsApi::makeGet([&]() {
                                              return dsApi::okResult(insideProperty(resource, property));
                                            })}}),
              dsApi::makeDocumentRoute(
                  "/value", {.methods = {{HttpMethod::GET,
                                          [&](auto&&) { return dsApi::okResult(insidePropertyValue(resource, property)); }},
                                         {HttpMethod::PUT,
                                          [&](auto&&) {
                                            insidePropertySetValue(resource, property,
                                                                std::move(request.httpArgs.content));
                                            return dsApi::noContentResult();
                                          }}}}),
              dsApi::makeDocumentRoute("/defaultValue", {.methods = {dsApi::makeGet([&]() {
                                                           return dsApi::okResult(
                                                               insidePropertyDefaultValue(resource, property));
                                                         })}}),
              dsApi::makeDocumentRoute(
                  "/resetValue", {.methods = {
                                         {HttpMethod::POST,
                                          [&](auto&&) {
                                                   insidePropertyResetValue(resource, property);
                                            return dsApi::noContentResult();
                                          }}}}),
          },                      std::move(request));
        }
      }

      // match resource method
      for (const auto& method : inside.methods) {
        if (auto remainingPath = startsWithSegment(path, std::string_view{method.id})) {
          DS_DEBUG(resource, method.id);
          requirePresent(method);
          path = std::move(*remainingPath);
          return dsApi::invoke(
              std::vector{
                  dsApi::makeDocumentRoute("", {.methods = {dsApi::makeGet([&]() {
                                                  return dsApi::okResult(insideMethod(resource, method));
                                                })}}),
                  dsApi::makeDocumentRoute("/invoke", {.methods = {{HttpMethod::POST,
                                                                    [&](auto&&) {
                                                                      insideMethodInvoke(resource, method);
                                                                      return dsApi::noContentResult();
                                                                    }}}}),
              },
              std::move(request));
        }
      }
      return dsApi::resourceNotFoundResult();
    });
  }

  InsideResource insideRoot() {
    // root collects children from all context roots
    auto paths = std::vector<std::string>{};
    for (auto& context : ds::asio::resourceContexts()) {
      boost::range::push_back(paths, context.get().invoke([&] { return context.get().root().childIds(); }));
    }
    return InsideResource{
        .id = "",
        .attributes = {
            .description =
                [&] {
                  switch (m_mode) {
                    case InsideMode::inside:
                      return R"(View inside resources for investigation by humans.

Breaking changes without notice.

Endpoints:
// resource
GET /api/v1/inside/{resourcePath}

// property
GET /api/v1/inside/{resourcePath}/{property}
GET /api/v1/inside/{resourcePath}/{property}/value
PUT /api/v1/inside/{resourcePath}/{property}/value
GET /api/v1/inside/{resourcePath}/{property}/defaultValue
POST /api/v1/inside/{resourcePath}/{property}/resetValue

// method
GET /api/v1/inside/{resourcePath}/{method}
POST /api/v1/inside/{resourcePath}/{method}/invoke
)";
                    case InsideMode::test:
                      return
                          R"(View inside resources for automated tests. Subset of inside api.

Breaking changes coordinated with tests.
)";
                    case InsideMode::none: return "";
                    default: return "";
                  }
                }(),
            .config = {},
            .children = std::move(paths)}};
  }

  InsideProperty insideProperty(ds::asio::ResourceBase& resource, const ds::asio::ResourceProperty& x) {
    return InsideProperty{.id = x.id,
                          .attributes = {.description = x.description ? x.description : "",
                                         .value = x.value ? x.value(resource) : ds::JsonFragment{},
                                         .defaultValue = x.defaultValue ? x.defaultValue(resource) : ds::JsonFragment{},
                                         .mutable_ = x.setValue != nullptr}};
  }

  ds::JsonFragment insidePropertyValue(ds::asio::ResourceBase& resource, const ds::asio::ResourceProperty& property) {
    DS_REQUIRE(property.value, "Property cannot be read.");
    return property.value(resource);
  }

  void insidePropertySetValue(ds::asio::ResourceBase& resource, const ds::asio::ResourceProperty& property,
                              ds::JsonFragment&& value) {
    DS_REQUIRE(property.setValue != nullptr, "Property is not mutable.");
    DS_REQUIRE(property.value, "Property cannot be read.");
    auto oldValue = property.value(resource);
    if (value == oldValue) {
      DS_NOTICE("Set resource property - nothing to do.", resource, property.id, value, oldValue);
    } else {
      DS_NOTICE("Set resource property.", resource, property.id, value, oldValue);
      property.setValue(resource, std::move(value));
    }
  }

  ds::JsonFragment insidePropertyDefaultValue(ds::asio::ResourceBase& resource,
                                              const ds::asio::ResourceProperty& property) {
    DS_REQUIRE(property.defaultValue, "Property has no default value.");
    return property.defaultValue(resource);
  }

  void insidePropertyResetValue(ds::asio::ResourceBase& resource, const ds::asio::ResourceProperty& property) {
    DS_REQUIRE(property.setValue != nullptr, "Property is not mutable.");
    DS_REQUIRE(property.value, "Property cannot be read.");
    DS_REQUIRE(property.defaultValue, "Property has no default value.");
    auto oldValue = property.value(resource);
    auto defaultValue = property.defaultValue(resource);
    if (defaultValue == oldValue) {
      DS_NOTICE("Reset resource property - nothing to do.", resource, property.id, defaultValue, oldValue);
    } else {
      DS_NOTICE("Reset resource property.", resource, property.id, defaultValue, oldValue);
      property.setValue(resource, std::move(defaultValue));
    }
  }

  InsideMethod insideMethod(ds::asio::ResourceBase& resource, const ds::asio::ResourceMethod& x) {
    return InsideMethod{.id = x.id, .attributes = {.description = x.description ? x.description : ""}};
  }

  void insideMethodInvoke(ds::asio::ResourceBase& resource, const ds::asio::ResourceMethod& method) {
    DS_REQUIRE(method.invoke, "Method cannot be invoked.");
    DS_NOTICE("Invoke resource method", resource, method.id);
    method.invoke(resource);
  }

  InsideResource insideResource(ds::asio::ResourceBase& resource) {
    const auto& inside = resource.inside();
    return InsideResource{
        .id = resource.path(),
        .attributes = {.description = inside.description ? inside.description : "",
                       .config = resource.jsonConfig(),
                       .children = resource.childIds(),
                       .properties = ds::makeVectorNoReserve(
                           inside.properties | boost::adaptors::filtered([&](const auto& x) { return present(x); }) |
                           boost::adaptors::transformed([&](const auto& x) { return std::string{x.id}; })),
                       .methods = ds::makeVectorNoReserve(
                           inside.methods | boost::adaptors::filtered([&](const auto& x) { return present(x); }) |
                           boost::adaptors::transformed([&](const auto& x) { return std::string{x.id}; }))},
        .included = {.properties = ds::makeVectorNoReserve(
                         inside.properties | boost::adaptors::filtered([&](const auto& x) { return present(x); }) |
                         boost::adaptors::transformed([&](const auto& x) { return insideProperty(resource, x); })),
                     .methods = ds::makeVectorNoReserve(
                         inside.methods | boost::adaptors::filtered([&](const auto& x) { return present(x); }) |
                         boost::adaptors::transformed([&](const auto& x) { return insideMethod(resource, x); }))}};
  }
};

} // namespace

std::unique_ptr<dsApi::Service> makeInsideService() {
  return std::make_unique<ServiceImpl>(InsideMode::inside);
}

std::unique_ptr<dsApi::Service> makeTestService() {
  return std::make_unique<ServiceImpl>(InsideMode::test);
}

} // namespace dsApi

DS_JSON_DECLARE_STRUCT((dsApi), InsideArgs, (path));
DS_JSON_DECLARE_STRUCT((dsApi), InsidePropertyAttributes, (description)(value)(defaultValue)(mutable_));
DS_JSON_DECLARE_STRUCT((dsApi), InsideProperty, (id)(attributes));
DS_JSON_DECLARE_STRUCT((dsApi), InsideMethodAttributes, (description));
DS_JSON_DECLARE_STRUCT((dsApi), InsideMethod, (id)(attributes));
DS_JSON_DECLARE_STRUCT((dsApi), InsideResourceAttributes, (description)(config)(children)(properties)(methods));
DS_JSON_DECLARE_STRUCT((dsApi), InsideResourceIncluded, (properties)(methods));
DS_JSON_DECLARE_STRUCT((dsApi), InsideResource, (id)(attributes)(included));

DS_JSON_DEFINE_STRUCT((dsApi), InsideArgs);
DS_JSON_DEFINE_STRUCT((dsApi), InsidePropertyAttributes);
DS_JSON_DEFINE_STRUCT((dsApi), InsideProperty);
DS_JSON_DEFINE_STRUCT((dsApi), InsideMethodAttributes);
DS_JSON_DEFINE_STRUCT((dsApi), InsideMethod);
DS_JSON_DEFINE_STRUCT((dsApi), InsideResourceAttributes);
DS_JSON_DEFINE_STRUCT((dsApi), InsideResourceIncluded);
DS_JSON_DEFINE_STRUCT((dsApi), InsideResource);
