/*
    Copyright (c) 2016 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/
#include "service.h"

#include <ds/catch/catch.h>
#include <ds/log.h>

#include <boost/algorithm/string/predicate.hpp>
#include <boost/serialization/strong_typedef.hpp>
#include <trompeloeil.hpp>

using namespace trompeloeil;

namespace {
static const char *TAGS = "[dsApiService][dsApi]";

BOOST_STRONG_TYPEDEF(std::string, Id);

class Method {
public:
  MAKE_MOCK0(call, void());
  MAKE_MOCK0(callWithReturn, bool());
};

struct Args {
  std::string a;
  std::string b;
};

class ArgsMethod {
public:
  MAKE_MOCK1(call, void(Args &&));
  MAKE_MOCK1(callWithReturn, bool(Args &&));
};

class ArgsByIdMethod {
public:
  MAKE_MOCK2(call, void(const Id &, Args &&));
};

class ResultArgsByIdMethod {
public:
  MAKE_MOCK2(call, int(const Id &, Args &&));
};

struct GetByIdMock {
public:
  MAKE_MOCK1(call, int(const Id &));
};

struct GetWithQueryMock {
public:
  MAKE_MOCK1(call, int(const dsApi::GetQuery &));
};

struct GetByIdWithQueryMock {
public:
  MAKE_MOCK2(call, int(const Id &, const dsApi::GetQuery &));
};

struct ByIdMock {
public:
  MAKE_MOCK1(call, void(const Id &));
};

SCENARIO("dsApiService", TAGS) {
  dsApi::Service service;
  auto requireResourceNotFound = [&](const dsApi::HttpResult &x) {
    INFO(x);
    CHECK(x.status == 404);
    CHECK(boost::starts_with(dsApi::error(x).detail, "Resource not found."));
  };
  auto requireReplyMethodNotAllowed = [&](const dsApi::HttpResult &x, const std::string &allow) {
    INFO(x);
    CHECK(x.status == 405);
    CHECK(boost::starts_with(dsApi::error(x).detail, "Method not allowed."));
    CHECK(x.headers == dsApi::HttpHeaders{{"Allow", {allow}}});
  };
  auto requireError = [&](const dsApi::HttpResult &x, auto &&message) {
    INFO(x);
    CHECK(x.status == 500);
    CHECK(boost::starts_with(dsApi::error(x).detail, message));
  };

  GIVEN("/echo route") {
    service.routes.push_back({.pattern = "/echo", .invoke = [&](auto &&context) {
                                CHECK(context.httpArgs.method == dsApi::HttpMethod::POST);
                                auto data = dsApi::data<std::string>(context.httpArgs);
                                CHECK(data == "test");
                                return dsApi::okDocumentResult(data);
                              }});

    THEN("request to /echo succeeds") {
      CHECK(dsApi::data<std::string>(
                service.invoke({.path = "/echo",
                                .httpArgs = {.method = dsApi::HttpMethod::POST,
                                             .content = ds::toJsonFragment(dsApi::document("test"))}})) == "test");
    }

    THEN("request to non-matching fails") {
      SECTION("non-matching prefix") {
        requireResourceNotFound(service.invoke({.path = "", .httpArgs = {.method = dsApi::HttpMethod::POST}}));
        requireResourceNotFound(service.invoke({.path = "/", .httpArgs = {.method = dsApi::HttpMethod::POST}}));
        requireResourceNotFound(service.invoke({.path = "/ech", .httpArgs = {.method = dsApi::HttpMethod::POST}}));
      }

      SECTION("matching prefix with suffix") {
        requireResourceNotFound(service.invoke({.path = "/echo/foo", .httpArgs = {.method = dsApi::HttpMethod::POST}}));
        requireResourceNotFound(service.invoke({.path = "/echo2", .httpArgs = {.method = dsApi::HttpMethod::POST}}));
      }
    }
  }

  GIVEN("literal route before and after arg route") {
    service.routes.push_back(
        {.pattern = "/literal1", .invoke = [&](auto &&context) { return dsApi::okDocumentResult("literal1"); }});
    service.routes.push_back(
        {.pattern = "/{arg}", .invoke = [&](auto &&context) { return dsApi::okDocumentResult("arg"); }});
    service.routes.push_back(
        {.pattern = "/literal2", .invoke = [&](auto &&context) { return dsApi::okDocumentResult("literal2"); }});

    THEN("request to /literal1 routes to literal1") {
      CHECK(dsApi::data<std::string>(
                service.invoke({.path = "/literal1", .httpArgs = {.method = dsApi::HttpMethod::GET}})) == "literal1");
    }

    THEN("request to /literal2 routes to literal1") {
      CHECK(dsApi::data<std::string>(
                service.invoke({.path = "/literal2", .httpArgs = {.method = dsApi::HttpMethod::GET}})) == "literal2");
    }

    THEN("request to /x routes to arg") {
      CHECK(dsApi::data<std::string>(service.invoke({.path = "/x", .httpArgs = {.method = dsApi::HttpMethod::GET}})) ==
            "arg");
    }
  }

  GIVEN("/wildcard route returning remaining path") {
    auto makeArgs = [] {
      return dsApi::HttpArgs{.method = dsApi::HttpMethod::POST, .content = ds::toJsonFragment("test")};
    };
    service.routes.push_back({.pattern = "/wildcard/{*path}", .invoke = [&](auto &&context) {
                                CHECK(context.httpArgs.method == dsApi::HttpMethod::POST);
                                CHECK(ds::fromJson<std::string>(context.httpArgs.content) == "test");
                                return dsApi::okDocumentResult(context.routeArgs["path"]);
                              }});

    THEN("request to /wildcard") {
      CHECK(dsApi::data<std::string>(service.invoke({.path = "/wildcard", .httpArgs = makeArgs()})) == "");
      ;
    }

    THEN("request to /wildcard/ succeeds") {
      CHECK(dsApi::data<std::string>(service.invoke({.path = "/wildcard/", .httpArgs = makeArgs()})) == "/");
      ;
    }

    THEN("request to /wildcard/* succeeds") {
      CHECK(dsApi::data<std::string>(service.invoke({.path = "/wildcard/1", .httpArgs = makeArgs()})) == "/1");
    }

    THEN("request to non-matching prefix fails") {
      requireResourceNotFound(service.invoke({.path = "", .httpArgs = makeArgs()}));
      requireResourceNotFound(service.invoke({.path = "/", .httpArgs = makeArgs()}));
      requireResourceNotFound(service.invoke({.path = "/wildcar", .httpArgs = makeArgs()}));
    }

    THEN("request to matching prefix without '/'") {
      requireResourceNotFound(service.invoke({.path = "/wildcards", .httpArgs = makeArgs()}));
    }
  }

  GIVEN("/fail route throwing exception") {
    service.routes.push_back(
        {.pattern = "/fail", .invoke = [&](auto &&...) -> dsApi::HttpResult { DS_FAIL_REQUIRE("Test failure."); }});

    THEN("request to /fail fails") {
      requireError(service.invoke({.path = "/fail", .httpArgs = {.method = dsApi::HttpMethod::POST}}), "Test failure.");
    }
  }

  GIVEN("method /test") {
    Method method;
    service.routes.push_back(dsApi::makeMethodRoute("/test", [&]() { method.call(); }));
    service.routes.push_back(dsApi::makeMethodRoute("/testReturn", [&]() { return method.callWithReturn(); }));
    auto call = [&](dsApi::HttpMethod method = dsApi::HttpMethod::POST) {
      return service.invoke({.path = "/test", .httpArgs = {.method = method}});
    };
    auto callWithReturn = [&](dsApi::HttpMethod method = dsApi::HttpMethod::POST) {
      return service.invoke({.path = "/testReturn", .httpArgs = {.method = method}});
    };
    THEN("POST succeeds with 204 and on content") {
      REQUIRE_CALL(method, call());
      CHECK(call() == dsApi::noContentResult());
    }

    THEN("POST succeeds with 200 and returns content") {
      REQUIRE_CALL(method, callWithReturn()).RETURN(true);
      CHECK(callWithReturn() == dsApi::okDocumentResult(true));
    }

    THEN("POST returns error on exception") {
      REQUIRE_CALL(method, call()).LR_SIDE_EFFECT(DS_FAIL_REQUIRE("Test failure"));
      requireError(call(), "Test failure");
    }

    THEN("Other methods fail") {
      for (auto method :
           {dsApi::HttpMethod::GET, dsApi::HttpMethod::PUT, dsApi::HttpMethod::DELETE, dsApi::HttpMethod::PATCH}) {
        requireReplyMethodNotAllowed(call(method), "POST");
      }
    }
  }

  GIVEN("method /test with args") {
    ArgsMethod method;
    service.routes.push_back(dsApi::makeMethodRoute<Args>("/test", [&](auto &&args) { method.call(DS_FWD(args)); }));
    service.routes.push_back(
        dsApi::makeMethodRoute<Args>("/testReturn", [&](auto &&args) { return method.callWithReturn(DS_FWD(args)); }));
    THEN("POST succeeds with 204 and no content") {
      REQUIRE_CALL(method, call(Args{.a = "7"}));
      CHECK(service.invoke({.path = "/test",
                            .httpArgs = {.method = dsApi::HttpMethod::POST,
                                         .content = ds::toJsonFragment(Args{.a = "7"})}}) == dsApi::noContentResult());
    }

    THEN("POST succeeds with 200 and returns content") {
      REQUIRE_CALL(method, callWithReturn(Args{.a = "7"})).RETURN(true);
      CHECK(service.invoke(
                {.path = "/testReturn",
                 .httpArgs = {.method = dsApi::HttpMethod::POST, .content = ds::toJsonFragment(Args{.a = "7"})}}) ==
            dsApi::okDocumentResult(true));
    }

    THEN("POST returns error on exception") {
      REQUIRE_CALL(method, call(_)).LR_SIDE_EFFECT(DS_FAIL_REQUIRE("Test failure"));
      requireError(service.invoke({.path = "/test", .httpArgs = {.method = dsApi::HttpMethod::POST}}), "Test failure");
    }

    THEN("Other methods fail") {
      for (auto method :
           {dsApi::HttpMethod::GET, dsApi::HttpMethod::PUT, dsApi::HttpMethod::DELETE, dsApi::HttpMethod::PATCH}) {
        requireReplyMethodNotAllowed(service.invoke({.path = "/test", .httpArgs = {.method = method}}), "POST");
      }
    }
  }

  GIVEN("methodById /collection/{id}/test") {
    ArgsByIdMethod method;
    service.routes.push_back(dsApi::makeMethodByIdRoute<Id, Args>(
        "/collection/{id}/test", [&](auto &&... args) { method.call(DS_FWD(args)...); }));

    THEN("POST passes id and arguments from content, succeeds with 204 and no content") {
      REQUIRE_CALL(method, call(Id{"8"}, Args{.a = "7"}));
      CHECK(service.invoke({.path = "/collection/8/test",
                            .httpArgs = {.method = dsApi::HttpMethod::POST,
                                         .content = ds::toJsonFragment(Args{.a = "7"})}}) == dsApi::noContentResult());
    }

    THEN("POST returns error on exception") {
      REQUIRE_CALL(method, call(_, _)).LR_SIDE_EFFECT(DS_FAIL_REQUIRE("Test failure"));
      requireError(service.invoke({.path = "/collection/7/test", .httpArgs = {.method = dsApi::HttpMethod::POST}}),
                   "Test failure");
    }

    THEN("Other methods fail") {
      for (auto method :
           {dsApi::HttpMethod::GET, dsApi::HttpMethod::PUT, dsApi::HttpMethod::DELETE, dsApi::HttpMethod::PATCH}) {
        requireReplyMethodNotAllowed(service.invoke({.path = "/collection/7/test", .httpArgs = {.method = method}}),
                                     "POST");
      }
    }
  }

  GIVEN("methodById with result /collection/{id}/test") {
    ResultArgsByIdMethod method;
    service.routes.push_back(dsApi::makeMethodByIdRoute<Id, Args>(
        "/collection/{id}/test", [&](auto &&... args) { return method.call(DS_FWD(args)...); }));

    THEN("POST passes id and arguments from content, succeeds with 200 and result") {
      REQUIRE_CALL(method, call(Id{"8"}, Args{.a = "7"})).RETURN(3);
      CHECK(service.invoke({.path = "/collection/8/test",
                            .httpArgs = {.method = dsApi::HttpMethod::POST,
                                         .content = ds::toJsonFragment(Args{.a = "7"})}}) == dsApi::okResult(3));
    }

    THEN("POST returns error on exception") {
      REQUIRE_CALL(method, call(_, _)).LR_SIDE_EFFECT(DS_FAIL_REQUIRE("Test failure")).RETURN(3);
      requireError(service.invoke({.path = "/collection/7/test", .httpArgs = {.method = dsApi::HttpMethod::POST}}),
                   "Test failure");
    }

    THEN("Other methods fail") {
      for (auto method :
           {dsApi::HttpMethod::GET, dsApi::HttpMethod::PUT, dsApi::HttpMethod::DELETE, dsApi::HttpMethod::PATCH}) {
        requireReplyMethodNotAllowed(service.invoke({.path = "/collection/7/test", .httpArgs = {.method = method}}),
                                     "POST");
      }
    }
  }

  GIVEN("document /test with GET") {
    auto called = 0;
    service.routes.push_back(dsApi::makeDocumentRoute("/test", {.methods = {dsApi::makeGet([&]() {
                                                                  called++;
                                                                  return 7;
                                                                })}}));

    THEN("GET method returns document") {
      CHECK(dsApi::data<int>(service.invoke({.path = "/test", .httpArgs = {.method = dsApi::HttpMethod::GET}})) == 7);
      CHECK(called == 1);
    }

    THEN("other methods fail") {
      for (auto method :
           {dsApi::HttpMethod::PUT, dsApi::HttpMethod::DELETE, dsApi::HttpMethod::PATCH, dsApi::HttpMethod::POST}) {
        requireReplyMethodNotAllowed(service.invoke({.path = "/test", .httpArgs = {.method = method}}), "GET");
      }
      CHECK(called == 0);
    }
  }

  GIVEN("document /test with GET with argument") {
    service.routes.push_back(
        dsApi::makeDocumentRoute("/test/{a}/{b}", {.methods = {dsApi::makeGet<Args>([&](auto &&args) {
                                                     CHECK(args.a == "1");
                                                     CHECK(args.b == "2");
                                                     return 7;
                                                   })}}));

    THEN("GET method returns document and parses arguments") {
      CHECK(dsApi::data<int>(service.invoke({.path = "/test/1/2", .httpArgs = {.method = dsApi::HttpMethod::GET}})) ==
            7);
    }
  }

  GIVEN("document /test/{id} with GET, DELETE") {
    service.routes.push_back(dsApi::makeDocumentRoute(
        "/test/{id}", {.methods = {
                           dsApi::makeGetById<Id>([&](const Id &id) -> bool { DS_FAIL_REQUIRE(""); }),
                           dsApi::makeDeleteById<Id>([&](const Id &id) { DS_FAIL_REQUIRE(""); }),
                       }}));

    THEN("POST methods fails with valid ALLOW header") {
      requireReplyMethodNotAllowed(service.invoke({.path = "/test/3", .httpArgs = {.method = dsApi::HttpMethod::POST}}),
                                   "GET, DELETE");
    }
  }

  GIVEN("document /test/{id} with GET, DELETE, PATCH") {
    service.routes.push_back(dsApi::makeDocumentRoute(
        "/test/{id}", {.methods = {dsApi::makeGetById<Id>([&](const Id &id) -> bool { DS_FAIL_REQUIRE(""); }),
                                   dsApi::makeDeleteById<Id>([&](const Id &id) { DS_FAIL_REQUIRE(""); })},
                       .patchRoutes = {{.pattern = "/{b}",
                                        .operations = {dsApi::makeReplacePatch<Args, double>(
                                            [&](auto &&args, auto &&value) { DS_FAIL_REQUIRE(""); })}}}}));

    THEN("POST methods fails with valid ALLOW header") {
      requireReplyMethodNotAllowed(service.invoke({.path = "/test/3", .httpArgs = {.method = dsApi::HttpMethod::POST}}),
                                   "GET, DELETE, PATCH");
    }
  }

  SECTION("makeGetQuery") {
    CHECK(dsApi::makeGetQuery(dsApi::HttpRequest{}) == dsApi::GetQuery{});
    CHECK(dsApi::makeGetQuery(dsApi::HttpRequest{.httpArgs = {.query = "include="}}) == dsApi::GetQuery{});
    CHECK(dsApi::makeGetQuery(dsApi::HttpRequest{.httpArgs = {.query = "include=aa"}}) ==
          dsApi::GetQuery{.include = {"aa"}});
    CHECK(dsApi::makeGetQuery(dsApi::HttpRequest{.httpArgs = {.query = "include=aa,bb"}}) ==
          dsApi::GetQuery{.include = {"aa", "bb"}});
    CHECK(dsApi::makeGetQuery(dsApi::HttpRequest{.httpArgs = {.query = "includeAll=true"}}) ==
          dsApi::GetQuery{.includeAll = true});
    CHECK(dsApi::makeGetQuery(dsApi::HttpRequest{.httpArgs = {.query = "include=aa&includeAll=true"}}) ==
          dsApi::GetQuery{.include = {"aa"}, .includeAll = true});
  }

  GIVEN("document /test with GET resource by id") {
    auto method = GetByIdMock{};
    service.routes.push_back(dsApi::makeDocumentRoute(
        "/test/{id}", {.methods = {dsApi::makeGetById<Id>([&](const Id &id) { return method.call(id); })}}));
    THEN("calls method") {
      REQUIRE_CALL(method, call(Id{"d"})).RETURN(7);
      CHECK(dsApi::data<int>(service.invoke({.path = "/test/d", .httpArgs = {.method = dsApi::HttpMethod::GET}})) == 7);
    }
  }

  GIVEN("makeGetWithQuery") {
    auto method = GetWithQueryMock{};
    service.routes.push_back(dsApi::makeDocumentRoute(
        "/test",
        {.methods = {dsApi::makeGetWithQuery([&](const dsApi::GetQuery &query) { return method.call(query); })}}));
    THEN("calls method") {
      REQUIRE_CALL(method, call(dsApi::GetQuery{.include = {"aa"}})).RETURN(7);
      CHECK(dsApi::data<int>(service.invoke(
                {.path = "/test",
                 .httpArgs = {.method = dsApi::HttpMethod::GET, .content = {}, .query = "include=aa"}})) == 7);
    }
  }

  GIVEN("makeGetByIdWithQuery") {
    auto method = GetByIdWithQueryMock{};
    service.routes.push_back(dsApi::makeDocumentRoute(
        "/test/{id}", {.methods = {dsApi::makeGetByIdWithQuery<Id>(
                           [&](const Id &id, const dsApi::GetQuery &query) { return method.call(id, query); })}}));
    THEN("calls method") {
      REQUIRE_CALL(method, call(Id{"d"}, dsApi::GetQuery{.include = {"aa"}})).RETURN(7);
      CHECK(dsApi::data<int>(service.invoke(
                {.path = "/test/d",
                 .httpArgs = {.method = dsApi::HttpMethod::GET, .content = {}, .query = "include=aa"}})) == 7);
    }
  }

  GIVEN("document /test with DELETE resource by id") {
    auto method = ByIdMock{};
    service.routes.push_back(dsApi::makeDocumentRoute(
        "/test/{id}", {.methods = {dsApi::makeDeleteById<Id>([&](const Id &id) { return method.call(id); })}}));
    THEN("calls method") {
      REQUIRE_CALL(method, call(Id{"d"}));
      dsApi::data<int>(service.invoke({.path = "/test/d", .httpArgs = {.method = dsApi::HttpMethod::DELETE}}));
    }
  }

  // TODO(someday): makeCreate

  GIVEN("document /test with makeMethod") {
    auto method = ArgsMethod{};
    service.routes.push_back(dsApi::makeDocumentRoute(
        "/test/{a}",
        {.methods = {dsApi::makeMethod<Args>([&](Args &&args) { return method.call(std::move(args)); })}}));
    THEN("calls method") {
      REQUIRE_CALL(method, call(Args{.a = "d"}));
      dsApi::data<int>(service.invoke({.path = "/test/d", .httpArgs = {.method = dsApi::HttpMethod::POST}}));
    }
  }

  GIVEN("document /{a} with PATCH routes") {
    auto called = 0;
    service.routes.push_back(dsApi::makeDocumentRoute(
        "/{a}",
        {.methods = {},
         .patchRoutes = {
             {.pattern = "/{b}", .operations = {dsApi::makeReplacePatch<Args, double>([&](auto &&args, auto &&value) {
                                   CHECK(args.a == "1");
                                   CHECK(args.b == "2");
                                   CHECK(value == 3);
                                   called++;
                                 })}},
             {.pattern = "/x/{b}", .operations = {dsApi::makeReplacePatch<Args, double>([&](auto &&args, auto &&value) {
                                     CHECK(args.a == "1");
                                     CHECK(args.b == "3");
                                     CHECK(value == 5);
                                     called++;
                                   })}}}}));

    THEN("replace PATCH is routed") {
      dsApi::PatchOperations patchOperations;
      patchOperations.push_back(
          {.op = dsApi::PatchOperationType::replace, .path = "/2", .value = ds::toJsonFragment(3.0)});
      patchOperations.push_back(
          {.op = dsApi::PatchOperationType::replace, .path = "/x/3", .value = ds::toJsonFragment(5.0)});
      CHECK(service.invoke({.path = "/1",
                            .httpArgs = {.method = dsApi::HttpMethod::PATCH,
                                         .content = ds::toJsonFragment(patchOperations)}}) == dsApi::noContentResult());
      CHECK(called == 2);
    }

    THEN("other PATCH operations fail") {
      for (auto op :
           {dsApi::PatchOperationType::add, dsApi::PatchOperationType::remove, dsApi::PatchOperationType::test}) {
        dsApi::PatchOperations patchOperations;
        patchOperations.push_back({.op = op, .path = "/2", .value = ds::toJsonFragment(3.0)});
        requireError(service.invoke({.path = "/1",
                                     .httpArgs = {.method = dsApi::HttpMethod::PATCH,
                                                  .content = ds::toJsonFragment(patchOperations)}}),
                     "Path operation not supported.");
      }
      CHECK(!called);
    }

    THEN("replace PATCH is routed more times till first error") {
      dsApi::PatchOperations patchOperations;
      patchOperations.push_back(
          {.op = dsApi::PatchOperationType::replace, .path = "/2", .value = ds::toJsonFragment(3.0)});
      patchOperations.push_back(
          {.op = dsApi::PatchOperationType::replace, .path = "/2", .value = ds::toJsonFragment(3.0)});
      patchOperations.push_back({.op = dsApi::PatchOperationType::add, .path = "/2", .value = ds::toJsonFragment(3.0)});
      patchOperations.push_back(
          {.op = dsApi::PatchOperationType::replace, .path = "/2", .value = ds::toJsonFragment(3.0)});
      requireError(service.invoke({.path = "/1",
                                   .httpArgs = {.method = dsApi::HttpMethod::PATCH,
                                                .content = ds::toJsonFragment(patchOperations)}}),
                   "Path operation not supported.");
      CHECK(called == 2);
    }

    THEN("other http methods fail") {
      for (auto method :
           {dsApi::HttpMethod::GET, dsApi::HttpMethod::PUT, dsApi::HttpMethod::DELETE, dsApi::HttpMethod::POST}) {
        requireReplyMethodNotAllowed(service.invoke({.path = "/7", .httpArgs = {.method = method}}), "PATCH");
        CHECK(!called);
      }
    }
  }

  GIVEN("document /{id} with PATCH routes by id") {
    auto called = 0;
    service.routes.push_back(dsApi::makeDocumentRoute(
        "/{id}",
        {.methods = {},
         .patchRoutes = {{.pattern = "/value",
                          .operations = {dsApi::makeReplaceByIdPatch<Id, double>([&](const Id &id, auto &&value) {
                            CHECK(id == Id{"1"});
                            CHECK(value == 3);
                            called++;
                          })}}}}));

    THEN("replace PATCH by id is routed") {
      dsApi::PatchOperations patchOperations;
      patchOperations.push_back(
          {.op = dsApi::PatchOperationType::replace, .path = "/value", .value = ds::toJsonFragment(3.0)});
      CHECK(service.invoke({.path = "/1",
                            .httpArgs = {.method = dsApi::HttpMethod::PATCH,
                                         .content = ds::toJsonFragment(patchOperations)}}) == dsApi::noContentResult());
      CHECK(called == 1);
    }
  }
}

} // namespace

DS_REFLECT_ANONYMOUS_STRUCT(Args, (a)(b));
