/*
    Copyright (c) 2019 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/
#include "uploader.h"

#include <ds-apartment/ds-devices.h>
#include <ds-backend/connector.h>
#include <ds/asio/ds-signal.h>
#include <ds/asio/promise.h>
#include <ds/asio/resource.h>

#include <deque>

using namespace std::literals;
static ds::LogChannel dsLogChannel("dsBackendUploader");

namespace dsBackend {

UploaderHooks uploaderHooks{};

class UploaderImpl : public Uploader {
public:
  UploaderImpl(ds::asio::ResourceBase& parent, dsBackend::Connector& bus, UploaderHooks& hooks)
      : Uploader{parent, "uploader"}, m_bus{bus}, m_hooks{hooks} {
    DS_INFO_ENTER(*this);
  }

private:
  dsBackend::Connector& m_bus;
  UploaderHooks& m_hooks;
  bool m_isBusConnected{m_bus.isConnected()};
  ds::Slot<bool> DS_UNIQUE_NAME(slot){*this, m_bus.isConnectedChanged(),
                                      [this](bool x) { onBusIsConnectedChanged(x); }};
  std::optional<dsApi::HttpRequest> m_inUpload{};
  std::deque<dsApi::HttpRequest> m_toUpload{};

  void asyncUpload(dsApi::HttpRequest&& x) override {
    DS_INFO_ENTER(*this, x, m_toUpload.size());
    assertSynchronized();
    if (m_toUpload.size() >= config().capacity) {
      DS_LOG(m_isBusConnected ? ds::LogSeverity::WARNING : ds::LogSeverity::INFO, *this,
             "Discarding record on overflow", m_toUpload.front());
      m_hooks.dropped(m_toUpload.front());
      m_toUpload.pop_front();
    }
    m_toUpload.push_back(std::move(x));
    triggerUpload();
  }

  void onBusIsConnectedChanged(bool isBusConnected) {
    if (m_isBusConnected == isBusConnected) {
      return;
    }
    DS_INFO_ENTER(*this, isBusConnected);
    m_isBusConnected = isBusConnected;
    if (isBusConnected) {
      triggerUpload();
    }
  }

  void triggerUpload() {
    DS_DEBUG_ENTER(*this, m_inUpload, m_isBusConnected, m_toUpload.size());
    if (m_inUpload || !m_isBusConnected || m_toUpload.empty()) {
      if (!m_inUpload && m_toUpload.empty()) {
        m_hooks.nothingToDo();
      } else if (!m_isBusConnected) {
        m_hooks.blockedOnDisconnected();
      }
      return;
    }
    m_inUpload = std::move(m_toUpload.front());
    m_toUpload.pop_front();
    m_bus.asyncInvoke(dsApi::HttpRequest{m_inUpload.value()},
                      ds::makePromise<dsApi::HttpResult>(io(), [this](std::optional<dsApi::HttpResult>&& result) {
                        DS_INFO_ENTER(*this, result);
                        if (!result || !result->isSuccess()) {
                          if (!m_isBusConnected && m_toUpload.size() < config().capacity) {
                            DS_WARNING(*this, "Upload failed, bus disconnected, retry later.", m_inUpload, result);
                            m_toUpload.push_front(std::move(m_inUpload.value()));
                            // will back off because bus is disconnected
                          } else {
                            DS_ERROR(*this, "Upload failed, dropping.", m_inUpload, result);
                            m_hooks.dropped(m_inUpload.value());
                          }
                        } else {
                          DS_INFO(*this, "Succeeded.", m_inUpload);
                          // consume pop_front on success and upload next one
                          m_hooks.succeeded(m_inUpload.value());
                        }
                        m_inUpload.reset();
                        triggerUpload();
                      }));
  }
};

std::unique_ptr<Uploader> makeUploader(ds::asio::ResourceBase& parent, dsBackend::Connector& bus,
                                       UploaderHooks& hooks) {
  return std::make_unique<UploaderImpl>(parent, bus, hooks);
}

} // namespace dsBackend

DS_JSON_DEFINE_STRUCT((dsBackend), UploaderConfig);
