#include "media-library-api.h"

#include <ds-api/service.h>
#include <ds-model/media-library.h>
#include <ds/json.h>
#include <ds/log.h>
#include <ds/system.h>

static ds::LogChannel dsLogChannel("dsMediaLibrary");

namespace dsMediaLibrary {

struct MediaLibraryApiImpl : public MediaLibraryApi {
  MediaLibraryApiImpl(ds::System& system)
      : m_system(system),
        m_clipsPath{ds::str(m_system.dirs().datadir, "/ds-media-library/clips")},
        m_resourcesPath{ds::str(m_system.dirs().datadir, "/ds-media-library/resources")} {
    loadClips("/");
    loadClips("/user");
  }

  void loadClips(std::string&& path) {
    DS_NOTICE("Load clips", path);
    try {
      const auto& clipsPath = ds::str(m_clipsPath, path);
      auto files = m_system.lsRegularFiles(clipsPath, ".json");
      for (const auto& file : files) {
        try {
          auto clip = m_system.readFile(file);
          m_clips.clips.push_back(ds::fromJson<dsModel::Clip>(clip));
        } catch (std::exception& e) {
          DS_NOTICE("Cannot create Clip from JSON file", file);
        }
      }
    } catch (std::exception& e) {
      DS_FAIL_REQUIRE("Cannot initialize Media library", e);
    }
  }

  [[nodiscard]] dsModel::Clip clip(const dsModel::ClipId& id) const override {
    const auto& clipIt =
        std::find_if(m_clips.clips.cbegin(), m_clips.clips.cend(), [&](const auto& _clip) { return _clip.id == id; });
    if (clipIt != m_clips.clips.cend()) {
      return *clipIt;
    }
    return dsModel::Clip{.id = id};
  }

  [[nodiscard]] auto getClipIterator(const dsModel::ClipId& id) {
    auto clipIt =
        std::find_if(m_clips.clips.begin(), m_clips.clips.end(), [&](const auto& _clip) { return _clip.id == id; });
    DS_REQUIRE(clipIt != m_clips.clips.end(), "Clip does not exists", id);
    return clipIt;
  }

  [[nodiscard]] auto clipFilePath(const dsModel::Clip& clip) {
    DS_REQUIRE(clip.attributes.category == dsModel::ClipCategoryId::user);
    return ds::str(m_clipsPath, "/user/", clip.id, ".json");
  }

  [[nodiscard]] auto clipResourcePath(const dsModel::Clip& clip) {
    DS_REQUIRE(clip.attributes.category == dsModel::ClipCategoryId::user);
    return ds::str(m_resourcesPath, "/user/", clip.id, ".mp3");
  }

  [[nodiscard]] auto clipResourceUri(const dsModel::Clip& clip) {
    DS_REQUIRE(clip.attributes.category == dsModel::ClipCategoryId::user);
    return ds::str("/media/user/", clip.id, ".mp3");
  }

  void validateTmpFilePath(const std::string& path) {
    boost::filesystem::path tmp{path};
    DS_REQUIRE(tmp.parent_path().string() == "/tmp");
  }

  void saveClipJsonFile(const dsModel::Clip& clip) { m_system.writeToFile(clipFilePath(clip), ds::toJson(clip)); }

  void deleteClip(const dsModel::ClipId& id) override {
    DS_NOTICE("Delete clip", id);
    auto clipIt = getClipIterator(id);
    DS_REQUIRE(clipIt->attributes.category == dsModel::ClipCategoryId::user, "Only user clip can be deleted.",
               clipIt->id);

    auto resourcePath = clipResourcePath(*clipIt);
    if (m_system.existsFile(resourcePath)) {
      m_system.remove(resourcePath);
    }

    auto filePath = clipFilePath(*clipIt);
    if (m_system.existsFile(filePath)) {
      m_system.remove(filePath);
    }

    m_clips.clips.erase(clipIt);
  }

  [[nodiscard]] dsModel::Clips clips() const override { return m_clips; }

  [[nodiscard]] dsModel::ClipId addClip(dsModel::Clip&& clip) override {
    DS_NOTICE("Add clip", clip);

    DS_REQUIRE(!clip.attributes.name.message.empty(), "Clip name cannot be empty.");
    DS_REQUIRE(clip.attributes.category == dsModel::ClipCategoryId::user,
               "Invalid clip type. Only user clip can be created.", clip.attributes.category);
    DS_REQUIRE(!clip.attributes.resources.empty(), "Clip resource cannot be empty.");
    DS_REQUIRE(clip.attributes.resources.size() == 1, "Clip resource should have only one resource.");
    DS_REQUIRE(!clip.attributes.resources.at(0).uri.empty(), "Clip resource uri cannot be empty.");

    const auto id = dsModel::ClipId{ds::str("clipId", m_system.randomId())};
    auto exists =
        std::any_of(m_clips.clips.cbegin(), m_clips.clips.cend(), [&](const auto& _clip) { return _clip.id == id; });
    DS_REQUIRE(!exists, "Clip with id already exists!", id);

    auto tmpFilePath = clip.attributes.resources.at(0).uri;
    validateTmpFilePath(tmpFilePath);

    auto newClip = dsModel::Clip{.id = id,
                                 .lastChanged = m_system.now(),
                                 .attributes = {.name = {.message = std::move(clip.attributes.name.message)},
                                                .category = std::move(dsModel::ClipCategoryId::user),
                                                .resources = std::move(clip.attributes.resources)}};

    newClip.attributes.resources.at(0).uri = clipResourceUri(newClip);

    saveClipJsonFile(newClip);
    m_clips.clips.push_back(newClip);

    boost::filesystem::copy_file(tmpFilePath, clipResourcePath(newClip));
    m_system.remove(tmpFilePath);

    return id;
  }

  void setClipName(const dsModel::ClipId& id, const std::string& name) override {
    DS_NOTICE("Set clip name", id, name);
    auto clipIt = getClipIterator(id);
    DS_REQUIRE(clipIt->attributes.category == dsModel::ClipCategoryId::user, "Only user clip can be modified.",
               clipIt->id);
    auto oldName = std::exchange(clipIt->attributes.name.message, name);
    try {
      saveClipJsonFile(*clipIt);
    } catch (const ds::Exception& e) {
      clipIt->attributes.name.message = oldName;
      DS_FAIL_REQUIRE("Could not save the file.", e.what());
    }
  }

  void setClipUri(const dsModel::ClipId& id, const std::string& uri) override {
    DS_NOTICE("Set clip uri", id, uri);
    validateTmpFilePath(uri);
    auto clipIt = getClipIterator(id);
    DS_REQUIRE(clipIt->attributes.category == dsModel::ClipCategoryId::user, "Only user clip can be modified.",
               clipIt->id);
    DS_REQUIRE(clipIt->attributes.resources.size() == 1, "Clip resource should have only one resource.");

    boost::filesystem::copy_file(uri, clipResourcePath(*clipIt), boost::filesystem::copy_option::overwrite_if_exists);
    m_system.remove(uri);
  }

  void setClipDuration(const dsModel::ClipId& id, const ds::Clock::duration& duration) override {
    DS_NOTICE("Set clip duration", id, duration);
    auto clipIt = getClipIterator(id);
    DS_REQUIRE(clipIt->attributes.category == dsModel::ClipCategoryId::user, "Only user clip can be modified.",
               clipIt->id);
    DS_REQUIRE(clipIt->attributes.resources.size() == 1, "Clip resource should have only one resource.");
    auto oldDuration = std::exchange(clipIt->attributes.resources.at(0).duration, duration);
    try {
      saveClipJsonFile(*clipIt);
    } catch (const ds::Exception& e) {
      clipIt->attributes.resources.at(0).duration = oldDuration;
      DS_FAIL_REQUIRE("Could not save the file.", e.what());
    }
  }

private:
  ds::System& m_system;
  dsModel::Clips m_clips;
  const std::string m_clipsPath;
  const std::string m_resourcesPath;
};

struct MediaLibraryService : public dsApi::Service {
  MediaLibraryService(MediaLibraryApi& api) : m_api(api) {
    routes.push_back(dsApi::makeDocumentRoute("/clips", {.methods = {dsApi::makeGet([&]() { return m_api.clips(); }),
                                                                     dsApi::makeCreate<dsModel::Clip>([&](auto&& clip) {
                                                                       DS_NOTICE("Create Clip.", clip);
                                                                       return m_api.addClip(std::move(clip));
                                                                     })}}));
    routes.push_back(dsApi::makeDocumentRoute(
        "/clips/{id}",
        {.methods = {dsApi::makeGetById<dsModel::ClipId>([&](const auto& clipId) { return m_api.clip(clipId); }),
                     dsApi::makeDeleteById<dsModel::ClipId>(
                         [&](const auto& clipId) { return m_api.deleteClip(clipId); })},
         .patchRoutes = {{.pattern = "/name",
                          .operations = {dsApi::makeReplaceByIdPatch<dsModel::ClipId, std::string>(
                              [&](auto&&... args) { m_api.setClipName(DS_FWD(args)...); })}},
                         {.pattern = "/uri",
                          .operations = {dsApi::makeReplaceByIdPatch<dsModel::ClipId, std::string>(
                              [&](auto&&... args) { api.setClipUri(DS_FWD(args)...); })}},
                         {.pattern = "/duration",
                          .operations = {dsApi::makeReplaceByIdPatch<dsModel::ClipId, ds::Clock::duration>(
                              [&](auto&&... args) { api.setClipDuration(DS_FWD(args)...); })}}}}));
  }

private:
  MediaLibraryApi& m_api;
};

std::unique_ptr<MediaLibraryApi> makeMediaLibraryApi(ds::System& system) {
  return std::make_unique<MediaLibraryApiImpl>(system);
}

std::unique_ptr<dsApi::Service> makeMediaLibraryService(MediaLibraryApi& api) {
  return std::make_unique<MediaLibraryService>(api);
}

} // namespace dsMediaLibrary
