/*
    Copyright (c) 2016 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/
#include <ds-api/service.h>
#include <ds/catch/catch.h>
#include <ds/log.h>

#include <boost/algorithm/string/predicate.hpp>

#include "apartment-api-mock.h"
#include "model.h"

using namespace trompeloeil;

namespace {

static const char *TAGS = "[dsModelApartment][dsModel]";

SCENARIO("dsModelScenariosApartmentApiRoutes", TAGS) {
  dsModel::ScenariosApiMock mock;
  dsApi::Service service{dsModel::apartmentApiRoutes(mock)};

  SECTION("invokeStandardScenario") {
    const auto args = dsModel::InvokeStandardScenarioArgs{.context = dsModel::InvokeContextId{"x"}};
    REQUIRE_CALL(mock, invokeStandardScenario(_)).LR_SIDE_EFFECT(CHECK(_1 == args););
    CHECK(service.invoke(dsApi::HttpRequest{
              .path = "/scenarios/invoke",
              .httpArgs = {.method = dsApi::HttpMethod::POST, .content = ds::toJsonFragment(args)}}) ==
          dsApi::noContentResult());
  }

  SECTION("createScenario") {
    const auto data = dsModel::Scenario{
        dsModel::ApplicationZoneScenario{.id = {}, .type = "applicationZoneScenario", .attributes = {.name = "name0"}}};
    REQUIRE_CALL(mock, createScenario(_)).LR_SIDE_EFFECT(CHECK(_1 == data);).LR_RETURN(dsModel::ScenarioId{"id0"});
    const auto result = service.invoke(dsApi::HttpRequest{
        .path = "/scenarios",
        .httpArgs = {.method = dsApi::HttpMethod::POST, .content = ds::toJsonFragment(dsApi::document(data))}});
    CHECK(result.status == 201);
    CHECK(result.headers == dsApi::HttpHeaders{{"Location", {"id0"}}});
  }

  SECTION("deleteScenario") {
    const auto id = dsModel::ScenarioId{"x"};
    REQUIRE_CALL(mock, deleteScenario(_)).LR_SIDE_EFFECT(CHECK(_1 == id));
    CHECK(service.invoke(dsApi::HttpRequest{.path = ds::str("/scenarios/", id),
                                            .httpArgs = {.method = dsApi::HttpMethod::DELETE}}) ==
          dsApi::noContentResult());
  }

  SECTION("setScenarioName") {
    const auto id = dsModel::ScenarioId{"x"};
    const auto value = "{newName}";
    REQUIRE_CALL(mock, setScenarioName(_, _)).LR_SIDE_EFFECT(CHECK(_1 == id); CHECK(_2 == value););
    dsApi::PatchOperations patchOperations{};
    patchOperations.push_back(
        {.op = dsApi::PatchOperationType::replace, .path = ds::str("/name"), .value = ds::toJsonFragment(value)});
    CHECK(service.invoke(dsApi::HttpRequest{
              .path = ds::str("/scenarios/", id),
              .httpArgs = {.method = dsApi::HttpMethod::PATCH, .content = ds::toJsonFragment(patchOperations)}}) ==
          dsApi::noContentResult());
  }

  SECTION("invokeScenario") {
    const auto id = dsModel::ScenarioId{"x"};
    REQUIRE_CALL(mock, invokeScenario(_)).LR_SIDE_EFFECT(CHECK(_1 == id););
    CHECK(service.invoke(
              dsApi::HttpRequest{.path = "/scenarios/x/invoke", .httpArgs = {.method = dsApi::HttpMethod::POST}}) ==
          dsApi::noContentResult());
  }

  SECTION("persistScenario") {
    const auto id = dsModel::ScenarioId{"x"};
    REQUIRE_CALL(mock, persistScenario(_)).LR_SIDE_EFFECT(CHECK(_1 == id););
    CHECK(service.invoke(
              dsApi::HttpRequest{.path = "/scenarios/x/persist", .httpArgs = {.method = dsApi::HttpMethod::POST}}) ==
          dsApi::noContentResult());
  }
}

SCENARIO("dsModelApartmentApiService", TAGS) {
  dsModel::ScenariosApiMock scenariosApi;
  dsModel::ApartmentApiMock apartmentApi;
  dsModel::MeteringsApiMock meteringsApi{};

  REQUIRE_CALL(apartmentApi, scenarios()).LR_RETURN(scenariosApi);
  ALLOW_CALL(apartmentApi, meterings()).LR_RETURN(meteringsApi);
  auto service = dsModel::makeApartmentService(apartmentApi);

  SECTION("scenarios smoke test") {
    // full coverage is in scenarios-test
    const auto id = dsModel::ScenarioId{"x"};
    REQUIRE_CALL(scenariosApi, deleteScenario(_)).LR_SIDE_EFFECT(CHECK(_1 == id));
    CHECK(service->invoke(dsApi::HttpRequest{.path = ds::str("/scenarios/", id),
                                             .httpArgs = {.method = dsApi::HttpMethod::DELETE}}) ==
          dsApi::noContentResult());
  }

  auto checkGet = [&](const auto &path, const auto &data) {
    const auto result =
        service->invoke(dsApi::HttpRequest{.path = path, .httpArgs = {.method = dsApi::HttpMethod::GET}});
    INFO(result);
    CHECK(dsApi::data<std::decay_t<decltype(data)>>(result) == data);
  };

  SECTION("zone") {
    const auto data = dsModel::Zone{.id = dsModel::ZoneId{"id0"}};
    REQUIRE_CALL(apartmentApi, zone(data.id)).LR_RETURN(data);
    checkGet("/zones/id0", data);
  }

  SECTION("zones") {
    const auto zones = std::vector<dsModel::Zone>{{.id = dsModel::ZoneId{"id0"}}};
    const auto data = dsModel::Zones{.zones = zones};
    REQUIRE_CALL(apartmentApi, zones()).LR_RETURN(zones);
    checkGet("/zones", data);
  }

  SECTION("zoneStatus") {
    const auto data = dsModel::ZoneStatus{.id = dsModel::ZoneId{"id0"}};
    REQUIRE_CALL(apartmentApi, zoneStatus(data.id)).LR_RETURN(data);
    checkGet("/zones/id0/status", data);
  }

  SECTION("zoneStatuses") {
    const auto data = std::vector<dsModel::ZoneStatus>{{.id = dsModel::ZoneId{"id0"}}};
    REQUIRE_CALL(apartmentApi, zoneStatuses()).LR_RETURN(data);
    checkGet("/zones/status", data);
  }

  SECTION("setTemperatureStatusEcoPreferred") {
    const auto args = dsModel::TemperatureStatusArgs{.zoneId = dsModel::ZoneId{"1"}};
    const auto value = true;
    REQUIRE_CALL(apartmentApi, setTemperatureStatusEcoPreferred(_, _))
        .LR_SIDE_EFFECT(CHECK(_1 == args); CHECK(_2 == value););
    dsApi::PatchOperations patchOperations{};
    patchOperations.push_back({.op = dsApi::PatchOperationType::replace,
                               .path = ds::str("/applications/temperature/ecoPreferred"),
                               .value = ds::toJsonFragment(value)});
    CHECK(service->invoke(dsApi::HttpRequest{
              .path = "/zones/1/status",
              .httpArgs = {.method = dsApi::HttpMethod::PATCH, .content = ds::toJsonFragment(patchOperations)}}) ==
          dsApi::noContentResult());
  }

  SECTION("setTemperatureStatusSetbackRequested") {
    const auto args = dsModel::TemperatureStatusArgs{.zoneId = dsModel::ZoneId{"1"}};
    const auto value = true;
    REQUIRE_CALL(apartmentApi, setTemperatureStatusSetbackRequested(_, _))
        .LR_SIDE_EFFECT(CHECK(_1 == args); CHECK(_2 == value););
    dsApi::PatchOperations patchOperations{};
    patchOperations.push_back({.op = dsApi::PatchOperationType::replace,
                               .path = ds::str("/applications/temperature/setbackRequested"),
                               .value = ds::toJsonFragment(value)});
    CHECK(service->invoke(dsApi::HttpRequest{
              .path = "/zones/1/status",
              .httpArgs = {.method = dsApi::HttpMethod::PATCH, .content = ds::toJsonFragment(patchOperations)}}) ==
          dsApi::noContentResult());
  }

  SECTION("setTemperatureStatusPassiveCooling") {
    const auto args = dsModel::TemperatureStatusArgs{.zoneId = dsModel::ZoneId{"1"}};
    const auto value = true;
    REQUIRE_CALL(apartmentApi, setTemperatureStatusPassiveCooling(_, _))
        .LR_SIDE_EFFECT(CHECK(_1 == args); CHECK(_2 == value););
    dsApi::PatchOperations patchOperations{};
    patchOperations.push_back({.op = dsApi::PatchOperationType::replace,
                               .path = ds::str("/applications/temperature/passiveCooling"),
                               .value = ds::toJsonFragment(value)});
    CHECK(service->invoke(dsApi::HttpRequest{
              .path = "/zones/1/status",
              .httpArgs = {.method = dsApi::HttpMethod::PATCH, .content = ds::toJsonFragment(patchOperations)}}) ==
          dsApi::noContentResult());
  }

  SECTION("setZoneName") {
    const auto args = dsModel::ZoneArgs{.zoneId = dsModel::ZoneId{"x"}};
    const auto value = "{newName}";
    REQUIRE_CALL(apartmentApi, setZoneName(_, _)).LR_SIDE_EFFECT(CHECK(_1 == args); CHECK(_2 == value););
    dsApi::PatchOperations patchOperations{};
    patchOperations.push_back(
        {.op = dsApi::PatchOperationType::replace, .path = ds::str("/name"), .value = ds::toJsonFragment(value)});
    CHECK(service->invoke(dsApi::HttpRequest{
              .path = ds::str("/zones/", args.zoneId),
              .httpArgs = {.method = dsApi::HttpMethod::PATCH, .content = ds::toJsonFragment(patchOperations)}}) ==
          dsApi::noContentResult());
  }

  SECTION("setOutputValue") {
    const auto args = dsModel::OutputStatusArgs{.deviceId = dsModel::DSDeviceId{"dev"}, .outputId = "output"};
    const auto value = 3.0;
    REQUIRE_CALL(apartmentApi, setOutputValue(_, _)).LR_SIDE_EFFECT(CHECK(_1 == args); CHECK(_2 == value););
    dsApi::PatchOperations patchOperations{};
    patchOperations.push_back({.op = dsApi::PatchOperationType::replace,
                               .path = ds::str("/functionBlocks/", args.deviceId, "/outputs/", args.outputId, "/value"),
                               .value = ds::toJsonFragment(value)});
    CHECK(service->invoke(dsApi::HttpRequest{
              .path = ds::str("/dsDevices/", args.deviceId, "/status"),
              .httpArgs = {.method = dsApi::HttpMethod::PATCH, .content = ds::toJsonFragment(patchOperations)}}) ==
          dsApi::noContentResult());
  }

  SECTION("apartment") {
    const auto apartment = [&] { return dsModel::Apartment{.id = dsModel::ApartmentId{"a"}}; };
    const auto reply = [&] { return dsApi::HttpResult{.status = 200, .content = ds::toJsonFragment(dsApi::document(apartment()))}; };
    REQUIRE_CALL(apartmentApi, apartmentCached(_)).LR_RETURN(reply());
    const auto result = service->invoke(dsApi::HttpRequest{.path = "", .httpArgs = {.method = dsApi::HttpMethod::GET}});
    INFO(result);
    CHECK(dsApi::data<dsModel::Apartment>(result) == apartment());
  }

  SECTION("setApartmentAlarm1Name") {
    const auto value = "{newName}";
    REQUIRE_CALL(apartmentApi, setApartmentAlarm1Name(value));
    dsApi::PatchOperations patchOperations{};
    patchOperations.push_back({.op = dsApi::PatchOperationType::replace,
                               .path = ds::str("/security/alarm1/name"),
                               .value = ds::toJsonFragment(value)});
    CHECK(service->invoke(dsApi::HttpRequest{
              .path = "/",
              .httpArgs = {.method = dsApi::HttpMethod::PATCH, .content = ds::toJsonFragment(patchOperations)}}) ==
          dsApi::noContentResult());
  }

  SECTION("setApartmentAlarm2Name") {
    const auto value = "{newName}";
    REQUIRE_CALL(apartmentApi, setApartmentAlarm2Name(value));
    dsApi::PatchOperations patchOperations{};
    patchOperations.push_back({.op = dsApi::PatchOperationType::replace,
                               .path = ds::str("/security/alarm2/name"),
                               .value = ds::toJsonFragment(value)});
    CHECK(service->invoke(dsApi::HttpRequest{
              .path = "/",
              .httpArgs = {.method = dsApi::HttpMethod::PATCH, .content = ds::toJsonFragment(patchOperations)}}) ==
          dsApi::noContentResult());
  }

  SECTION("setApartmentAlarm3Name") {
    const auto value = "{newName}";
    REQUIRE_CALL(apartmentApi, setApartmentAlarm3Name(value));
    dsApi::PatchOperations patchOperations{};
    patchOperations.push_back({.op = dsApi::PatchOperationType::replace,
                               .path = ds::str("/security/alarm3/name"),
                               .value = ds::toJsonFragment(value)});
    CHECK(service->invoke(dsApi::HttpRequest{
              .path = "/",
              .httpArgs = {.method = dsApi::HttpMethod::PATCH, .content = ds::toJsonFragment(patchOperations)}}) ==
          dsApi::noContentResult());
  }

  SECTION("setApartmentAlarm4Name") {
    const auto value = "{newName}";
    REQUIRE_CALL(apartmentApi, setApartmentAlarm4Name(value));
    dsApi::PatchOperations patchOperations{};
    patchOperations.push_back({.op = dsApi::PatchOperationType::replace,
                               .path = ds::str("/security/alarm4/name"),
                               .value = ds::toJsonFragment(value)});
    CHECK(service->invoke(dsApi::HttpRequest{
              .path = "/",
              .httpArgs = {.method = dsApi::HttpMethod::PATCH, .content = ds::toJsonFragment(patchOperations)}}) ==
          dsApi::noContentResult());
  }

  SECTION("pollApartment") {
    REQUIRE_CALL(apartmentApi, pollApartment());
    CHECK(service->invoke(dsApi::HttpRequest{.path = "/poll", .httpArgs = {.method = dsApi::HttpMethod::POST}}) ==
          dsApi::noContentResult());
  }

  SECTION("apartmentStatus") {
    const auto apartmentStatus = [&] { return dsModel::ApartmentStatus{.id = dsModel::ApartmentId{"a"}}; };
    const auto reply = [&] {
      return dsApi::HttpResult{.status = 200, .content = ds::toJsonFragment(dsApi::document(apartmentStatus()))};
    };
    REQUIRE_CALL(apartmentApi, apartmentStatusCached(_)).LR_RETURN(reply());
    const auto result =
        service->invoke(dsApi::HttpRequest{.path = "/status", .httpArgs = {.method = dsApi::HttpMethod::GET}});
    INFO(result);
    CHECK(dsApi::data<dsModel::ApartmentStatus>(result) == apartmentStatus());
  }

  SECTION("setApartmentTemperatureStatusEcoPreferred") {
    const auto args = dsModel::ApartmentArgs{};
    const auto value = true;
    REQUIRE_CALL(apartmentApi, setApartmentTemperatureStatusEcoPreferred(_, _))
        .LR_SIDE_EFFECT(CHECK(_1 == args); CHECK(_2 == value););
    dsApi::PatchOperations patchOperations{};
    patchOperations.push_back({.op = dsApi::PatchOperationType::replace,
                               .path = ds::str("/temperature/ecoPreferred"),
                               .value = ds::toJsonFragment(value)});
    CHECK(service->invoke(dsApi::HttpRequest{
              .path = "/status",
              .httpArgs = {.method = dsApi::HttpMethod::PATCH, .content = ds::toJsonFragment(patchOperations)}}) ==
          dsApi::noContentResult());
  }

  SECTION("pollApartmentStatus") {
    REQUIRE_CALL(apartmentApi, pollApartmentStatus());
    CHECK(service->invoke(dsApi::HttpRequest{
              .path = "/status/poll", .httpArgs = {.method = dsApi::HttpMethod::POST}}) == dsApi::noContentResult());
  }

  SECTION("setUserDefinedState") {
    const auto args = dsModel::UserDefinedStateStatusArgs{.userDefinedStateId = "dev"};
    const auto value = ds::str("active");
    REQUIRE_CALL(apartmentApi, setUserDefinedStateStatus(_, _)).LR_SIDE_EFFECT(CHECK(_1 == args); CHECK(_2 == value););
    dsApi::PatchOperations patchOperations{};
    patchOperations.push_back(
        {.op = dsApi::PatchOperationType::replace, .path = ds::str("/status"), .value = ds::toJsonFragment(value)});
    CHECK(service->invoke(dsApi::HttpRequest{
              .path = ds::str("/userDefinedStates/", args.userDefinedStateId, "/status"),
              .httpArgs = {.method = dsApi::HttpMethod::PATCH, .content = ds::toJsonFragment(patchOperations)}}) ==
          dsApi::noContentResult());
  }

  SECTION("floor") {
    const auto args = dsModel::FloorArgs{.floorId = dsModel::FloorId{"x"}};
    const auto floor = [&] { return dsModel::Floor{.id = args.floorId}; };
    REQUIRE_CALL(apartmentApi, floor(args)).LR_RETURN(floor());
    const auto result = service->invoke(
        dsApi::HttpRequest{.path = ds::str("/floors/", args.floorId), .httpArgs = {.method = dsApi::HttpMethod::GET}});
    INFO(result);
    CHECK(dsApi::data<dsModel::Floor>(result) == floor());
  }

  SECTION("floors") {
    const auto floors = [&] { return dsModel::Floors{}; };
    REQUIRE_CALL(apartmentApi, floors()).LR_RETURN(floors());
    const auto result =
        service->invoke(dsApi::HttpRequest{.path = "/floors", .httpArgs = {.method = dsApi::HttpMethod::GET}});
    INFO(result);
    CHECK(dsApi::data<dsModel::Floors>(result) == floors());
  }

  SECTION("setFloorName") {
    const auto args = dsModel::FloorArgs{.floorId = dsModel::FloorId{"x"}};
    const auto value = "{newName}";
    REQUIRE_CALL(apartmentApi, setFloorName(_, _)).LR_SIDE_EFFECT(CHECK(_1 == args); CHECK(_2 == value););
    dsApi::PatchOperations patchOperations{};
    patchOperations.push_back(
        {.op = dsApi::PatchOperationType::replace, .path = ds::str("/name"), .value = ds::toJsonFragment(value)});
    CHECK(service->invoke(dsApi::HttpRequest{
              .path = ds::str("/floors/", args.floorId),
              .httpArgs = {.method = dsApi::HttpMethod::PATCH, .content = ds::toJsonFragment(patchOperations)}}) ==
          dsApi::noContentResult());
  }

  SECTION("cluster") {
    const auto data = dsModel::Cluster{.id = dsModel::ClusterId{"id0"}};
    REQUIRE_CALL(apartmentApi, cluster(data.id)).LR_RETURN(data);
    checkGet("/clusters/id0", data);
  }

  SECTION("clusters") {
    const auto data = std::vector<dsModel::Cluster>{{.id = dsModel::ClusterId{"id0"}}};
    REQUIRE_CALL(apartmentApi, clusters()).LR_RETURN(data);
    checkGet("/clusters", data);
  }

  SECTION("clusterStatus") {
    const auto data = dsModel::ClusterStatus{.id = dsModel::ClusterId{"id0"}};
    REQUIRE_CALL(apartmentApi, clusterStatus(data.id)).LR_RETURN(data);
    checkGet("/clusters/id0/status", data);
  }

  SECTION("clusterStatuses") {
    const auto data = std::vector<dsModel::ClusterStatus>{{.id = dsModel::ClusterId{"id0"}}};
    REQUIRE_CALL(apartmentApi, clusterStatuses()).LR_RETURN(data);
    checkGet("/clusters/status", data);
  }

  SECTION("set cluster name") {
    const auto args = dsModel::ClusterArgs{.clusterId = dsModel::ClusterId{"{id}"}};
    const auto name = "{newName}";
    REQUIRE_CALL(apartmentApi, setClusterName(_, _)).LR_SIDE_EFFECT(CHECK(_1 == args); CHECK(_2 == name));
    dsApi::PatchOperations patchOperations{};
    patchOperations.push_back(
        {.op = dsApi::PatchOperationType::replace, .path = ds::str("/name"), .value = ds::toJsonFragment(name)});
    CHECK(service->invoke(dsApi::HttpRequest{
              .path = "/clusters/{id}",
              .httpArgs = {.method = dsApi::HttpMethod::PATCH, .content = ds::toJsonFragment(patchOperations)}}) ==
          dsApi::noContentResult());
  }

  SECTION("dsDevice") {
    const auto data = dsModel::DSDevice{.id = dsModel::DSDeviceId{"id0"}};
    REQUIRE_CALL(apartmentApi, device(data.id)).LR_RETURN(data);
    checkGet("/dsDevices/id0", data);

    SECTION("setShadeAttributes") {
      const auto id0 = dsModel::DSDeviceId{"id0"};
      dsModel::ShadeAttributes value{.shadeType = dsModel::ShadeType::windowBlind};
      REQUIRE_CALL(apartmentApi, setShadeAttributes(_, _)).LR_SIDE_EFFECT(CHECK(_1 == id0); CHECK(_2 == value););
      dsApi::PatchOperations patchOperations{};
      patchOperations.push_back({.op = dsApi::PatchOperationType::replace,
                                 .path = ds::str("/shadeAttributes"),
                                 .value = ds::toJsonFragment(value)});
      CHECK(service->invoke(dsApi::HttpRequest{
                .path = ds::str("/dsDevices/", id0),
                .httpArgs = {.method = dsApi::HttpMethod::PATCH, .content = ds::toJsonFragment(patchOperations)}}) ==
            dsApi::noContentResult());
    }
  }

  SECTION("dsDevices") {
    const auto data = std::vector<dsModel::DSDevice>{{.id = dsModel::DSDeviceId{"id0"}}};
    REQUIRE_CALL(apartmentApi, devices()).LR_RETURN(data);
    checkGet("/dsDevices", data);
  }

  SECTION("dsDeviceStatus") {
    const auto data = dsModel::DSDeviceStatus{.id = dsModel::DSDeviceId{"id0"}};
    REQUIRE_CALL(apartmentApi, deviceStatus(data.id)).LR_RETURN(data);
    checkGet("/dsDevices/id0/status", data);
  }

  SECTION("dsDeviceStatuses") {
    const auto data = std::vector<dsModel::DSDeviceStatus>{{.id = dsModel::DSDeviceId{"id0"}}};
    REQUIRE_CALL(apartmentApi, deviceStatuses()).LR_RETURN(data);
    checkGet("/dsDevices/status", data);
  }

  SECTION("submodule") {
    const auto data = dsModel::Submodule{.id = dsModel::SubmoduleId{"id0"}};
    REQUIRE_CALL(apartmentApi, submodule(data.id)).LR_RETURN(data);
    checkGet("/submodules/id0", data);
  }

  SECTION("submodules") {
    const auto data = std::vector<dsModel::Submodule>{{.id = dsModel::SubmoduleId{"id0"}}};
    REQUIRE_CALL(apartmentApi, submodules()).LR_RETURN(data);
    checkGet("/submodules", data);
  }

  SECTION("functionBlock") {
    const auto data = dsModel::FunctionBlock{.id = dsModel::FunctionBlockId{"id0"}};
    REQUIRE_CALL(apartmentApi, functionBlock(data.id)).LR_RETURN(data);
    checkGet("/functionBlocks/id0", data);
  }

  SECTION("functionBlocks") {
    const auto data = std::vector<dsModel::FunctionBlock>{{.id = dsModel::FunctionBlockId{"id0"}}};
    REQUIRE_CALL(apartmentApi, functionBlocks()).LR_RETURN(data);
    checkGet("/functionBlocks", data);
  }

  SECTION("setTemperatureSetpoint") {
    const auto args = dsModel::TemperatureSetpointArgs{.applicationId = dsModel::ApplicationId{"temperature"},
                                                       .temperatureLevel = dsModel::TemperatureLevel::heatingComfort};
    const auto value = 15.0;
    REQUIRE_CALL(apartmentApi, setTemperatureSetpoint(_, _)).LR_SIDE_EFFECT(CHECK(_1 == args); CHECK(_2 == value););
    dsApi::PatchOperations patchOperations{};
    patchOperations.push_back({.op = dsApi::PatchOperationType::replace,
                               .path = "/setpoints/heatingComfort/value",
                               .value = ds::toJsonFragment(value)});
    CHECK(service->invoke(dsApi::HttpRequest{
              .path = "/applications/temperature",
              .httpArgs = {.method = dsApi::HttpMethod::PATCH, .content = ds::toJsonFragment(patchOperations)}}) ==
          dsApi::noContentResult());
  }

  SECTION("installation") {
    // The code for /apartment/installation and /installation is shared.
    // Full coverage is in dsModelInstallationApiService
    SECTION("metering") {
      const auto metering = [&] { return dsModel::Metering{.id = dsModel::MeteringId{"a"}}; };
      const auto args = dsModel::MeteringArgs{.meteringId = dsModel::MeteringId{"dsm-x-power"}};
      ALLOW_CALL(apartmentApi, apiRevision()).LR_RETURN(dsModel::ApiRevision{});
      REQUIRE_CALL(meteringsApi, metering(_)).LR_SIDE_EFFECT(CHECK(_1 == args);).LR_RETURN(metering());
      const auto result = service->invoke(dsApi::HttpRequest{.path = "/installation/meterings/dsm-x-power",
                                                             .httpArgs = {.method = dsApi::HttpMethod::GET}});
      INFO(result);
      CHECK(dsApi::data<dsModel::Metering>(result) == metering());
    }

    SECTION("installation") {
      const auto installation = [&] { return dsModel::Installation{.id = dsModel::InstallationId{"a"}}; };
      REQUIRE_CALL(apartmentApi, installation()).LR_RETURN(installation());
      const auto result =
          service->invoke(dsApi::HttpRequest{.path = "/installation", .httpArgs = {.method = dsApi::HttpMethod::GET}});
      INFO(result);
      CHECK(dsApi::data<dsModel::Installation>(result) == installation());
    }
  }

  SECTION("dsServer") {
    const auto server = [&] { return dsModel::DSServer{.id = dsModel::DSServerId{"a"}}; };
    REQUIRE_CALL(apartmentApi, dsServer()).LR_RETURN(server());
    const auto result =
        service->invoke(dsApi::HttpRequest{.path = "/dsServer", .httpArgs = {.method = dsApi::HttpMethod::GET}});
    INFO(result);
    CHECK(dsApi::data<dsModel::DSServer>(result) == server());
  }

  SECTION("controllers") {
    const auto returnedControllers = dsModel::Controllers{.controllers = {{.id = dsModel::ControllerId{"a"}}}};

    REQUIRE_CALL(apartmentApi, controllers()).LR_RETURN(returnedControllers);
    const auto result =
        service->invoke(dsApi::HttpRequest{.path = "/controllers", .httpArgs = {.method = dsApi::HttpMethod::GET}});
    INFO(result);
    CHECK(dsApi::data<dsModel::Controllers>(result) == returnedControllers);
  }

  SECTION("authorizeController") {
    const auto id = dsModel::ControllerId{"x"};
    const auto args = dsModel::AuthorizeControllerArgs{.authScope = "scope"};
    const auto result = dsModel::AuthorizeControllerResult{.location = "http://example.com/auth"};
    REQUIRE_CALL(apartmentApi, authorizeController(_, _))
        .LR_SIDE_EFFECT(CHECK(_1 == id); CHECK(_2 == args);)
        .LR_RETURN(result);
    CHECK(service->invoke(dsApi::HttpRequest{
              .path = "/controllers/x/authorize",
              .httpArgs = {.method = dsApi::HttpMethod::POST, .content = ds::toJsonFragment(args)}}) ==
          dsApi::okResult(result));
  }

  SECTION("scenarios") {
    const auto returnedScenarios = std::vector<dsModel::Scenario>{
        {dsModel::ApplicationZoneScenario{.id = dsModel::ApplicationZoneScenarioId{"a"}}}};
    REQUIRE_CALL(scenariosApi, scenarios()).LR_RETURN(returnedScenarios);
    const auto result =
        service->invoke(dsApi::HttpRequest{.path = "/scenarios", .httpArgs = {.method = dsApi::HttpMethod::GET}});
    INFO(result);
    CHECK(dsApi::data<dsModel::Scenarios>(result) == dsModel::Scenarios{.scenarios = returnedScenarios});
  }

  SECTION("userDefinedStates") {
    const auto returnedStates = dsModel::UserDefinedStates{.userDefinedStates = {{.id = dsModel::UserDefinedStateId{"a"}}}};

    REQUIRE_CALL(apartmentApi, userDefinedStates()).LR_RETURN(returnedStates);
    const auto result = service->invoke(
        dsApi::HttpRequest{.path = "/userDefinedStates", .httpArgs = {.method = dsApi::HttpMethod::GET}});
    INFO(result);
    CHECK(dsApi::data<dsModel::UserDefinedStates>(result) == returnedStates);
  }

  SECTION("api revision") {
    const auto revision = [&] { return dsModel::ApiRevision{.id = dsModel::ApiRevisionId{"apiRevisionId"}}; };
    REQUIRE_CALL(apartmentApi, apiRevision()).LR_RETURN(revision());
    const auto result =
        service->invoke(dsApi::HttpRequest{.path = "/apiRevision", .httpArgs = {.method = dsApi::HttpMethod::GET}});
    INFO(result);
    CHECK(dsApi::data<dsModel::ApiRevision>(result) == revision());
  }

  SECTION("metering") {
    const auto metering = [&] { return dsModel::Metering{.id = dsModel::MeteringId{"a"}}; };
    const auto args = dsModel::MeteringArgs{.meteringId = dsModel::MeteringId{"dsm-x-power"}};
    REQUIRE_CALL(meteringsApi, metering(_)).LR_SIDE_EFFECT(CHECK(_1 == args);).LR_RETURN(metering());
    const auto result = service->invoke(
        dsApi::HttpRequest{.path = "/meterings/dsm-x-power", .httpArgs = {.method = dsApi::HttpMethod::GET}});
    INFO(result);
    CHECK(dsApi::data<dsModel::Metering>(result) == metering());
  }

  SECTION("meterings") {
    const auto meterings = dsModel::Meterings{.meterings = {{.id = dsModel::MeteringId{"a"}}}};
    REQUIRE_CALL(meteringsApi, meterings()).LR_RETURN(meterings);
    const auto result =
        service->invoke(dsApi::HttpRequest{.path = "/meterings", .httpArgs = {.method = dsApi::HttpMethod::GET}});
    INFO(result);
    CHECK(dsApi::data<dsModel::Meterings>(result) == meterings);
  }

  SECTION("meteringValue") {
    const auto meteringValue = [&] { return dsModel::MeteringValue{.id = dsModel::MeteringId{"a"}}; };
    const auto args = dsModel::MeteringArgs{.meteringId = dsModel::MeteringId{"dsm-x-power"}};
    REQUIRE_CALL(meteringsApi, meteringValue(_)).LR_SIDE_EFFECT(CHECK(_1 == args);).LR_RETURN(meteringValue());
    const auto result = service->invoke(
        dsApi::HttpRequest{.path = "/meterings/values/dsm-x-power", .httpArgs = {.method = dsApi::HttpMethod::GET}});
    INFO(result);
    CHECK(dsApi::data<dsModel::MeteringValue>(result) == meteringValue());
  }

  SECTION("meteringValues") {
    const auto meteringValues = [&] { return dsModel::MeteringValues{.values = {{.id = dsModel::MeteringId{"a"}}}}; };
    REQUIRE_CALL(meteringsApi, meteringValues()).LR_RETURN(meteringValues());
    const auto result = service->invoke(
        dsApi::HttpRequest{.path = "/meterings/values", .httpArgs = {.method = dsApi::HttpMethod::GET}});
    INFO(result);
    CHECK(dsApi::data<dsModel::MeteringValues>(result) == meteringValues());
  }
}

SCENARIO("dsModelInstallationApiService", TAGS) {
  dsModel::ApartmentApiMock apartmentApi;
  dsModel::MeteringsApiMock meteringsApi{};

  ALLOW_CALL(apartmentApi, meterings()).LR_RETURN(meteringsApi);
  auto service = dsModel::makeInstallationService(apartmentApi);

  SECTION("metering") {
    const auto metering = [&] { return dsModel::Metering{.id = dsModel::MeteringId{"a"}}; };
    const auto args = dsModel::MeteringArgs{.meteringId = dsModel::MeteringId{"dsm-x-power"}};
    REQUIRE_CALL(apartmentApi, apiRevision()).LR_RETURN(dsModel::ApiRevision{});
    REQUIRE_CALL(meteringsApi, metering(_)).LR_SIDE_EFFECT(CHECK(_1 == args);).LR_RETURN(metering());
    const auto result = service->invoke(
        dsApi::HttpRequest{.path = "/meterings/dsm-x-power", .httpArgs = {.method = dsApi::HttpMethod::GET}});
    INFO(result);
    CHECK(dsApi::data<dsModel::Metering>(result) == metering());
  }

  SECTION("meterings") {
    const dsModel::Metering meteringA = {.id = dsModel::MeteringId{"a"}};
    const dsModel::Metering meteringB = {.id = dsModel::MeteringId{"b"}};
    const auto meterings = [&] { return dsModel::Meterings{.meterings = {meteringA, meteringB}}; };
    REQUIRE_CALL(apartmentApi, apiRevision()).LR_RETURN(dsModel::ApiRevision{});
    REQUIRE_CALL(meteringsApi, meterings()).LR_RETURN(meterings());
    const auto result =
        service->invoke(dsApi::HttpRequest{.path = "/meterings", .httpArgs = {.method = dsApi::HttpMethod::GET}});
    INFO(result);
    CHECK(dsApi::data<dsModel::Meterings>(result) == meterings());
  }

  SECTION("meteringValue") {
    const auto meteringValue = [&] { return dsModel::MeteringValue{.id = dsModel::MeteringId{"a"}}; };
    const auto args = dsModel::MeteringArgs{.meteringId = dsModel::MeteringId{"dsm-x-power"}};
    REQUIRE_CALL(apartmentApi, apiRevision()).LR_RETURN(dsModel::ApiRevision{});
    REQUIRE_CALL(meteringsApi, meteringValue(_)).LR_SIDE_EFFECT(CHECK(_1 == args);).LR_RETURN(meteringValue());
    const auto result = service->invoke(
        dsApi::HttpRequest{.path = "/meterings/values/dsm-x-power", .httpArgs = {.method = dsApi::HttpMethod::GET}});
    INFO(result);
    CHECK(dsApi::data<dsModel::MeteringValue>(result) == meteringValue());
  }

  SECTION("meteringValues") {
    const auto meteringValues = [&] { return dsModel::MeteringValues{.values = {{.id = dsModel::MeteringId{"a"}}}}; };
    REQUIRE_CALL(apartmentApi, apiRevision()).LR_RETURN(dsModel::ApiRevision{});
    REQUIRE_CALL(meteringsApi, meteringValues()).LR_RETURN(meteringValues());
    const auto result = service->invoke(
        dsApi::HttpRequest{.path = "/meterings/values", .httpArgs = {.method = dsApi::HttpMethod::GET}});
    INFO(result);
    CHECK(dsApi::data<dsModel::MeteringValues>(result) == meteringValues());
  }

  SECTION("deprecated meterings") {
    auto checkDeprecatedError = [&](const dsApi::HttpResult &x) {
      INFO(x);
      CHECK(x.status == 500);
      CHECK(boost::starts_with(dsApi::error(x).detail, "Deprecated"));
    };

    ALLOW_CALL(apartmentApi, apiRevision()).LR_RETURN(dsModel::ApiRevision{.attributes = {.removals = 1}});
    checkDeprecatedError(service->invoke(
        dsApi::HttpRequest{.path = "/meterings/dsm-x-power", .httpArgs = {.method = dsApi::HttpMethod::GET}}));
    checkDeprecatedError(
        service->invoke(dsApi::HttpRequest{.path = "/meterings", .httpArgs = {.method = dsApi::HttpMethod::GET}}));
    checkDeprecatedError(service->invoke(
        dsApi::HttpRequest{.path = "/meterings/values/dsm-x-power", .httpArgs = {.method = dsApi::HttpMethod::GET}}));
    checkDeprecatedError(service->invoke(
        dsApi::HttpRequest{.path = "/meterings/values", .httpArgs = {.method = dsApi::HttpMethod::GET}}));
  }

  SECTION("updateParameterCache") {
    const auto args = dsModel::UpdateDeviceParameterCacheArgs{.mode = dsModel::ParameterCacheUpdateMode::slow};
    REQUIRE_CALL(apartmentApi, updateParameterCache(_)).LR_SIDE_EFFECT(CHECK(_1 == args););
    CHECK(service->invoke(dsApi::HttpRequest{
              .path = "/parameterCache/update",
              .httpArgs = {.method = dsApi::HttpMethod::POST, .content = ds::toJsonFragment(args)}}) ==
          dsApi::noContentResult());
  }

  SECTION("parameterCacheStatus") {
    const auto parameterCacheStatus = [&] { return dsModel::ParameterCache{.id = dsModel::ParameterCacheId{"a"}}; };
    REQUIRE_CALL(apartmentApi, parameterCacheStatus()).LR_RETURN(parameterCacheStatus());
    const auto result =
        service->invoke(dsApi::HttpRequest{.path = "/parameterCache", .httpArgs = {.method = dsApi::HttpMethod::GET}});
    INFO(result);
    CHECK(dsApi::data<dsModel::ParameterCache>(result) == parameterCacheStatus());
  }

  SECTION("installation") {
    const auto installation = [&] { return dsModel::Installation{.id = dsModel::InstallationId{"a"}}; };
    REQUIRE_CALL(apartmentApi, installation()).LR_RETURN(installation());
    const auto result =
        service->invoke(dsApi::HttpRequest{.path = "/", .httpArgs = {.method = dsApi::HttpMethod::GET}});
    INFO(result);
    CHECK(dsApi::data<dsModel::Installation>(result) == installation());
  }
}

SCENARIO("dsModelDssApiService", TAGS) {
  dsModel::ApartmentApiMock apartmentApi;
  auto service = dsModel::makeDssService(apartmentApi);

  SECTION("storeAccessToken") {
    const auto args = dsModel::StoreControllerAccessTokenArgs{.authData = "data", .authScope = "scope", .params = {}};
    REQUIRE_CALL(apartmentApi, storeControllerAccessToken(_,_)).LR_SIDE_EFFECT(CHECK(_1 == dsModel::ControllerId{"1234"}); CHECK(_2 == args););
    CHECK(service->invoke(dsApi::HttpRequest{
              .path = "/controllers/1234/storeAccessToken",
              .httpArgs = {.method = dsApi::HttpMethod::POST, .content = ds::toJsonFragment(args)}}) ==
          dsApi::noContentResult());
  }
}

} // namespace
