/*
    Copyright (c) 2018 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/
#include "ds-signal.h"

#include <ds/asio/test.h>
#include <ds/catch/catch.h>

#include "ds-signal-mock.h"

using namespace std::literals;

static const char* TAGS = "[dsSignal][ds]";

using namespace trompeloeil;

namespace {
DS_SLOT_MOCK1(IntSlotMock, int);

TEST_CASE("dsSignal", TAGS) {
  auto loop = ds::asio::TestLoopResource{};

  SECTION("Emit without slots does nothing.") {
    {
      ds::Signal<> signal;
      signal.emit();
      CHECK(true);
    }
    {
      ds::Signal<int> signal;
      signal.emit(5);
      CHECK(true);
    }
  }

  SECTION("Linked slot is emitted.") {
    {
      int x{};
      ds::Signal<> signal;
      auto slot = ds::Slot<>(loop, signal, [&] {
        ++x;
        loop.stop();
      });
      signal.emit();
      loop.run();
      CHECK(x == 1);
    }
    {
      int sum{};
      ds::Signal<int> signal;
      auto slot = ds::Slot<int>(loop, signal, [&](auto&& x) {
        sum += x;
        loop.stop();
      });
      signal.emit(5);
      loop.run();
      CHECK(sum == 5);
    }
  }

  SECTION("Emit works with lvalue.") {
    ds::Signal<int> signal;
    auto x = 5;
    signal.emit(x);
    SUCCEED();
  }

  SECTION("Linked mocked slot is emitted .") {
    {
      ds::Signal<> signal;
      auto slot = ds::VoidSlotMock(loop, signal);
      REQUIRE_CALL(slot, call()).LR_SIDE_EFFECT(loop.stop());
      signal.emit();
      loop.run();
    }
    {
      ds::Signal<int> signal;
      auto slot = IntSlotMock(loop, signal);
      REQUIRE_CALL(slot, call(5)).LR_SIDE_EFFECT(loop.stop());
      signal.emit(5);
      loop.run();
    }
  }

  SECTION("2 linked mocked slots are emitted .") {
    {
      ds::Signal<> signal;
      auto slot1 = ds::VoidSlotMock(loop, signal);
      auto slot2 = ds::VoidSlotMock(loop, signal);
      REQUIRE_CALL(slot1, call()).LR_SIDE_EFFECT(loop.stop());
      REQUIRE_CALL(slot2, call()).LR_SIDE_EFFECT(loop.stop());
      signal.emit();
      loop.run();
      loop.run();
    }
    {
      ds::Signal<int> signal;
      auto slot1 = IntSlotMock(loop, signal);
      auto slot2 = IntSlotMock(loop, signal);
      REQUIRE_CALL(slot1, call(5)).LR_SIDE_EFFECT(loop.stop());
      ;
      REQUIRE_CALL(slot2, call(5)).LR_SIDE_EFFECT(loop.stop());
      ;
      signal.emit(5);
      loop.run();
      loop.run();
    }
  }

  SECTION("Deleted slot is unlinked.") {
    ds::Signal<> signal;
    {
      auto slot = ds::Slot<>(loop, signal, [&] { CHECK(false); });
    }
    signal.emit();
    loop.postPostStopRun();
  }

  SECTION("Deleted slot does not invoke the queued callback.") {
    ds::Signal<> signal;
    {
      auto slot = ds::Slot<>(loop, signal, [&] { CHECK(false); });
      signal.emit();
    }
    loop.postPostStopRun();
  }

  SECTION("Linked slots are emitted in link order.") {
    int x = {};
    ds::Signal<> signal;
    auto slot1 = ds::Slot<>(loop, signal, [&] {
      CHECK(x == 0);
      ++x;
      loop.stop();
    });
    auto slot2 = std::make_unique<ds::Slot<>>(loop, signal, [&] { ++x; });
    auto slot3 = ds::Slot<>(loop, signal, [&] {
      CHECK(x == 1);
      ++x;
      loop.stop();
    });
    slot2.reset();
    signal.emit();
    loop.run();
    loop.run();
    CHECK(x == 2);
  }
}

} // namespace
