/*
    Copyright (c) 2020 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/
#include "box.h"

#include <ds/catch/catch.h>

namespace {

using namespace std::literals;
static const char* TAGS = "[dsBox][ds]";

struct X {
  int i;
};

TEST_CASE("dsBox", TAGS) {
  SECTION("&& constructor") {
    SECTION("std::string") {
      const auto x = ds::Box<std::string>{"a"};
      CHECK(*x == "a");
    }

    SECTION("std::vector<int>") {
      const auto x = ds::Box<std::vector<int>>{{1, 2, 3}};
      CHECK(*x == std::vector{1, 2, 3});
    }

    SECTION("aggregate X") {
      const auto x = ds::Box<X>{{.i = 7}};
      CHECK(x->i == 7);
    }
  }

  SECTION("class template argument deduction") {
    static_assert(std::is_same_v<decltype(ds::Box{"a"s}), ds::Box<std::string>>);
  }

  SECTION("T&& assignment") {
    auto x = ds::Box{"a"s};
    CHECK(*x == "a");
    x = "b"s;
    CHECK(*x == "b");
    x = "c";
    CHECK(*x == "c");
  }

  SECTION("copy/move constructor") {
    auto x = ds::Box{"a"s};
    const auto xCopied = x;
    auto xMoved = std::move(x);
    CHECK(*xCopied == *xMoved);
  }

  SECTION("copy/move assigment") {
    auto x = ds::Box("a"s);
    auto xCopied = ds::Box("b"s);
    auto xMoved = ds::Box("c"s);
    xCopied = x;
    xMoved = std::move(x);
    CHECK(*xCopied == *xMoved);
  }

  SECTION("move only type can be copied") {
    const auto x = ds::Box(std::make_unique<int>(4));
    const auto x2 = x;
    CHECK(**x == 4);
    CHECK(**x2 == 4);
    CHECK(*x == *x2);
  }

  SECTION("accessors") {
    const auto x = ds::Box("a"s);
    CHECK(*x == "a");
    CHECK(*x.get() == "a");
    CHECK(x->size() == 1);
  }

  SECTION("swap") {
    auto a = ds::Box("a"s);
    auto b = ds::Box("b"s);
    swap(a, b);
    CHECK(*a == "b");
  }

  SECTION("implicit const & cast") {
    auto a = ds::Box("a"s);
    const std::string& b = a;
    CHECK(*a == b);
  }

  SECTION("operator ==, !=") {
    const auto aBox1 = ds::Box("a"s);
    const auto aBox2 = ds::Box("a"s);
    const auto bBox = ds::Box("b"s);

    SECTION("compare boxes") {
      CHECK(aBox1 == aBox2);
      CHECK(aBox1 != bBox);
    }

    SECTION("compare with T") {
      const auto a = "a"s;
      const auto b = "b"s;
      CHECK(aBox1 == a);
      CHECK(aBox1 != b);
      CHECK(a == aBox2);
      CHECK(a != bBox);
    }

    SECTION("compare with subclass of Box<T>") {
      struct SubBox : ds::Box<std::string> {
        using ds::Box<std::string>::Box;
      };
      const auto a = SubBox{"a"s};
      const auto b = SubBox{"b"};
      CHECK(aBox1 == a);
      CHECK(aBox1 != b);
      CHECK(a == aBox2);
      CHECK(a != bBox);
    }

    SECTION("compare with comparable with T") {
      const auto a = "a";
      const auto b = "b";
      CHECK(aBox1 == a);
      CHECK(aBox1 != b);
      CHECK(a == aBox2);
      CHECK(a != bBox);
    }
  }

  SECTION("unbox") {
    const auto box = ds::Box("a"s);
    auto x = box.unbox();
    x.push_back('b');
    const auto box2 = ds::Box{std::move(x)};
    CHECK(box2 == "ab");
  }
}

} // namespace
