/*
    Copyright (c) 2018 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/
#include "clock.h"
#include <ds/log.h>
#include <boost/asio/wait_traits.hpp>
#include <mutex>
#include "clock-mock.h"

static ds::LogChannel dsLogChannel("dsClock");

namespace ds {

// No locking.
// ClockMocker shall be activated before the clock is used.
// ClockMocker shall be deactivate after clock is used.

static bool s_isMocked;
static Clock::duration s_mockedNow;
static int s_mockedToWaitDurationCount;

Clock::time_point Clock::now() noexcept {
  auto result = [&] {
    if (s_isMocked) {
      return time_point{s_mockedNow};
    } else {
      return time_point{std::chrono::steady_clock::now().time_since_epoch()};
    }
  }();
  DS_DEBUG_LEAVE(result.time_since_epoch());
  return result;
}

Clock::duration ClockWaitTraits::to_wait_duration(const Clock::duration& d) {
  if (s_isMocked) {
    // Detect infinite loops in test and break them.
    if (++s_mockedToWaitDurationCount >= 1000) {
      // We cannot fail with an exception here,
      // because calling code (timer?, asio?) is not exception safe.
      // Throwing leads to dangling and undefined behavior.
      // As this code here is only to support unit tests,
      // and it means error, we assert.
      // It saves investigations of further errors.
      DS_FAIL_ASSERT("Mocked clock is not advancing.", s_mockedToWaitDurationCount);
    }
    return {};
  } else {
    return boost::asio::wait_traits<Clock>::to_wait_duration(d);
  }
}

ClockMocker::ClockMocker() {
  DS_INFO_ENTER();
  DS_REQUIRE(!s_isMocked, "Only one ClockMocker instance can be active at once.");
  s_isMocked = true;
  s_mockedNow = {};
}

ClockMocker::~ClockMocker() {
  DS_INFO_ENTER();
  s_isMocked = false;
}

void ClockMocker::advance(Clock::duration x) {
  DS_REQUIRE(x.count() >= 0, "Failed to advance clock backwards.", x.count());
  DS_INFO_ENTER(x);
  s_mockedNow = s_mockedNow + x;
  s_mockedToWaitDurationCount = {};
}

} // namespace ds
