/*
    Copyright (c) 2016 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/
#include "date-time.h"
#include <ds/catch/catch.h>
#include <ds/log.h>

static const char* TAGS = "[ds][dsDateTime]";

namespace {
constexpr auto dateTime = ds::DateTime(std::chrono::milliseconds(1505731033777));

TEST_CASE("dsToLocalMsIsoString", TAGS) {
  auto string = ds::toLocalMsIsoString(dateTime);
  // 2017-09-18T12:37:13.777+02:00 but time zone may differ
  CHECK(string.size() == 29);

  tm tm_ = {};
  int ms;
  char tzSign;
  int tzH, tzM;
  CHECK(::sscanf(string.c_str(), "%d-%d-%dT%d:%d:%d.%d%c%d:%d", &tm_.tm_year, &tm_.tm_mon, &tm_.tm_mday, &tm_.tm_hour,
                 &tm_.tm_min, &tm_.tm_sec, &ms, &tzSign, &tzH, &tzM) == 10);
  CHECK((tzSign == '+' || tzSign == '-'));
  tm_.tm_gmtoff = (tzH * 60 * 60 + tzM * 60) * (tzSign == '-' ? -1 : 1);
  CHECK(tm_.tm_year == 2017);
  CHECK(tm_.tm_mon == 9);

  // convert the rest (days, hours, etc) to duration and compare
  auto duration1 = std::chrono::hours(tm_.tm_mday) * 24 + std::chrono::hours(tm_.tm_hour) +
                   std::chrono::minutes(tm_.tm_min) + std::chrono::seconds(tm_.tm_sec) + std::chrono::milliseconds(ms) -
                   std::chrono::seconds(tm_.tm_gmtoff);
  auto duration2 = std::chrono::hours(18) * 24 + std::chrono::hours(12) + std::chrono::minutes(37) +
                   std::chrono::seconds(13) + std::chrono::milliseconds(777) - std::chrono::seconds(2 * 3600);
  CHECK(duration1.count() == duration2.count());
}

TEST_CASE("dsToString(ds::DateTime)", TAGS) {
  CHECK(ds::toString(dateTime) == "2017-09-18T10:37:13.777Z");
}

TEST_CASE("dsToDateTime(std::string)", TAGS) {
  CHECK(ds::toDateTime("2017-09-18T10:37:13.777Z") == dateTime);
  SECTION("works without milliseconds") {
    CHECK(ds::toString(ds::toDateTime("2017-09-18T10:37:13Z")) == "2017-09-18T10:37:13.000Z");
  }
  SECTION("works with various precision for milliseconds") {
    CHECK(ds::toString(ds::toDateTime("2017-09-18T10:37:13.Z")) == "2017-09-18T10:37:13.000Z");
    CHECK(ds::toString(ds::toDateTime("2017-09-18T10:37:13.1Z")) == "2017-09-18T10:37:13.100Z");
    CHECK(ds::toString(ds::toDateTime("2017-09-18T10:37:13.12Z")) == "2017-09-18T10:37:13.120Z");
    CHECK(ds::toString(ds::toDateTime("2017-09-18T10:37:13.123Z")) == "2017-09-18T10:37:13.123Z");
  }
  SECTION("throws without Z suffix") {
    CHECK_THROWS(ds::toDateTime("2017-09-18T10:37:13.111"));
    CHECK_THROWS(ds::toDateTime("2017-09-18T10:37:13.11"));
    CHECK_THROWS(ds::toDateTime("2017-09-18T10:37:13.1"));
    CHECK_THROWS(ds::toDateTime("2017-09-18T10:37:13."));
    CHECK_THROWS(ds::toDateTime("2017-09-18T10:37:13"));
  }
  SECTION("throws without time up to seconds") {
    CHECK_THROWS(ds::toDateTime("2017-09-18T10:37Z"));
    CHECK_THROWS(ds::toDateTime("2017-09-18T10:37"));
    CHECK_THROWS(ds::toDateTime("2017-09-18Z"));
    CHECK_THROWS(ds::toDateTime("2017-09-18"));
  }
  SECTION("throws without T in middle") { CHECK_THROWS(ds::toDateTime("2017-09-18 10:37:13.111Z")); }
}

TEST_CASE("dsSystemTime works with ds::str", TAGS) {
  CHECK(ds::str(dateTime) == "2017-09-18T10:37:13.777Z");
}

TEST_CASE("dsMax(ds::DateTime,ds::DateTime)", TAGS) {
  CHECK(ds::max(ds::toDateTime("2017-09-18T10:37:13.123Z"), ds::toDateTime("2017-09-18T10:37:13.123Z")) ==
        ds::toDateTime("2017-09-18T10:37:13.123Z"));
  CHECK(ds::max(ds::toDateTime("2017-09-18T10:37:13.120Z"), ds::toDateTime("2017-09-18T10:37:13.121Z")) ==
        ds::toDateTime("2017-09-18T10:37:13.121Z"));
  CHECK(ds::max(ds::toDateTime("2017-09-18T10:37:13.124Z"), ds::toDateTime("2017-09-18T10:37:13.126Z")) ==
        ds::toDateTime("2017-09-18T10:37:13.126Z"));
}

} // namespace
