#include "json.h"

#include <mutex>

using namespace std::literals;

namespace ds {

JsonFragment makeJsonFragment(json_object* x) {
  return JsonFragment(x, json_object_put);
}

std::string toJson(const JsonFragment& x) {
  // Thread safety on const object is typical requirement in c++ world.
  // The json_object_to_json_string allocates the returned
  // string inside the json_object and returns unowned `const char *`.
  // That makes it thread unsafe.
  //
  // Lock to work around.
  static std::mutex s_mutex;
  auto lock = std::scoped_lock{s_mutex};
  return json_object_to_json_string_ext(x.get(), 0);
}

namespace detail {
void Str<JsonFragment>::f(std::ostream& ostream, const JsonFragment& x) {
  ostream << toJson(x);
}
} // namespace detail

JsonFragment parseJsonFragment(gsl::czstring text) {
  json_tokener* tok = json_tokener_new();
  DS_REQUIRE(tok);
  DS_DEFER(json_tokener_free(tok));

  // json-c is not able to parse bool and numbers
  // with explicit len instead of -1
  if (text == "null"sv) {
    return {};
  }
  json_object* object = json_tokener_parse_ex(tok, text, -1);
  DS_REQUIRE(object, "Failed to parse json.", ds::hideArg(json_tokener_error_desc(json_tokener_get_error(tok))));
  return makeJsonFragment(object);
}

ds::PrimitiveJsonValue FromJson<ds::PrimitiveJsonValue>::call(json_object* json) {
  const auto& type = json_object_get_type(json);
  switch (type) {
    case json_type_null:
      return {};
    case json_type_string:
      return static_cast<std::string>(json_object_get_string(json));
    case json_type_double:
    case json_type_int:
      return json_object_get_double(json);
    case json_type_boolean:
      return static_cast<bool>(json_object_get_boolean(json));
    case json_type_object:
    case json_type_array:
    default:
      DS_FAIL_REQUIRE("Expecting json primitive value.", type);
  }
}

JsonFragment ToJsonFragment<ds::PrimitiveJsonValue>::call(const ds::PrimitiveJsonValue& x) {
  return std::visit([](auto&& y) { return ToJsonFragment<std::decay_t<decltype(y)>>::call(y); }, x);
}

} // namespace ds

std::ostream& operator<<(std::ostream& ostream, const ds::JsonFragment& x) {
  return ostream << ds::toJson(x);
}
bool operator==(const ds::JsonFragment& a, const ds::JsonFragment& b) {
  return json_object_equal(a.get(), b.get());
}
bool operator!=(const ds::JsonFragment& a, const ds::JsonFragment& b) {
  return !(a == b);
}
