#include "system.h"

#include <sys/time.h>

#include <boost/algorithm/string.hpp>
#include <boost/filesystem.hpp>
#include <boost/uuid/random_generator.hpp>
#include <boost/uuid/uuid_io.hpp>
#include <cerrno>
#include <fstream>
#include <string>

#include "conf.h"
#include "log.h"
#include "memory.h"
#include "random.h"

static ds::LogChannel dsLogChannel("dsSystem");
using namespace std::literals;

namespace ds {

System::~System() {}

class SystemImpl final : public System {
public:
  SystemImpl() : m_dirs(Dirs{DS_SYSCONFDIR, DS_DATADIR}) {}
  SystemImpl(Dirs&& dirs) : m_dirs(std::move(dirs)) {}

  const Dirs& dirs() const override { return m_dirs; }

  bool existsFile(const std::string& filename) const override { return boost::filesystem::exists(filename); }

  std::string readFileImpl(const std::string& filename) const {
    try {
      std::ifstream in(filename, std::ios::in | std::ios::binary);
      in.exceptions(std::ifstream::failbit | std::ifstream::badbit);
      DS_REQUIRE(in, "");
      in.seekg(0, std::ios::end);
      std::string result;
      result.resize(in.tellg());
      in.seekg(0, std::ios::beg);
      in.read(&result[0], result.size());
      in.close();
      DS_DEBUG_LEAVE(filename, result);
      return (result);
    } catch (std::exception &e) {
      DS_FAIL_REQUIRE("Failed to read file", filename, e);
    }
  }

  void writeToFileImpl(const std::string& filename, const std::string& data) const {
    try {
      std::ofstream ofs{filename, std::ofstream::out | std::ofstream::trunc};
      DS_REQUIRE(ofs, "");
      ofs << data << std::endl;
      ofs.close();
    } catch (std::exception &e) {
      DS_FAIL_REQUIRE("Failed to write to file", filename, e);
    }
  }

  std::string readFile(const std::string& filename) const override {
    DS_DEBUG_ENTER(filename);
    return readFileImpl(filename);
  }

  std::optional<std::string> readFileIfExists(const std::string& filename) const override {
    DS_DEBUG_ENTER(filename);
    if (boost::filesystem::exists(filename)) {
      return readFileImpl(filename);
    }
    return {};
  }

  void writeToFile(const std::string& filename, const std::string& data) const override {
    DS_DEBUG_ENTER(filename, data);
    writeToFileImpl(filename, data);
  }

  std::vector<std::string> lsRegularFiles(const std::string& path, const std::string& extension) const override {
    DS_DEBUG_ENTER(path, extension);
    try {
      if (!boost::filesystem::exists(path)) {
        return {};
      }
      std::vector<std::string> files;
      boost::filesystem::directory_iterator endItr;
      for (boost::filesystem::directory_iterator itr(path); itr != endItr; ++itr) {
        if (boost::filesystem::is_regular_file(itr->path()) && (itr->path().extension() == extension)) {
          files.push_back(itr->path().string());
        }
      }
      return files;
    } catch (boost::filesystem::filesystem_error& e) {
      DS_FAIL_REQUIRE("Failed to iterate files", path, e.what());
    }
  }

  void rename(const std::string& old_p, const std::string& new_p) const override {
    DS_DEBUG_ENTER(old_p, new_p);
    boost::filesystem::rename(old_p, new_p);
  }

  void remove(const std::string& path) const override {
    DS_DEBUG_ENTER(path);
    boost::filesystem::remove(path);
  }

  ds::DateTime now() const override { return std::chrono::system_clock::now(); }

  ds::Conf& conf() const override { return *m_conf; }

  Release release() const override { return m_release; }

  bool isInternalRelease() const override {
    DS_DEBUG_ENTER(release());
    switch (release()) {
      case Release::fieldtest: [[fallthrough]];
      case Release::production: return false;
      case Release::devel: [[fallthrough]];
      case Release::testing: [[fallthrough]];
      default: return true;
    }
  };

  std::string randomId() const override { return ds::randomId(); }
  std::string randomUuidString() const override {
    thread_local boost::uuids::random_generator generator{};
    return to_string(generator());
  }

  std::string timeZone() const override {
    /// No portable C / C++ function exists.
    /// Reading /etc/localtime symlink is the status quo.
    /// Once we have systemd, we can consider timedate over dbus
    boost::system::error_code ec;
    const auto path = boost::filesystem::read_symlink("/etc/localtime", ec).native();
    for (auto prefix : {"/usr/share/zoneinfo/"sv, "../usr/share/zoneinfo/"sv}) {
      if (boost::starts_with(path, prefix)) {
        return path.substr(prefix.size());
      }
    }
    return {};
  }

  std::chrono::seconds timeZoneOffset() const override {
    struct timeval timeval;
    if (gettimeofday(&timeval, NULL)) {
      throw std::runtime_error(strerror(errno));
    }
    struct tm tm;
    localtime_r(&timeval.tv_sec, &tm);
    return std::chrono::seconds{tm.tm_gmtoff};
  }

private:
  const Dirs m_dirs;
  const std::unique_ptr<ds::Conf> m_conf{ds::makeConf(*this)};
  const Release m_release{ds::toEnum<Release>(
      boost::trim_copy(readFileIfExists(ds::str(dirs().sysconfdir, "/release-type")).value_or(std::string{})))};
};

System& system() {
  static SystemImpl x;
  return x;
}

std::unique_ptr<System> makeSystem() {
  return std::make_unique<SystemImpl>();
}

std::unique_ptr<System> makeSystem(System::Dirs&& dirs) {
  return std::make_unique<SystemImpl>(std::move(dirs));
}

} // namespace ds
