/*
    Copyright (c) 2019 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/
#pragma once

#include <ds/clock.h>
#include <ds/reflect.h>
#include <memory>
#include <vector>

namespace ds {

/// Trace is basic record for latency tracing.
/// Weakly follows span in [opentracing specification](https://opentracing.io/specification/).
///
/// Trace may collect children traces to track child-of dependencies.
///
/// Measurements are done relatively to parent trace.
struct Trace {
  std::string name{};                 /// Local name of traced code without arguments, relative to parent trace.
  ds::Clock::duration parentOffset{}; /// Offset in parent trace
  ds::Clock::duration duration{};     /// Measured trace duration

  // Child-of relation - parent waits for outcome of the child.
  /// - children are inside parent.
  /// - children not neccessary fill parent, can overlap (e.g. parallel workers).
  std::vector<Trace> children{};
};

/// Tracer measuring Trace from constructor to desctructor.
class Tracer {
public:
  /// Tracer adding the measured trace to the `parent` Tracer.
  Tracer(std::string &&name, Tracer &parent);

  /// Tracer adding the measured trace to `maybeParent` Tracer.
  /// Discards the measurement if parent == nullptr.
  Tracer(std::string &&name, Tracer *maybeParent);

  Tracer(const Tracer &) = delete;

  /// Tracer passing the measured trace.
  ///
  /// Current measure is received by consume()
  /// If `onFinished` is passed, measured trace is passed to it in destructor.
  Tracer(std::string &&name, ds::Clock::duration parentOffset = {}, std::function<void(Trace &&)> onFinished = {});

  ~Tracer();

  /// Current duration measured from construction
  ds::Clock::duration duration() const;

  /// Consume the accumulated trace.
  /// Throws if already consumed or created with parent or onFinished
  Trace consume();

  /// Add child record
  void addChild(Trace &&);

private:
  Trace m_trace{};
  ds::Clock::time_point m_startedAt = ds::Clock::now();
  std::function<void(Trace &&)> m_onFinished;
};

/// Tracer pushed into thread local stack.
///
/// Very convenient for normal blocking functions.
///
/// Must _not_ be allocated on heap.
///
class ThreadLocalTracer : public Tracer {
public:
  /// ThreadLocalTracer adding the measured trace to thread local tracer.
  /// Does nothing if thread local tracer does not exist.
  ThreadLocalTracer(std::string &&name);

  /// ThreadLocalTracer adding the measured trace to `parent` Tracer.
  ThreadLocalTracer(std::string &&name, Tracer &parent);

  /// ThreadLocalTracer adding the measured trace to `maybeParent` Tracer.
  /// Discards the measurement if parent == nullptr.
  ThreadLocalTracer(std::string &&name, Tracer *parent);

  /// ThreadLocalTracer passing the measured trace to `result` instance.
  ThreadLocalTracer(std::string &&name, Trace &result);

  /// ThreadLocalTracer adding the measured trace to `onFinished` function.
  ThreadLocalTracer(std::string &&name, ds::Clock::duration parentOffset, std::function<void(Trace &&)> onFinished);

  ~ThreadLocalTracer();

private:
  ThreadLocalTracer *m_next{};

  void push();
};

} // namespace ds

DS_REFLECT_STRUCT((ds), Trace, (name)(parentOffset)(duration)(children))
