/*
    Copyright (c) 2009,2011,2022 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

*/

#pragma once

#include <ds/reflect.h>
#include <bitset>

#include "datetools.h"
#include "ds485types.h"
#include "ds485/specification-fwd.h"
#include "model/modelconst.h"

namespace dss {

struct DSMeterSpec {
  dsuid_t DSID;
  BusMemberDevice_t DeviceType;
  uint32_t SoftwareRevisionARM;
  uint32_t SoftwareRevisionDSP;
  uint32_t HardwareVersion;
  uint16_t APIVersion;
  std::string Name;
  std::bitset<8> flags;
  uint8_t ApartmentState;
};

struct GroupSpec {
  uint8_t GroupID;
  ApplicationType applicationType;
  uint16_t NumberOfDevices;
  std::string Name;
  uint32_t applicationConfiguration;
};

struct ClusterSpec : public GroupSpec {
  bool canHaveStateMachine;
  uint16_t protectionClass;
  uint16_t location;
  uint16_t floor;
  bool configurationLocked;
  std::vector<int> lockedScenes;
};

struct DeviceBinaryInputSpec {
  uint8_t TargetGroup;
  BinaryInputType InputType;
  BinaryInputId InputID;
  bool Independent;
};

struct DeviceStateSpec {
  std::string Name;
  std::vector<std::string> Values;
};

struct CircuitPowerStateSpec {
  uint8_t Index;
  uint8_t State;
  uint32_t ActiveThreshold;
  uint32_t InactiveThreshold;
};

struct DeviceSensorSpec {
  SensorType sensorType;
  uint32_t SensorPollInterval;
  uint8_t SensorBroadcastFlag;
  uint8_t SensorConversionFlag;
};

struct DeviceSensorValue {
  SensorType sensorType;
  uint8_t sensorIndex;
  double value;
  uint32_t valueAge;
  uint32_t contextId;
  std::string contextMsg;
  DateTime timestamp;
  DeviceSensorValue()
      : sensorType(SensorType::UnknownType), sensorIndex(0), value(0), valueAge(0), contextId(0), timestamp(0) {}
};

struct DeviceSpec {
  devid_t ShortAddress;
  uint16_t FunctionID;
  uint16_t ProductID;
  uint16_t VendorID;
  uint16_t revisionId;
  bool Locked;
  uint8_t ActiveState;
  uint8_t OutputMode;
  ButtonInputMode LTMode;
  std::vector<int> Groups;
  dsuid_t DSID;
  uint8_t buttonID;
  uint8_t buttonActiveGroup;
  bool buttonSetsLocalPriority;
  bool buttonCallsPresent;
  std::string Name;
  uint16_t ZoneID;
  std::vector<DeviceBinaryInputSpec> binaryInputs;
  bool binaryInputsValid;
  std::vector<DeviceSensorSpec> sensorInputs;
  bool sensorInputsValid;
  std::vector<ChannelType> outputChannels;
  bool outputChannelsValid;
  uint8_t activeGroup;
  uint8_t defaultGroup;
};

struct DSMeterHash {
  uint32_t Hash;
  uint32_t ModificationCount;
  uint32_t EventCount;
};

struct ZoneHeatingConfigSpec {
  HeatingControlMode mode;
  uint16_t Kp;
  uint8_t Ts;
  uint16_t Ti;
  uint16_t Kd;
  int16_t Imin;
  int16_t Imax;
  uint8_t Ymin;
  uint8_t Ymax;
  uint8_t AntiWindUp;
  uint8_t KeepFloorWarm;
  uint16_t SourceZoneId;
  int16_t Offset;
  uint8_t EmergencyValue;
  uint8_t ManualValue;
};

struct ZoneHeatingInternalsSpec {
  uint16_t Trecent;
  uint16_t Treference;
  int16_t TError;
  int16_t TErrorPrev;
  int32_t Integral;
  int32_t Yp;
  int32_t Yi;
  int32_t Yd;
  uint8_t Y;
  uint8_t AntiWindUp;
};

struct ZoneHeatingStateSpec {
  uint8_t State;
};

struct ZoneHeatingOperationModeSpec {
  // `uint16_t opModes[16]` loses size in reflection
  // `std::array<uint16_t,16>` has problems with template specialization on oe-core
  std::array<uint16_t, 16> opModes;
}
;

struct MeteringConfigurationSpec {
  MeteringProducerCategory category;
  uint8_t scaleFactor;
};
}

DS_REFLECT_DECLARE_STRUCT((dss), DSMeterSpec,
                          (DSID)(DeviceType)(SoftwareRevisionARM)(SoftwareRevisionDSP)(HardwareVersion)(APIVersion)(
                              Name)(flags)(ApartmentState));
DS_REFLECT_DECLARE_STRUCT((dss), GroupSpec,
                          (GroupID)(applicationType)(NumberOfDevices)(Name)(applicationConfiguration));
DS_REFLECT_DECLARE_STRUCT(
    (dss), ClusterSpec,
    (GroupID)(applicationType)(NumberOfDevices)(Name)(applicationConfiguration)(canHaveStateMachine)(protectionClass)(
        location)(floor)(configurationLocked)(lockedScenes));
DS_REFLECT_DECLARE_STRUCT((dss), DeviceBinaryInputSpec, (TargetGroup)(InputType)(InputID)(Independent));
DS_REFLECT_DECLARE_STRUCT((dss), DeviceStateSpec, (Name)(Values));
DS_REFLECT_DECLARE_STRUCT((dss), CircuitPowerStateSpec, (Index)(State)(ActiveThreshold)(InactiveThreshold));
DS_REFLECT_DECLARE_STRUCT((dss), DeviceSensorSpec,
                          (sensorType)(SensorPollInterval)(SensorBroadcastFlag)(SensorConversionFlag));
DS_REFLECT_DECLARE_STRUCT((dss), DeviceSensorValue,
                          (sensorType)(sensorIndex)(value)(valueAge)(contextId)(contextMsg)(timestamp));
DS_REFLECT_DECLARE_STRUCT((dss), DeviceSpec,
                          (ShortAddress)(FunctionID)(ProductID)(VendorID)(revisionId)(Locked)(ActiveState)(OutputMode)(
                              LTMode)(Groups)(DSID)(buttonID)(buttonActiveGroup)(buttonSetsLocalPriority)(
                              buttonCallsPresent)(Name)(ZoneID)(binaryInputs)(binaryInputsValid)(sensorInputs)(
                              sensorInputsValid)(outputChannels)(outputChannelsValid)(activeGroup)(defaultGroup));
DS_REFLECT_DECLARE_STRUCT((dss), DSMeterHash, (Hash)(ModificationCount)(EventCount));
DS_REFLECT_DECLARE_STRUCT((dss), MeteringConfigurationSpec, (category)(scaleFactor));
DS_REFLECT_DECLARE_STRUCT((dss), ZoneHeatingConfigSpec,
                          (mode)(Kp)(Ts)(Ti)(Kd)(Imin)(Imax)(Ymin)(Ymax)(AntiWindUp)(KeepFloorWarm)(SourceZoneId)(
                              Offset)(EmergencyValue)(ManualValue));
DS_REFLECT_DECLARE_STRUCT((dss), ZoneHeatingInternalsSpec,
                          (Trecent)(Treference)(TError)(TErrorPrev)(Integral)(Yp)(Yi)(Yd)(Y)(AntiWindUp));
DS_REFLECT_DECLARE_STRUCT((dss), ZoneHeatingStateSpec, (State));
DS_REFLECT_DECLARE_STRUCT((dss), ZoneHeatingOperationModeSpec, (opModes));

