/*
    Copyright (c) 2016 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <boost/bind.hpp>

#include "dss.h"
#include "eventinterpreterjsplugins.h"
#include "eventinterpreterplugins.h"
#include "internaleventrelaytarget.h"
#include "model/apartment.h"
#include "model/group.h"
#include "model/state.h"
#include "scripting/jscluster.h"
#include "scripting/jscurl.h"
#include "scripting/jsdatabase.h"
#include "scripting/jsevent.h"
#include "scripting/jslogger.h"
#include "scripting/jsmodel.h"
#include "scripting/jsproperty.h"
#include "scripting/jswebservice.h"
#include "scripting/scriptobject.h"
#include "subscription_profiler.h"

static ds::LogChannel dsLogChannel("eventInterpreterJsPlugins");

namespace dss {

//================================================== ScriptContextWrapper

ScriptContextWrapper::ScriptContextWrapper(std::shared_ptr<ScriptContext> _pContext, PropertyNodePtr _pRootNode,
                                           const std::string& _identifier, bool _uniqueNode)
    : m_pContext(_pContext), m_Identifier(_identifier), m_UniqueNode(_uniqueNode) {
  if ((_pRootNode != NULL) && !_identifier.empty()) {
    if (_uniqueNode) {
      m_pPropertyNode = _pRootNode->createProperty(_identifier + "+");
    } else {
      m_pPropertyNode = _pRootNode->createProperty(_identifier);
    }
  }
}

ScriptContextWrapper::~ScriptContextWrapper() {
  if ((m_pPropertyNode != NULL) && (m_pPropertyNode->getParentNode() != NULL)) {
    if (m_UniqueNode) {
      m_pPropertyNode->getParentNode()->removeChild(m_pPropertyNode);
    } else {
      m_pPropertyNode->removeChild(m_StartedAtNode);
      m_pPropertyNode->removeChild(m_StopNode);
      m_pPropertyNode->removeChild(m_AttachedObjectsNode);
      m_pPropertyNode->removeChild(m_FilesNode);
    }
  }
}

void ScriptContextWrapper::init() {
  if (m_pPropertyNode) {
    m_StopNode = m_pPropertyNode->createProperty("stopScript+");
    m_StopNode->linkToProxy(
        PropertyProxyMemberFunction<ScriptContextWrapper, bool>(*this, NULL, &ScriptContextWrapper::stopScript));
    m_StartedAtNode = m_pPropertyNode->createProperty("startedAt+");
    m_StartedAtNode->linkToProxy(
        PropertyProxyMemberFunction<DateTime, std::string, false>(m_StartTime, &DateTime::toString));
    m_AttachedObjectsNode = m_pPropertyNode->createProperty("attachedObjects+");
    m_AttachedObjectsNode->linkToProxy(
        PropertyProxyMemberFunction<ScriptContext, int>(*m_pContext, &ScriptContext::getAttachedObjectsCount));
  }
}

void ScriptContextWrapper::destroy() {
  m_LoadedFiles.clear();
  if (m_pPropertyNode) {
    m_pPropertyNode->removeChild(m_StartedAtNode);
    m_pPropertyNode->removeChild(m_StopNode);
    m_pPropertyNode->removeChild(m_AttachedObjectsNode);
    m_pPropertyNode->removeChild(m_FilesNode);
  }
}

std::shared_ptr<ScriptContext> ScriptContextWrapper::get() {
  return m_pContext;
}

void ScriptContextWrapper::addFile(const std::string& _name) {
  m_LoadedFiles.push_back(_name);
  if (m_pPropertyNode != NULL) {
    if (m_FilesNode == NULL) {
      m_FilesNode = m_pPropertyNode->createProperty("files+");
    }
    m_FilesNode->createProperty("file+")->setStringValue(_name);
  }
}

PropertyNodePtr ScriptContextWrapper::getPropertyNode() {
  return m_pPropertyNode;
}

const std::string& ScriptContextWrapper::getIdentifier() const {
  return m_Identifier;
}

void ScriptContextWrapper::stopScript(bool _value) {
  DS_INFO("Stop of script '", m_pPropertyNode->getName(), "' requested.");
  m_pContext->stop();
}

class JSStoreEvent : public Event {
private:
  PropertyNodePtr m_pNode;

public:
  JSStoreEvent(const std::string& _name) : Event(_name) {}
  void setPropertyNodePtr(PropertyNodePtr& _node) { m_pNode = _node; }
  PropertyNodePtr& getPropertyNodePtr() { return m_pNode; }
};

class WrapperAwarePropertyScriptExtension : public PropertyScriptExtension, public EventInterpreterPlugin {
public:
  WrapperAwarePropertyScriptExtension(PropertySystem& _propertySystem, EventInterpreter& _pInterpreter,
                                      const std::string& _storeDirectory)
      : PropertyScriptExtension(_propertySystem),
        EventInterpreterPlugin("WrapperAwarePropertyScriptExtension", &_pInterpreter),
        m_StoreDirectory(addTrailingBackslash(_storeDirectory)) {}

  virtual void subscribe() {
    auto&& interpreter = getEventInterpreter();
    interpreter.subscribe(std::make_shared<EventSubscription>(*this, EventName::JSPropertyStore));
  }

  virtual void handleEvent(Event& _event, const EventSubscription& _subscription) { performStore(&_event); }

  virtual PropertyNodePtr getProperty(ScriptContext* _context, const std::string& _path) {
    if (beginsWith(_path, "/")) {
      return PropertyScriptExtension::getProperty(_context, _path);
    } else {
      std::shared_ptr<ScriptContextWrapper> wrapper = _context->getWrapper();
      DS_ASSERT(wrapper != NULL, "");
      return wrapper->getPropertyNode()->getProperty(_path);
    }
  }

  virtual PropertyNodePtr createProperty(ScriptContext* _context, const std::string& _name) {
    if (beginsWith(_name, "/")) {
      return PropertyScriptExtension::createProperty(_context, _name);
    } else {
      std::shared_ptr<ScriptContextWrapper> wrapper = _context->getWrapper();
      DS_ASSERT(wrapper != NULL, "");
      return wrapper->getPropertyNode()->createProperty(_name);
    }
  }

  virtual bool store(ScriptContext* _context, PropertyNodePtr _node) {
    std::shared_ptr<ScriptContextWrapper> wrapper = _context->getWrapper();
    DS_ASSERT(wrapper != NULL, "");
    std::string fileName(wrapper->getIdentifier());
    PropertyNodePtr pNode;
    if (_node) {
      pNode = _node;
      fileName += "_" + pNode->getDisplayName();
    } else {
      pNode = wrapper->getPropertyNode();
    }

    std::shared_ptr<Event> pEvent;
    if (wrapper->getIdentifier().find("user-defined-actions") != std::string::npos) {
      DSS::getInstance()->getApartment().emitStructureChanged(); // user-defined actions property tree
      pEvent = ModelChangedEvent::createUdaChanged();
    } else if (wrapper->getIdentifier().find("system-addon-timed-events") != std::string::npos) {
      DSS::getInstance()->getApartment().emitStructureChanged(); // system-addon-events property tree
      pEvent = ModelChangedEvent::createTimedEventChanged();
    }

    if (pEvent) {
      DSS::getInstance()->getEventQueue().pushEvent(pEvent);
    }

    // TODO: sanitize filename to prevent world-domination

    std::string path = m_StoreDirectory + fileName + ".xml";
    std::shared_ptr<JSStoreEvent> pJSEvent = std::make_shared<JSStoreEvent>(EventName::JSPropertyStore);
    pJSEvent->setContext(wrapper->getIdentifier());
    pJSEvent->setLocation(path);
    pJSEvent->setProperty(EventProperty::Time, "+15");
    pJSEvent->setProperty(EventProperty::Unique, "Yes");
    pJSEvent->setPropertyNodePtr(pNode);

    DSS::getInstance()->getEventQueue().pushEvent(pJSEvent);

    return true;
  }

  virtual bool load(ScriptContext* _context) {
    std::shared_ptr<ScriptContextWrapper> wrapper = _context->getWrapper();
    DS_ASSERT(wrapper != NULL, "");
    std::string path = m_StoreDirectory + wrapper->getIdentifier() + ".xml";
    {
      JSStoreEvent JSEvent(EventName::JSPropertyStore);
      JSEvent.setContext(wrapper->getIdentifier());
      JSEvent.setLocation(path);
      std::vector<std::shared_ptr<Event> > removedEvents =
          DSS::getInstance()->getEventRunner().removeReplacementEvents(JSEvent);
      if (removedEvents.size() >= 1) {
        performStore(removedEvents[0].get());
      }
    }
    // TODO: sanitize filename to prevent world-domination
    return loadFromXML(path, wrapper->getPropertyNode());
  }

private:
  bool performStore(Event* _event) {
    bool result = false;
    JSStoreEvent* pEvent = dynamic_cast<JSStoreEvent*>(_event);
    if (pEvent) {
      std::string path = pEvent->getPropertyByName(EventProperty::Location);
      PropertyNodePtr pNode = pEvent->getPropertyNodePtr();
      DS_NOTICE("Writing script config to", path);
      result = saveToXML(path, pNode);
      DS_DEBUG("Finished writing script config to", path);
    }
    return result;
  }

private:
  std::string m_StoreDirectory;
}; // WrapperAwarePropertyScriptExtension

//================================================== EventInterpreterPluginJavascript

EventInterpreterPluginJavascript::EventInterpreterPluginJavascript(EventInterpreter* _pInterpreter)
    : EventInterpreterPlugin("javascript", _pInterpreter) {} // ctor

EventInterpreterPluginJavascript::~EventInterpreterPluginJavascript() {
  DS_NOTICE("Terminating all scripts...");
  typedef std::vector<std::shared_ptr<ScriptContextWrapper> >::iterator tScriptContextWrapperIterator;
  tScriptContextWrapperIterator ipScriptContextWrapper = m_WrappedContexts.begin();
  int shutdownTimeout = 0;
  while (ipScriptContextWrapper != m_WrappedContexts.end()) {
    (*ipScriptContextWrapper)->get()->stop();
    while ((*ipScriptContextWrapper)->get()->hasAttachedObjects()) {
      sleepMS(10);
      shutdownTimeout += 10;
      if (shutdownTimeout > 2500) {
        DS_WARNING("Forced to shutdown pending scripts, timeout reached");
        break;
      }
    }
    (*ipScriptContextWrapper)->destroy();
    (*ipScriptContextWrapper)->get()->detachWrapper();
    ipScriptContextWrapper = m_WrappedContexts.erase(ipScriptContextWrapper);
  }
  DS_INFO("All scripts Terminated");
}

void EventInterpreterPluginJavascript::handleEvent(Event& _event, const EventSubscription& _subscription) {
  if (_subscription.getOptions()->hasParameter("filename1")) {
    StopWatch stopWatch(_event.getName());
    bool timingEnabled = true;

    stopWatch.startSubscription(); /* start before environment setup */

    if (m_pEnvironment == NULL) {
      initializeEnvironment();
    }

    /* TODO this should be done one in SubscriptionProfiler */
    if (!m_pEnvironment->isTimingEnabled()) {
      timingEnabled = false;
      stopWatch.cancel();
    }

    std::string scriptID;
    if (_subscription.getOptions()->hasParameter("script_id")) {
      scriptID = _subscription.getOptions()->getParameter("script_id");
    }
    bool uniqueNode = false;
    if (scriptID.empty()) {
      uniqueNode = true;
      scriptID = _event.getName() + _subscription.getID();
    }

    if (timingEnabled) {
      stopWatch.setSubscriptionName(scriptID);
    }

    std::shared_ptr<ScriptContext> ctx(m_pEnvironment->getContext());
    auto wrapper = std::make_shared<ScriptContextWrapper>(ctx, m_pScriptRootNode, scriptID, uniqueNode);
    ctx->attachWrapper(wrapper); // cyclic shared_ptr
    wrapper->init();

    {
      ScriptLock lock(ctx);
      JSContextThread th(ctx.get());

      ScriptObject raisedEvent(*ctx, NULL);
      raisedEvent.setProperty<const std::string&>("name", _event.getName());
      ctx->getRootObject().setProperty("raisedEvent", &raisedEvent);
      if (_event.hasPropertySet(EventProperty::Location)) {
        raisedEvent.setProperty<const std::string&>("location", _event.getPropertyByName(EventProperty::Location));
      }
      EventRaiseLocation raiseLocation = _event.getRaiseLocation();
      ScriptObject source(*ctx, NULL);
      if ((raiseLocation == erlGroup) || (raiseLocation == erlApartment)) {
        if (DSS::hasInstance()) {
          std::shared_ptr<const Group> group = _event.getRaisedAtGroup(DSS::getInstance()->getApartment());
          source.setProperty(
              "set", ".zone(" + intToString(group->getZoneID()) + ").group(" + intToString(group->getID()) + ")");
          source.setProperty("groupID", group->getID());
          source.setProperty("zoneID", group->getZoneID());
          source.setProperty("isApartment", raiseLocation == erlApartment);
          source.setProperty("isGroup", raiseLocation == erlGroup);
          source.setProperty("isDevice", false);
        }
      } else if (raiseLocation == erlDevice) {
        std::shared_ptr<const DeviceReference> device = _event.getRaisedAtDevice();
        try {
          source.setProperty("set", "dsuid(" + dsuid2str(device->getDSID()) + ")");
          dsid_t dsid;
          if (dsuid_to_dsid(device->getDSID(), &dsid)) {
            source.setProperty("dsid", dsid2str(dsid));
          } else {
            source.setProperty("dsid", "");
          }
          source.setProperty("dsuid", dsuid2str(device->getDSID()));
          source.setProperty("zoneID", device->getDevice()->getZoneID());
        } catch (ItemNotFoundException& e) {
        }
        source.setProperty("isApartment", false);
        source.setProperty("isGroup", false);
        source.setProperty("isDevice", true);
      } else if (raiseLocation == erlState) {
        std::shared_ptr<const State> state = _event.getRaisedAtState();
        if (state->getType() == StateType_Device) {
          std::weak_ptr<Device> weakDevice = state->getProviderDevice();
          if (std::shared_ptr<Device> device = weakDevice.lock()) {
            source.setProperty("set", "dsuid(" + dsuid2str(device->getDSID()) + ")");
            dsid_t dsid;
            if (dsuid_to_dsid((const dsuid_t)device->getDSID(), &dsid)) {
              source.setProperty("dsid", dsid2str(dsid));
            } else {
              source.setProperty("dsid", "");
            }
            source.setProperty("dsuid", dsuid2str(device->getDSID()));
            source.setProperty("zoneID", device->getZoneID());
            source.setProperty("isApartment", false);
            source.setProperty("isGroup", false);
            source.setProperty("isDevice", true);
          }
        } else if (state->getType() == StateType_Apartment) {
          source.setProperty("isApartment", true);
          source.setProperty("isGroup", false);
          source.setProperty("isDevice", false);
        } else if (state->getType() == StateType_Group) {
          std::weak_ptr<Group> weakGroup = state->getProviderGroup();
          if (std::shared_ptr<Group> group = weakGroup.lock()) {
            source.setProperty("isApartment", false);
            source.setProperty("isGroup", true);
            source.setProperty("isDevice", false);
            source.setProperty("groupID", group->getID());
            source.setProperty("zoneID", group->getZoneID());
          }
        } else if (state->getType() == StateType_Service) {
          source.setProperty("isService", true);
          source.setProperty("isGroup", false);
          source.setProperty("isDevice", false);
          source.setProperty("serviceName", state->getProviderService());
        } else if (state->getType() == StateType_Script) {
          source.setProperty("isScript", true);
          source.setProperty("isGroup", false);
          source.setProperty("isDevice", false);
          source.setProperty("serviceName", state->getProviderService());
        }
      }
      raisedEvent.setProperty("source", &source);

      // add raisedEvent.parameter
      ScriptObject param(*ctx, NULL);
      for (auto&& elt : _event.getProperties().getContainer()) {
        DS_INFO("JavaScript Event Handler: setting parameter ", elt.first, " to ", elt.second);
        param.setProperty<const std::string&>(elt.first, elt.second);
      }
      raisedEvent.setProperty("parameter", &param);

      // add raisedEvent.subscription
      ScriptObject subscriptionObj(*ctx, NULL);
      raisedEvent.setProperty("subscription", &subscriptionObj);
      subscriptionObj.setProperty<const std::string&>("name", _subscription.getEventName());
    }

    std::string scripts;

    for (int i = 0; i < UCHAR_MAX; i++) {
      if (timingEnabled) {
        stopWatch.startScript();
      }

      std::string paramName = std::string("filename") + intToString(i + 1);
      if (!_subscription.getOptions()->hasParameter(paramName)) {
        break;
      }
      std::string scriptName = _subscription.getOptions()->getParameter(paramName);

      if (!beginsWith(scriptName, "/")) {
        if (DSS::hasInstance()) {
          scriptName = DSS::getInstance()->getDataDirectory() + "/" + scriptName;
        }
      }

      wrapper->addFile(scriptName);

      if (timingEnabled) {
        stopWatch.setScriptName(scriptName);
      }

      try {
        DS_DEBUG("JavaScript Event Handler: running script ", scriptName);
        ctx->evaluateScript<void>(scriptName);

      } catch (ScriptException& ex) {
        DS_ERROR("JavaScript Event Handler: Script error running/parsing script '", scriptName,
                 "'. Message: ", ex.what());
        break;
      } catch (DSSException& ex) {
        DS_ERROR("JavaScript Event Handler: dSS/Bus error running/parsing script '", scriptName,
                 "'. Message: ", ex.what());
        break;
      } catch (boost::thread_resource_error& ex) {
        DS_ERROR("JavaScript Event Handler: Fatal ressource error running/parsing script '", scriptName,
                 "'. Message: ", ex.what());
        return;
      } catch (std::runtime_error& ex) {
        DS_ERROR("JavaScript Event Handler: Fatal error running/parsing script '", scriptName,
                 "'. Message: ", ex.what());
        return;
      }

      scripts = scripts + scriptName + " ";

      if (timingEnabled) {
        stopWatch.stopScript();
      }
    }

    if (ctx->hasAttachedObjects()) {
      m_WrappedContexts.push_back(wrapper);
      DS_INFO("JavaScript Event Handler: keep ", scripts, " in memory");
    } else {
      ctx->detachWrapper();
      wrapper->destroy();
    }

    if (timingEnabled) {
      stopWatch.stopSubscription();
    }

  } else {
    throw std::runtime_error("JavaScript Event Handler: missing argument filename1");
  }
} // handleEvent

const std::string EventInterpreterPluginJavascript::kCleanupScriptsEventName =
    "EventInterpreterPluginJavascript_cleanupScripts";

void EventInterpreterPluginJavascript::initializeEnvironment() {
  if (DSS::hasInstance()) {
    m_pEnvironment.reset(new ScriptEnvironment(&DSS::getInstance()->getSecurity()));
    m_pEnvironment->initialize();

    ScriptExtension* ext;
    ext = new ScriptLoggerExtension(DSS::getInstance()->getJSLogDirectory(), DSS::getInstance()->getEventInterpreter());
    m_pEnvironment->addExtension(ext);

    ext = new EventScriptExtension(DSS::getInstance()->getEventQueue(), getEventInterpreter());
    m_pEnvironment->addExtension(ext);

    WrapperAwarePropertyScriptExtension* wapExt = new WrapperAwarePropertyScriptExtension(
        DSS::getInstance()->getPropertySystem(), getEventInterpreter(), DSS::getInstance()->getSavedPropsDirectory());
    m_pEnvironment->addExtension(wapExt);
    getEventInterpreter().addPlugin(wapExt);
    wapExt->subscribe();

    ext = new CurlScriptContextExtension();
    m_pEnvironment->addExtension(ext);

    ext = new WebserviceConnectionScriptContextExtension();
    m_pEnvironment->addExtension(ext);

    ext = new ModelScriptContextExtension(DSS::getInstance()->getApartment());
    m_pEnvironment->addExtension(ext);

    ext = new ModelConstantsScriptExtension();
    m_pEnvironment->addExtension(ext);

    ext = new ClusterScriptExtension(DSS::getInstance()->getApartment());
    m_pEnvironment->addExtension(ext);

    ext = new DatabaseScriptExtension();
    m_pEnvironment->addExtension(ext);

    setupCleanupEvent();
    m_pScriptRootNode = DSS::getInstance()->getPropertySystem().createProperty("/scripts");

    // Create compatibility link for add-ons states
    // TODO(someday): can be removed when system addons are integrated into core dss component
    PropertyNodePtr aliasNode =
        DSS::getInstance()->getPropertySystem().createProperty("/usr/addon-states/heating-controller/");
    PropertyNodePtr addonNode =
        DSS::getInstance()->getPropertySystem().createProperty("/usr/addon-states/system-addon-heating-controller/");
    aliasNode->alias(addonNode);
    aliasNode = DSS::getInstance()->getPropertySystem().createProperty("/scripts/heating-controller/");
    addonNode = DSS::getInstance()->getPropertySystem().createProperty("/scripts/system-addon-heating-controller/");
    aliasNode->alias(addonNode);

  } else {
    m_pEnvironment.reset(new ScriptEnvironment());
    m_pEnvironment->initialize();
  }
} // initializeEnvironment

void EventInterpreterPluginJavascript::setupCleanupEvent() {
  EventInterpreterInternalRelay* pRelay = dynamic_cast<EventInterpreterInternalRelay*>(
      getEventInterpreter().getPluginByName(EventInterpreterInternalRelay::getPluginName()));
  m_pRelayTarget = std::make_shared<InternalEventRelayTarget>(boost::ref<EventInterpreterInternalRelay>(*pRelay));

  auto cleanupEventSubscription =
      std::make_shared<EventSubscription>(kCleanupScriptsEventName, EventInterpreterInternalRelay::getPluginName(),
                                            getEventInterpreter(), std::shared_ptr<SubscriptionOptions>());
  m_pRelayTarget->subscribeTo(cleanupEventSubscription);
  m_pRelayTarget->setCallback(boost::bind(&EventInterpreterPluginJavascript::cleanupTerminatedScripts, this, _1, _2));

  auto pEvent = std::make_shared<Event>(kCleanupScriptsEventName);
  pEvent->setProperty(EventProperty::ICalStartTime, DateTime().toRFC2445IcalDataTime());
  pEvent->setProperty(EventProperty::ICalRRule, "FREQ=SECONDLY;INTERVAL=20");
  pEvent->setProperty(EventProperty::Unique, "Yes");
  getEventInterpreter().getQueue().pushEvent(pEvent);
} // setupCleanupEvent

void EventInterpreterPluginJavascript::cleanupTerminatedScripts(Event& _event, const EventSubscription& _subscription) {
  typedef std::vector<std::shared_ptr<ScriptContextWrapper> >::iterator tScriptContextWrapperIterator;
  tScriptContextWrapperIterator ipScriptContextWrapper = m_WrappedContexts.begin();
  while (ipScriptContextWrapper != m_WrappedContexts.end()) {
    // if the context of the wrapper does not have active objects anymore we
    // can clean it up; in the non-caching mode the wrapper has the last
    // reference to the context shared-ptr
    if (!(*ipScriptContextWrapper)->get()->hasAttachedObjects()) {
      DS_DEBUG("JavaScript cleanup: erasing script " + (*ipScriptContextWrapper)->getIdentifier());
      (*ipScriptContextWrapper)->destroy();
      (*ipScriptContextWrapper)->get()->detachWrapper();
      ipScriptContextWrapper = m_WrappedContexts.erase(ipScriptContextWrapper);
    } else {
      DS_DEBUG("JavaScript cleanup: script " + (*ipScriptContextWrapper)->getIdentifier() +
               " still has objects: " + intToString((*ipScriptContextWrapper)->get()->getAttachedObjectsCount()));
      ++ipScriptContextWrapper;
    }
  }
} // cleanupTerminatedScripts
} // namespace dss
