/*
 *  Copyright (c) 2018 digitalSTROM.org, Zurich, Switzerland
 *
 *  This file is part of digitalSTROM Server.
 *
 *  digitalSTROM Server is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  digitalSTROM Server is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
 */

#include "hal-device.h"

#include <chrono>
#include <cmath>

#include "base.h"
#include "cache/cache-updater.h"
#include "ds-api/route-segments.h"
#include "dss.h"
#include "event.h"
#include "model/apartment.h"
#include "model/device.h"
#include "model/scenehelper.h"
#include "util.h"

namespace dss::hal {

static ds::LogChannel dsLogChannel("halShadeDevice");

static const int HalfWavePerSecond = 100;
static const int MillisecondsPerHalfWave = 10;
static constexpr int DefaultPositionTime = 120 * HalfWavePerSecond;
static constexpr int DefaultDimStepLen = static_cast<int>(0.2 * HalfWavePerSecond);
static constexpr int DefaultDimStepPause = static_cast<int>(0.81 * HalfWavePerSecond);
static constexpr int DefaultDimStepNum = 4;
static constexpr int DefaultIncInterval = 250;

void HalShadeDevice::write(const Device& device, const std::string& parameter, double value) const {
  // "/scene/5/brightness" -> 128, 5
  // "/scene/6/dontCare"   -> 128, 134 (=128+6)
  // "/scene/7/brightness/dontCare" -> 128, 135 (=128+7)
  DeviceCache devCache = m_sceneCacheDb.tryLookup(device.getDSID());
  if (auto route = dsApi::matchRoute("/scene/{sceneNr}/{channelOrFlag}", parameter)) {
    auto& scene = route.args["sceneNr"];
    auto& param = route.args["channelOrFlag"];
    unsigned int sceneNr = strToUInt(scene);
    DS_REQUIRE(sceneNr <= MaxSceneNumber, "Scene number is out of range.", sceneNr);
    SceneEntry& sceneCache = devCache.scene.sceneTable[sceneNr];
    if (param == "shadePositionOutside") {
      auto channelSpec = getOutputChannelIdAndSize("shadePositionOutside");
      auto newValue = convertToOutputChannelValue(channelSpec.first, value);
      auto lowValue = device.getDeviceConfig(CfgClassSceneExtention, sceneNr / 2);
      if ((sceneNr % 2) == 0) {
        lowValue &= 0xF0;
        lowValue |= (newValue >> 4) & 0x0F;
      } else {
        lowValue &= 0x0F;
        lowValue |= newValue & 0xF0;
      }
      device.setDeviceConfig(CfgClassScene, sceneNr, (newValue >> 8) & 0xFF, false);
      device.setDeviceConfig(CfgClassSceneExtention, sceneNr / 2, lowValue, false);
      SceneChannel& shadePositionOutside = sceneCache.channels[ChannelType::shadePositionOutside];
      shadePositionOutside.value = value;
      sceneCache.lastChanged = std::chrono::system_clock::now();
      devCache.lastChanged = sceneCache.lastChanged;
    } else if (param == "shadeOpeningAngleOutside") {
      DeviceFeatures_t features = device.getDeviceFeatures();
      DS_REQUIRE(features.hasOutputAngle, "Device dows not have channel", param);
      auto channelSpec = getOutputChannelIdAndSize("shadeOpeningAngleOutside");
      auto newValue = convertToOutputChannelValue(channelSpec.first, value);
      device.setDeviceConfig(CfgClassSceneAngle, sceneNr, newValue, false);
      SceneChannel& shadeOpeningAngleOutside = sceneCache.channels[ChannelType::shadeOpeningAngleOutside];
      shadeOpeningAngleOutside.value = value;
      sceneCache.lastChanged = std::chrono::system_clock::now();
      devCache.lastChanged = sceneCache.lastChanged;
    } else {
      DS_REQUIRE(false, "Unknown channel requested", param);
    }
    m_sceneCacheDb.update(device.getDSID(), devCache);
  } else if (auto route = dsApi::matchRoute("/property/{propertyName}", parameter)) {
    auto& propertyName = route.args["propertyName"];
    auto property = ds::tryToEnum<DeviceProperty>(propertyName);
    if (property) {
      unsigned int intValueLow = 0;
      std::optional<unsigned int> intValueHigh;
      if (*property == DeviceProperty::dimStepNum) {
        DS_REQUIRE(value >= 0, "Value too low");
        DS_REQUIRE(value <= 255, "Value too high");
        intValueLow = static_cast<unsigned int>(value);
      } else if (*property == DeviceProperty::dimStepLen || *property == DeviceProperty::dimStepPause) {
        DS_REQUIRE(value >= 0, "Value too low");
        DS_REQUIRE(value <= 2.55, "Value too high");
        intValueLow = static_cast<unsigned int>(value / 0.01);
      } else if (*property == DeviceProperty::posTimeUp || *property == DeviceProperty::posTimeDown) {
        DS_REQUIRE(value >= 0, "Value too low");
        DS_REQUIRE(value <= 655.35, "Value too high");
        unsigned int tempValue = static_cast<unsigned int>(value / 0.01);
        intValueLow = tempValue & 0xFF;
        intValueHigh = (tempValue >> 8) & 0xFF;
      } else {
        DS_FAIL_REQUIRE("No value conversion found for ", propertyName);
      }
      int bank;
      int offset;
      int length;
      std::tie(bank, offset, length) = getDeviceParameter(device, *property);
      device.setDeviceConfig(bank, offset, intValueLow, false);
      devCache.property[*property] = intValueLow;
      if (intValueHigh) {
        device.setDeviceConfig(bank, offset + 1, *intValueHigh, false);
        devCache.property[*property] = *intValueHigh << 8 | intValueLow;
      }
      devCache.lastChanged = std::chrono::system_clock::now();
      m_sceneCacheDb.update(device.getDSID(), devCache);
    } else {
      DS_FAIL_REQUIRE("Unknown property requested", propertyName);
    }
  } else {
    DS_FAIL_REQUIRE("Unknown parameter requested", parameter);
  }
}

// TODO does not save to db, should be method not function
double savePositionOutside(SceneEntry& sceneEntry, uint8_t hi, uint8_t lo) {
  lo |= lo << 4; // low nibble is doubled, why?
  auto rawValue = static_cast<uint16_t>(hi) << 8 | lo;

  auto channelType = ChannelType::shadePositionOutside;
  sceneEntry.channels[channelType].value = convertFromOutputChannelValue(channelType, rawValue);
  return sceneEntry.channels[channelType].value.value();
}

// TODO does not save to db, should be method not function
double saveAngleOutside(SceneEntry& sceneEntry, uint8_t raw) {
  auto channelType = ChannelType::shadeOpeningAngleOutside;
  sceneEntry.channels[channelType].value = convertFromOutputChannelValue(channelType, raw);
  return sceneEntry.channels[channelType].value.value();
}

double HalShadeDevice::readDoubleRAM(const Device& device, int sceneId, ChannelType channelType) const {
  auto state = m_StateCache.tryLookup(device.getDSID());
  auto sceneEntry = m_sceneCacheDb.tryLookup(device.getDSID()).scene.sceneTable[sceneId];
  double ret;

  if (sceneId != state.sceneInRAMBank.value_or(-1)) {
    // typically cache-updater, lots of noise
    DS_DEBUG("readDoubleRAM sceneInRAMBank mismatch", device, sceneId, channelType, state.sceneInRAMBank);
    return std::nan("Scene values not in RAM");
  }

  if (!sceneEntry.flags.dontCare || sceneEntry.flags.dontCare.value() || !sceneEntry.channels[channelType].dontCare ||
      sceneEntry.channels[channelType].dontCare.value()) {
    // we shouldn't end up here, if cache-updater fetches the flags first
    DS_NOTICE("readDoubleRAM dontCare unkwnown/set", device, sceneId, channelType, sceneEntry);
    return std::nan("dontCare flags missing/set");
  }

  // roll-out the red carpet...
  DS_NOTICE("readDoubleRAM fetching values", device, sceneId, channelType, sceneEntry);
  switch (channelType) {
    case ChannelType::none: break;
    case ChannelType::shadePositionOutside: {
      auto lo = device.getDeviceConfig(CfgClassRuntime, 0x2) & 0x0F;
      auto hi = device.getDeviceConfig(CfgClassRuntime, 0x3);
      // Note: mind race condition. DSD-6855

      auto devSceneCache = m_sceneCacheDb.tryLookup(device.getDSID());
      auto& sceneEntry = devSceneCache.scene.sceneTable[sceneId];
      ret = savePositionOutside(sceneEntry, hi, lo);
      sceneEntry.lastChanged = std::chrono::system_clock::now();
      devSceneCache.lastChanged = sceneEntry.lastChanged;
      m_sceneCacheDb.update(device.getDSID(), devSceneCache);
    }
      return ret;

    case ChannelType::shadeOpeningAngleOutside: {
      auto value = device.getDeviceConfig(CfgClassRuntime, 0x4);
      // Note: mind race condition. DSD-6855

      auto devSceneCache = m_sceneCacheDb.tryLookup(device.getDSID());
      auto& sceneEntry = devSceneCache.scene.sceneTable[sceneId];
      ret = saveAngleOutside(sceneEntry, value);
      sceneEntry.lastChanged = std::chrono::system_clock::now();
      devSceneCache.lastChanged = sceneEntry.lastChanged;
      m_sceneCacheDb.update(device.getDSID(), devSceneCache);
    }
      return ret;

    case ChannelType::hue:
    case ChannelType::saturation:
    case ChannelType::colortemp:
    case ChannelType::x:
    case ChannelType::y:
    case ChannelType::brightness:
    case ChannelType::shadePositionIndoor:
    case ChannelType::shadeOpeningAngleIndoor:
    case ChannelType::transparency:
    case ChannelType::airFlowIntensity:
    case ChannelType::airFlowDirection:
    case ChannelType::airFlapPosition:
    case ChannelType::airLouverPosition:
    case ChannelType::heatingPower:
    case ChannelType::coolingCapacity:
    case ChannelType::audioVolume:
    case ChannelType::powerState:
    case ChannelType::airLouverAuto:
    case ChannelType::airFlowAuto:
    case ChannelType::waterTemperature:
    case ChannelType::waterFlowRate:
    case ChannelType::powerLevel:
    case ChannelType::videoStation:
    case ChannelType::videoInputSource: break; DS_FAIL_REQUIRE("Non applicable", channelType);
  }
  return std::nan("Invalid ChannelType"); // not reached
}

double HalShadeDevice::readDouble(const Device& device, const std::string& parameter) const {
  // "/scene/5/brightness" -> 128, 5
  // "/scene/6/dontcare"   -> 128, 134 (=128+6)
  // "/scene/7/brightness/dontcare" -> 128, 135 (=128+7)
  double returnValue = 0;
  DeviceCache devCache = m_sceneCacheDb.tryLookup(device.getDSID());
  if (auto route = dsApi::matchRoute("/scene/{sceneNr}/{channelOrFlag}", parameter)) {
    auto& scene = route.args["sceneNr"];
    auto& param = route.args["channelOrFlag"];
    unsigned int sceneNr = strToUInt(scene);
    DS_REQUIRE(sceneNr <= MaxSceneNumber, "Scene number is out of range.", sceneNr);

    if (device.getRevisionID() == 0x387) {
      auto channelType = ds::tryToEnum<ChannelType>(param);
      if (!channelType.has_value()) {
        DS_FAIL_REQUIRE("Unknown channel requested", param);
      }
      return readDoubleRAM(device, sceneNr, *channelType);
    }

    SceneEntry& sceneCache = devCache.scene.sceneTable[sceneNr];
    if (param == "shadePositionOutside") {
      auto value = device.getDeviceConfig(CfgClassScene, sceneNr);
      auto extValue = device.getDeviceConfig(CfgClassSceneExtention, sceneNr / 2);
      if ((sceneNr % 2) == 1) {
        extValue &= 0xF0; // high bits for odd scene numbers
        extValue >>= 4;
      } else {
        extValue &= 0x0F; // lower bits for even scene numbers
      }
      returnValue = savePositionOutside(sceneCache, value, extValue);
    } else if (param == "shadeOpeningAngleOutside") {
      DeviceFeatures_t features = device.getDeviceFeatures();
      DS_REQUIRE(features.hasOutputAngle, "Device dows not have channel", param);
      auto value = device.getDeviceConfig(CfgClassSceneAngle, sceneNr);
      returnValue = saveAngleOutside(sceneCache, value);
    } else {
      DS_FAIL_REQUIRE("Unknown channel requested", param);
    }
    sceneCache.lastChanged = std::chrono::system_clock::now();
    devCache.lastChanged = sceneCache.lastChanged;
    // m_sceneCacheDb.update(device.getDSID(), devCache);  // TODO: Fix race condition. DSD-6955
  } else if (auto route = dsApi::matchRoute("/property/{propertyName}", parameter)) {
    auto& propertyName = route.args["propertyName"];
    auto property = ds::tryToEnum<DeviceProperty>(propertyName);
    if (property && (*property == DeviceProperty::dimStepNum || *property == DeviceProperty::dimStepLen ||
                     *property == DeviceProperty::dimStepPause || *property == DeviceProperty::posTimeUp ||
                     *property == DeviceProperty::posTimeDown)) {
      int bank;
      int offset;
      int length;
      std::tie(bank, offset, length) = getDeviceParameter(device, *property);
      auto valueWord = device.getDeviceConfigWord(bank, offset & 0xFE);
      auto valueLow = valueWord & 0x00FF;
      auto valueHigh = (valueWord & 0xFF00) >> 8;
      auto value = (offset % 2 == 0) ? valueLow : valueHigh;
      if (*property == DeviceProperty::dimStepNum) {
        returnValue = value;
        devCache.property[*property] = value;
      } else if (*property == DeviceProperty::dimStepLen || *property == DeviceProperty::dimStepPause) {
        returnValue = value * 0.01;
        devCache.property[*property] = value;
      } else if (*property == DeviceProperty::posTimeUp || *property == DeviceProperty::posTimeDown) {
        returnValue = valueWord * 0.01;
        devCache.property[*property] = valueWord;
      } else {
        DS_FAIL_REQUIRE("No value conversion found for ", propertyName);
      }
      devCache.lastChanged = std::chrono::system_clock::now();
      m_sceneCacheDb.update(device.getDSID(), devCache);
    } else {
      DS_FAIL_REQUIRE("Unknown property requested", propertyName);
    }
  } else {
    DS_FAIL_REQUIRE("Unknown parameter requested", parameter);
  }
  return returnValue;
}

void HalShadeDevice::addPositionOutsideRegister(const Device& device, int scene, int value, bool firstRegister) const {
  DeviceCache devCache = m_sceneCacheDb.tryLookup(device.getDSID());
  SceneEntry& sceneCache = devCache.scene.sceneTable[scene];
  SceneChannel& sceneChannel = sceneCache.channels[ChannelType::shadePositionOutside];

  /* Set up the array upon first time access */
  if (sceneChannel.registerValues.empty()) {
    sceneChannel.registerValues.push_back(std::nullopt);
    sceneChannel.registerValues.push_back(std::nullopt);
  }

  if (firstRegister) {
    sceneChannel.registerValues[0] = value;
  } else {
    sceneChannel.registerValues[1] = value;
  }

  /* Save combined logical value only if both registers are valid */
  if (sceneChannel.registerValues.at(0) && sceneChannel.registerValues.at(1)) {
    (void)savePositionOutside(sceneCache, *sceneChannel.registerValues[0], *sceneChannel.registerValues[1]);
  }

  sceneCache.lastChanged = std::chrono::system_clock::now();
  devCache.lastChanged = sceneCache.lastChanged;
  m_sceneCacheDb.update(device.getDSID(), devCache);
}

void HalShadeDevice::updateDeviceConfig(const Device& device, int bank, int offset, int value) const {
  DeviceFeatures_t features = device.getDeviceFeatures();

  /* Shade devices have 16-bit position values which needs to be taken
   * into account upon incoming device config requests that can only
   * write 8-bit at a time.
   *
   * If both values have never been read, we will not update the cache,
   * but save the register values separately. Only if both values have been
   * read at some point in time, then the combined value will be saved in
   * the cache.
   */
  if ((bank == CfgClassScene) && (offset < 128)) {
    addPositionOutsideRegister(device, offset, value, true);
  } else if ((bank == CfgClassScene) && (offset >= 128)) {
    // scene config / flags
    HalDevice::updateDeviceConfig(device, bank, offset, value);
  } else if ((bank == CfgClassSceneExtention) && (offset < 64)) {
    int scene = offset * 2;
    int scene_odd = scene + 1;
    int value_next = value;

    // even scenes
    value &= 0x0F; // lower bits for even scene numbers
    addPositionOutsideRegister(device, scene, value, false);

    // odd scenes
    value_next &= 0xF0; // high bits for odd scene numbers
    value_next >>= 4;
    addPositionOutsideRegister(device, scene_odd, value_next, false);

  } else if ((bank == CfgClassSceneAngle) && features.hasOutputAngle && (offset < 128)) {
    DeviceCache devCache = m_sceneCacheDb.tryLookup(device.getDSID());
    SceneEntry& sceneCache = devCache.scene.sceneTable[offset];
    (void)saveAngleOutside(sceneCache, value);
    sceneCache.lastChanged = std::chrono::system_clock::now();
    devCache.lastChanged = sceneCache.lastChanged;
    m_sceneCacheDb.update(device.getDSID(), devCache);
  } else if ((bank == CfgClassRuntime) && (offset >= 0x02) && (offset <= 0x04)) {
    initializeStatus(device);
    auto now = std::chrono::system_clock::now();
    auto state = m_StateCache.tryLookup(device.getDSID());
    switch (offset) {
      case 0x02: {
        auto target = convertToOutputChannelValue(
            ChannelType::shadePositionOutside, state.channels[ChannelType::shadePositionOutside].current.value_or(0));
        target &= 0xFF00;
        target |= (value & 0xFF);
        state.channels[ChannelType::shadePositionOutside].target =
            convertFromOutputChannelValue(ChannelType::shadePositionOutside, target);
        break;
      }
      case 0x03: {
        auto target = convertToOutputChannelValue(
            ChannelType::shadePositionOutside, state.channels[ChannelType::shadePositionOutside].current.value_or(0));
        target &= 0x00FF;
        target |= value << 8;
        state.channels[ChannelType::shadePositionOutside].target =
            convertFromOutputChannelValue(ChannelType::shadePositionOutside, target);
        break;
      }
      case 0x04: {
        state.channels[ChannelType::shadeOpeningAngleOutside].target =
            convertFromOutputChannelValue(ChannelType::shadeOpeningAngleOutside, value);
        break;
      }
    }
    if (!state.terminatesAt) {
      state.terminatesAt = now;
    }
    updateCurrentValues(device, now, state);
    m_StateCache.update(device.getDSID(), state);
    publishToPropertyTree(device);
  } else {
    for (const auto& property : getDeviceParameters(device)) {
      int propBank;
      int propOffset;
      int length;
      std::tie(propBank, propOffset, length) = property.second;
      if ((propBank == bank) && (propOffset == offset)) {
        DeviceCache devCache = m_sceneCacheDb.tryLookup(device.getDSID());
        int prevValue = 0;
        if (devCache.property[property.first]) {
          prevValue = *devCache.property[property.first] & 0xFF00;
        }
        devCache.property[property.first] = prevValue | value;
        devCache.lastChanged = std::chrono::system_clock::now();
        m_sceneCacheDb.update(device.getDSID(), devCache);
      } else if ((length == 2) && (propBank == bank) && ((propOffset + 1) == offset)) {
        DeviceCache devCache = m_sceneCacheDb.tryLookup(device.getDSID());
        int prevValue = 0;
        if (devCache.property[property.first]) {
          prevValue = *devCache.property[property.first] & 0x00FF;
        }
        devCache.property[property.first] = prevValue | (value << 8);
        devCache.lastChanged = std::chrono::system_clock::now();
        m_sceneCacheDb.update(device.getDSID(), devCache);
      }
    }
  }
}

const std::vector<ChannelType> HalShadeDevice::getSceneChannels(const Device& device) const {
  DeviceFeatures_t features = device.getDeviceFeatures();
  if (features.hasOutputAngle) {
    return std::vector<ChannelType>({ChannelType::shadePositionOutside, ChannelType::shadeOpeningAngleOutside});
  } else {
    return std::vector<ChannelType>({ChannelType::shadePositionOutside});
  }
}

const std::vector<DeviceProperty> HalShadeDevice::getDeviceProperties(const Device& device) const {
  return std::vector<DeviceProperty>({DeviceProperty::dimStepNum, DeviceProperty::dimStepLen,
                                      DeviceProperty::dimStepPause, DeviceProperty::posTimeUp,
                                      DeviceProperty::posTimeDown});
}

bool HalShadeDevice::handleDimming(int sceneId, DeviceCache& cache, const ds::DateTime& now, DeviceInternalState& state,
                                   DeviceInternalOutputState& outputState, bool up) const {
  if (outputState.current && *outputState.current > 0 && state.lastSceneCalled &&
      cache.scene.sceneTable[*state.lastSceneCalled].channels[ChannelType::shadePositionOutside].value.value_or(0) >
          0) {
    if (state.lastDimmCommand != sceneId) {
      state.fineDimmSteps = 0;
      state.dimmStatus = DeviceDimmingStatus::none;
    }
    state.dimmingTimeout = now + std::chrono::milliseconds(2 * DefaultIncInterval);
    state.lastDimmCommand = sceneId;
    if (state.dimmStatus == DeviceDimmingStatus::none ||
        (!state.terminatesAt && (state.dimmStatus == DeviceDimmingStatus::stepping))) {
      // once terminatesAt has expired we can accept the next step
      auto nextStepInMs = (cache.property[DeviceProperty::dimStepLen].value_or(DefaultDimStepLen) +
                           cache.property[DeviceProperty::dimStepPause].value_or(DefaultDimStepPause)) *
                          MillisecondsPerHalfWave;
      state.terminatesAt = now + std::chrono::milliseconds(static_cast<int64_t>(nextStepInMs));
      auto posTime = (up ? cache.property[DeviceProperty::posTimeUp].value_or(DefaultPositionTime)
                         : -1 * cache.property[DeviceProperty::posTimeDown].value_or(DefaultPositionTime));
      outputState.target =
          *outputState.current +
          ((double)cache.property[DeviceProperty::dimStepLen].value_or(DefaultDimStepLen) / posTime) * 100;
      if (state.fineDimmSteps < cache.property[DeviceProperty::dimStepNum].value_or(DefaultDimStepNum)) {
        state.fineDimmSteps++;
        state.dimmStatus = DeviceDimmingStatus::stepping;
      } else {
        state.fineDimmSteps = 0;
        state.dimmStatus = DeviceDimmingStatus::dimming;
      }
    }
    if (state.dimmStatus == DeviceDimmingStatus::dimming) {
      state.terminatesAt = now + std::chrono::milliseconds(2 * DefaultIncInterval); // two INC-intervals
      auto posTime = (up ? cache.property[DeviceProperty::posTimeUp].value_or(DefaultPositionTime)
                         : -1 * cache.property[DeviceProperty::posTimeDown].value_or(DefaultPositionTime));
      DS_DEBUG("handleDimming:", state.dimmStatus, *outputState.current, *outputState.target, posTime);
      outputState.target =
          *outputState.current + ((2.0 * DefaultIncInterval) / (posTime * MillisecondsPerHalfWave)) * 100;
    }
    if (!up) {
      outputState.target = std::max(*outputState.target, 0.001);
    } else {
      outputState.target = limitOutputChannelValue(ChannelType::shadePositionOutside, *outputState.target);
    }
    return true;
  }
  return false;
}

double HalShadeDevice::calculateRuntime(DeviceCache& cache, DeviceInternalState& state) const {
  if (state.channels[ChannelType::shadePositionOutside].current &&
      state.channels[ChannelType::shadePositionOutside].target) {
    int totalRuntime = 0;
    if (*state.channels[ChannelType::shadePositionOutside].current <
        *state.channels[ChannelType::shadePositionOutside].target) {
      totalRuntime = cache.property[DeviceProperty::posTimeUp].value_or(DefaultPositionTime) * MillisecondsPerHalfWave;
    } else {
      totalRuntime =
          cache.property[DeviceProperty::posTimeDown].value_or(DefaultPositionTime) * MillisecondsPerHalfWave;
    }
    return std::abs((*state.channels[ChannelType::shadePositionOutside].target -
                     *state.channels[ChannelType::shadePositionOutside].current) /
                    100.0 * totalRuntime);
  }
  return 0;
}

bool HalShadeDevice::setOutputValue(const Device& device, ChannelType channel, double value, ds::DateTime now) const {
  initializeStatus(device);
  auto state = m_StateCache.tryLookup(device.getDSID());
  auto cache = m_sceneCacheDb.tryLookup(device.getDSID());
  if (cache.property[DeviceProperty::outputMode] &&
      *cache.property[DeviceProperty::outputMode] == OUTPUT_MODE_DEACTIVATED) {
    m_StateCache.remove(device.getDSID());
    return false;
  }

  if (channel == ChannelType::shadeOpeningAngleOutside) {
    return HalDevice::setOutputValue(device, channel, value, now);
  }

  auto channels = getSceneChannels(device);
  if (std::find(channels.begin(), channels.end(), channel) == channels.end()) {
    DS_FAIL_REQUIRE("No such channel", channel);
  }

  updateCurrentValues(device, now, state);
  state.channels[channel].target = value;
  state.lastDimmCommand.reset();
  state.localPriority = false;
  state.lastSceneCalled.reset();
  state.sceneInRAMBank.reset();
  double dimTime = calculateRuntime(cache, state);
  if (dimTime <= 1000) {
    dimTime = 0;
  }
  if (!state.terminatesAt) {
    state.terminatesAt = now + std::chrono::milliseconds(static_cast<int64_t>(dimTime));
  }
  state.lastChanged = now;
  updateCurrentValues(device, now, state);
  if (state.terminatesAt &&
      std::chrono::duration_cast<std::chrono::milliseconds>(*state.terminatesAt - state.lastChanged).count() > 1000) {
    auto devRef = DeviceReference(device.getDSID(), &device.getApartment());
    auto event = std::make_shared<Event>(EventName::ApartmentProxyDeviceTimeout, devRef);
    event->setProperty(
        "time",
        ds::str("+",
                std::chrono::duration_cast<std::chrono::seconds>(*state.terminatesAt - state.lastChanged).count() + 2));
    device.getApartment().pushEvent(event);
  }
  m_StateCache.update(device.getDSID(), state);
  publishToPropertyTree(device);
  return true;
}

bool HalShadeDevice::callScene(const Device& device, int sceneId, bool forced, ds::DateTime now) const {
  DS_REQUIRE(sceneId <= MaxSceneNumber, "Scene number is out of range.", sceneId);

  bool needsSaving = false;

  initializeStatus(device);

  auto cache = m_sceneCacheDb.tryLookup(device.getDSID());
  auto sceneCache = cache.scene.sceneTable[sceneId];

  if (cache.property[DeviceProperty::outputMode] &&
      *cache.property[DeviceProperty::outputMode] == OUTPUT_MODE_DEACTIVATED) {
    m_StateCache.remove(device.getDSID());
    return false;
  }

  auto state = m_StateCache.tryLookup(device.getDSID());

  if (state.operationLocked) {
    return false;
  }

  needsSaving = updateCurrentValues(device, now, state) || needsSaving;

  // err on the safe side
  if (state.sceneInRAMBank) {
    state.sceneInRAMBank.reset();
    needsSaving = true;
  }

  if (SceneHelper::isDimScene(sceneId)) {
    auto& channel = state.channels[getDefaultChannel()];
    if (isStop(sceneId, (forced || !state.localPriority), cache)) {
      if (channel.current) {
        state.lastDimmCommand.reset();
        channel.target = *channel.current;
        state.terminatesAt.reset();
        state.dimmingTimeout.reset();
      }
    } else if (isDimmingDown(sceneId, (forced || !state.localPriority), cache)) {
      needsSaving = handleDimming(sceneId, cache, now, state, channel, false) || needsSaving;
    } else if (isDimmingUp(sceneId, (forced || !state.localPriority), cache)) {
      needsSaving = handleDimming(sceneId, cache, now, state, channel, true) || needsSaving;
    } else if (sceneId == SceneDimArea) {
      if (state.lastDimmCommand && (*state.lastDimmCommand == SceneDecA1 || *state.lastDimmCommand == SceneDecA2 ||
                                    *state.lastDimmCommand == SceneDecA3 || *state.lastDimmCommand == SceneDecA4)) {
        needsSaving = handleDimming(*state.lastDimmCommand, cache, now, state, channel, false) || needsSaving;
      }
      if (state.lastDimmCommand && (*state.lastDimmCommand == SceneIncA1 || *state.lastDimmCommand == SceneIncA2 ||
                                    *state.lastDimmCommand == SceneIncA3 || *state.lastDimmCommand == SceneIncA4)) {
        needsSaving = handleDimming(*state.lastDimmCommand, cache, now, state, channel, true) || needsSaving;
      }
    } else {
      return false;
    }
  } else {
    if (!sceneCache.flags.dontCare) {
      // we don't know yet if the device participates in the scene
      if (needsSaving) {
        state.lastChanged = now;
        m_StateCache.update(device.getDSID(), state);
      }
      return false;
    }

    if (*sceneCache.flags.dontCare) {
      // device ignores scene call, nothing changes
      return false;
    }

    if (forced || *sceneCache.flags.ignoreLocalPrio || !state.localPriority || (sceneId == SceneLocalOff)) {
      bool firstChannel = true;
      for (auto& channel : state.channels) {
        if (sceneCache.channels[channel.first].dontCare && *sceneCache.channels[channel.first].dontCare) {
          // the channel does not react to the scene for sure
          continue;
        }
        state.lastDimmCommand.reset();
        state.dimmingTimeout.reset();

        channel.second.target = sceneCache.channels[channel.first].value;
        if (channel.second.target) {
          if (device.hardwareType() == HardwareType::GR_KL300) {
            if (device.getShadeType() == dsModel::ShadeType::awningTerrace || \
                device.getShadeType() == dsModel::ShadeType::awningFacadeHorizontal) {
              //Invert value
              *channel.second.target = 100 - *channel.second.target;
              DS_DEBUG("Inverted channel value for GR-KL300 awning device", *channel.second.target);
            }
          }
          channel.second.target = limitOutputChannelValue(channel.first, *channel.second.target);
        }

        needsSaving = true;

        if (firstChannel) {
          state.terminatesAt.reset();
          state.beforeLastSceneCalled = state.lastSceneCalled;
          state.lastSceneCalled = sceneId;
          state.sceneInRAMBank = sceneId;
          firstChannel = false;
        }
      }
      if (sceneId == SceneLocalOn && *sceneCache.flags.specialHandling) {
        state.localPriority = true;
      } else {
        state.localPriority = false;
      }
      needsSaving = true;
    } else {
      return false;
    }
  }
  double dimTime = calculateRuntime(cache, state);
  if (dimTime <= 1000) {
    dimTime = 0;
  }
  if (!state.terminatesAt) {
    state.terminatesAt = now + std::chrono::milliseconds(static_cast<int64_t>(dimTime));
  }
  state.lastChanged = now;
  needsSaving = updateCurrentValues(device, now, state) || needsSaving;
  if (state.terminatesAt &&
      std::chrono::duration_cast<std::chrono::milliseconds>(*state.terminatesAt - state.lastChanged).count() > 1000) {
    auto devRef = DeviceReference(device.getDSID(), &device.getApartment());
    auto event = std::make_shared<Event>(EventName::ApartmentProxyDeviceTimeout, devRef);
    event->setProperty(
        "time",
        ds::str("+",
                std::chrono::duration_cast<std::chrono::seconds>(*state.terminatesAt - state.lastChanged).count() + 2));
    event->setProperty(EventProperty::Unique, "Yes");
    device.getApartment().pushEvent(event);
  } else if (state.dimmingTimeout) {
    auto devRef = DeviceReference(device.getDSID(), &device.getApartment());
    auto event = std::make_shared<Event>(EventName::ApartmentProxyDeviceDimmingTimeout, devRef);
    event->setProperty("time", "+5");
    event->setProperty(EventProperty::Unique, "Yes");
    device.getApartment().pushEvent(event);
  }

  if (needsSaving) {
    state.lastChanged = now;
    m_StateCache.update(device.getDSID(), state);
    publishToPropertyTree(device);
  }

  if (device.getRevisionID() == 0x387) {
    if (!state.sceneInRAMBank) {
      // dimming scene, lots of noise
      return needsSaving;
    }

    // RAM bank now in sync with SceneConfig bank, possibility to cache channel outputs
    auto missingValue = false;
    for (auto channel : sceneCache.channels) {
      if (channel.second.dontCare.value_or(false) || channel.second.value) {
        // if the channel is dontCare, the RAM bank will never reveal the channel configuration
        continue;
      }
      missingValue = true;
      break;
    }

    if (missingValue) {
      DS_NOTICE("callScene on FW 3.8.7 -- poke updater", sceneId, sceneCache);
      auto event = std::make_shared<Event>(CacheUpdatePlugin::CacheUpdateEventSingleDevice);
      event->setProperty(EventProperty::DSUID, ds::str(device.getDSID()));
      device.getApartment().pushEvent(event);
    } else {
      DS_NOTICE("callScene on FW 3.8.7 -- already fetched", sceneId, sceneCache);
    }
  }

  return needsSaving;
}

bool HalShadeDevice::setOutputValue(const Device& device, uint8_t value, ds::DateTime now) const {
  initializeStatus(device);
  auto state = m_StateCache.tryLookup(device.getDSID());
  auto cache = m_sceneCacheDb.tryLookup(device.getDSID());
  if (cache.property[DeviceProperty::outputMode] &&
      *cache.property[DeviceProperty::outputMode] == OUTPUT_MODE_DEACTIVATED) {
    m_StateCache.remove(device.getDSID());
    return false;
  }

  // DSD-1032: ..since it is a 8-bit value, the value is shifted and doubled to
  // get a 16-bit value.  Example: 0x6A will become position 0x6A6A.
  state.channels[ChannelType::shadePositionOutside].target =
      convertFromOutputChannelValue(ChannelType::shadePositionOutside, (value << 8 | value));
  double dimTime = calculateRuntime(cache, state);
  state.lastChanged = now;
  state.terminatesAt = now + std::chrono::milliseconds(static_cast<int64_t>(dimTime));
  updateCurrentValues(device, now, state);
  state.lastDimmCommand.reset();
  state.dimmingTimeout.reset();
  state.localPriority = false;
  state.lastSceneCalled.reset();
  m_StateCache.update(device.getDSID(), state);
  publishToPropertyTree(device);
  return true;
}

// DSD-7558
DeviceCacheState HalShadeDevice::cacheStatusFw387(const Device& device) const {
  DeviceCacheState result{};

  result.id = ds::str(device.getDSID());
  auto cache = m_sceneCacheDb.tryLookup(device.getDSID());
  populateCache(device, cache);

  for (auto& property : cache.property) {
    result.total++;
    if (property.second) {
      result.completed++;
    }
  }

  for (auto& scene : cache.scene.sceneTable) {
    result.total++;
    if (scene.second.flags.dontCare) {
      result.completed++;
    }
    // just ignore the channels, if we have them then fine, but we can't schedule it
  }

  result.completion = (result.total > 0) ? (static_cast<double>(result.completed) / result.total) * 100 : 100;

  return result;
}

DeviceCacheState HalShadeDevice::cacheStatus(const Device& device) const {
  if (device.getRevisionID() == 0x387) { // DSD-7558, fw bug, read bank64
    return cacheStatusFw387(device);
  }
  return HalDevice::cacheStatus(device);
}

} // namespace dss::hal
