/*
    Copyright (c) 2020 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

*/
#include "action-execute.h"

#include <trompeloeil.hpp>

#include "businterface-mock.h"
#include "ds/catch/catch.h"
#include "ds/log.h"
#include "model/apartment.h"
#include "model/dsmeter.h"
#include "tests/util/dss_instance_fixture.h"

static ds::LogChannel dsLogChannel("actionExecuteTest", ds::LogSeverity::DEBUG);
static const char* TAGS = "[dss][actionExecute]";

using namespace dss;
using namespace trompeloeil;
using trompeloeil::_;

namespace {
static const Dsuid dsuid1{"0000000000000000000000000000000001"};
static const Dsuid dsuid2{"0000000000000000000000000000000002"};
static const Dsuid dsuid3{"0000000000000000000000000000000003"};

void testPlayAudioActivityPbuf(const uint8_t* _data, uint _size, const std::string& _methodName,
                               const std::string& _dsuid) {
  vdcapi::Message msg;
  msg.set_type(vdcapi::VDSM_REQUEST_GENERIC_REQUEST);
  auto result = msg.ParseFromArray(_data, _size);
  REQUIRE(result == true);
  auto genericRequest = msg.mutable_vdsm_request_generic_request();
  REQUIRE(_methodName == genericRequest->methodname());
  REQUIRE(_dsuid == genericRequest->dsuid());
  auto pbufStr = msg.ShortDebugString();
  dsLogChannel.log(ds::LogSeverity::NOTICE, "Play audio activity pbuf " + pbufStr);
}

SCENARIO("ActionExecute play-audio-activity", TAGS) {
  DSSInstanceFixture fixture;
  Apartment& apartment(DSS::getInstance()->getApartment());
  std::shared_ptr<Event> pEvent;
  BusInterfaceMock busInterface;
  apartment.setBusInterface(&busInterface);
  std::shared_ptr<DSMeter> meter1 = apartment.allocateDSMeter(dsuid1);
  std::shared_ptr<DSMeter> meter2 = apartment.allocateDSMeter(dsuid2);

  meter1->setBusMemberType(BusMember_vDC);

  PropertySystem& propertySystem = *apartment.getPropertySystem();
  {
    auto sceneResponderEntryNode{propertySystem.createProperty("/scripts/system-addon-scene-responder/entries/0")};
    auto actionsNode{sceneResponderEntryNode->createProperty("actions")};
    auto actionNode{actionsNode->createProperty("0")};
    actionNode->setChildValue("type", std::string("play-audio-activity"));
    actionNode->setChildValue("dsuid", dsuid2str(dsuid1));
  }
  {
    auto sceneResponderEntryNode{propertySystem.createProperty("/scripts/system-addon-scene-responder/entries/1")};
    auto actionsNode{sceneResponderEntryNode->createProperty("actions")};
    auto actionNode{actionsNode->createProperty("0")};
    actionNode->setChildValue("type", std::string("play-audio-activity"));
    actionNode->setChildValue("dsuid", dsuid2str(dsuid2));
  }
  {
    auto sceneResponderEntryNode{propertySystem.createProperty("/scripts/system-addon-scene-responder/entries/2")};
    auto actionsNode{sceneResponderEntryNode->createProperty("actions")};
    auto actionNode{actionsNode->createProperty("0")};
    actionNode->setChildValue("type", std::string("play-audio-activity"));
    actionNode->setChildValue("dsuid", dsuid2str(dsuid3));
  }
  {
    auto sceneResponderEntryNode{propertySystem.createProperty("/scripts/system-addon-scene-responder/entries/3")};
    auto actionsNode{sceneResponderEntryNode->createProperty("actions")};
    auto actionNode{actionsNode->createProperty("0")};
    actionNode->setChildValue("type", std::string("play-audio-activity"));
    actionNode->setChildValue("dsuid", std::string());
  }
  {
    auto sceneResponderEntryNode{propertySystem.createProperty("/scripts/system-addon-scene-responder/entries/4")};
    auto actionsNode{sceneResponderEntryNode->createProperty("actions")};
    auto actionNode{actionsNode->createProperty("0")};
    actionNode->setChildValue("type", std::string("play-audio-activity"));
  }

  REQUIRE_CALL(busInterface.structureQuery, protobufMessageRequest(dsuid1, _, _, _, _))
      .TIMES(1)
      .SIDE_EFFECT(testPlayAudioActivityPbuf(_3, _2, "playAudioActivity", dsuid2str(dsuid1)));

  ActionExecute actionExecute(Properties{});
  actionExecute.execute("/scripts/system-addon-scene-responder/entries/0");

  actionExecute.execute("/scripts/system-addon-scene-responder/entries/1");
  actionExecute.execute("/scripts/system-addon-scene-responder/entries/2");
  actionExecute.execute("/scripts/system-addon-scene-responder/entries/3");
  actionExecute.execute("/scripts/system-addon-scene-responder/entries/4");
  FORBID_CALL(busInterface.structureQuery, protobufMessageRequest(_, _, _, _, _));
}

} // namespace
