/*
    Copyright (c) 2009,2010,2011 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

*/

#include "archives-storage.h"

#include <ds/log.h>

#include <boost/algorithm/string/predicate.hpp>
#include <filesystem>

#include "ds485types.h"
#include "metering/archives-connection.h"

static ds::LogChannel dsLogChannel("dssMeteringArchivesStorage");

using namespace std::literals;

using namespace dss;
using namespace dss::metering;

namespace dss::metering {

std::string filename(const ArchiveId& archiveId) {
  return ds::str(std::get<Dsuid>(archiveId),
                 (std::get<Category>(archiveId) == Category::consumption) ? "" : "-production", ".rrd");
}

constexpr auto omsPrefix = "open-metering-system-"sv;
constexpr auto modbusPrefix = "modbus-"sv;


std::string filename(const OpenMeteringSystemId& openMeteringSystemId) {
  return ds::str(omsPrefix, openMeteringSystemId, ".rrd");
}

std::string filename(const archives::ModbusMeteringArchiveId& modbusMeteringId) {
  return ds::str(modbusPrefix, modbusMeteringId.first, "-", modbusMeteringId.second, ".rrd");
}

namespace {

auto dsidLength = ds::str(dsid_t{}).size();
auto dsuidLength = ds::str(dsuid_t{}).size();
auto dsuidDashProductionLength = filename(ArchiveId{Dsuid{}, Category::production}).size() - ".rrd"sv.length();

class ArchivesStorageImpl final : public ArchivesStorage {
  ds::System& m_system;
  std::string m_storageLocation{"/invalid-store/"s}; // std::filesystem::path

public:
  ArchivesStorageImpl(ds::System& system) : m_system(system) {}
  virtual ~ArchivesStorageImpl() override = default;

  void setMeteringStorageLocation(std::string&& path) override { m_storageLocation = std::move(path); }

  bool hasArchive(const ArchiveId& archiveId) const override { return m_system.existsFile(filepath(archiveId)); }
  bool hasArchive(const OpenMeteringSystemId& openMeteringSystemId) const override {
    return m_system.existsFile(filepath(openMeteringSystemId));
  }

  bool hasArchive(const archives::ModbusMeteringArchiveId& modbusMeteringId) const override {
    return m_system.existsFile(filepath(modbusMeteringId));
  }

  std::string filepath(const ArchiveId& archiveId) const override {
    auto path = std::filesystem::path(m_storageLocation);
    path /= std::filesystem::path(filename(archiveId));
    return path.string();
  }

  std::string filepath(const OpenMeteringSystemId& openMeteringSystemId) const override {
    auto path = std::filesystem::path(m_storageLocation);
    path /= std::filesystem::path(filename(openMeteringSystemId));
    return path.string();
  }

  std::string filepath(const archives::ModbusMeteringArchiveId& modbusMeteringId) const override {
    auto path = std::filesystem::path(m_storageLocation);
    path /= std::filesystem::path(filename(modbusMeteringId));
    return path.string();
  }

  std::vector<OpenMeteringSystemId> searchOmsArchives() const override;

  std::string storageLocation() const override { return m_storageLocation; };

  std::vector<std::filesystem::path> searchRrdFiles() const override {
    decltype(searchRrdFiles()) result{};
    for (auto file : m_system.lsRegularFiles(m_storageLocation, ".rrd"s)) {
      result.push_back(std::filesystem::path{file});
    }
    return result;
  }

  std::vector<std::pair<CompoundId, std::string>> searchArchives() const override {
    decltype(searchArchives()) result{};
    for (auto file : searchRrdFiles()) {
      try {
        auto stem = std::string(file.stem());
        if (boost::starts_with(stem, omsPrefix)) {
          // TODO check serial non null, or drop serial number for name
          auto serial = stem.substr(omsPrefix.length());
          result.push_back({OpenMeteringSystemId{serial}, file});
        } else if (boost::starts_with(stem, modbusPrefix)) {
          // TODO check serial non null, or drop serial number for name
          auto serial = stem.substr(modbusPrefix.length());
          result.push_back({ModbusMeteringId{serial}, file});
        } else if (stem.size() == dsidLength) {
          result.push_back({str2dsid(stem), file});
        } else if (stem.size() == dsuidLength) {
          result.push_back({ArchiveId{str2dsuid(stem), Category::consumption}, file});
        } else if (stem.size() == dsuidDashProductionLength) {
          auto dsuid = str2dsuid(stem.substr(0, dsuidLength));
          if (stem.rfind("-production"sv) == std::string::npos) {
            DS_NOTICE("suffix mismatch", file);
            continue;
          }
          result.push_back({ArchiveId{dsuid, Category::production}, file});
        } else {
          DS_WARNING("searchArchives: unknown rrd", file);
        }
      } catch (std::exception& e) {
        // ignore file, could be downgrade from future version
        DS_WARNING("searchArchives", file, e);
      }
    }
    return result;
  }

private:
};

std::vector<OpenMeteringSystemId> ArchivesStorageImpl::searchOmsArchives() const {
  std::vector<OpenMeteringSystemId> ret;

  for (auto file : m_system.lsRegularFiles(m_storageLocation, ".rrd"s)) {
    auto stem = std::string(std::filesystem::path(file).stem());
    // check for dsuid archives, and remove others
    if (boost::starts_with(stem, omsPrefix)) {
      auto serial = stem.substr(omsPrefix.length());
      ret.push_back(OpenMeteringSystemId{serial});
    }
  }
  return ret;
}

} // namespace

std::unique_ptr<ArchivesStorage> makeArchivesStorage(ds::System& system) {
  return std::make_unique<ArchivesStorageImpl>(system);
}

} // namespace dss::metering
