/*
    Copyright (c) 2022 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/

#include "controller-direction.h"

#include <ds/catch/catch.h>
#include <ds/log.h>
#include <ds/system-mock.h>

#include "archives-connection-mock.h"
#include "archives-storage-mock.h"
#include "archives/controller-mock.h"
#include "config.h"
#include "model/apartment-mock.h"

using namespace trompeloeil;
using namespace std::literals;

using dss::Dsuid;
using namespace dss::metering;
using namespace dss::metering::archives;

using system_clock = std::chrono::system_clock;

static ds::LogChannel dsLogChannel("dssMeteringControllderTest");

namespace {

static const char* TAGS = "[dSS][metering][controllder]";

SCENARIO("controller", TAGS) {
  dss::ApartmentMock apartment{};
  ArchivesStorageMock archivesStorage{};
  auto controllerArchives_ = std::make_unique<ControllerArchivesMock>();
  auto& controllerArchives = *controllerArchives_;
  auto system = ds::SystemMock();

  auto controllerDirection = makeMockedControllerDirection(apartment, std::move(controllerArchives_), archivesStorage,
                                                           Category::consumption, system);

  auto createController = [&apartment](const dsModel::ControllerId& controllerId) {
    static int ct{};
    auto controller = apartment.allocateDSMeter(dss::Dsuid{controllerId});
    controller->setName(ds::str("controller", ct++));
    controller->setIsPresent(true);
    controller->setIsConnected(true); // isPresent checks isConnected internally
    controller->setCapability_HasMetering(true);
    return controller;
  };

  auto controllerId1 = dsModel::ControllerId{"1111111111111111111111111111111100"};
  auto controller1 = createController(controllerId1);

  auto populateArchive = [&controllerDirection, &archivesStorage, &controllerArchives](auto& controller,
                                                                                       ArchiveValue&& archiveValue) {
    REQUIRE_CALL(archivesStorage, hasArchive(ArchiveId{controller.id(), Category::consumption}))
        .RETURN(true); // -> m_meters.push_back
    REQUIRE_CALL(controllerArchives, lastValue(_)).RETURN(archiveValue);
    controllerDirection->getEnergy(controller.getDSID());
  };

  auto mockPath = [](const auto& archiveId) { return ds::str("/storage/", archiveId, ".rrd"s); };

  SECTION("needsFlushing") {
    populateArchive(*controller1, {20, 3600, ds::DateTime(10s)});
    ElectricityMeteringConfig config{};

    WHEN("interval expired") {
      REQUIRE_CALL(controllerArchives, lastTimestamp(ArchiveId{Dsuid{controllerId1}, Category::consumption}))
          .RETURN(ds::DateTime{0s});
      REQUIRE_CALL(archivesStorage, filepath(ArchiveId{Dsuid{controllerId1}, Category::consumption}))
          .RETURN(mockPath(_1));
      CHECK(controllerDirection->needsFlushing(ds::DateTime(config.diskFlushInterval + 1s)) ==
            std::vector{"/storage/(1111111111111111111111111111111100,consumption).rrd"s});
    }
  }

  SECTION("validateAndConvert") {
    auto controllerId1 = dsModel::ControllerId{"1111111111111111111111111111111100"};
    auto y = GENERATE(Category::consumption, Category::production);
    GIVEN("archive is up-to-date") {
      REQUIRE_CALL(archivesStorage, hasArchive(ArchiveId{controllerId1, y})).RETURN(true);
      REQUIRE_CALL(controllerArchives, validateDBSeries(_, _)).RETURN(true);
      controllerDirection->validateAndConvertArchive({Dsuid{controllerId1}, y});
    }
    GIVEN("archive is invalid") {
      REQUIRE_CALL(archivesStorage, hasArchive(ArchiveId{controllerId1, y})).RETURN(true);
      REQUIRE_CALL(controllerArchives, validateDBSeries(_, _)).RETURN(false);
      REQUIRE_CALL(archivesStorage, filepath(ArchiveId{controllerId1, y}))
          .RETURN(ds::str("/storage-location/", _1, ".rrd"));
      REQUIRE_CALL(system, remove(_));
      controllerDirection->validateAndConvertArchive({Dsuid{controllerId1}, y});
    }
    GIVEN("archive needs conversion ") {
      REQUIRE_CALL(archivesStorage, hasArchive(ArchiveId{controllerId1, y})).RETURN(true);
      REQUIRE_CALL(controllerArchives, validateDBSeries(_, _)).LR_SIDE_EFFECT(_2 = true).RETURN(true);
      REQUIRE_CALL(controllerArchives, tuneDBPowerSettings(_));
      controllerDirection->validateAndConvertArchive({Dsuid{controllerId1}, y});
    }
  }

  SECTION("getSeries") {
    GIVEN("valid archives") {
      REQUIRE_CALL(archivesStorage, hasArchive(ArchiveId{controllerId1, Category::consumption})).RETURN(true);
      REQUIRE_CALL(controllerArchives, flush(std::vector{ArchiveId{controllerId1, Category::consumption}}));
      REQUIRE_CALL(controllerArchives, xport(1, 601, 1, std::vector{ArchiveId{controllerId1, Category::consumption}},
                                             SeriesTypes::power, false))
          .RETURN([]() -> ValueRange {
            std::vector<Value> ret;
            while (ret.size() < 600) {
              ret.push_back(Value{1, ds::DateTime{1s}}); // all valid data, incl. last entry
            }
            return ValueRange{.step = 1, .values = std::move(ret)};
          }());

      WHEN("single valid controller") {
        auto range = controllerDirection->getSeries({Dsuid{controllerId1}}, 1s, SeriesTypes::power, false,
                                                    {ds::DateTime{1s}, ds::DateTime{600s}}, 0);
        CHECK(range.values.size() == 600);
      }

      WHEN("add ghost controllers") {
        // metering disabled -> no data
        auto controllerId2 = dsModel::ControllerId{"1111111111111111111111111111111200"};
        REQUIRE_CALL(archivesStorage, hasArchive(ArchiveId{controllerId2, Category::consumption})).RETURN(false);
        auto controller2 = apartment.allocateDSMeter(dss::Dsuid{controllerId2});

        // metering but no data
        auto controllerId3 = dsModel::ControllerId{"1111111111111111111111111111111300"};
        auto controller3 = apartment.allocateDSMeter(dss::Dsuid{controllerId3});
        controller3->setCapability_HasMetering(true);
        REQUIRE_CALL(archivesStorage, hasArchive(ArchiveId{controllerId3, Category::consumption})).RETURN(false);

        THEN("same as above") {
          auto range =
              controllerDirection->getSeries({Dsuid{controllerId1}, Dsuid{controllerId2}, Dsuid{controllerId3}}, 1s,
                                             SeriesTypes::power, false, {ds::DateTime{1s}, ds::DateTime{600s}}, 0);
          CHECK(range.values.size() == 600);
        }
      }
    }
    WHEN("empty list of dsmid given") {
      apartment.removeDSMeter(dss::Dsuid{controllerId1});
      THEN("no data returned") {
        auto range = controllerDirection->getSeries({}, 1s, SeriesTypes::power, false,
                                                    {ds::DateTime{1s}, ds::DateTime{600s}}, 0);
        CHECK(range == ValueRange{});
      }
    }
  }

  GIVEN("new measurements") {
    // first only sets baseline to compute delta, only RAM
    controllerDirection->addEnergy(Dsuid{controllerId1}, {75, 100, ds::DateTime(100s)});

    WHEN("archive already exists") {
      THEN("value is added") {
        REQUIRE_CALL(archivesStorage, hasArchive(ArchiveId{controllerId1, Category::consumption})).RETURN(true);
        REQUIRE_CALL(controllerArchives, lastValue(_))
            .RETURN(ArchiveValue{.power = 0.0, .energy = 50.0, .lastChanged = ds::DateTime{0s}});
        REQUIRE_CALL(controllerArchives,
                     addValue(ArchiveId{Dsuid{"1111111111111111111111111111111100"s}, Category::consumption},
                              ArchiveValue{.power = 75.0, .energy = 250.0, .lastChanged = ds::DateTime{105s}}));
        controllerDirection->addEnergy(Dsuid{controllerId1}, {75, 300, ds::DateTime(105s)});
      }
    }
    WHEN("archive needs creation") {
      REQUIRE_CALL(archivesStorage, hasArchive(ArchiveId{controllerId1, Category::consumption}))
          .RETURN(false)
          .TIMES(2); // getEnergy, getOrCreateCachedSeries
      REQUIRE_CALL(controllerArchives, create(_));
      REQUIRE_CALL(controllerArchives,
                   addValue(ArchiveId{Dsuid{"1111111111111111111111111111111100"s}, Category::consumption},
                            ArchiveValue{.power = 75, .energy = 200.0, .lastChanged = ds::DateTime{101s}}));
      controllerDirection->addEnergy(Dsuid{controllerId1}, {75, 300, ds::DateTime(101s)});
    }

    SECTION("jumps in consumed energy detected") {
      WHEN("timestamp of new measurement predates previous") {
        THEN("nothing added to archive") {
          ALLOW_CALL(archivesStorage, hasArchive(ArchiveId{controllerId1, Category::consumption}))
              .RETURN(false); // property tree update
          controllerDirection->addEnergy(Dsuid{controllerId1}, {75, 300, ds::DateTime(99s)});
        }
      }
      WHEN("new energy decreases") {
        THEN("nothing added to archive") {
          ALLOW_CALL(archivesStorage, hasArchive(ArchiveId{controllerId1, Category::consumption}))
              .RETURN(false); // property tree update
          controllerDirection->addEnergy(Dsuid{controllerId1}, {75, 99, ds::DateTime(101s)});
        }
      }
      THEN("nothing added to archive if limit exceeded") {
        ALLOW_CALL(archivesStorage, hasArchive(ArchiveId{controllerId1, Category::consumption}))
            .RETURN(false); // property tree update
        controllerDirection->addEnergy(Dsuid{controllerId1}, {75, 8000 + 100, ds::DateTime(101s)});
      }
      THEN("archive is updated if below limit") {
        REQUIRE_CALL(controllerArchives, create(_));
        REQUIRE_CALL(controllerArchives, addValue(_, _));
        ALLOW_CALL(archivesStorage, hasArchive(ArchiveId{controllerId1, Category::consumption}))
            .RETURN(false); // property tree update
        controllerDirection->addEnergy(Dsuid{controllerId1}, {75, 8000 + 99, ds::DateTime(101s)});
      }
      WHEN("updates have been lost") {
        THEN("nothing added to archive if more then 5s") {
          ALLOW_CALL(archivesStorage, hasArchive(ArchiveId{controllerId1, Category::consumption}))
              .RETURN(false); // property tree update
          controllerDirection->addEnergy(Dsuid{controllerId1}, {75, 300, ds::DateTime(106s)});
        }
      }
    }
  }

  GIVEN("getEnergy") {
    WHEN("archive does not exist") {
      REQUIRE_CALL(archivesStorage, hasArchive(ArchiveId{controllerId1, Category::consumption})).RETURN(false);
      THEN("no data return") { CHECK(!controllerDirection->getEnergy(dss::Dsuid{controllerId1})); }
    }
    WHEN("archive exists") {
      REQUIRE_CALL(archivesStorage, hasArchive(ArchiveId{controllerId1, Category::consumption})).RETURN(true);

      WHEN("accessed first time") {
        REQUIRE_CALL(controllerArchives, lastValue(_))
            .RETURN(ArchiveValue{.power = 0.0, .energy = 50.0, .lastChanged = ds::DateTime{10s}});
        THEN("data returned from DB") {
          CHECK(controllerDirection->getEnergy(dss::Dsuid{controllerId1}) ==
                ArchiveValue{.power = 0, .energy = 50.0, .lastChanged = ds::DateTime(10s)});
        }
      }
      WHEN("lastValue fails") {
        REQUIRE_CALL(controllerArchives, lastValue(_)).RETURN(std::nullopt);
        THEN("return is nullopt") { CHECK(!controllerDirection->getEnergy(dss::Dsuid{controllerId1})); }
      }
    }
  }

  SECTION("getLastReading") {
    using dsModel::MeteringType;

    populateArchive(*controller1, {20, 3600, ds::DateTime(10s)});

    auto controllerId2 = dsModel::ControllerId{"1111111111111111111111111111111200"};
    auto controller2 = createController(controllerId2);
    populateArchive(*controller2, {10, 1800, ds::DateTime(10s)});

    WHEN("single dsuid / EnergyMetering") {
      CHECK(controllerDirection->getLastReading(controller1->getDSID(), MeteringType::energyMetering) ==
            LastReading{MeteringType::energyMetering, 1.0, ds::DateTime{10s}}); // Wh is default
      CHECK(controllerDirection->getLastReading(controller1->getDSID(), MeteringType::energyMetering, EnergyUnit::Ws) ==
            LastReading{MeteringType::energyMetering, 3600.0, ds::DateTime{10s}});
      CHECK(controllerDirection->getLastReading(controller1->getDSID(), MeteringType::energyMetering, EnergyUnit::Wh) ==
            LastReading{MeteringType::energyMetering, 1.0, ds::DateTime{10s}});
    }
    WHEN("apartment / EnergyMetering") {
      CHECK(controllerDirection->getLastReadingApartment(MeteringType::energyMetering, EnergyUnit::Ws) ==
            LastReading{MeteringType::energyMetering, 5400.0, ds::DateTime{10s}});
      CHECK(controllerDirection->getLastReadingApartment(MeteringType::energyMetering, EnergyUnit::Wh) ==
            LastReading{MeteringType::energyMetering, 1.5, ds::DateTime{10s}});
      CHECK(controllerDirection->getLastReadingApartment(MeteringType::powerMetering) ==
            LastReading{MeteringType::powerMetering, 30.0, ds::DateTime{10s}});
    }
  }

  SECTION("15min average uploads") {
    REQUIRE_CALL(archivesStorage, hasArchive(ArchiveId{controllerId1, Category::consumption})).RETURN(true);
    REQUIRE_CALL(controllerArchives, flush(std::vector{ArchiveId{controllerId1, Category::consumption}}));

    WHEN("no data present") {
      REQUIRE_CALL(controllerArchives,
                   xport(_, _, std::chrono::duration_cast<std::chrono::seconds>(15min).count(),
                         std::vector{ArchiveId{controllerId1, Category::consumption}}, SeriesTypes::energyDelta, true))
          .RETURN(ValueRange{.start = 1, .end = 901, .step = 900, .values = {}}); // empty
      auto ret =
          controllerDirection->get15MinuteAverage({dss::Dsuid{controllerId1}}, SeriesTypes::energyDelta, Unit::Wh);
      CHECK(ret == 0.0);
    }
    WHEN("NaN value found") {
      THEN("exception thrown") { // rely on getSeries to filter them out
        REQUIRE_CALL(controllerArchives, xport(_, _, std::chrono::duration_cast<std::chrono::seconds>(15min).count(),
                                               std::vector{ArchiveId{controllerId1, Category::consumption}},
                                               SeriesTypes::energyDelta, true))
            .RETURN(ValueRange{1, 901, 900, {{std::nan("myNaN"), ds::DateTime{0s}}}});
        CHECK_THROWS(
            controllerDirection->get15MinuteAverage({dss::Dsuid{controllerId1}}, SeriesTypes::energyDelta, Unit::Wh));
      }
    }
  }

  SECTION("meterings") {
    using dsModel::MeteringType;

    GIVEN("consumption meter") {
      populateArchive(*controller1, {75, 100, ds::DateTime(100s)});

      THEN("consumptionn meterings exist") {
        CHECK(controllerDirection->meterings() ==
              std::vector{std::pair<Dsuid, MeteringType>{Dsuid{controllerId1}, MeteringType::powerMetering},
                          std::pair<Dsuid, MeteringType>{Dsuid{controllerId1}, MeteringType::energyMetering}});
      }
      WHEN("meter is not present") {
        controller1->setIsPresent(false);
        CHECK(controllerDirection->meterings() == std::vector<std::pair<Dsuid, MeteringType>>{});
      }
      SECTION("meter does not have metering capability") {
        controller1->setCapability_HasMetering(false);
        CHECK(controllerDirection->meterings() == std::vector<std::pair<Dsuid, MeteringType>>{});
      }
    }
    GIVEN("production meter") {
      auto controllerArchives2_ = std::make_unique<ControllerArchivesMock>();
      auto& controllerArchives2 = *controllerArchives2_;

      auto controllerDirectionProduction = makeMockedControllerDirection(apartment, std::move(controllerArchives2_),
                                                                         archivesStorage, Category::production, system);

      auto populateProductionArchive = [&controllerDirectionProduction, &archivesStorage, &controllerArchives2](
                                           auto& controller, ArchiveValue&& archiveValue) {
        // controller.setCapability_HasMeteringProducerEnabled(false);
        REQUIRE_CALL(archivesStorage, hasArchive(ArchiveId{controller.id(), Category::production})).RETURN(true);
        REQUIRE_CALL(controllerArchives2, lastValue(_)).RETURN(archiveValue);
        controllerDirectionProduction->getEnergy(controller.getDSID());
      };

      populateProductionArchive(*controller1, {75, 100, ds::DateTime(101s)});
      controller1->setCapability_HasMeteringProducer(true);
      controller1->setCapability_HasMeteringProducerEnabled(true);

      THEN("production meterings exist") {
        CHECK(controllerDirectionProduction->meterings() ==
              std::vector{std::pair<Dsuid, MeteringType>{Dsuid{controllerId1}, MeteringType::powerProducedMetering},
                          std::pair<Dsuid, MeteringType>{Dsuid{controllerId1}, MeteringType::energyProducedMetering}});
      }
      WHEN("meter does not have metering capability") {
        controller1->setCapability_HasMetering(false);
        // no change
        CHECK(controllerDirectionProduction->meterings() ==
              std::vector{std::pair<Dsuid, MeteringType>{Dsuid{controllerId1}, MeteringType::powerProducedMetering},
                          std::pair<Dsuid, MeteringType>{Dsuid{controllerId1}, MeteringType::energyProducedMetering}});
      }
      WHEN("meter does not have metering producer enabled") {
        controller1->setCapability_HasMeteringProducerEnabled(false);
        CHECK(controllerDirectionProduction->meterings() == std::vector<std::pair<Dsuid, MeteringType>>{});
      }
    }
  }

  GIVEN("empty rrd / this test is never modifying to round-robin archive") {
    WHEN("1st sensor value is received") {
      controllerDirection->addEnergy(Dsuid{controllerId1}, {.power = 0, .energy = 10, .timestamp = ds::DateTime(300s)});

      THEN("total energy counter is not updated yet") {
        // cache is not populated upon 1st value, also rrd file does not exist
        REQUIRE_CALL(archivesStorage, hasArchive(ArchiveId{controllerId1, Category::consumption})).RETURN(false);
        CHECK(!controllerDirection->getEnergy(Dsuid(controllerId1)).has_value());
      }

      WHEN("2nd sensor value is reiceved") {
        REQUIRE_CALL(archivesStorage, hasArchive(ArchiveId{controllerId1, Category::consumption})).RETURN(true);
        REQUIRE_CALL(controllerArchives, lastValue(_))
            .RETURN(ArchiveValue{.power = 0.0, .energy = 0.0, .lastChanged = ds::DateTime{0s}});
        REQUIRE_CALL(controllerArchives,
                     addValue(_, ArchiveValue{.power = 1, .energy = 2, .lastChanged = ds::DateTime(301s)}));
        controllerDirection->addEnergy(Dsuid{controllerId1},
                                       {.power = 1, .energy = 12, .timestamp = ds::DateTime(301s)});
        THEN("energy counter is increased by (2nd - 1st) sensor value") {
          CHECK(controllerDirection->getEnergy(Dsuid(controllerId1)) ==
                ArchiveValue{.power = 1, .energy = 2.0, .lastChanged = ds::DateTime(301s)});
        }

        WHEN("sensor value is smaller then previous sensor value") {
          // sets new baseline
          controllerDirection->addEnergy(Dsuid{controllerId1},
                                         {.power = 1, .energy = 2, .timestamp = ds::DateTime(302s)});

          THEN("energy counter is not changed") {
            CHECK(controllerDirection->getEnergy(Dsuid(controllerId1)) ==
                  ArchiveValue{.power = 1, .energy = 2.0, .lastChanged = ds::DateTime(301s)});
          }

          WHEN("next sensor value is bigger then previous") {
            REQUIRE_CALL(controllerArchives,
                         addValue(_, ArchiveValue{.power = 2, .energy = 3.0, .lastChanged = ds::DateTime(303s)}));
            controllerDirection->addEnergy(Dsuid{controllerId1},
                                           {.power = 2, .energy = 3, .timestamp = ds::DateTime(303s)}); // 1 - 0
            THEN("energy counter is increased again") {
              CHECK(controllerDirection->getEnergy(Dsuid(controllerId1)) ==
                    ArchiveValue{.power = 2, .energy = 3.0, .lastChanged = ds::DateTime(303s)});
            }
          }
        }
      }
    }
  }

  GIVEN("non-empty rrd / this test is never modifying to round-robin archive") {
    REQUIRE_CALL(archivesStorage, hasArchive(ArchiveId{controllerId1, Category::consumption})).RETURN(true);
    REQUIRE_CALL(controllerArchives, lastValue(_))
        .RETURN(ArchiveValue{.power = 0.0, .energy = 12.5, .lastChanged = ds::DateTime{300s}});

    THEN("last value is known") {
      CHECK(controllerDirection->getEnergy(Dsuid(controllerId1)) ==
            ArchiveValue{.power = 0, .energy = 12.5, .lastChanged = ds::DateTime(300s)});
    }
  }

  GIVEN("archive with UINT_MAX") {
    REQUIRE_CALL(archivesStorage, hasArchive(ArchiveId{controllerId1, Category::consumption})).RETURN(true);
    REQUIRE_CALL(controllerArchives, lastValue(_))
        .RETURN(ArchiveValue{.power = 0.0, .energy = UINT_MAX, .lastChanged = ds::DateTime{0s}});
    REQUIRE_CALL(controllerArchives, addValue(_, _))
        .LR_SIDE_EFFECT(CHECK(_2.energy == static_cast<double>(UINT_MAX) + 1));

    THEN("energy counter is increased by (2nd - 1st) sensor value") {
      controllerDirection->addEnergy(Dsuid{controllerId1}, {.power = 1, .energy = 12, .timestamp = ds::DateTime(300s)});
      controllerDirection->addEnergy(Dsuid{controllerId1}, {.power = 1, .energy = 13, .timestamp = ds::DateTime(301s)});
    }
  }
}

SCENARIO("energy production 2", TAGS) {
  dss::ApartmentMock apartment{};
  ArchivesStorageMock archivesStorage{};
  auto system = ds::SystemMock();

  auto controllerArchives_ = std::make_unique<ControllerArchivesMock>();
  auto& controllerArchives = *controllerArchives_;

  auto controllerDirection = makeMockedControllerDirection(apartment, std::move(controllerArchives_), archivesStorage,
                                                           Category::production, system);

  auto createController = [&apartment](const dsModel::ControllerId& controllerId) {
    static int ct{};
    auto controller = apartment.allocateDSMeter(dss::Dsuid{controllerId});
    controller->setName(ds::str("controller", ct++));
    controller->setIsPresent(true);
    controller->setIsConnected(true); // isPresent checks isConnected internally
    controller->setCapability_HasMetering(true);
    return controller;
  };

  auto controllerId1 = dsModel::ControllerId{"1111111111111111111111111111111100"};
  auto controller1 = createController(controllerId1);

  SECTION("addProducedEnergy") {
    // online sets baseline
    controllerDirection->addEnergy(Dsuid{controllerId1}, {75, 100, ds::DateTime(100s)}); // baseline

    // second call will start adding values
    REQUIRE_CALL(archivesStorage, hasArchive(ArchiveId{controllerId1, Category::production})).RETURN(true);
    REQUIRE_CALL(controllerArchives, lastValue(_))
        .RETURN(ArchiveValue{.power = 0.0, .energy = 50.0, .lastChanged = ds::DateTime{0s}});
    REQUIRE_CALL(controllerArchives,
                 addValue(ArchiveId{Dsuid{"1111111111111111111111111111111100"s}, Category::production},
                          ArchiveValue{.power = 115.0, .energy = 260.0, .lastChanged = ds::DateTime{101s}}));
    controllerDirection->addEnergy(Dsuid{controllerId1}, {115, 310, ds::DateTime(101s)});

    THEN("getEnergyValue last value is known") {
      CHECK(controllerDirection->getEnergy(Dsuid(controllerId1)) ==
            ArchiveValue{.power = 115, .energy = 260, .lastChanged = ds::DateTime(101s)});
    }

    SECTION("jumps in consumed energy detected") {
      WHEN("timestamp of new measurement predates previous") {
        THEN("nothing added to archive") {
          controllerDirection->addEnergy(Dsuid{controllerId1}, {75, 10000 + 310, ds::DateTime(99s)});
        }
      }
      WHEN("timestamp of new measurement equals previous") {
        THEN("nothing added to archive") {
          controllerDirection->addEnergy(Dsuid{controllerId1}, {75, 49999 + 310, ds::DateTime(101s)});
        }
      }
      WHEN("new timestamp is a few millisconds less than a full second") {
        THEN("value is added to archive") {
          REQUIRE_CALL(controllerArchives, addValue(_, _));
          controllerDirection->addEnergy(
              Dsuid{controllerId1},
              {75, gsl::narrow_cast<unsigned int>(670.0 * 49999 / 1000 + 310), ds::DateTime(101s + 670ms)});
        }
      }
      WHEN("when more then 5s pass bewtween 2 updates") {
        THEN("nothing added to archive") {
          controllerDirection->addEnergy(Dsuid{controllerId1}, {75, 49999 + 310, ds::DateTime(101s + 6s)});
        }
      }
      THEN("nothing added to archive if limit exceeded") {
        controllerDirection->addEnergy(Dsuid{controllerId1}, {75, 50000 + 310, ds::DateTime(102s)});
      }
      THEN("archive is updated if below limit") {
        REQUIRE_CALL(controllerArchives, addValue(_, _));
        controllerDirection->addEnergy(Dsuid{controllerId1}, {75, 49999 + 310, ds::DateTime(102s)});
      }
      WHEN("an update is lost") {
        auto x = GENERATE(2s, 3s, 4s, 5s);
        THEN("limit is multiplied by the delay upon next update") {
          INFO(ds::str(x));
          REQUIRE_CALL(controllerArchives, addValue(_, _));
          controllerDirection->addEnergy(
              Dsuid{controllerId1},
              {75, gsl::narrow_cast<double>(x.count() * 50000 + 310 - 1), ds::DateTime(101s + x)});
        }
      }
    }
  }

  WHEN("getProducedSeries") {
    REQUIRE_CALL(archivesStorage, hasArchive(ArchiveId{controllerId1, Category::production})).RETURN(true);
    REQUIRE_CALL(controllerArchives, flush(std::vector{ArchiveId{Dsuid{controllerId1}, Category::production}}));
    REQUIRE_CALL(controllerArchives, xport(1, 601, 1, std::vector{ArchiveId(controllerId1, Category::production)},
                                           SeriesTypes::power, false))
        .RETURN([]() -> ValueRange {
          std::vector<Value> ret;
          while (ret.size() < 600) {
            ret.push_back(Value{1, ds::DateTime{1s}}); // all valid data, incl. last entry
          }
          return ValueRange{.step = 1, .values = std::move(ret)};
        }());

    WHEN("single valid controller") {
      auto range = controllerDirection->getSeries({Dsuid{controllerId1}}, 1s, SeriesTypes::power, false,
                                                  {ds::DateTime{1s}, ds::DateTime{600s}}, 0);
      CHECK(range.values.size() == 600);
    }
  }
}

SCENARIO("getSeries start/end", TAGS) {
  dss::ApartmentMock apartment{};
  ArchivesStorageMock archivesStorage{};
  auto system = ds::SystemMock();

  auto controllerArchives_ = std::make_unique<ControllerArchivesMock>();
  auto& controllerArchives = *controllerArchives_;
  auto controllerDirection = makeMockedControllerDirection(apartment, std::move(controllerArchives_), archivesStorage,
                                                           Category::consumption, system);

  auto controllerId1 = dsModel::ControllerId{"1111111111111111111111111111111100"};
  auto controller1 = apartment.allocateDSMeter(dss::Dsuid{controllerId1});

  ALLOW_CALL(controllerArchives, flush(std::vector{ArchiveId{Dsuid{controllerId1}, Category::consumption}}));
  REQUIRE_CALL(archivesStorage, hasArchive(ArchiveId{controllerId1, Category::consumption})).RETURN(true);

  GIVEN("start/end") {
    WHEN("resolution=1") {
      REQUIRE_CALL(controllerArchives, xport(1, 601, 1, _, _, _))
          .RETURN(ValueRange{.step = 1, .values = std::vector<Value>(600)});
      auto values = controllerDirection->getSeries({Dsuid{controllerId1}}, 1s, SeriesTypes::power, false,
                                                   {ds::DateTime{1s}, ds::DateTime{600s}}, 0);
      WHEN("values < max values for given resolution") {
        REQUIRE_CALL(controllerArchives, xport(1, 601, 1, _, _, _))
            .RETURN(ValueRange{.step = 1, .values = std::vector<Value>(10, {1, ds::DateTime{1s}})});
        auto range = controllerDirection->getSeries({Dsuid{controllerId1}}, 1s, SeriesTypes::power, false,
                                                    {ds::DateTime{1s}, ds::DateTime{600s}}, 10);
        CHECK(range.values.size() == 10); // no change, more values than valueCount is returned
      }
    }

    WHEN("resolution=60") {
      REQUIRE_CALL(controllerArchives, xport(0, 660, 60, _, _, _))
          .RETURN(ValueRange{.step = 60, .values = std::vector<Value>(720, {1, ds::DateTime{1s}})});
      auto range = controllerDirection->getSeries({Dsuid{controllerId1}}, 60s, SeriesTypes::power, false,
                                                  {ds::DateTime{1s}, ds::DateTime{600s}}, 0);
      CHECK(range.values.size() == 720);
    }
  }

  GIVEN("start/values") {
    WHEN("resolution=1") {
      REQUIRE_CALL(controllerArchives, xport(1, 102, 1, _, _, _))
          .RETURN(ValueRange{.step = 1, .values = std::vector<Value>(100, {1, ds::DateTime{1s}})});
      auto range = controllerDirection->getSeries({Dsuid{controllerId1}}, 1s, SeriesTypes::power, false,
                                                  {ds::DateTime{1s}, std::nullopt}, 100);
      CHECK(range.step == 1);
      CHECK(range.values.size() == 100);

      WHEN("unlimited values") {
        auto now = system_clock::now();

        // bug: end time is derived from now, start is not adjusted
        REQUIRE_CALL(controllerArchives, xport(1, _, 1, _, _, _))
            .LR_SIDE_EFFECT(CHECK(system_clock::to_time_t(now) <= _2))
            .RETURN(ValueRange{.step = 1, .values = std::vector<Value>(600, {1, ds::DateTime{1s}})});
        auto range = controllerDirection->getSeries({Dsuid{controllerId1}}, 1s, SeriesTypes::power, false,
                                                    {ds::DateTime{1s}, std::nullopt}, 0);
        CHECK(range.step == 1);
        CHECK(range.values.size() == 600); // real check is start/end in REQUIRE_CALL
      }

      WHEN("more values then available") {
        REQUIRE_CALL(controllerArchives, xport(1, _, 1, _, _, _))
            .RETURN(ValueRange{.step = 1, .values = std::vector<Value>(600)});
        auto range = controllerDirection->getSeries({Dsuid{controllerId1}}, 1s, SeriesTypes::power, false,
                                                    {ds::DateTime{1s}, std::nullopt}, 1000);
        CHECK(range.step == 1);
        CHECK(range.values.size() == 599); // off-by-one due to null Value
      }
    }

    WHEN("resolution=60") {
      REQUIRE_CALL(controllerArchives, xport(0, 6060, 60, _, _, _))
          .RETURN(ValueRange{.step = 60, .values = std::vector<Value>(100, {1, ds::DateTime{1s}})});
      auto range = controllerDirection->getSeries({Dsuid{controllerId1}}, 60s, SeriesTypes::power, false,
                                                  {ds::DateTime{1s}, std::nullopt}, 100);
      CHECK(range.step == 60);
      CHECK(range.values.size() == 100);

      WHEN("unlimited values") {
        auto now = system_clock::now();

        // bug: end is derived form now(), not relative to end
        REQUIRE_CALL(controllerArchives, xport(0, _, 60, _, _, _))
            .LR_SIDE_EFFECT(CHECK(system_clock::to_time_t(now) - (720 + 1) * 60 < _2))
            .RETURN(ValueRange{.step = 60, .values = std::vector<Value>(720, {1, ds::DateTime{1s}})});
        auto range = controllerDirection->getSeries({Dsuid{controllerId1}}, 60s, SeriesTypes::power, false,
                                                    {ds::DateTime{1s}, std::nullopt}, 0);
        CHECK(range.values.size() == 720); // real check is start/end in REQUIRE_CALL
        CHECK(range.step == 60);
      }
    }
  }

  GIVEN("end/values") {
    WHEN("resolution=1") {
      REQUIRE_CALL(controllerArchives, xport(500, 601, 1, _, _, _))
          .RETURN(ValueRange{.step = 1, .values = std::vector<Value>(100, {1, ds::DateTime{1s}})});
      auto range = controllerDirection->getSeries({Dsuid{controllerId1}}, 1s, SeriesTypes::power, false,
                                                  {std::nullopt, ds::DateTime{600s}}, 100);
      CHECK(range.step == 1);
      CHECK(range.values.size() == 100);

      WHEN("request more values then avail") {
        REQUIRE_CALL(controllerArchives, xport(0, 601, 1, _, _, _))
            .RETURN(ValueRange{.step = 1, .values = std::vector<Value>(600, {1, ds::DateTime{1s}})});
        auto range = controllerDirection->getSeries({Dsuid{controllerId1}}, 1s, SeriesTypes::power, false,
                                                    {std::nullopt, ds::DateTime{600s}}, 1000);
        CHECK(range.step == 1);
        CHECK(range.values.size() == 600);
      }
    }
    WHEN("resolution=60") {
      REQUIRE_CALL(controllerArchives, xport(960, 7020, 60, _, _, _))
          .RETURN(ValueRange{.step = 60, .values = std::vector<Value>(100, {1, ds::DateTime{1s}})});
      auto range = controllerDirection->getSeries({Dsuid{controllerId1}}, 60s, SeriesTypes::power, false,
                                                  {std::nullopt, ds::DateTime{7000s}}, 100);
      CHECK(range.step == 60);
      CHECK(range.values.size() == 100);

      WHEN("unlimited values") {
        auto now = system_clock::now();

        // bug: start is derived form now(), not relative to end
        REQUIRE_CALL(controllerArchives, xport(_, 43320, 60, _, _, _))
            .LR_SIDE_EFFECT(CHECK(system_clock::to_time_t(now) - (720 + 1) * 60 < _1))
            .RETURN(ValueRange{.step = 60, .values = std::vector<Value>(720, {1, ds::DateTime{1s}})});
        auto range = controllerDirection->getSeries({Dsuid{controllerId1}}, 60s, SeriesTypes::power, false,
                                                    {std::nullopt, ds::DateTime{43260s}}, 0);
        CHECK(range.step == 60);
        CHECK(range.values.size() == 720);
      }
    }
  }

  WHEN("intermediate resolution") {
    WHEN("resolution=2s") {
      CHECK_THROWS(controllerDirection->getSeries({Dsuid{controllerId1}}, 2s, SeriesTypes::power, false,
                                                  {std::nullopt, ds::DateTime{600s}}, INT_MAX));
    }
    WHEN("resolution=61s") {
      CHECK_THROWS(controllerDirection->getSeries({Dsuid{controllerId1}}, 61s, SeriesTypes::power, false,
                                                  {std::nullopt, ds::DateTime{600s}}, INT_MAX));
    }
    WHEN("resolution=15min+1s") {
      CHECK_THROWS(controllerDirection->getSeries({Dsuid{controllerId1}}, 15min + 1s, SeriesTypes::power, false,
                                                  {std::nullopt, ds::DateTime{600s}}, INT_MAX));
    }
    WHEN("resolution=24h+1s") {
      CHECK_THROWS(controllerDirection->getSeries({Dsuid{controllerId1}}, 24h + 1s, SeriesTypes::power, false,
                                                  {std::nullopt, ds::DateTime{600s}}, INT_MAX));
    }
    WHEN("resolution=24h+1s") {
      CHECK_THROWS(controllerDirection->getSeries({Dsuid{controllerId1}}, 7 * 24h + 1s, SeriesTypes::power, false,
                                                  {std::nullopt, ds::DateTime{600s}}, INT_MAX));
    }
    WHEN("resolution=30d+1s") {
      CHECK_THROWS(controllerDirection->getSeries({Dsuid{controllerId1}}, 30 * 24h + 1s, SeriesTypes::power, false,
                                                  {std::nullopt, ds::DateTime{600s}}, INT_MAX));
    }
  }
}

SCENARIO("total power 2", TAGS) {
  dss::ApartmentMock apartment{};
  ArchivesStorageMock archivesStorage{};
  auto system = ds::SystemMock();

  auto controllerArchives1_ = std::make_unique<ControllerArchivesMock>();
  auto& controllerArchives1 = *controllerArchives1_;
  auto consumption = makeMockedControllerDirection(apartment, std::move(controllerArchives1_), archivesStorage,
                                                   Category::consumption, system);

  auto controllerArchives2_ = std::make_unique<ControllerArchivesMock>();
  auto& controllerArchives2 = *controllerArchives2_;
  auto production = makeMockedControllerDirection(apartment, std::move(controllerArchives2_), archivesStorage,
                                                  Category::production, system);

  auto controllerId1 = dsModel::ControllerId{"1111111111111111111111111111111100"};
  auto controller1 = apartment.allocateDSMeter(dss::Dsuid{controllerId1});

  GIVEN("no metering trippled in yet") {
    THEN("getApartmentPower") {
      CHECK(production->getApartmentPower() == 0);
      CHECK(consumption->getApartmentPower() == 0);
    }
  }

  GIVEN("consumption/production power") {
    REQUIRE_CALL(archivesStorage, hasArchive(ArchiveId{controllerId1, Category::consumption}))
        .TIMES(AT_LEAST(2))
        .RETURN(false);
    ALLOW_CALL(controllerArchives1, create(_));
    ALLOW_CALL(controllerArchives1, addValue(_, _));
    REQUIRE_CALL(archivesStorage, hasArchive(ArchiveId{controllerId1, Category::production}))
        .TIMES(AT_LEAST(2))
        .RETURN(false);
    ALLOW_CALL(controllerArchives2, create(_));
    ALLOW_CALL(controllerArchives2, addValue(_, _));

    auto now = system_clock::now();
    production->addEnergy(Dsuid{controllerId1}, {75, 100, now - 1s}); // first baseline is set
    production->addEnergy(Dsuid{controllerId1}, {75, 120, now});      // 2nd call populates cache

    consumption->addEnergy(Dsuid{controllerId1}, {25, 100, now - 1s});
    consumption->addEnergy(Dsuid{controllerId1}, {25, 120, now});

    THEN("getApartmentPower") {
      CHECK(consumption->getApartmentPower() == 25);
      CHECK(production->getApartmentPower() == 75);
    }

    GIVEN("2nd controller") {
      auto controllerId2 = dsModel::ControllerId{"1111111111111111111111111111111200"};
      auto controller2 = apartment.allocateDSMeter(dss::Dsuid{controllerId2});

      REQUIRE_CALL(archivesStorage, hasArchive(ArchiveId{controllerId2, Category::consumption}))
          .TIMES(AT_LEAST(2))
          .RETURN(false);
      REQUIRE_CALL(archivesStorage, hasArchive(ArchiveId{controllerId2, Category::production}))
          .TIMES(AT_LEAST(2))
          .RETURN(false);

      production->addEnergy(Dsuid{controllerId2}, {75, 100, now - 1s});
      production->addEnergy(Dsuid{controllerId2}, {75, 120, now});

      consumption->addEnergy(Dsuid{controllerId2}, {25, 100, now - 1s});
      consumption->addEnergy(Dsuid{controllerId2}, {25, 120, now});

      THEN("getApartmentPower doubles") {
        CHECK(consumption->getApartmentPower() == 50);
        CHECK(production->getApartmentPower() == 150);
      }
    }
  }
}

SCENARIO("get instant power 2", TAGS) {
  dss::ApartmentMock apartment{};
  ArchivesStorageMock archivesStorage{};
  auto system = ds::SystemMock();

  auto controllerId1 = dsModel::ControllerId{"1111111111111111111111111111111100"};
  auto controller1 = apartment.allocateDSMeter(dss::Dsuid{controllerId1});

  auto controllerArchives1_ = std::make_unique<ControllerArchivesMock>();
  auto& controllerArchives1 = *controllerArchives1_;
  auto consumption = makeMockedControllerDirection(apartment, std::move(controllerArchives1_), archivesStorage,
                                                   Category::consumption, system);

  auto controllerArchives2_ = std::make_unique<ControllerArchivesMock>();
  auto& controllerArchives2 = *controllerArchives2_;
  auto production = makeMockedControllerDirection(apartment, std::move(controllerArchives2_), archivesStorage,
                                                  Category::production, system);

  GIVEN("current data") {
    REQUIRE_CALL(archivesStorage, hasArchive(ArchiveId{controllerId1, Category::production}))
        .TIMES(AT_LEAST(2))
        .RETURN(false);
    ALLOW_CALL(controllerArchives1, create(_));
    ALLOW_CALL(controllerArchives1, addValue(_, _));
    REQUIRE_CALL(archivesStorage, hasArchive(ArchiveId{controllerId1, Category::consumption}))
        .TIMES(AT_LEAST(2))
        .RETURN(false);
    ALLOW_CALL(controllerArchives2, create(_));
    ALLOW_CALL(controllerArchives2, addValue(_, _));

    auto now = std::chrono::system_clock::now();
    consumption->addEnergy(Dsuid{controllerId1}, {0, 1000, now - 1s});
    production->addEnergy(Dsuid{controllerId1}, {0, 100, now - 1s}); // first baseline is set

    consumption->addEnergy(Dsuid{controllerId1}, {25, 2000, now});
    production->addEnergy(Dsuid{controllerId1}, {75, 200, now}); // 2nd call populates cache

    THEN("instant power is set") {
      CHECK(25 == consumption->getInstantPower(dss::Dsuid{controllerId1}));
      CHECK(75 == production->getInstantPower(dss::Dsuid{controllerId1}));
    }
    THEN("instant energy is set") {
      CHECK(consumption->getLastEnergy(dss::Dsuid{controllerId1}) == 1000);
      CHECK(production->getLastEnergy(dss::Dsuid{controllerId1}) == 100);
    }
  }

  GIVEN("old sensor data") {
    REQUIRE_CALL(archivesStorage, hasArchive(ArchiveId{controllerId1, Category::consumption}))
        .TIMES(AT_LEAST(2))
        .RETURN(false);
    ALLOW_CALL(controllerArchives1, create(_));
    ALLOW_CALL(controllerArchives1, addValue(_, _));
    REQUIRE_CALL(archivesStorage, hasArchive(ArchiveId{controllerId1, Category::production}))
        .TIMES(AT_LEAST(2))
        .RETURN(false);
    ALLOW_CALL(controllerArchives2, create(_));
    ALLOW_CALL(controllerArchives2, addValue(_, _));

    auto now = std::chrono::system_clock::now();
    consumption->addEnergy(Dsuid{controllerId1}, {0, 1000, now - 10s});
    production->addEnergy(Dsuid{controllerId1}, {0, 100, now - 10s}); // first baseline is set

    consumption->addEnergy(Dsuid{controllerId1}, {25, 2000, now - 9s});
    production->addEnergy(Dsuid{controllerId1}, {75, 200, now - 9s}); // 2nd call populates cache

    WHEN("no update for 3s") {
      THEN("instant power is 0") {
        CHECK(0 == consumption->getInstantPower(dss::Dsuid{controllerId1}));
        CHECK(0 == production->getInstantPower(dss::Dsuid{controllerId1}));
      }
      THEN("last energy is preserved") {
        CHECK(consumption->getLastEnergy(dss::Dsuid{controllerId1}) == 1000);
        CHECK(production->getLastEnergy(dss::Dsuid{controllerId1}) == 100);
      }
    }
  }
}

} // namespace
