#include "modbusmeteringsystem.h"

#include <ds/conf.h>
#include <ds/log.h>
#include <ds/system.h>

#include "archives-storage.h"
#include "archives/modbusmeteringsystem.h"
#include "ds-model/meterings.h"
#include "metering/readings/modbusmeteringsystem.h"

#include <unordered_map>

static ds::LogChannel dsLogChannel("dssMeteringModbusMeteringSystem");

DS_JSON_DEFINE_STRUCT((dss::metering), SerialPortConfigFile);
DS_JSON_DEFINE_STRUCT((dss::metering), UserMeterConfigFile);
DS_JSON_DEFINE_STRUCT((dss::metering), AttachedMetersConfigFile);

namespace dss::metering {

namespace {

using system_clock = std::chrono::system_clock;

/*auto modbusMeteringId = [](const ModbusMeteringSystem::Reading& reading) -> ModbusMeteringId {
  return ModbusMeteringId{reading.id};
};

auto emptyId = [](const ModbusMeteringId& id) { return id == ModbusMeteringId{}; };*/

class ModbusMeteringSystemImpl final : public ModbusMeteringSystem {
public:

  ModbusMeteringSystemImpl(std::unique_ptr<archives::ModbusMeteringArchives>&& modbusMeteringArchives,
                         ArchivesStorage& archivesStorage,
                         ds::System& system) 
        : m_archives(std::move(modbusMeteringArchives)),
          m_archivesStorage(archivesStorage),
          m_system(system) {
    loadSystemConfig();
    loadSerialConfig();
    loadUserConfig();
    loadAttachedMeterConfig();
  }

  virtual ~ModbusMeteringSystemImpl() override = default;

  std::vector<std::tuple<ModbusMeteringId, uint8_t, bool, std::string, dsModel::MeteringType>> meterings() const override {
    using dsModel::MeteringType;
    std::vector<std::tuple<ModbusMeteringId, uint8_t, bool, std::string, dsModel::MeteringType>> meterings{};

    for (auto& reading : m_attachedMeterReadings) {
      uint8_t slaveAddress = reading.second.slaveAddress;
      auto meteringId = ModbusMeteringId(std::string(reading.second.id));
      bool isGlobal = m_attachedMeters.at(reading.first).isGlobal;
      std::string application = ds::str(m_attachedMeters.at(reading.first).application);

      auto timeDiff = std::chrono::duration_cast<std::chrono::seconds>( \
                          ds::DateTime::clock::now() - reading.second.lastChanged);

      if (std::string(reading.second.id) != "" && timeDiff < 900s) {
        meterings.push_back({meteringId, slaveAddress, isGlobal, application, dsModel::MeteringType::powerMetering});
        meterings.push_back({meteringId, slaveAddress, isGlobal, application, dsModel::MeteringType::powerProducedMetering});
        meterings.push_back({meteringId, slaveAddress, isGlobal, application, dsModel::MeteringType::energyMetering});
        meterings.push_back({meteringId, slaveAddress, isGlobal, application, dsModel::MeteringType::energyProducedMetering});

        if (reading.second.tariffEnabled) {
          if (reading.second.t1EnergyConsumed > 0) {
            meterings.push_back({meteringId, slaveAddress, isGlobal, application, dsModel::MeteringType::energyConsumedTariff1Metering});
          }

          if (reading.second.t2EnergyConsumed > 0) {
            meterings.push_back({meteringId, slaveAddress, isGlobal, application, dsModel::MeteringType::energyConsumedTariff2Metering});
          }

          if (reading.second.t3EnergyConsumed > 0) {
            meterings.push_back({meteringId, slaveAddress, isGlobal, application, dsModel::MeteringType::energyConsumedTariff3Metering});
          }

          if (reading.second.t4EnergyConsumed > 0) {
            meterings.push_back({meteringId, slaveAddress, isGlobal, application, dsModel::MeteringType::energyConsumedTariff4Metering});
          }

          if (reading.second.t6EnergyConsumed > 0) {
            meterings.push_back({meteringId, slaveAddress, isGlobal, application, dsModel::MeteringType::energyConsumedTariff5Metering});
          }

          if (reading.second.t6EnergyConsumed > 0) {
            meterings.push_back({meteringId, slaveAddress, isGlobal, application, dsModel::MeteringType::energyConsumedTariff6Metering});
          }

          if (reading.second.t1EnergyProduced > 0) {
            meterings.push_back({meteringId, slaveAddress, isGlobal, application, dsModel::MeteringType::energyProducedTariff1Metering});
          }

          if (reading.second.t2EnergyProduced > 0) {
            meterings.push_back({meteringId, slaveAddress, isGlobal, application, dsModel::MeteringType::energyProducedTariff2Metering});
          }

          if (reading.second.t3EnergyProduced > 0) {
            meterings.push_back({meteringId, slaveAddress, isGlobal, application, dsModel::MeteringType::energyProducedTariff3Metering});
          }

          if (reading.second.t4EnergyProduced > 0) {
            meterings.push_back({meteringId, slaveAddress, isGlobal, application, dsModel::MeteringType::energyProducedTariff4Metering});
          }

          if (reading.second.t5EnergyProduced > 0) {
            meterings.push_back({meteringId, slaveAddress, isGlobal, application, dsModel::MeteringType::energyProducedTariff5Metering});
          }

          if (reading.second.t6EnergyProduced > 0) {
            meterings.push_back({meteringId, slaveAddress, isGlobal, application, dsModel::MeteringType::energyProducedTariff6Metering});
          }
        }
      }


    }
    return meterings;
  }

  dsModel::SerialPortConfig getSerialPortConfig() const {
    return m_serialPortConfig;
  }

  void setSerialPortConfig(const dsModel::SerialPortConfig& serialPortConfig) {
    m_serialPortConfig = serialPortConfig;
    saveSerialConfig();
  }

  std::vector<dsModel::ModbusMeter> getModbusAttachedMeters() const {
    std::vector<dsModel::ModbusMeter> meters;
    for (auto& meter: m_attachedMeters) {
        meters.push_back(meter.second);
    }
    return meters;
  }

  void addModbusAttachedMeter(const dsModel::ModbusMeter& meter) {
    DS_NOTICE("Adding meter", meter);
    bool globalMeterAlreadyExists = false;
    for (const auto& meter: m_attachedMeters) {
      if (meter.second.isGlobal) {
        globalMeterAlreadyExists = true;
      }
    }
    bool canAddMeter = true;
    if (globalMeterAlreadyExists && meter.isGlobal) {
      canAddMeter = false;
    }
    DS_REQUIRE(canAddMeter, "Modbus introduction meter already exists");

    std::string id = meter.manufacturer + "-" + meter.model + "-" + meter.name + "-address-" + std::to_string(meter.slaveAddress);

    m_attachedMeters[id] = meter;
    m_attachedMeters[id].id = id;
    auto supportedMeters = getModbusSupportedMeters();
    for (const auto& supportedMeter : supportedMeters) {
      if (supportedMeter.id == m_attachedMeters[id].configId) {
        m_attachedMeters[id].model = supportedMeter.model;
        m_attachedMeters[id].manufacturer = supportedMeter.manufacturer;
      }
    }

    saveAttachedMeterConfig();
  }

  void deleteModbusAttachedMeter(const dsModel::MeteringId& meter) {
    auto it = m_attachedMeters.find(meter);
    if (m_attachedMeterReadings.find((*it).second.id) != m_attachedMeterReadings.end()) {
      DS_DEBUG("Also deleted reading for", (*it).second.id);
      m_attachedMeterReadings.erase((*it).second.id);
    }
    if (m_attachedMeterReadingsSynthetic.find((*it).second.id) != m_attachedMeterReadingsSynthetic.end()) {
      DS_DEBUG("Also deleted reading for", (*it).second.id);
      m_attachedMeterReadingsSynthetic.erase((*it).second.id);
    }

    if (it != m_attachedMeters.end()) {
        m_attachedMeters.erase(it);
    }
    saveAttachedMeterConfig();
  }

  virtual std::vector<dsModel::ElectricityMeterConfig> getModbusSupportedMeters() const {
    std::vector<dsModel::ElectricityMeterConfig> configs;
    for (auto& config: m_userConfigs) {
        configs.push_back(config.second);
    }

    for (auto& config: m_systemConfigs) {
        configs.push_back(config.second);
    }
    return configs;
  }

  virtual std::vector<dsModel::ElectricityMeterConfig> getModbusUserMeters() const {
    std::vector<dsModel::ElectricityMeterConfig> configs;
    for (auto& config: m_userConfigs) {
        configs.push_back(config.second);
    }
    return configs;
  }

  virtual void addModbusSupportedMeter(const dsModel::ElectricityMeterConfig& meterConfig) {
    std::string id = "userConfig-" + meterConfig.manufacturer + "-" + meterConfig.model;


    m_userConfigs[id] = meterConfig;
    m_userConfigs[id].id = id;
    m_userConfigs[id].displayName = ds::str(m_userConfigs[id].manufacturer, " - ", m_userConfigs[id].model);

    saveUserConfig();
  }

  virtual void deleteModbusSupportedMeter(const dsModel::MeteringConfigId& meterConfigId) {
    auto it = m_userConfigs.find(meterConfigId);
    if (it != m_userConfigs.end()) {
        m_userConfigs.erase(it);
    }

    saveUserConfig();
  }

  void addReading(Reading&& reading) {
    for (auto& meter : m_attachedMeters) {
      if (reading.slaveAddress == meter.second.slaveAddress) {
        DS_NOTICE("Adding modbus reading to found meter.", meter, reading);
        std::optional<Reading> previousReading;
        if (m_attachedMeterReadings.find(std::string(meter.second.id)) != m_attachedMeterReadings.end()) {
          previousReading = m_attachedMeterReadings[meter.second.id];
        }

        float factor = 1;
        if (previousReading) {
          auto timeDiff = reading.lastChanged - (*previousReading).lastChanged;
          factor = std::chrono::duration_cast<std::chrono::milliseconds>(timeDiff).count() / 1000.0F;
          DS_DEBUG("Time difference in seconds to previous reading", factor);
        }

        if (m_attachedMeterReadingsSynthetic.find(meter.second.id) == m_attachedMeterReadingsSynthetic.end()) {
          auto meterIdConsumption = dss::metering::archives::ModbusMeteringArchiveId(std::string(reading.id), Category::consumption);
          auto lastValue = m_archives->lastValue(meterIdConsumption);

          if (lastValue) {
            m_attachedMeterReadingsSynthetic[meter.second.id].energyConsumed = (*lastValue).energy;
            DS_DEBUG("Fetched consumption energy values from db ", (*lastValue).energy);
          }

          auto meterIdProduction = dss::metering::archives::ModbusMeteringArchiveId(std::string(reading.id), Category::production);
          lastValue = m_archives->lastValue(meterIdProduction);

          if (lastValue) {
            m_attachedMeterReadingsSynthetic[meter.second.id].energyProduced = (*lastValue).energy;
            DS_DEBUG("Fetched production energy values from db ", (*lastValue).energy);
          }
        }

        m_attachedMeterReadings[meter.second.id] = reading;
        m_attachedMeterReadingsSynthetic[meter.second.id].id = reading.id;
        m_attachedMeterReadingsSynthetic[meter.second.id].slaveAddress = reading.slaveAddress;
        m_attachedMeterReadingsSynthetic[meter.second.id].meteringType = reading.meteringType;
        m_attachedMeterReadingsSynthetic[meter.second.id].lastChanged = reading.lastChanged;
        m_attachedMeterReadingsSynthetic[meter.second.id].tariffEnabled = reading.tariffEnabled;
        m_attachedMeterReadingsSynthetic[meter.second.id].powerConsumed = reading.powerConsumed;
        m_attachedMeterReadingsSynthetic[meter.second.id].powerProduced = reading.powerProduced;
        m_attachedMeterReadingsSynthetic[meter.second.id].t1EnergyConsumed = reading.t1EnergyConsumed;
        m_attachedMeterReadingsSynthetic[meter.second.id].t2EnergyConsumed = reading.t2EnergyConsumed;
        m_attachedMeterReadingsSynthetic[meter.second.id].t3EnergyConsumed = reading.t3EnergyConsumed;
        m_attachedMeterReadingsSynthetic[meter.second.id].t4EnergyConsumed = reading.t4EnergyConsumed;
        m_attachedMeterReadingsSynthetic[meter.second.id].t5EnergyConsumed = reading.t5EnergyConsumed;
        m_attachedMeterReadingsSynthetic[meter.second.id].t6EnergyConsumed = reading.t6EnergyConsumed;
        m_attachedMeterReadingsSynthetic[meter.second.id].t1EnergyProduced = reading.t1EnergyProduced;
        m_attachedMeterReadingsSynthetic[meter.second.id].t2EnergyProduced = reading.t2EnergyProduced;
        m_attachedMeterReadingsSynthetic[meter.second.id].t3EnergyProduced = reading.t3EnergyProduced;
        m_attachedMeterReadingsSynthetic[meter.second.id].t4EnergyProduced = reading.t4EnergyProduced;
        m_attachedMeterReadingsSynthetic[meter.second.id].t5EnergyProduced = reading.t5EnergyProduced;
        m_attachedMeterReadingsSynthetic[meter.second.id].t6EnergyProduced = reading.t6EnergyProduced;
        
        m_attachedMeterReadingsSynthetic[meter.second.id].energyConsumed += reading.powerConsumed * factor;
        
        m_attachedMeterReadingsSynthetic[meter.second.id].energyProduced += reading.powerProduced * factor;

        meter.second.serial = reading.id;
        meter.second.tariffEnabled = reading.tariffEnabled;

        auto meterId = dss::metering::archives::ModbusMeteringArchiveId(std::string(reading.id), Category::consumption);
        if (!m_archivesStorage.hasArchive(meterId)) {
          m_archives->create(meterId);
        }

        DS_DEBUG("Metering archive should be there");
        //TODO add tariffs
        //double energyConsumed = reading.powerConsumed;
        //double energyProduced = reading.powerProduced;
        /*if (previousReading) {
          energyConsumed = reading.energyConsumed - (*previousReading).energyConsumed;
          energyProduced = reading.energyProduced - (*previousReading).energyProduced;
        }*/
        ArchiveValue value{.power = reading.powerConsumed, 
                           .energy = m_attachedMeterReadingsSynthetic[meter.second.id].energyConsumed, 
                           .lastChanged = reading.lastChanged};
        DS_DEBUG("Saving consumption", value);
        DS_DEBUG("Saving consumption", static_cast<int64_t>(value.energy));
        m_archives->addValue(meterId, std::move(value));

        auto meterIdProduction = dss::metering::archives::ModbusMeteringArchiveId(std::string(reading.id), Category::production);
        if (!m_archivesStorage.hasArchive(meterIdProduction)) {
          m_archives->create(meterIdProduction);
        }
        value = {.power = reading.powerProduced, 
                 .energy = m_attachedMeterReadingsSynthetic[meter.second.id].energyProduced,
                 .lastChanged = reading.lastChanged};
        DS_DEBUG("Saving production", value);
        DS_DEBUG("Saving production", static_cast<int64_t>(value.energy));
        m_archives->addValue(meterIdProduction, std::move(value));

        
      }
    }
  }

  void unmarkGlobalAll() {
    for (auto& meter : m_attachedMeters) {
      if (meter.second.isGlobal) {
        DS_NOTICE("Unmarked meter as introduction", meter.second.id);
        meter.second.isGlobal = false;
      }
    }
  };

  std::optional<LastReading> getLastReading(ModbusMeteringId id, dsModel::MeteringType meteringType,
                                            EnergyUnit energyUnit) const {
    DS_DEBUG_ENTER();
    DS_DEBUG(id, meteringType, energyUnit);
    auto convertEnergyUnit = [energyUnit](double value) {
      if (energyUnit == EnergyUnit::Wh) {
        return value / (3600);
      } else {
        return value;
      }
    };

    for (const auto& reading: m_attachedMeterReadingsSynthetic) {
      if (std::string(reading.second.id) == std::string(id)) {
        Reading currentReading;
        for (const auto& aReading : m_attachedMeterReadings) {
          if (std::string(aReading.second.id) == std::string(id)) {
            currentReading = aReading.second;
          }
        }
        switch (meteringType) {
        case dsModel::MeteringType::powerMetering:
          return LastReading{meteringType, reading.second.powerConsumed, reading.second.lastChanged};
        case dsModel::MeteringType::powerProducedMetering:
          return LastReading{meteringType, reading.second.powerProduced, reading.second.lastChanged};
        case dsModel::MeteringType::energyMetering:
          return LastReading{meteringType, convertEnergyUnit(currentReading.energyConsumed), reading.second.lastChanged};
        case dsModel::MeteringType::energyConsumedTariff1Metering:
          return LastReading{meteringType, convertEnergyUnit(reading.second.t1EnergyConsumed), reading.second.lastChanged};
        case dsModel::MeteringType::energyConsumedTariff2Metering:
          return LastReading{meteringType, convertEnergyUnit(reading.second.t2EnergyConsumed), reading.second.lastChanged};
        case dsModel::MeteringType::energyConsumedTariff3Metering:
          return LastReading{meteringType, convertEnergyUnit(reading.second.t3EnergyConsumed), reading.second.lastChanged};
        case dsModel::MeteringType::energyConsumedTariff4Metering:
          return LastReading{meteringType, convertEnergyUnit(reading.second.t4EnergyConsumed), reading.second.lastChanged};
        case dsModel::MeteringType::energyConsumedTariff5Metering:
          return LastReading{meteringType, convertEnergyUnit(reading.second.t5EnergyConsumed), reading.second.lastChanged};
        case dsModel::MeteringType::energyConsumedTariff6Metering:
          return LastReading{meteringType, convertEnergyUnit(reading.second.t6EnergyConsumed), reading.second.lastChanged};
        case dsModel::MeteringType::energyProducedMetering:
          return LastReading{meteringType, convertEnergyUnit(currentReading.energyProduced), reading.second.lastChanged}; 
        case dsModel::MeteringType::energyProducedTariff1Metering:
          return LastReading{meteringType, convertEnergyUnit(reading.second.t1EnergyProduced), reading.second.lastChanged};
        case dsModel::MeteringType::energyProducedTariff2Metering:
          return LastReading{meteringType, convertEnergyUnit(reading.second.t2EnergyProduced), reading.second.lastChanged};
        case dsModel::MeteringType::energyProducedTariff3Metering:
          return LastReading{meteringType, convertEnergyUnit(reading.second.t3EnergyProduced), reading.second.lastChanged};
        case dsModel::MeteringType::energyProducedTariff4Metering:
          return LastReading{meteringType, convertEnergyUnit(reading.second.t4EnergyProduced), reading.second.lastChanged};
        case dsModel::MeteringType::energyProducedTariff5Metering:
          return LastReading{meteringType, convertEnergyUnit(reading.second.t5EnergyProduced), reading.second.lastChanged};
        case dsModel::MeteringType::energyProducedTariff6Metering:
          return LastReading{meteringType, convertEnergyUnit(reading.second.t6EnergyProduced), reading.second.lastChanged};
        case dsModel::MeteringType::none:
        default: break;
      }
      }
    }
    return std::nullopt;
  }

  ValueRange getSeries(std::vector<ModbusMeteringId> ids, dsModel::MeteringType meteringType, SeriesTypes type, std::chrono::seconds resolution,
                               const TimeDuration& timeDuration, int valueCount, EnergyUnit energyUnit) const {
    DS_DEBUG_ENTER();

    auto category = [](auto meteringType) {
      switch (meteringType) {
        case dsModel::MeteringType::powerMetering:
        case dsModel::MeteringType::energyMetering:
        case dsModel::MeteringType::energyConsumedTariff2Metering:
        case dsModel::MeteringType::energyConsumedTariff1Metering: 
        case dsModel::MeteringType::energyConsumedTariff3Metering:
        case dsModel::MeteringType::energyConsumedTariff4Metering: 
        case dsModel::MeteringType::energyConsumedTariff5Metering:
        case dsModel::MeteringType::energyConsumedTariff6Metering: return Category::consumption;
        case dsModel::MeteringType::powerProducedMetering:
        case dsModel::MeteringType::energyProducedMetering:
        case dsModel::MeteringType::energyProducedTariff1Metering:
        case dsModel::MeteringType::energyProducedTariff2Metering: 
        case dsModel::MeteringType::energyProducedTariff3Metering:
        case dsModel::MeteringType::energyProducedTariff4Metering:
        case dsModel::MeteringType::energyProducedTariff5Metering:
        case dsModel::MeteringType::energyProducedTariff6Metering: return Category::production;
        case dsModel::MeteringType::none:
        default: DS_FAIL_REQUIRE("unhandled", meteringType);
      }
    };

    std::vector<archives::ModbusMeteringArchiveId> modbusMeteringIds;
    for (const auto& id : ids) {
      modbusMeteringIds.push_back(std::make_pair(std::string(id), category(meteringType)));
    }

    ElectricityMeteringConfig config;
    auto& archiveConfigs = config.timeSeries;
    auto it = std::find_if(std::rbegin(archiveConfigs), std::rend(archiveConfigs),
                          [resolution](const auto& element) { return element.resolution == resolution; });
    DS_REQUIRE(it != std::rend(archiveConfigs), "unsupported resolution");
    int numberOfValues = it->numberOfValues;


    // adjust limits by start/end/values
    valueCount = std::min(valueCount, numberOfValues);

    long unsigned int step = resolution.count();
    auto align = [step](time_t value) { return value - value % step; };


    auto now = system_clock::now();
    time_t maxEnd = align(system_clock::to_time_t(now));
    time_t maxStart = align(system_clock::to_time_t(now) - (step * numberOfValues));
    time_t start{}, end{};

    if (timeDuration == TimeDuration{}) {
      end = maxEnd;
      if (valueCount == 0) {
        start = maxStart;
      } else {
        start = end - (step * valueCount);
      }
    } else if (!timeDuration.start) {
      end = align(system_clock::to_time_t(*timeDuration.end));
      if (valueCount == 0) {
        start = maxStart;
      } else {
        start = end - (step * valueCount);
      }
    } else if (!timeDuration.end) {
      start = align(system_clock::to_time_t(*timeDuration.start));
      if (valueCount == 0) {
        end = maxEnd;
      } else {
        end = start + (step * valueCount);
      }
    } else {
      start = align(system_clock::to_time_t(*timeDuration.start));
      end = align(system_clock::to_time_t(*timeDuration.end));
    }
    end += step;
    DS_DEBUG("Doing export", start, end, resolution.count(), modbusMeteringIds, type);

    auto valueRange = m_archives->xport(start,
                      end, 
                      resolution.count(), modbusMeteringIds, type, true);
    int valueSize = valueRange.values.size();
    if (valueSize > 20) {
      for (int i = 20; i > 0 ; i--) {
        if (valueRange.values[valueSize - i].value < 0.001 && valueRange.values[valueSize - i - 1].value > 0.1) {
          valueRange.values[valueSize - i].value = valueRange.values[valueSize - i - 1].value;
        }
      }
    }
    return valueRange;
  }

  double get15MinuteAverage(std::vector<ModbusMeteringId> ids, dsModel::MeteringType meteringType,
                            metering::SeriesTypes type, Unit unit) const override {
    auto range = getSeries(ids, meteringType, type, 15min, {}, 1, unit == Unit::Wh ? EnergyUnit::Wh : EnergyUnit::Ws);
    auto& values = range.values;
    if (values.size() != 1) {
      DS_NOTICE("unexpected value count received", values.size(), range);
      return 0;
    }
    DS_REQUIRE(values.size() == 1);
    DS_REQUIRE(std::isfinite(values[0].value));
    return values[0].value;
  };

private:
    void loadSerialConfig() {
      auto& system = ds::system();
      auto filePath = std::filesystem::path(system.dirs().sysconfdir) / "modbus-serial.json";
      DS_NOTICE("Load serial configuration", filePath.string());
      auto file = system.readFileIfExists(filePath.string());
      if (!file) {
          DS_NOTICE("File not found. Load default configuration.");
          return;
      }
      auto serialConfigFile = ds::fromJson<SerialPortConfigFile>(*file);
      m_serialPortConfig = serialConfigFile.serialPortConfig;
    }

    void saveSerialConfig() {
      auto& system = ds::system();
      auto filePath = std::filesystem::path(system.dirs().sysconfdir) / "modbus-serial.json";
      DS_NOTICE("save serial configuration", filePath.string(), m_serialPortConfig);

      SerialPortConfigFile serialPortConfig{.serialPortConfig = m_serialPortConfig};
      system.writeToFile(filePath.string(), ds::toJson(serialPortConfig));
    }

    void loadUserConfig() {
      auto& system = ds::system();
      auto filePath = std::filesystem::path(system.dirs().sysconfdir) / "modbus-user-config.json";
      DS_NOTICE("Load serial configuration", filePath.string());
      auto file = system.readFileIfExists(filePath.string());
      if (!file) {
          DS_NOTICE("File not found. Load  default configuration.");
          return;
      }
      m_userConfigs.clear();
      auto userConfigFile = ds::fromJson<UserMeterConfigFile>(*file);
      for (auto& config : userConfigFile.userMeterConfigs) {
        if (std::string(config.id) != "") {
          m_userConfigs[config.id] = config;
        }
      }
    }

    void loadSystemConfig() {
      auto& system = ds::system();
      auto filePath = std::filesystem::path(system.dirs().sysconfdir) / "modbus-system-config.json";
      DS_NOTICE("Load serial configuration", filePath.string());
      auto file = system.readFileIfExists(filePath.string());
      if (!file) {
          DS_NOTICE("File not found. Load  default configuration.");
          return;
      }
      m_systemConfigs.clear();
      auto systemConfigFile = ds::fromJson<UserMeterConfigFile>(*file);
      for (auto& config : systemConfigFile.userMeterConfigs) {
        if (std::string(config.id) != "") {
          m_systemConfigs[config.id] = config;
        }
      }
    }

    void saveUserConfig() {
      UserMeterConfigFile userMeterConfigFile{};
      for (auto& config : m_userConfigs) {
        userMeterConfigFile.userMeterConfigs.push_back(config.second);
      }
      auto& system = ds::system();
      auto filePath = std::filesystem::path(system.dirs().sysconfdir) / "modbus-user-config.json";
      DS_NOTICE("save serial configuration", filePath.string(), userMeterConfigFile);

      system.writeToFile(filePath.string(), ds::toJson(userMeterConfigFile));

    }

    void loadAttachedMeterConfig() {
      auto& system = ds::system();
      auto filePath = std::filesystem::path(system.dirs().sysconfdir) / "modbus-attached-config.json";
      DS_NOTICE("Load serial configuration", filePath.string());
      auto file = system.readFileIfExists(filePath.string());
      if (!file) {
          DS_NOTICE("File not found. Load default configuration.");
          return;
      }
      auto attachedMetersFile = ds::fromJson<AttachedMetersConfigFile>(*file);
      m_attachedMeters.clear();
      for (auto& attachedMeter : attachedMetersFile.attachedMeters) {
        if (std::string(attachedMeter.id) != "") {
          m_attachedMeters[attachedMeter.id] = attachedMeter;
        }
      }
    }

    void saveAttachedMeterConfig() {
      AttachedMetersConfigFile attachedMeters{};
      for (auto& attachedMeter : m_attachedMeters) {
        attachedMeters.attachedMeters.push_back(attachedMeter.second);
      }

      auto& system = ds::system();
      auto filePath = std::filesystem::path(system.dirs().sysconfdir) / "modbus-attached-config.json";
      DS_NOTICE("save serial configuration", filePath.string(), attachedMeters);

      system.writeToFile(filePath.string(), ds::toJson(attachedMeters));
    }

  dsModel::SerialPortConfig m_serialPortConfig;
  std::unordered_map<std::string, dsModel::ModbusMeter> m_attachedMeters;
  std::unordered_map<std::string, Reading> m_attachedMeterReadings;
  std::unordered_map<std::string, Reading> m_attachedMeterReadingsSynthetic;
  std::unordered_map<std::string, dsModel::ElectricityMeterConfig> m_userConfigs;
  std::unordered_map<std::string, dsModel::ElectricityMeterConfig> m_systemConfigs;

  std::unique_ptr<archives::ModbusMeteringArchives> m_archives;
  ArchivesStorage& m_archivesStorage;
  ds::System& m_system;
};

} // namespace

std::unique_ptr<ModbusMeteringSystem> makeModbusMeteringSystem(ArchivesConnection& archivesConnection,
                                                               ArchivesStorage& archivesStorage) {
  auto modbusMeteringArchive =
      archives::makeModbusArchive(archivesConnection, archivesStorage, ds::system());
  return std::make_unique<ModbusMeteringSystemImpl>(std::move(modbusMeteringArchive), archivesStorage, ds::system());
}

std::unique_ptr<ModbusMeteringSystem> makeModbusMeteringSystemMocked(std::unique_ptr<archives::ModbusMeteringArchives> modbusMeteringArchives, 
                                                                     ArchivesStorage& archivesStorage,
                                                                     ds::System& system) {
  return std::make_unique<ModbusMeteringSystemImpl>(std::move(modbusMeteringArchives), archivesStorage, system);
}

} // namespace dss::metering


