#include "power-states.h"

#include <numeric>
#include <tuple>

#include "metering.h"
#include "model/dsmeter.h"
#include "model/state.h"

static ds::LogChannel dsLogChannel("dssMeteringPowerStates");

namespace dss::metering {

namespace {

static auto createStateName = [](std::string meteringId, const PowerStateConfig& config) {
  // TODO Cover DS_REQUIRE_FAIL of dsm state
  return ds::str(meteringId, "-", config.category, "-on", config.setThreshold, "-off", config.resetThreshold);
};

class PowerStatesImpl : public PowerStates {
  std::vector<std::tuple<std::shared_ptr<State>, std::string, PowerStateConfig, int>> m_states{};
  Apartment& m_apartment;

public:
  PowerStatesImpl(Apartment& apartment) : m_apartment(apartment) {}
  std::shared_ptr<State> createState(std::string meteringId, PowerStateConfig&& powerStateConfig) override;
  void removeState(std::string meteringId, const PowerStateConfig& powerStateConfig) override;
  void trigger(std::string meteringId, TotalPower&& totalPower) override;
  void checkHysteresis(unsigned int value, PowerStateConfig& config, std::shared_ptr<State> state);
};

std::shared_ptr<State> PowerStatesImpl::createState(std::string meteringId,
                                                    PowerStateConfig&& powerStateConfig) {
  DS_INFO_ENTER(meteringId, powerStateConfig);
  if (powerStateConfig.resetThreshold > powerStateConfig.setThreshold) {
    DS_WARNING("Powerstate malconfiguration detected, set threshold lower than reset threshold.", meteringId, powerStateConfig);
  }
  auto it = std::find_if(
      std::begin(m_states), std::end(m_states), [meteringId, config = powerStateConfig](const auto& element) {
        return std::get<PowerStateConfig>(element) == config && std::get<std::string>(element) == meteringId;
      });
  if (it != std::end(m_states)) {
    auto& [state, meterSysten, config, references] = *it;
    // in clang variables from structured binding are only aliases they are not
    // variables, thus you cannot capture it in DS_REQUIRE lambda by reference
    auto scopeIssue = references++;
    DS_REQUIRE(scopeIssue > 0);
    DS_INFO_LEAVE();
    return state;
  }

  auto state = m_apartment.allocateState(StateType_Circuit, createStateName(meteringId, powerStateConfig), {});
  m_states.push_back(std::make_tuple(state, meteringId, std::move(powerStateConfig), 1));
  DS_INFO_LEAVE();
  return state;
}

void PowerStatesImpl::removeState(std::string meteringId, const PowerStateConfig& powerStateConfig) {
  DS_INFO_ENTER(meteringId, powerStateConfig);
  auto it = std::find_if(
      std::begin(m_states), std::end(m_states), [meteringId, config = powerStateConfig](const auto& element) {
        return std::get<PowerStateConfig>(element) == config && std::get<std::string>(element) == meteringId;
      });
  DS_REQUIRE(it != std::end(m_states));

  auto& [state, _, config, references] = *it;
  auto scopeIssue = references--;
  DS_REQUIRE(scopeIssue > 0);

  if (references == 0) {
    DS_NOTICE("removeState", std::get<std::string>(*it), config, state->getName());
    m_apartment.removeState(state); // NOTE: state is reference if you erase, iterator becomes invalid
    m_states.erase(it);
  }
  DS_INFO_LEAVE();
}

void PowerStatesImpl::checkHysteresis(unsigned int meteringValue, PowerStateConfig& config,
                                      std::shared_ptr<State> state) {
  if (state->getState() != State_Active) {
    if (config.resetThreshold > config.setThreshold) {
      state->setState(CallOrigin::system, State_Inactive);
      DS_INFO("Wrong configuration  forced ->inactive", state);
    } else if (meteringValue > config.setThreshold) {
      DS_NOTICE("  ->active", state);
      state->setState(CallOrigin::system, State_Active);
    }
  } else {
    if (meteringValue < config.resetThreshold) {
      DS_NOTICE("->inactive", state);
      state->setState(CallOrigin::system, State_Inactive);
    }
  }
}

void PowerStatesImpl::trigger(std::string meteringId, TotalPower&& totalPower) {
  DS_DEBUG(meteringId, totalPower);

  for (auto [state, meteringSystemInner, config, _] : m_states) {
    if (meteringSystemInner != meteringId) {
      continue;
    }

    DS_DEBUG(config, state, state->getState(), totalPower);
    if (config.category == Category::consumption) {
      checkHysteresis(totalPower.consumption, config, state);
    } else {
      checkHysteresis(totalPower.production, config, state);
    }
  }
}

} // namespace

std::shared_ptr<PowerStates> makePowerStates(Apartment& apartment) {
  return std::make_shared<PowerStatesImpl>(apartment);
}

} // namespace dss::metering
