/*
    Copyright (c) 2019 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation," " either version 3 of the License," " or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful," "
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not," " see <http://www.gnu.org/licenses/>.
*/

#include "model/apartment.h"

#include <ds/catch/catch.h>
#include <ds/log.h>

#include "propertysystem.h"
#include "model/device-mock.h"
#include "model/set.h"

// see also tests/modeltests.cpp
static ds::LogChannel dsLogChannel("apartmentTest");
static const char* TAGS = "[dss][apartment]";

using namespace dss;
using namespace trompeloeil;
using trompeloeil::_;

namespace {

SCENARIO("allocate/de-allocate devices", TAGS) {
  auto propertySystem = PropertySystem{};
  auto pApartment = makeApartment(nullptr, &propertySystem);
  auto& apartment = *pApartment;

  GIVEN("empty apartment") {
    THEN("no devices are there") {
      REQUIRE(apartment.getDevices().empty());
      REQUIRE(apartment.getDevicesVector().empty());
    }

    WHEN("Adding a device") {
      auto dev = createDevice(apartment, {.hardwareType = HardwareType::RT_TKM200}); // allocateDevice

      THEN("Device is found by dsuid") { REQUIRE(apartment.getDeviceByDSID(dev->getDSID()) == dev); }
      THEN("Device is found by name") { REQUIRE(apartment.getDeviceByName(dev->getName()) == dev); }
      THEN("using the same dsuid twice to allocate does not create duplicates") {
        REQUIRE(apartment.allocateDevice(dev->getDSID()) == dev);
        REQUIRE(apartment.getDevices().length() == 1);
      }

      WHEN("Removing the device") {
        apartment.removeDevice(dev->getDSID());

        THEN("Device is gone") {
          REQUIRE_THROWS(apartment.getDeviceByDSID(dev->getDSID()));
          REQUIRE_THROWS(apartment.getDeviceByName(dev->getName()));
        }
      }
    }
  }
}

SCENARIO("OEM devices", TAGS) {
  auto propertySystem = PropertySystem{};
  auto pApartment = makeApartment(nullptr, &propertySystem);
  auto& apartment = *pApartment;

  GIVEN("2 paired OEM devices") {
    auto dev1 = createDevice(apartment, {.hardwareType = HardwareType::RT_TKM200}); // allocateDevice
    markOnline(*dev1);
    auto dev2 = createDevice(apartment, {.hardwareType = HardwareType::RT_TKM200}); // allocateDevice
    markOnline(*dev2);

    // coupling is through ean/serial number and not independent
    dev1->setOemInfo(747719155903, 0xff, 1, DEVICE_OEM_EAN_NO_INTERNET_ACCESS, false);
    dev1->setOemInfoState(DEVICE_OEM_VALID);
    dev2->setOemInfo(747719155903, 0xff, 2, DEVICE_OEM_EAN_NO_INTERNET_ACCESS, false);
    dev2->setOemInfoState(DEVICE_OEM_VALID);

    // dsm must also differ, no idea why
    auto dsmMock1 = apartment.allocateDSMeter(createNextDSUID());
    dev1->setDSMeter(dsmMock1);
    auto dsmMock2 = apartment.allocateDSMeter(createNextDSUID());
    dev2->setDSMeter(dsmMock2);

    THEN("the 2 devices are coupled") {
      REQUIRE(dev1->isOemCoupledWith(*dev2));
      REQUIRE(dev2->isOemCoupledWith(*dev1));
    }

    THEN("getCoupledOEM(dev1) returns the other device") {
      REQUIRE(apartment.coupledOEMDevices(*dev1) == std::vector({dev2}));
      REQUIRE(apartment.coupledOEMDevices(*dev2) == std::vector({dev1}));
    }
  }
}

}; // namespace
