/*
    Copyright (c) 2010 digitalSTROM.org, Zurich, Switzerland

    Author: Patrick Staehlin, futureLAB AG <pstaehlin@futurelab.ch>

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

*/

#pragma once

#include <ds-model/apartment-fwd.h>
#include <ds-model/groups.h>
#include <ds-model/installation-fwd.h>
#include <ds-model/model-fwd.h>
#include <ds/asio/ds-signal.h>
#include <regex.h>

#include <boost/thread/recursive_mutex.hpp>
#include <memory>
#include <set>
#include <vector>

#include "datetools.h"
#include "devicecontainer.h"
#include "dssfwd.h"
#include "exception.h"
#include "metering/metering-fwd.h"
#include "model/device.h"
#include "model/modelconst.h"
#include "sensoraggregator.h"
#include "subsystem.h"
#include "web/webrequests.h"
#include "web/handler/jsonhelper.h"

namespace dss {
class PropertyNode;
typedef std::shared_ptr<PropertyNode> PropertyNodePtr;
class DSMeter;
class Cluster;
class ModelMaintenance;
class BusInterface;
class DeviceBusInterface;
class ActionRequestInterface;
class Event;
class EventSink;
class ApartmentProxy;
class Floor;
class ModelFeatures;

typedef struct ApartmentSensorStatus {
  ApartmentSensorStatus()
      : m_TemperatureValue(0),
        m_TemperatureTS(0),
        m_HumidityValue(0),
        m_HumidityTS(0),
        m_BrightnessValue(0),
        m_BrightnessTS(0),
        m_WindSpeedValue(0),
        m_WindSpeedTS(0),
        m_WindDirectionValue(0),
        m_WindDirectionTS(0),
        m_WindGustValue(0),
        m_WindGustTS(0),
        m_PrecipitationValue(0),
        m_PrecipitationTS(0),
        m_AirPressureValue(0),
        m_AirPressureTS(0),
        m_SunAzimuthValue(0),
        m_SunAzimuthTS(0),
        m_SunAzimuthLastLogTS(0),
        m_SunElevationValue(0),
        m_SunElevationTS(0),
        m_SunElevationLastLogTS(0),
        m_WeatherTS(0) {}
  double m_TemperatureValue;
  DateTime m_TemperatureTS;
  double m_HumidityValue;
  DateTime m_HumidityTS;
  double m_BrightnessValue;
  DateTime m_BrightnessTS;
  double m_WindSpeedValue;
  DateTime m_WindSpeedTS;
  double m_WindDirectionValue;
  DateTime m_WindDirectionTS;
  double m_WindGustValue;
  DateTime m_WindGustTS;
  double m_PrecipitationValue;
  DateTime m_PrecipitationTS;
  double m_AirPressureValue;
  DateTime m_AirPressureTS;
  double m_SunAzimuthValue;
  DateTime m_SunAzimuthTS;
  DateTime m_SunAzimuthLastLogTS;
  double m_SunElevationValue;
  DateTime m_SunElevationTS;
  DateTime m_SunElevationLastLogTS;
  // see https://openweathermap.org/weather-conditions
  std::string m_WeatherIconId;
  std::string m_WeatherConditionId;
  std::string m_WeatherServiceId;
  DateTime m_WeatherTS;
} ApartmentSensorStatus_t;

typedef struct {
  DeviceClasses_t deviceClass;
  std::set<int> lockedScenes;
} ClassLock_t;

typedef struct {
  int zoneID;
  std::vector<ClassLock_t> deviceClassLocks;
} ZoneLock_t;

typedef struct {
  dsuid_t dsuid;
  std::vector<int> lockedScenes;
} DeviceLock_t;

struct ModelSnapshot {};

/** Represents an Apartment
 * This is the root of the datamodel of the dss. The Apartment is responsible for delivering
 * and loading all subitems.
 */
class Apartment : public DeviceContainer {
private:
  DSS* m_dss;
  std::vector<std::shared_ptr<Zone>> m_Zones;
  std::vector<std::shared_ptr<DSMeter>> m_DSMeters;
  std::vector<std::shared_ptr<Device>> m_Devices;
  std::vector<std::shared_ptr<State>> m_States;
  std::vector<std::shared_ptr<Floor>> m_Floors;
  ApartmentSensorStatus_t m_SensorStatus;

  BusInterface* m_pBusInterface;
  PropertyNodePtr m_pPropertyNode;
  ModelMaintenance* m_pModelMaintenance;
  PropertySystem* m_pPropertySystem;
  EventSink* m_eventSink;
  mutable boost::recursive_mutex m_mutex;
  std::unique_ptr<ApartmentProxy> m_proxy;
  ds::Signal<std::shared_ptr<ModelSnapshot>> m_modelStructureChanged; // only Apartment subset of structure changed
  ds::Signal<> m_structureChanged;                                    // includes addons changes to property tree
  ds::Signal<> m_statusChanged;
  ds::Signal<> m_changed;
  struct Impl;
  friend struct Impl;
  const std::unique_ptr<Impl> m_impl;
  std::string m_cache;

  void addDefaultGroupsToZone(std::shared_ptr<Zone> _zone);

public:
  // Apartment is abstract base class, can only be instantiated via subclass
  Apartment(DSS* = nullptr, PropertySystem* = nullptr, BusInterface* = nullptr);
  ~Apartment() override;

  boost::recursive_mutex& getMutex() const { return m_mutex; }
  DSS* getDss() { return m_dss; }

  virtual dsModel::ApartmentId id() const;
  std::vector<dsModel::ZoneId> zoneIds() const;
  std::vector<dsModel::ClusterId> clusterIds() const;
  std::vector<dsModel::DSDeviceId> deviceIds() const;
  std::vector<dsModel::SubmoduleId> submoduleIds() const;
  std::vector<dsModel::FunctionBlockId> functionBlockIds() const;
  std::vector<dsModel::FloorId> floorIds() const;
  std::vector<dsModel::ControllerId> controllerIds() const;

  /// Push event to event queue
  void pushEvent(const std::shared_ptr<Event>&);

  void setName(const std::string& name) override;
  Set getDevices() const override;

  /// Returns a reference to the device with the DSID \a _dsid, nullptr if not found.
  std::shared_ptr<Device> tryGetDeviceByDSID(const dsuid_t&) const;
  /// Returns a reference to the device with the DSID \a _dsid, throws if not found
  std::shared_ptr<Device> getDeviceByDSID(const dsuid_t&) const;
  /** Returns a reference to the device with the name \a _name*/
  std::shared_ptr<Device> getDeviceByName(const std::string& _name);
  // returns copy of internal structure
  std::vector<std::shared_ptr<Device>> getDevicesVector() const { return m_Devices; }
  std::shared_ptr<Device> device(const dsModel::DSDeviceId&) const;
  std::vector<std::shared_ptr<Device>> coupledOEMDevices(const Device& dev);

  /** Returns the Zone by name */
  std::shared_ptr<Zone> getZoneByName(const std::string& _zoneName) const;
  std::weak_ptr<Zone> tryGetZone(const int id) const;
  /** Returns the Zone by its id */
  std::shared_ptr<Zone> getZone(const int id) const;
  std::shared_ptr<Zone> zone(const dsModel::ZoneId&) const;

  /** Returns a vector of all zones */
  std::vector<std::shared_ptr<Zone>> getZones() const;

  /** Return the next free zoneId. -1 if nothing found */
  int getNextFreeZoneId() const;

  std::weak_ptr<Floor> tryGetFloor(const int id) const;
  std::shared_ptr<Floor> getFloor(const int id) const;
  std::shared_ptr<Floor> floor(const dsModel::FloorId&) const;
  std::vector<std::shared_ptr<Floor>> getFloors() const;
  std::shared_ptr<Floor> allocateFloor(int id = FLOOR_ID_NONE);
  void removeFloor(int id);
  /** Return the next free floorId. -1 if nothing found */
  int getNextFreeFloorID() const;
  int getNextFloorOrderID();

  /** Returns a DSMeter by name */
  std::shared_ptr<DSMeter> getDSMeter(const std::string& _modName) const;
  std::shared_ptr<DSMeter> tryGetDSMeterByDSID(const dsuid_t&) const;
  std::shared_ptr<DSMeter> getDSMeterByDSID(const dsuid_t&) const;
  std::shared_ptr<DSMeter> controller(const dsModel::ControllerId&) const;
  std::shared_ptr<DSMeter> dsMeterByDeviceDsuid(const dsuid_t&) const;
  /** Returns a vector of dsMeters without vDSMs */
  std::vector<std::shared_ptr<DSMeter>> getDSMeters() const;
  /** Returns a vector of all dsMeters */
  std::vector<std::shared_ptr<DSMeter>> getDSMetersAll() const;

  /** Returns a Group by name */
  std::shared_ptr<Group> getGroup(const std::string& _name);
  /** Returns a Group by id */
  std::shared_ptr<Group> getGroup(const int _id);

  /** Returns a Cluster by id */
  std::weak_ptr<Cluster> tryGetCluster(const int _id) const;
  std::shared_ptr<Cluster> getCluster(const int _id) const;
  std::shared_ptr<Cluster> cluster(const dsModel::ClusterId& _id);
  /** Returns a vector of Clusters. */
  std::vector<std::shared_ptr<Cluster>> getClusters() const;
  /** Return an Empty Cluster, if available. */
  std::shared_ptr<Cluster> getEmptyCluster();

  /** Returns a vector of Global Application Groups */
  std::vector<std::shared_ptr<Group>> getGlobalApps();

  /// @return nullptr on failure
  std::shared_ptr<State> tryGetState(const eStateType _type, const std::string& _stateName) const;
  /// @throw ItemNotFoundException on failure
  std::shared_ptr<State> getState(const eStateType _type, const std::string& _stateName) const;
  std::shared_ptr<State> tryGetNonScriptState(const std::string& stateName) const;
  std::shared_ptr<State> getNonScriptState(const std::string& _stateName) const;
  /** Returns the state by type, identifier and name */
  std::shared_ptr<State> getState(const eStateType _type, const std::string& _identifier,
                                  const std::string& _stateName) const;

  /// Returns the state of any type, with matching identifier and name
  std::shared_ptr<State> getState(const std::string& identifier, const std::string& name) const;

  void updateSensorStates(const dsuid_t& dsuid, SensorType sensorType, double value, CallOrigin origin);
  void updateSensorStates(int zoneId, int groupId, SensorType sensorType, double value, CallOrigin origin);

  /** Allocates a device and returns a reference to it. */
  std::shared_ptr<Device> allocateDevice(const dsuid_t _dsid);
  /** Allocates a zone and returns a reference to it. Should a zone with
   * the given _zoneID already exist, a reference to the existing zone will
   * be returned.
   * NOTE: Outside code should never call this function
   */
  std::shared_ptr<Zone> allocateZone(int _zoneID);
  std::shared_ptr<DSMeter> allocateDSMeter(const dsuid_t _dsid);
  std::shared_ptr<State> allocateState(eStateType _type, const std::string& _name, const std::string& _scriptId);
  std::shared_ptr<State> allocateState(const std::string& _identifier, const std::string& _scriptId,
                                       std::shared_ptr<Device> _device, SensorType _sensorType,
                                       const std::string& _activateCondition, const std::string& _deactivateCondition);
  std::shared_ptr<State> allocateState(const std::string& _identifier, const std::string& _scriptId,
                                       std::shared_ptr<Group> _group, SensorType _sensorType,
                                       const std::string& _activateCondition, const std::string& _deactivateCondition);
  std::shared_ptr<State> allocateDeviceState(std::shared_ptr<Device> _device, const std::string& stateName);
  std::shared_ptr<State> allocateMeterState(std::shared_ptr<DSMeter> _meter, int _inputIndex);
  std::shared_ptr<State> allocateBinaryInputState(std::shared_ptr<Device> _device, int inputIndex);
  std::shared_ptr<State> allocateGroupState(std::shared_ptr<Group> _group, const std::string& name);

  void removeZone(int _zoneID);
  void removeDevice(dsuid_t _device);
  void removeDSMeter(dsuid_t _dsMeter);
  void removeInactiveMeters();
  void removeState(std::shared_ptr<State> _state);
  std::vector<std::shared_ptr<State>> getStates();

  void aggregateSensorValue(SensorType _sensorType, const dsuid_t& source, double _value);
  std::optional<SensorSourceValue> getAggregatedSensorValue(SensorType _sensorType);
  ApartmentSensorStatus_t getSensorStatus() const;
  void setTemperature(double, DateTime);
  void setHumidityValue(double, DateTime);
  void setBrightnessValue(double, DateTime);
  void setWindSpeed(double, DateTime);
  void setWindDirection(double, DateTime);
  void setWindGust(double, DateTime);
  void setPrecipitation(double, DateTime);
  void setAirPressure(double, DateTime);
  void setSunAzimuth(double, DateTime);
  void setSunAzimuthLastLogTS(DateTime);
  void setSunElevation(double, DateTime);
  void setSunElevationLastLogTS(DateTime);
  void setWeatherInformation(std::string&& iconId, std::string&& conditionId, std::string&& serviceId, DateTime ts);
  void updateWeatherProtectedClusters(const double _sensorValue);
  std::pair<std::vector<DeviceLock_t>, std::vector<ZoneLock_t>> getClusterLocks();

  bool setDevicesFirstSeen(const DateTime& dateTime);

  void setBusInterface(BusInterface* _value) { m_pBusInterface = _value; }
  BusInterface* getBusInterface() { return m_pBusInterface; }
  ActionRequestInterface* getActionRequestInterface();
  DeviceBusInterface* getDeviceBusInterface();
  /** Returns the root-node for the apartment tree */
  PropertyNodePtr getPropertyNode() { return m_pPropertyNode; }
  void setModelMaintenance(ModelMaintenance* _value) { m_pModelMaintenance = _value; }
  ModelMaintenance* getModelMaintenance() { return m_pModelMaintenance; }
  virtual void setMetering(metering::Metering* _value) = 0;
  virtual metering::Metering& getMetering() = 0;
  void setEventSink(EventSink& x) { m_eventSink = &x; }
  PropertySystem* getPropertySystem() { return m_pPropertySystem; }

  virtual ModelFeatures* getModelFeatures() = 0;

  std::vector<dsModel::ApartmentApplicationType> apartmentApplications() const;

  void setProxy(std::unique_ptr<ApartmentProxy> proxy);
  ApartmentProxy& proxy() const;

  ds::Slots<>& structureChanged() { return m_structureChanged; }
  void emitStructureChanged() { m_structureChanged.emit(); }
  ds::Slots<std::shared_ptr<ModelSnapshot>>& modelStructureChanged() { return m_modelStructureChanged; }
  void emitModelStructureChanged(std::shared_ptr<ModelSnapshot> modelSnapshot) {
    m_modelStructureChanged.emit(std::move(modelSnapshot));
  }

  ds::Slots<>& statusChanged() { return m_statusChanged; }
  void emitStatusChanged() { m_statusChanged.emit(); }

  // apartment attributes changed
  ds::Slots<>& changed() { return m_changed; }

  void refreshCache() { 
    m_cache = "";
    JSONWriter json;
    toJSON(*this, json);
    m_cache = json.successJSON();
  }
  std::string getCache() { return m_cache; }

  AggregatedWindSpeed m_aggregateWindSpeed;
  AggregatedWindGust m_aggregateWindGust;
  AggregatedTemperature m_aggregateTemperature;
  AggregatedBrightness m_aggregateBrightness;
  AggregatedHumidity m_aggregateHumidity;
  AggregatedPrecipitation m_aggregatePrecipitation;
  AggregatedAirPressure m_aggregateAirPressure;
}; // Apartment

/** Exception that will be thrown if a given item could not be found */
class ItemNotFoundException : public DSSException {
public:
  ItemNotFoundException(const std::string& name);
};

std::unique_ptr<Apartment> makeApartment(DSS* = nullptr, PropertySystem* = nullptr, BusInterface* = nullptr);

} // namespace dss
