/*
    Copyright (c) 2019 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation," " either version 3 of the License," " or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful," "
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not," " see <http://www.gnu.org/licenses/>.
*/

#include "model/buttonattributes.h"

#include <ds/catch/catch.h>
#include <ds/log.h>

#include "model/device.h"

static ds::LogChannel dsLogChannel("dssModelButtonAttributesTest");
static const char* TAGS = "[model][device][attributes]";

using namespace dss;
using namespace dss::model;
using namespace trompeloeil;
using namespace std::literals;
using trompeloeil::_;

// AKM JSON API strings, don't change!
#define BUTTONINPUT_AKM_STANDARD "standard"
#define BUTTONINPUT_AKM_INVERTED "inverted"
#define BUTTONINPUT_AKM_ON_RISING_EDGE "on_rising_edge"
#define BUTTONINPUT_AKM_ON_FALLING_EDGE "on_falling_edge"
#define BUTTONINPUT_AKM_OFF_RISING_EDGE "off_rising_edge"
#define BUTTONINPUT_AKM_OFF_FALLING_EDGE "off_falling_edge"
#define BUTTONINPUT_AKM_RISING_EDGE "rising_edge"
#define BUTTONINPUT_AKM_FALLING_EDGE "falling_edge"

namespace {

SCENARIO("Attributes", TAGS) {
  SECTION("automation input") {
    SECTION("json api parameters remain supported") {
      struct Y {
        std::string legacyArgument;
        AutomationInputMode automationInput;
      };
      const auto& y = GENERATE(
          values(std::initializer_list<Y>{{BUTTONINPUT_AKM_STANDARD, AutomationInputMode::standard},
                                          {BUTTONINPUT_AKM_INVERTED, AutomationInputMode::inverted},
                                          {BUTTONINPUT_AKM_ON_RISING_EDGE, AutomationInputMode::on_rising_edge},
                                          {BUTTONINPUT_AKM_ON_FALLING_EDGE, AutomationInputMode::on_falling_edge},
                                          {BUTTONINPUT_AKM_OFF_RISING_EDGE, AutomationInputMode::off_rising_edge},
                                          {BUTTONINPUT_AKM_OFF_FALLING_EDGE, AutomationInputMode::off_falling_edge},
                                          {BUTTONINPUT_AKM_RISING_EDGE, AutomationInputMode::rising_edge},
                                          {BUTTONINPUT_AKM_FALLING_EDGE, AutomationInputMode::falling_edge}}));

      INFO(ds::str(y.legacyArgument));
      CHECK(ds::toEnum<AutomationInputMode>(y.legacyArgument) == y.automationInput);
    }
  }
  SECTION("convert ButtonInputMode enum to AutomationInputType") {
    struct Y {
      ButtonInputMode buttonInputMode;
      std::optional<AutomationInputMode> automationInputMode;
    };
    const auto& y = GENERATE(
        values(std::initializer_list<Y>{{ButtonInputMode::AKM_STANDARD, AutomationInputMode::standard},
                                        {ButtonInputMode::AKM_INVERTED, AutomationInputMode::inverted},
                                        {ButtonInputMode::AKM_ON_RISING_EDGE, AutomationInputMode::on_rising_edge},
                                        {ButtonInputMode::AKM_ON_FALLING_EDGE, AutomationInputMode::on_falling_edge},
                                        {ButtonInputMode::AKM_OFF_RISING_EDGE, AutomationInputMode::off_rising_edge},
                                        {ButtonInputMode::AKM_OFF_FALLING_EDGE, AutomationInputMode::off_falling_edge},
                                        {ButtonInputMode::AKM_RISING_EDGE, AutomationInputMode::rising_edge},
                                        {ButtonInputMode::AKM_FALLING_EDGE, AutomationInputMode::falling_edge},
                                        {ButtonInputMode::STANDARD, std::nullopt},
                                        {ButtonInputMode::TURBO, std::nullopt},
                                        {ButtonInputMode::SWITCHED, std::nullopt},
                                        {ButtonInputMode::TWO_WAY_DW_WITH_INPUT1, std::nullopt},
                                        {ButtonInputMode::TWO_WAY_DW_WITH_INPUT2, std::nullopt},
                                        {ButtonInputMode::TWO_WAY_DW_WITH_INPUT3, std::nullopt},
                                        {ButtonInputMode::TWO_WAY_DW_WITH_INPUT4, std::nullopt},
                                        {ButtonInputMode::TWO_WAY_UP_WITH_INPUT1, std::nullopt},
                                        {ButtonInputMode::TWO_WAY_UP_WITH_INPUT2, std::nullopt},
                                        {ButtonInputMode::TWO_WAY_UP_WITH_INPUT3, std::nullopt},
                                        {ButtonInputMode::TWO_WAY_UP_WITH_INPUT4, std::nullopt},
                                        {ButtonInputMode::TWO_WAY, std::nullopt},
                                        {ButtonInputMode::ONE_WAY, std::nullopt},
                                        {ButtonInputMode::HEATING_PUSHBUTTON, std::nullopt},
                                        {ButtonInputMode::DEACTIVATED, std::nullopt}}));
    INFO(ds::str(y.buttonInputMode));
    CHECK(ds::tryToEnum<AutomationInputMode>(y.buttonInputMode) == y.automationInputMode);
  }

  SECTION("convert AutomationInputType enum to ButtonInputMode") {
    struct Y {
      AutomationInputMode automationInputMode;
      std::optional<ButtonInputMode> buttonInputMode;
    };
    const auto& y = GENERATE(
        values(std::initializer_list<Y>{{AutomationInputMode::standard, ButtonInputMode::AKM_STANDARD},
                                        {AutomationInputMode::inverted, ButtonInputMode::AKM_INVERTED},
                                        {AutomationInputMode::on_rising_edge, ButtonInputMode::AKM_ON_RISING_EDGE},
                                        {AutomationInputMode::on_falling_edge, ButtonInputMode::AKM_ON_FALLING_EDGE},
                                        {AutomationInputMode::off_rising_edge, ButtonInputMode::AKM_OFF_RISING_EDGE},
                                        {AutomationInputMode::off_falling_edge, ButtonInputMode::AKM_OFF_FALLING_EDGE},
                                        {AutomationInputMode::rising_edge, ButtonInputMode::AKM_RISING_EDGE},
                                        {AutomationInputMode::falling_edge, ButtonInputMode::AKM_FALLING_EDGE},
                                        {AutomationInputMode::none, std::nullopt}}));

    INFO(ds::str(y.automationInputMode));
    CHECK(ds::tryToEnum<ButtonInputMode>(y.automationInputMode) == y.buttonInputMode);
  }
}

} // namespace
