/*
    Copyright (c) 2019 digitalSTROM AG, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/

#include "model/device-mock.h"

#include <ds/asio/test.h>
#include <ds/catch/catch.h>
#include <ds/log.h>

#include "businterface-mock.h"
#include "ds-model.h"
#include "model/apartment.h"
#include "model/modelmaintenance-mock.h"
#include "model/shademanipulator.h"

static ds::LogChannel dsLogChannel("deviceMockTest");

using namespace dss;
using namespace dss::model;
using namespace trompeloeil;
using trompeloeil::_;

namespace {

static const char* TAGS = "[model][device][mock]";

TEST_CASE("enforce basic specification and input/output count", TAGS) {
  auto pApartment = makeApartment(nullptr);
  auto& apartment = *pApartment;

  GIVEN("BL-KM200") {
    auto dev = createDevice(apartment, {.hardwareType = HardwareType::BL_KM200});
    CHECK(dev->getProductID() == 0x00C8);
    CHECK(dev->getFunctionID() == 0x3115);
    CHECK(dev->getVendorID() == 2);
    CHECK(dev->getDeviceClass() == DeviceClasses_t::DEVICE_CLASS_BL);
    CHECK(dev->getDeviceType() == DEVICE_TYPE_KM);
    CHECK(dev->deviceCount() == 1);
    CHECK(dev->hasOutput() == true);
    CHECK(dev->hasInput() == true);
    CHECK(dev->hasMultibuttons() == false);
    CHECK(dev->hardwareType() == HardwareType::BL_KM200);
    CHECK(dev->getIconPath() == "km_blue.png");
  }

  GIVEN("BL-KM300") {
    auto dev = createDevice(apartment, {.hardwareType = HardwareType::BL_KM300});
    CHECK(dev->getProductID() == 0x012C);
    CHECK(dev->getFunctionID() == 0x31D0);
    CHECK(dev->getDeviceClass() == DeviceClasses_t::DEVICE_CLASS_BL);
    CHECK(dev->getDeviceType() == DEVICE_TYPE_KM);
    CHECK(dev->deviceCount() == 1);
    CHECK(dev->hasOutput() == true);
    CHECK(dev->hasInput() == false);
    CHECK(dev->hardwareType() == HardwareType::BL_KM300);
    CHECK(dev->getIconPath() == "km300_blue.png");
  }

  GIVEN("BL-SDS200") {
    auto dev = createDevice(apartment, {.hardwareType = HardwareType::BL_SDS200});
    CHECK(dev->getProductID() == 0x18C8);
    CHECK(dev->getFunctionID() == 0x3102);
    CHECK(dev->getDeviceClass() == DeviceClasses_t::DEVICE_CLASS_BL);
    CHECK(dev->getDeviceType() == DEVICE_TYPE_SDS);
    CHECK(dev->deviceCount() == 2);
    CHECK(dev->hasOutput() == false);
    CHECK(dev->hasInput() == true);
    CHECK(dev->hasMultibuttons() == true);
    CHECK(dev->hardwareType() == HardwareType::BL_SDS200);
    CHECK(dev->getIconPath() == "sds_blue.png");
  }

  GIVEN("BL-SK204") {
    auto dev = createDevice(apartment, {.hardwareType = HardwareType::BL_SK204});
    CHECK(dev->getProductID() == 0x1CCC);
    CHECK(dev->getFunctionID() == 0x31C9);
    CHECK(dev->getDeviceClass() == DeviceClasses_t::DEVICE_CLASS_BL);
    CHECK(dev->getDeviceType() == DEVICE_TYPE_SK);
    CHECK(dev->deviceCount() == 2);
    CHECK(dev->hasOutput() == false);
    CHECK(dev->hasInput() == true);
    CHECK(dev->hasMultibuttons() == false);
    CHECK(dev->hardwareType() == HardwareType::BL_SK204);
    CHECK(dev->getIconPath() == "sk_blue.png");
  }

  GIVEN("GE-KM200") {
    auto dev = createDevice(apartment, {.hardwareType = HardwareType::GE_KM200});
    CHECK(dev->getProductID() == 0x00C8);
    CHECK(dev->getFunctionID() == 0x1111);
    CHECK(dev->getDeviceClass() == DeviceClasses_t::DEVICE_CLASS_GE);
    CHECK(dev->getDeviceType() == DEVICE_TYPE_KM);
    CHECK(dev->hasOutput() == true);
    CHECK(dev->hasInput() == true);
    CHECK(dev->hasMultibuttons() == false);
    CHECK(dev->hardwareType() == HardwareType::GE_KM200);
    CHECK(dev->getIconPath() == "km_yellow.png");
  }

  GIVEN("GE-KM300") {
    auto dev = createDevice(apartment, {.hardwareType = HardwareType::GE_KM300});
    CHECK(dev->getProductID() == 0x012C);
    CHECK(dev->getFunctionID() == 0x1110);
    CHECK(dev->getDeviceClass() == DeviceClasses_t::DEVICE_CLASS_GE);
    CHECK(dev->getDeviceType() == DEVICE_TYPE_KM);
    CHECK(dev->deviceCount() == 1);
    CHECK(dev->hasOutput() == true);
    CHECK(dev->hasInput() == false);
    CHECK(dev->hardwareType() == HardwareType::GE_KM300);
    CHECK(dev->getIconPath() == "km300_yellow.png");
  }

  GIVEN("GE-TKM300") {
    auto dev = createDevice(apartment, {.hardwareType = HardwareType::GE_TKM300});
    CHECK(dev->getProductID() == 0x052C);
    CHECK(dev->getFunctionID() == 0x1119);
    CHECK(dev->getDeviceClass() == DeviceClasses_t::DEVICE_CLASS_GE);
    CHECK(dev->getDeviceType() == DEVICE_TYPE_TKM);
    CHECK(dev->deviceCount() == 1);
    CHECK(dev->hasOutput() == true);
    CHECK(dev->hasInput() == true);
    CHECK(dev->hasMultibuttons() == false);
    CHECK(dev->hardwareType() == HardwareType::GE_TKM300);
    CHECK(dev->getIconPath() == "tkm300_yellow.png");
  }
  GIVEN("GE-KL200") {
    auto dev = createDevice(apartment, {.hardwareType = HardwareType::GE_KL200});
    CHECK(dev->getProductID() == 0x0CC8);
    CHECK(dev->getFunctionID() == 0x1111);
    CHECK(dev->getDeviceClass() == DeviceClasses_t::DEVICE_CLASS_GE);
    CHECK(dev->getDeviceType() == DEVICE_TYPE_KL);
    CHECK(dev->deviceCount() == 1);
    CHECK(dev->hasOutput() == true);
    CHECK(dev->hasInput() == true);
    CHECK(dev->hasMultibuttons() == false);
    CHECK(dev->hardwareType() == HardwareType::GE_KL200);
    CHECK(dev->getIconPath() == "kl_yellow.png");
  }

  GIVEN("GE-TKM210") {
    auto dev = createDevice(apartment, {.hardwareType = HardwareType::GE_TKM210});
    CHECK(dev->getProductID() == 0x04D2);
    CHECK(dev->getFunctionID() == 0x1111);
    CHECK(dev->getDeviceClass() == DeviceClasses_t::DEVICE_CLASS_GE);
    CHECK(dev->getDeviceType() == DEVICE_TYPE_TKM);
    CHECK(dev->deviceCount() == 1);
    CHECK(dev->hasOutput() == true);
    CHECK(dev->hasInput() == true);
    CHECK(dev->hasMultibuttons() == false);
    CHECK(dev->hardwareType() == HardwareType::GE_TKM210);
    CHECK(dev->getIconPath() == "tkm_yellow.png");
  }

  GIVEN("GE-TKM220") {
    auto dev = createDevice(apartment, {.hardwareType = HardwareType::GE_TKM220});
    CHECK(dev->getProductID() == 0x04DC);
    CHECK(dev->getFunctionID() == 0x1101);
    CHECK(dev->getDeviceClass() == DeviceClasses_t::DEVICE_CLASS_GE);
    CHECK(dev->getDeviceType() == DEVICE_TYPE_TKM);
    CHECK(dev->deviceCount() == 1);
    CHECK(dev->hasOutput() == false);
    CHECK(dev->hasInput() == true);
    CHECK(dev->hasMultibuttons() == false);
    CHECK(dev->hardwareType() == HardwareType::GE_TKM220);
    CHECK(dev->getIconPath() == "tkm_yellow.png");
  }

  GIVEN("GE-TKM230") {
    auto dev = createDevice(apartment, {.hardwareType = HardwareType::GE_TKM230});
    CHECK(dev->getProductID() == 0x04E6);
    CHECK(dev->getFunctionID() == 0x1102);
    CHECK(dev->getDeviceClass() == DeviceClasses_t::DEVICE_CLASS_GE);
    CHECK(dev->getDeviceType() == DEVICE_TYPE_TKM);
    CHECK(dev->deviceCount() == 2);
    CHECK(dev->hasOutput() == false);
    CHECK(dev->hasInput() == true);
    CHECK(dev->hasMultibuttons() == true);
    CHECK(dev->hardwareType() == HardwareType::GE_TKM230);
    CHECK(dev->getIconPath() == "tkm_yellow.png");
  }

  GIVEN("GE-SDM200") {
    auto dev = createDevice(apartment, {.hardwareType = HardwareType::GE_SDM200});
    CHECK(dev->getProductID() == 0x08C8);
    CHECK(dev->getFunctionID() == 0x1111);
    CHECK(dev->getDeviceClass() == DeviceClasses_t::DEVICE_CLASS_GE);
    CHECK(dev->getDeviceType() == DEVICE_TYPE_SDM);
    CHECK(dev->deviceCount() == 1);
    CHECK(dev->hasOutput() == true);
    CHECK(dev->hasInput() == true);
    CHECK(dev->hasMultibuttons() == false);
    CHECK(dev->hardwareType() == HardwareType::GE_SDM200);
    CHECK(dev->getIconPath() == "sdm_yellow.png");
  }

  GIVEN("GE-SDM200JS") {
    auto dev = createDevice(apartment, {.hardwareType = HardwareType::GE_SDM200JS});
    CHECK(dev->getProductID() == 0x08C9);
    CHECK(dev->getFunctionID() == 0x1111);
    CHECK(dev->getDeviceClass() == DeviceClasses_t::DEVICE_CLASS_GE);
    CHECK(dev->getDeviceType() == DEVICE_TYPE_SDM);
    CHECK(dev->deviceCount() == 1);
    CHECK(dev->hasOutput() == true);
    CHECK(dev->hasInput() == true);
    CHECK(dev->hasMultibuttons() == false);
    CHECK(dev->hardwareType() == HardwareType::GE_SDM200JS);
    CHECK(dev->getIconPath() == "sdm_plug_yellow.png");
  }

  GIVEN("GE-SDM200CS") {
    auto dev = createDevice(apartment, {.hardwareType = HardwareType::GE_SDM200CS});
    CHECK(dev->getProductID() == 0x08CA);
    CHECK(dev->getFunctionID() == 0x1111);
    CHECK(dev->getDeviceClass() == DeviceClasses_t::DEVICE_CLASS_GE);
    CHECK(dev->getDeviceType() == DEVICE_TYPE_SDM);
    CHECK(dev->deviceCount() == 1);
    CHECK(dev->hasOutput() == true);
    CHECK(dev->hasInput() == true);
    CHECK(dev->hasMultibuttons() == false);
    CHECK(dev->hardwareType() == HardwareType::GE_SDM200CS);
    CHECK(dev->getIconPath() == "sdm_plug_yellow.png");
  }

  GIVEN("GE-SDS200CW") {
    auto dev = createDevice(apartment, {.hardwareType = HardwareType::GE_SDS200CW});
    CHECK(dev->getProductID() == 0x18C8);
    CHECK(dev->getFunctionID() == 0x1112);
    CHECK(dev->getDeviceClass() == DeviceClasses_t::DEVICE_CLASS_GE);
    CHECK(dev->getDeviceType() == DEVICE_TYPE_SDS);
    CHECK(dev->deviceCount() == 2);
    CHECK(dev->hasOutput() == true);
    CHECK(dev->hasInput() == true);
    CHECK(dev->hasMultibuttons() == true);
    CHECK(dev->hardwareType() == HardwareType::GE_SDS200CW);
    CHECK(dev->getIconPath() == "sds_yellow.png");
  }

  GIVEN("GE-SDS200CS") {
    auto dev = createDevice(apartment, {.hardwareType = HardwareType::GE_SDS200CS});
    CHECK(dev->getProductID() == 0x18C9);
    CHECK(dev->getFunctionID() == 0x1112);
    CHECK(dev->getDeviceClass() == DeviceClasses_t::DEVICE_CLASS_GE);
    CHECK(dev->getDeviceType() == DEVICE_TYPE_SDS);
    CHECK(dev->deviceCount() == 2);
    CHECK(dev->hasOutput() == true);
    CHECK(dev->hasInput() == true);
    CHECK(dev->hasMultibuttons() == true);
    CHECK(dev->hardwareType() == HardwareType::GE_SDS200CS);
    CHECK(dev->getIconPath() == "sds_yellow.png");
  }

  GIVEN("GE-SDS220CT") {
    auto dev = createDevice(apartment, {.hardwareType = HardwareType::GE_SDS220CT});
    CHECK(dev->getProductID() == 0x18DC);
    CHECK(dev->getFunctionID() == 0x1112);
    CHECK(dev->getVendorID() == 2);
    CHECK(dev->getDeviceClass() == DeviceClasses_t::DEVICE_CLASS_GE);
    CHECK(dev->getDeviceType() == DEVICE_TYPE_SDS);
    CHECK(dev->deviceCount() == 2);
    CHECK(dev->hasOutput() == true);
    CHECK(dev->hasInput() == true);
    CHECK(dev->hasMultibuttons() == true);
    CHECK(dev->hardwareType() == HardwareType::GE_SDS220CT);
    CHECK(dev->getIconPath() == "sds_yellow.png");
  }

  GIVEN("GE-ZWD300W") {
    auto dev = createDevice(apartment, {.hardwareType = HardwareType::GE_ZWD300W});
    CHECK(dev->getProductID() == 0x3D2C);
    CHECK(dev->getFunctionID() == 0x11D0);
    CHECK(dev->getDeviceClass() == DeviceClasses_t::DEVICE_CLASS_GE);
    CHECK(dev->getDeviceType() == DEVICE_TYPE_ZWD);
    CHECK(dev->hardwareType() == HardwareType::GE_ZWD300W);
    CHECK(dev->deviceCount() == 1);
    CHECK(dev->hasOutput() == true);
    CHECK(dev->hasInput() == false);
    CHECK(dev->getIconPath() == "zwd_yellow.png");
  }

  GIVEN("GE-SDM300W") {
    auto dev = createDevice(apartment, {.hardwareType = HardwareType::GE_SDM300W});
    CHECK(dev->getProductID() == 0x092C);
    CHECK(dev->getFunctionID() == 0x11D9);
    CHECK(dev->getDeviceClass() == DeviceClasses_t::DEVICE_CLASS_GE);
    CHECK(dev->getDeviceType() == DEVICE_TYPE_SDM);
    CHECK(dev->hardwareType() == HardwareType::GE_SDM300W);
    CHECK(dev->deviceCount() == 1);
    CHECK(dev->hasOutput() == true);
    CHECK(dev->hasInput() == true);
    CHECK(dev->hasMultibuttons() == false);
    CHECK(dev->getIconPath() == "sdm300_yellow.png");
  }

  GIVEN("GE-SDM300S") {
    auto dev = createDevice(apartment, {.hardwareType = HardwareType::GE_SDM300S});
    CHECK(dev->getProductID() == 0x092D);
    CHECK(dev->getFunctionID() == 0x11D9);
    CHECK(dev->getDeviceClass() == DeviceClasses_t::DEVICE_CLASS_GE);
    CHECK(dev->getDeviceType() == DEVICE_TYPE_SDM);
    CHECK(dev->hardwareType() == HardwareType::GE_SDM300S);
    CHECK(dev->hasOutput() == true);
    CHECK(dev->hasInput() == true);
    CHECK(dev->hasMultibuttons() == false);
    CHECK(dev->getIconPath() == "sdm300_yellow.png");
  }

  GIVEN("GE-UMV200 input:button") {
    auto dev = createDevice(apartment, DeviceMockConf{.hardwareType = HardwareType::GE_UMV200});
    CHECK(dev->getProductID() == 0x2CC8);
    CHECK(dev->getFunctionID() == 0x1111);
    CHECK(dev->getDeviceClass() == DeviceClasses_t::DEVICE_CLASS_GE);
    CHECK(dev->getDeviceType() == DEVICE_TYPE_UMV);
    CHECK(dev->hardwareType() == HardwareType::GE_UMV200);
    CHECK(dev->deviceCount() == 4);
    CHECK(dev->hasOutput() == true);
    CHECK(dev->hasInput() == true);
    CHECK(dev->hasMultibuttons() == false);
    CHECK(dev->getActiveGroup() == ColorIDYellow);
    CHECK(dev->getIconPath() == "umv_yellow.png");
  }

  GIVEN("BL-UMV200 input:button") {
    auto dev = createDevice(apartment, DeviceMockConf{.hardwareType = HardwareType::BL_UMV200});
    CHECK(dev->getFunctionID() == 0x3111);
    CHECK(dev->getProductID() == 0x2CC8);
    CHECK(dev->getDeviceClass() == DeviceClasses_t::DEVICE_CLASS_BL);
    CHECK(dev->getDeviceType() == DEVICE_TYPE_UMV);
    CHECK(dev->hardwareType() == HardwareType::BL_UMV200);
    CHECK(dev->deviceCount() == 4);
    CHECK(dev->hasOutput() == true);
    CHECK(dev->hasInput() == true);
    CHECK(dev->hasMultibuttons() == false);
    CHECK(dev->getActiveGroup() == ColorIDBlue);
    CHECK(dev->getIconPath() == "umv_blue.png");
  }

  GIVEN("RT-TKM200") {
    auto dev = createDevice(apartment, {.hardwareType = HardwareType::RT_TKM200});
    CHECK(dev->getProductID() == 0x04C8);
    CHECK(dev->getFunctionID() == 0x6001);
    CHECK(dev->getDeviceClass() == DeviceClasses_t::DEVICE_CLASS_RT);
    CHECK(dev->getDeviceType() == DEVICE_TYPE_TKM);
    CHECK(dev->hardwareType() == HardwareType::RT_TKM200);
    CHECK(dev->deviceCount() == 1);
    CHECK(dev->hasOutput() == true);
    CHECK(dev->hasInput() == false);
    CHECK(dev->getIconPath() == "tkm_red.png");
  }

  GIVEN("RT-SDM200") {
    auto dev = createDevice(apartment, {.hardwareType = HardwareType::RT_SDM200});
    CHECK(dev->getProductID() == 0x08C8);
    CHECK(dev->getFunctionID() == 0x6001);
    CHECK(dev->getDeviceClass() == DeviceClasses_t::DEVICE_CLASS_RT);
    CHECK(dev->getDeviceType() == DEVICE_TYPE_SDM);
    CHECK(dev->hardwareType() == HardwareType::RT_SDM200);
    CHECK(dev->deviceCount() == 1);
    CHECK(dev->hasOutput() == true);
    CHECK(dev->hasInput() == false);
    CHECK(dev->getIconPath() == "sdm_red.png");
  }

  GIVEN("SW-AKM200") {
    auto dev = createDevice(apartment, {.hardwareType = HardwareType::SW_AKM200});
    CHECK(dev->getProductID() == 0x20C8);
    CHECK(dev->getFunctionID() == 0x810b);
    CHECK(dev->getDeviceClass() == DeviceClasses_t::DEVICE_CLASS_SW);
    CHECK(dev->getDeviceType() == DEVICE_TYPE_AKM);
    CHECK(dev->hardwareType() == HardwareType::SW_AKM200);
    CHECK(dev->deviceCount() == 4);
    CHECK(dev->hasOutput() == false);
    CHECK(dev->hasInput() == true);
    CHECK(dev->hasMultibuttons() == true);
    CHECK(dev->isBinaryInput() == true);
    CHECK(dev->getIconPath() == "akm_black_black.png");
  }

  GIVEN("SW-KL200") {
    auto dev = createDevice(apartment, {.hardwareType = HardwareType::SW_KL200});
    CHECK(dev->getProductID() == 0x0CC8);
    CHECK(dev->getFunctionID() == 0x8111);
    CHECK(dev->getDeviceClass() == DeviceClasses_t::DEVICE_CLASS_SW);
    CHECK(dev->getDeviceType() == DEVICE_TYPE_KL);
    CHECK(dev->hardwareType() == HardwareType::SW_KL200);
    CHECK(dev->deviceCount() == 1);
    CHECK(dev->hasOutput() == true);
    CHECK(dev->hasInput() == true);
    CHECK(dev->hasMultibuttons() == false);
    CHECK(dev->getIconPath() == "kl_black_black.png");
  }

  GIVEN("SW-SSL200JS") {
    auto dev = createDevice(apartment, {.hardwareType = HardwareType::SW_SSL200JS});
    CHECK(dev->getProductID() == 0x0CD5);
    CHECK(dev->getFunctionID() == 0x8110);
    CHECK(dev->getDeviceClass() == DeviceClasses_t::DEVICE_CLASS_SW);
    CHECK(dev->getDeviceType() == DEVICE_TYPE_KL);
    CHECK(dev->hardwareType() == HardwareType::SW_SSL200JS);
    CHECK(dev->deviceCount() == 1);
    CHECK(dev->hasOutput() == true);
    CHECK(dev->hasInput() == false);
    CHECK(dev->getIconPath() == "ssl_black_black.png");
  }

  GIVEN("SW-SSL200FS") {
    auto dev = createDevice(apartment, {.hardwareType = HardwareType::SW_SSL200FS});
    CHECK(dev->getProductID() == 0x0CD6);
    CHECK(dev->getFunctionID() == 0x8110);
    CHECK(dev->getDeviceClass() == DeviceClasses_t::DEVICE_CLASS_SW);
    CHECK(dev->getDeviceType() == DEVICE_TYPE_KL);
    CHECK(dev->hardwareType() == HardwareType::SW_SSL200FS);
    CHECK(dev->deviceCount() == 1);
    CHECK(dev->hasOutput() == true);
    CHECK(dev->hasInput() == false);
    CHECK(dev->getIconPath() == "ssl_black_black.png");
  }

  GIVEN("SW-SK200") {
    auto dev = createDevice(apartment, {.hardwareType = HardwareType::SW_SK200});
    CHECK(dev->getProductID() == 0x1CC8);
    CHECK(dev->getFunctionID() == 0x8108);
    CHECK(dev->getVendorID() == 2);
    CHECK(dev->getDeviceClass() == DeviceClasses_t::DEVICE_CLASS_SW);
    CHECK(dev->getDeviceType() == DEVICE_TYPE_SK);
    CHECK(dev->hardwareType() == HardwareType::SW_SK200);
    CHECK(dev->deviceCount() == 1);
    CHECK(dev->hasOutput() == false);
    CHECK(dev->hasInput() == false);
    CHECK(dev->getIconPath() == "sk_black_black.png");
  }

  GIVEN("SW-SK201") {
    auto dev = createDevice(apartment, {.hardwareType = HardwareType::SW_SK201});
    CHECK(dev->getProductID() == 0x1CC9);
    CHECK(dev->getFunctionID() == 0x8108);
    CHECK(dev->getDeviceClass() == DeviceClasses_t::DEVICE_CLASS_SW);
    CHECK(dev->getDeviceType() == DEVICE_TYPE_SK);
    CHECK(dev->hardwareType() == HardwareType::SW_SK201);
    CHECK(dev->deviceCount() == 1);
    CHECK(dev->hasOutput() == false);
    CHECK(dev->hasInput() == false);
    CHECK(dev->getIconPath() == "sk_black_black.png");
  }

  GIVEN("SW-SK202") {
    auto dev = createDevice(apartment, {.hardwareType = HardwareType::SW_SK202});
    CHECK(dev->getProductID() == 0x1CCA);
    CHECK(dev->getFunctionID() == 0x8108);
    CHECK(dev->getDeviceClass() == DeviceClasses_t::DEVICE_CLASS_SW);
    CHECK(dev->getDeviceType() == DEVICE_TYPE_SK);
    CHECK(dev->hardwareType() == HardwareType::SW_SK202);
    CHECK(dev->deviceCount() == 1);
    CHECK(dev->hasOutput() == false);
    CHECK(dev->hasInput() == false);
    CHECK(dev->getIconPath() == "sk_black_black.png");
  }
  GIVEN("SW-SDS200CW") {
    auto dev = createDevice(apartment, {.hardwareType = HardwareType::SW_SDS200CW});
    CHECK(dev->getProductID() == 0x18C8);
    CHECK(dev->getFunctionID() == 0x8102);
    CHECK(dev->getDeviceClass() == DeviceClasses_t::DEVICE_CLASS_SW);
    CHECK(dev->getDeviceType() == DEVICE_TYPE_SDS);
    CHECK(dev->hardwareType() == HardwareType::SW_SDS200CW);
    CHECK(dev->deviceCount() == 2);
    CHECK(dev->hasOutput() == false);
    CHECK(dev->hasInput() == true);
    CHECK(dev->hasMultibuttons() == true);
    CHECK(dev->getIconPath() == "sds_black_black.png");
  }

  GIVEN("SW-SDS200CS") {
    auto dev = createDevice(apartment, {.hardwareType = HardwareType::SW_SDS200CS});
    CHECK(dev->getProductID() == 0x18C9);
    CHECK(dev->getFunctionID() == 0x8102);
    CHECK(dev->getDeviceClass() == DeviceClasses_t::DEVICE_CLASS_SW);
    CHECK(dev->getDeviceType() == DEVICE_TYPE_SDS);
    CHECK(dev->hardwareType() == HardwareType::SW_SDS200CS);
    CHECK(dev->deviceCount() == 2);
    CHECK(dev->hasOutput() == false);
    CHECK(dev->hasInput() == true);
    CHECK(dev->hasMultibuttons() == true);
    CHECK(dev->getIconPath() == "sds_black_black.png");
  }

  GIVEN("SW-SDS220CT") {
    auto dev = createDevice(apartment, {.hardwareType = HardwareType::SW_SDS220CT});
    CHECK(dev->getProductID() == 0x18DC);
    CHECK(dev->getFunctionID() == 0x8102);
    CHECK(dev->getDeviceClass() == DeviceClasses_t::DEVICE_CLASS_SW);
    CHECK(dev->getDeviceType() == DEVICE_TYPE_SDS);
    CHECK(dev->hardwareType() == HardwareType::SW_SDS220CT);
    CHECK(dev->deviceCount() == 2);
    CHECK(dev->hasOutput() == false);
    CHECK(dev->hasInput() == true);
    CHECK(dev->getIconPath() == "sds_black_black.png");
  }

  GIVEN("SW-AKM210") {
    auto dev = createDevice(apartment, {.hardwareType = HardwareType::SW_AKM210});
    CHECK(dev->getProductID() == 0x20D2);
    CHECK(dev->getFunctionID() == 0x810A);
    CHECK(dev->getDeviceClass() == DeviceClasses_t::DEVICE_CLASS_SW);
    CHECK(dev->getDeviceType() == DEVICE_TYPE_AKM);
    CHECK(dev->hardwareType() == HardwareType::SW_AKM210);
    CHECK(dev->deviceCount() == 2);
    CHECK(dev->hasOutput() == false);
    CHECK(dev->hasInput() == true);
    CHECK(dev->isBinaryInput() == true);
    CHECK(dev->hasMultibuttons() == true);
    CHECK(dev->getIconPath() == "akm_black_black.png");
  }

  GIVEN("SW-UMR200 input:button") {
    auto terminalBlocks = createSWUMR200(apartment, DeviceMockConf{.hardwareType = HardwareType::SW_UMR200});
    for (auto terminalBlock : terminalBlocks) {
      CHECK(terminalBlock->getDeviceClass() == DeviceClasses_t::DEVICE_CLASS_SW);
      CHECK(terminalBlock->getDeviceType() == DEVICE_TYPE_UMR);
      CHECK(terminalBlock->getProductID() == 0x30C8);
      CHECK(terminalBlock->hardwareType() == HardwareType::SW_UMR200);
      CHECK(terminalBlock->deviceCount() == 4);
      CHECK(terminalBlock->getIconPath() == "umr_black_black.png");
      DS_DEBUG(terminalBlock->multiDeviceIndex());
      switch (terminalBlock->multiDeviceIndex()) {
        // functionID read out from physical device
        case 0:
        case 1:
          CHECK(terminalBlock->getFunctionID() == 0x81ca);
          CHECK(terminalBlock->hasOutput() == false);
          CHECK(terminalBlock->hasInput() == true);
          CHECK(terminalBlock->hasMultibuttons() == true);
          break;
        case 2:
          CHECK(terminalBlock->getFunctionID() == 0x81d8);
          CHECK(terminalBlock->hasOutput() == true);
          CHECK(terminalBlock->hasInput() == false);
          break;
        case 3:
          CHECK(terminalBlock->getFunctionID() == 0x81d0);
          CHECK(terminalBlock->hasOutput() == true);
          CHECK(terminalBlock->hasInput() == false);
          break;
      }
      // needs to be configured after creation
      CHECK(terminalBlock->isBinaryInput() == false);
    }
  }

  GIVEN("SW-TKM200") {
    auto dev = createDevice(apartment, {.hardwareType = HardwareType::SW_TKM200});
    CHECK(dev->getProductID() == 0x04C8);
    CHECK(dev->getFunctionID() == 0x8103);
    CHECK(dev->getDeviceClass() == DeviceClasses_t::DEVICE_CLASS_SW);
    CHECK(dev->getDeviceType() == DEVICE_TYPE_TKM);
    CHECK(dev->hardwareType() == HardwareType::SW_TKM200);
    CHECK(dev->deviceCount() == 4);
    CHECK(dev->hasOutput() == false);
    CHECK(dev->hasInput() == true);
    CHECK(dev->hasMultibuttons() == true);
    CHECK(dev->isBinaryInput() == false);
    CHECK(dev->getIconPath() == "tkm_black_black.png");
  }

  GIVEN("SW-TKM210") {
    auto dev = createDevice(apartment, {.hardwareType = HardwareType::SW_TKM210});
    CHECK(dev->getProductID() == 0x04D2);
    CHECK(dev->getFunctionID() == 0x8102);
    CHECK(dev->getDeviceClass() == DeviceClasses_t::DEVICE_CLASS_SW);
    CHECK(dev->getDeviceType() == DEVICE_TYPE_TKM);
    CHECK(dev->hardwareType() == HardwareType::SW_TKM210);
    CHECK(dev->deviceCount() == 2);
    CHECK(dev->hasOutput() == false);
    CHECK(dev->hasInput() == true);
    CHECK(dev->hasMultibuttons() == true);
    CHECK(dev->isBinaryInput() == false);
    CHECK(dev->getIconPath() == "tkm_black_black.png");
  }

  GIVEN("SW-ZWS200") {
    auto dev = createDevice(apartment, {.hardwareType = HardwareType::SW_ZWS200});
    CHECK(dev->getProductID() == 0x14C8);
    CHECK(dev->getFunctionID() == 0x8111);
    CHECK(dev->getDeviceClass() == DeviceClasses_t::DEVICE_CLASS_SW);
    CHECK(dev->getDeviceType() == DEVICE_TYPE_ZWS);
    CHECK(dev->hardwareType() == HardwareType::SW_ZWS200);
    CHECK(dev->deviceCount() == 1);
    CHECK(dev->hasOutput() == true);
    CHECK(dev->hasInput() == true);
    CHECK(dev->hasMultibuttons() == false);
    CHECK(dev->getIconPath() == "zws_black_black.png");
  }

  GIVEN("SW-ZWS200F") {
    auto dev = createDevice(apartment, {.hardwareType = HardwareType::SW_ZWS200F});
    CHECK(dev->getProductID() == 0x14C9);
    CHECK(dev->getFunctionID() == 0x8111);
    CHECK(dev->getDeviceClass() == DeviceClasses_t::DEVICE_CLASS_SW);
    CHECK(dev->getDeviceType() == DEVICE_TYPE_ZWS);
    CHECK(dev->hardwareType() == HardwareType::SW_ZWS200F);
    CHECK(dev->deviceCount() == 1);
    CHECK(dev->hasOutput() == true);
    CHECK(dev->hasInput() == true);
    CHECK(dev->hasMultibuttons() == false);
    CHECK(dev->getIconPath() == "zws_black_black.png");
  }

  GIVEN("SW-ZWS200EF") {
    auto dev = createDevice(apartment, {.hardwareType = HardwareType::SW_ZWS200EF});
    CHECK(dev->getProductID() == 0x14CA);
    CHECK(dev->getFunctionID() == 0x8111);
    CHECK(dev->getDeviceClass() == DeviceClasses_t::DEVICE_CLASS_SW);
    CHECK(dev->getDeviceType() == DEVICE_TYPE_ZWS);
    CHECK(dev->hardwareType() == HardwareType::SW_ZWS200EF);
    CHECK(dev->deviceCount() == 1);
    CHECK(dev->hasOutput() == true);
    CHECK(dev->hasInput() == true);
    CHECK(dev->hasMultibuttons() == false);
    CHECK(dev->getIconPath() == "zws_black_black.png");
  }

  GIVEN("SW-ZWS205") {
    auto dev = createDevice(apartment, {.hardwareType = HardwareType::SW_ZWS205});
    CHECK(dev->getProductID() == 0x14CD);
    CHECK(dev->getFunctionID() == 0x8111);
    CHECK(dev->getDeviceClass() == DeviceClasses_t::DEVICE_CLASS_SW);
    CHECK(dev->getDeviceType() == DEVICE_TYPE_ZWS);
    CHECK(dev->hardwareType() == HardwareType::SW_ZWS205);
    CHECK(dev->deviceCount() == 1);
    CHECK(dev->hasOutput() == true);
    CHECK(dev->hasInput() == true);
    CHECK(dev->hasMultibuttons() == false);
    CHECK(dev->getIconPath() == "zws205_black_black.png");
  }

  GIVEN("GR-KL200") {
    auto dev = createDevice(apartment, {.hardwareType = HardwareType::GR_KL200});
    CHECK(dev->getProductID() == 0x0CC8);
    CHECK(dev->getFunctionID() == 0x2131);
    CHECK(dev->getDeviceClass() == DeviceClasses_t::DEVICE_CLASS_GR);
    CHECK(dev->getDeviceType() == DEVICE_TYPE_KL);
    CHECK(dev->hardwareType() == HardwareType::GR_KL200);
    CHECK(dev->deviceCount() == 1);
    CHECK(dev->hasOutput() == true);
    CHECK(dev->hasInput() == true);
    CHECK(dev->hasMultibuttons() == false);
    CHECK(dev->getIconPath() == "kl_grey.png");
  }

  GIVEN("GR-KL210") {
    auto dev = createDevice(apartment, {.hardwareType = HardwareType::GR_KL210});
    CHECK(dev->getProductID() == 0x0CD2);
    CHECK(dev->getFunctionID() == 0x2131);
    CHECK(dev->getDeviceClass() == DeviceClasses_t::DEVICE_CLASS_GR);
    CHECK(dev->hardwareType() == HardwareType::GR_KL210);
    CHECK(dev->getDeviceType() == DEVICE_TYPE_KL);
    CHECK(dev->deviceCount() == 1);
    CHECK(dev->hasOutput() == true);
    CHECK(dev->hasInput() == true);
    CHECK(dev->hasMultibuttons() == false);
    CHECK(dev->getIconPath() == "kl_grey.png");
  }

  GIVEN("GR-KL220") {
    auto dev = createDevice(apartment, {.hardwareType = HardwareType::GR_KL220});
    CHECK(dev->getProductID() == 0x0CDC);
    CHECK(dev->getFunctionID() == 0x2131);
    CHECK(dev->getVendorID() == 1);
    CHECK(dev->getDeviceClass() == DeviceClasses_t::DEVICE_CLASS_GR);
    CHECK(dev->getDeviceType() == DEVICE_TYPE_KL);
    CHECK(dev->hardwareType() == HardwareType::GR_KL220);
    CHECK(dev->deviceCount() == 1);
    CHECK(dev->hasOutput() == true);
    CHECK(dev->hasInput() == true);
    CHECK(dev->hasMultibuttons() == false);
    auto features = dev->getDeviceFeatures();
    CHECK(features.hasOutputAngle);
    CHECK(dev->getIconPath() == "kl_grey.png");
  }

  GIVEN("GR-KL300") {
    auto terminalBlocks = createGRKL300(apartment, DeviceMockConf{.hardwareType = HardwareType::GR_KL300});
    for (auto terminalBlock : terminalBlocks) {
      CHECK(terminalBlock->getProductID() == 0x0D2C);
      CHECK(terminalBlock->getDeviceClass() == DeviceClasses_t::DEVICE_CLASS_GR);
      CHECK(terminalBlock->getDeviceType() == DEVICE_TYPE_KL);
      CHECK(terminalBlock->hardwareType() == HardwareType::GR_KL300);
      CHECK(terminalBlock->deviceCount() == 2);
      CHECK(terminalBlock->getActiveGroup() == ColorIDGray);
      CHECK(terminalBlock->getDefaultGroup() == static_cast<int>(ApplicationType::blinds));
      CHECK_THROWS(terminalBlock->getSWThresholdAddress()); // for shades there's no threshold register, 0 off, >0 on
      CHECK(terminalBlock->getButtonInputCount() == 1);
      CHECK(terminalBlock->getButtonInputIndex() == 0); // both are 0!
      switch (terminalBlock->multiDeviceIndex()) {
        case 0:
          CHECK(terminalBlock->getFunctionID() == 0x21f1);
          CHECK(terminalBlock->hasInput());
          CHECK(terminalBlock->hasOutput());
          CHECK(terminalBlock->hasMultibuttons() == false);
          CHECK(terminalBlock->getGroupsCount() == 1);
          CHECK(terminalBlock->isInGroup(ColorIDGray));
          CHECK(terminalBlock->getButtonInputMode() == ButtonInputMode::TWO_WAY_DW_WITH_INPUT2);
          CHECK(terminalBlock->getIconPath() == "kl300_black_grey.png");
          break;
        case 1:
          CHECK(terminalBlock->getFunctionID() == 0x21c9);
          CHECK(terminalBlock->hasInput());
          CHECK(!terminalBlock->hasOutput());
          CHECK(terminalBlock->hasMultibuttons() == false);
          CHECK(terminalBlock->getBinaryInputCount());
          CHECK(terminalBlock->getGroupsCount() == 2); // target group of binary input
          CHECK(terminalBlock->isInGroup(ColorIDGray));
          CHECK(terminalBlock->isInGroup(ColorIDBlack));
          CHECK(terminalBlock->getButtonInputMode() == ButtonInputMode::TWO_WAY_UP_WITH_INPUT1);
          CHECK(terminalBlock->getIconPath() == "kl300_black_grey.png");
          break;
      }
    }
  }

  GIVEN("GR-HKL330") {
    auto dev = createDevice(apartment, {.hardwareType = HardwareType::GR_HKL330});
    CHECK(dev->getProductID() == 0x0D4A);
    CHECK(dev->getFunctionID() == 0x21f0);
    CHECK(dev->getVendorID() == 2);
    CHECK(dev->getDeviceClass() == DeviceClasses_t::DEVICE_CLASS_GR);
    CHECK(dev->getDeviceType() == DEVICE_TYPE_KL);
    CHECK(dev->deviceCount() == 1);
    CHECK(dev->hasOutput());
    CHECK(!dev->hasInput());
    CHECK(dev->hardwareType() == HardwareType::GR_HKL330);
    CHECK(dev->getIconPath() == "hkl330_grey.png");
    CHECK_THROWS(dev->getSWThresholdAddress()); // for shades there's no threshold register, 0 off, >0 on
  }

  GIVEN("GR-HKL230") {
    auto dev = createDevice(apartment, {.hardwareType = HardwareType::GR_HKL230});
    CHECK(dev->getProductID() == 0x0CE6);
    CHECK(dev->getFunctionID() == 0x2130);
    CHECK(dev->getDeviceClass() == DeviceClasses_t::DEVICE_CLASS_GR);
    CHECK(dev->getDeviceType() == DEVICE_TYPE_KL);
    CHECK(dev->hardwareType() == HardwareType::GR_HKL230);
    CHECK(dev->deviceCount() == 1);
    CHECK(dev->hasOutput() == true);
    CHECK(dev->hasInput() == false);
    auto features = dev->getDeviceFeatures();
    CHECK(features.hasOutputAngle);
    CHECK(dev->getIconPath() == "kl_grey.png");
  }

  GIVEN("GR-TKM200") {
    auto dev = createDevice(apartment, {.hardwareType = HardwareType::GR_TKM200});
    CHECK(dev->getProductID() == 0x04C8);
    CHECK(dev->getFunctionID() == 0x2103);
    CHECK(dev->getDeviceClass() == DeviceClasses_t::DEVICE_CLASS_GR);
    CHECK(dev->getDeviceType() == DEVICE_TYPE_TKM);
    CHECK(dev->hardwareType() == HardwareType::GR_TKM200);
    CHECK(dev->deviceCount() == 4);
    CHECK(dev->hasOutput() == false);
    CHECK(dev->hasInput() == true);
    CHECK(dev->hasMultibuttons() == true);
    CHECK(dev->isBinaryInput() == false);
    CHECK(dev->getIconPath() == "tkm_grey.png");
  }

  GIVEN("GR-TKM210") {
    auto dev = createDevice(apartment, {.hardwareType = HardwareType::GR_TKM210});
    CHECK(dev->getProductID() == 0x04D2);
    CHECK(dev->getFunctionID() == 0x2101);
    CHECK(dev->getDeviceClass() == DeviceClasses_t::DEVICE_CLASS_GR);
    CHECK(dev->getDeviceType() == DEVICE_TYPE_TKM);
    CHECK(dev->hardwareType() == HardwareType::GR_TKM210);
    CHECK(dev->deviceCount() == 1);
    CHECK(dev->hasOutput() == false);
    CHECK(dev->hasInput() == true);
    CHECK(dev->hasMultibuttons() == false);
    CHECK(dev->isBinaryInput() == false);
    CHECK(dev->getIconPath() == "tkm_grey.png");
  }

  GIVEN("GN-TKM200") {
    auto dev = createDevice(apartment, {.hardwareType = HardwareType::GN_TKM200});
    CHECK(dev->getProductID() == 0x04C8);
    CHECK(dev->getFunctionID() == 0x7103);
    CHECK(dev->getDeviceClass() == DeviceClasses_t::DEVICE_CLASS_GN);
    CHECK(dev->getDeviceType() == DEVICE_TYPE_TKM);
    CHECK(dev->hardwareType() == HardwareType::GN_TKM200);
    CHECK(dev->deviceCount() == 4);
    CHECK(dev->hasOutput() == false);
    CHECK(dev->hasInput() == true);
    CHECK(dev->hasMultibuttons() == true);
    CHECK(dev->isBinaryInput() == false);
    CHECK(dev->getIconPath() == "tkm_green.png");
  }

  GIVEN("GN-TKM210") {
    auto dev = createDevice(apartment, {.hardwareType = HardwareType::GN_TKM210});
    CHECK(dev->getProductID() == 0x04D2);
    CHECK(dev->getFunctionID() == 0x7101);
    CHECK(dev->getDeviceClass() == DeviceClasses_t::DEVICE_CLASS_GN);
    CHECK(dev->getDeviceType() == DEVICE_TYPE_TKM);
    CHECK(dev->hardwareType() == HardwareType::GN_TKM210);
    CHECK(dev->deviceCount() == 1);
    CHECK(dev->hasOutput() == false);
    CHECK(dev->hasInput() == true);
    CHECK(dev->hasMultibuttons() == false);
    CHECK(dev->isBinaryInput() == false);
    CHECK(dev->getIconPath() == "tkm_green.png");
  }

  GIVEN("GN-KM200") {
    auto dev = createDevice(apartment, {.hardwareType = HardwareType::GN_KM200});
    CHECK(dev->getProductID() == 0x00C8);
    CHECK(dev->getFunctionID() == 0x7010);
    CHECK(dev->getDeviceClass() == DeviceClasses_t::DEVICE_CLASS_GN);
    CHECK(dev->getDeviceType() == DEVICE_TYPE_KM);
    CHECK(dev->hardwareType() == HardwareType::GN_KM200);
    CHECK(dev->deviceCount() == 1);
    CHECK(dev->hasOutput() == true);
    CHECK(dev->hasInput() == false);
    CHECK(dev->getIconPath() == "km_green.png");
  }

  GIVEN("SW-SKM300") {
    auto terminalBlocks = createSWSKM300(apartment, DeviceMockConf{});
    CHECK(terminalBlocks.size() == 4);
    for (auto terminalBlock : terminalBlocks) {
      const auto& dev = terminalBlock;
      CHECK(dev->getProductID() == 0x392c);
      CHECK(dev->getFunctionID() == 0x81cb);
      CHECK(dev->getDeviceClass() == DeviceClasses_t::DEVICE_CLASS_SW);
      CHECK(dev->getDeviceType() == DEVICE_TYPE_SKM);
      CHECK(dev->hardwareType() == HardwareType::SW_SKM300);
      CHECK(dev->deviceCount() == 4);
      CHECK(dev->hasOutput() == false);
      CHECK(dev->hasInput() == true);
      CHECK(dev->hasMultibuttons() == true);
      CHECK(dev->getActiveGroup() == ColorIDYellow);
      CHECK(dev->actions().size() == 0);                      // input only device
      CHECK(dev->getIconPath() == "skm300_black_yellow.png"); // default
      CHECK(tryDefaultShadeType(*dev) == std::nullopt);       // input only, does not apply
      CHECK_THROWS(dev->getSWThresholdAddress());             // input only, not used
    }
  }

  GIVEN("not implemented device") { CHECK_THROWS(createDevice(apartment, {.hardwareType = HardwareType::unknown})); };
};

TEST_CASE("2way pairing", TAGS) {
  auto pApartment = makeApartment(nullptr);
  auto& apartment = *pApartment;

  GIVEN("SW-SKM300") {
    auto dev = createSWSKM300(apartment, {});

    WHEN("default") {
      THEN("device is not paired") {
        REQUIRE(!dev[0]->is2WaySlave());
        REQUIRE(!dev[0]->is2WayMaster());
      }
    }
    WHEN("when input mode is 2way with even input") {
      dev[0]->setButtonInputMode(ButtonInputMode::TWO_WAY_DW_WITH_INPUT2);
      THEN("device is 2way master") { REQUIRE(dev[0]->is2WayMaster()); }
    }
    WHEN("button input is 2way with with odd input") {
      dev[0]->setButtonInputMode(ButtonInputMode::TWO_WAY_DW_WITH_INPUT1);
      THEN("device is 2way slave") { REQUIRE(dev[0]->is2WaySlave()); }
    }
  }

  GIVEN("GR-KL300") {
    auto dev =
        createDevice(apartment, {.hardwareType = HardwareType::GR_KL300, .buttonInputMode = ButtonInputMode::STANDARD});

    WHEN("default") {
      THEN("device is not paired") {
        REQUIRE(dev->getDeviceFeatures().pairing);
        REQUIRE(!dev->is2WaySlave());
        REQUIRE(!dev->is2WayMaster());
      }
    }
    WHEN("when input mode is 2way with even input") {
      dev->setButtonInputMode(ButtonInputMode::TWO_WAY_DW_WITH_INPUT2);
      THEN("device is 2way master") {
        REQUIRE(dev->is2WayMaster());
        THEN("partner device remains unconfigured") { REQUIRE(!dev->getPartnerDevice()->is2WaySlave()); }
      }
    }
    WHEN("button input is 2way with with odd input") {
      dev->setButtonInputMode(ButtonInputMode::TWO_WAY_DW_WITH_INPUT1);
      THEN("device is 2way slave") {
        REQUIRE(dev->is2WaySlave());
        THEN("partner device remains unconfigured") { REQUIRE(!dev->getPartnerDevice()->is2WayMaster()); }
      }
    }
  }
}

TEST_CASE("Test Extra Options", TAGS) {
  auto pApartment = makeApartment(nullptr);
  auto& apartment = *pApartment;

  GIVEN("Custom RevisionID") {
    auto dev1 = createDevice(apartment, DeviceMockConf{.hardwareType = HardwareType::GE_TKM210, .revisionId = 0x902});
    auto dev2 = createDevice(apartment, DeviceMockConf{.hardwareType = HardwareType::GR_KL220, .revisionId = 0x387});
    THEN("Custom RevisionID is returned") {
      CHECK(dev1->getRevisionID() == 0x902);
      CHECK(dev2->getRevisionID() == 0x387);
    }
  }
}

} // namespace
