/*
    Copyright (c) 2010 digitalSTROM.org, Zurich, Switzerland

    Author: Patrick Staehlin, futureLAB AG <pstaehlin@futurelab.ch>

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include <ds/locale/ds-locale.h>
#include <ds/log.h>

#include "apartment.h"
#include "group.h"
#include "model/modelconst.h"
#include "propertysystem.h"
#include "scenehelper.h"
#include "set.h"
#include "state.h"
#include "status-field.h"
#include "status.h"
#include "web/webrequests.h"
#include "zone.h"

static ds::LogChannel dsLogChannel{"modelGroup"};

namespace dss {
namespace {

bool isApplicationWithStatusForNonBroadcastZone(ApplicationType a) {
  return a == ApplicationType::apartmentVentilation || a == ApplicationType::ventilation ||
         a == ApplicationType::recirculation;
}
bool isApplicationWithStatusForBroadcastZone(ApplicationType a) {
  return a >= ApplicationType::apartmentVentilation;
}
bool isApplicationWithStatus(ApplicationType a, int zoneId) {
  return zoneId == ZONE_ID_BROADCAST ? isApplicationWithStatusForBroadcastZone(a)
                                     : isApplicationWithStatusForNonBroadcastZone(a);
}

auto callScenePreset1 = [](Group& group, const CallOrigin origin, const SceneAccessCategory category,
                           bool isAwnings) { group.callSceneBlinds(origin, category, Scene1, "", false, isAwnings); };

auto callScenePreset0 = [](Group& group, const CallOrigin origin, const SceneAccessCategory category,
                           bool isAwnings) { group.callSceneBlinds(origin, category, SceneOff, "", false, isAwnings); };

auto callSceneStop = [](Group& group, const CallOrigin origin, const SceneAccessCategory category, bool isAwnings) {
  group.callSceneBlinds(origin, category, SceneStop, "", false, isAwnings);
};

auto callSceneIncrement = [](Group& group, const CallOrigin origin, const SceneAccessCategory category,
                             bool isAwnings) {
  group.callSceneBlinds(origin, category, SceneInc, "", false, isAwnings);
};

auto callSceneDecrement = [](Group& group, const CallOrigin origin, const SceneAccessCategory category,
                             bool isAwnings) {
  group.callSceneBlinds(origin, category, SceneDec, "", false, isAwnings);
};

auto callSceneSunProtection = [](Group& group, const CallOrigin origin, const SceneAccessCategory category,
                                 bool isAwnings) {
  group.callSceneBlinds(origin, category, SceneSunProtection, "", false, isAwnings);
};

const std::vector<GroupAction> groupActions = {
    {"app.moveUp", dsApiTranslation("dss-mainline", "applicationAction", "Blinds Move Up"), callScenePreset1,
     ApplicationType::blinds},
    {"app.moveDown", dsApiTranslation("dss-mainline", "applicationAction", "Blinds Move Down"), callScenePreset0,
     ApplicationType::blinds},
    {"app.stepUp", dsApiTranslation("dss-mainline", "applicationAction", "Blinds Step Up"), callSceneIncrement,
     ApplicationType::blinds},
    {"app.stepDown", dsApiTranslation("dss-mainline", "applicationAction", "Blinds Step Down"), callSceneDecrement,
     ApplicationType::blinds},
    {"app.sunProtection", dsApiTranslation("dss-mainline", "applicationAction", "Blinds Sun Protection"),
     callSceneSunProtection, ApplicationType::blinds},
    {"app.stop", dsApiTranslation("dss-mainline", "applicationAction", "Blinds Stop"), callSceneStop,
     ApplicationType::blinds},
    {"app.moveIn", dsApiTranslation("dss-mainline", "applicationAction", "Awnings Move In"), callScenePreset0,
     ApplicationType::awnings},
    {"app.moveOut", dsApiTranslation("dss-mainline", "applicationAction", "Awnings Move Out"), callScenePreset1,
     ApplicationType::awnings},
    {"app.stepIn", dsApiTranslation("dss-mainline", "applicationAction", "Awnings Step In"), callSceneDecrement,
     ApplicationType::awnings},
    {"app.stepOut", dsApiTranslation("dss-mainline", "applicationAction", "Awnings Step Out"), callSceneIncrement,
     ApplicationType::awnings},
};
} // namespace

//============================================= Group

Group::Group(const int _id, std::shared_ptr<Zone> _pZone)
    : AddressableModelItem(&_pZone->getApartment()),
      m_ZoneID(_pZone->getID()),
      m_GroupID(_id),
      m_ApplicationType(ApplicationType::none),
      m_pApplicationBehavior(new DefaultBehavior(m_pPropertyNode, SceneOff)),
      m_LastCalledScene(SceneOff),
      m_LastButOneCalledScene(SceneOff),
      m_IsValid(false),
      m_SyncPending(false),
      m_readFromDsm(false),
      m_connectedDevices(0) {} // ctor

Group::~Group() = default;

bool Group::isValid() const {
  if (isDefaultGroup(m_GroupID) || isCluster(m_GroupID)) {
    return m_IsValid && (m_ApplicationType != ApplicationType::none);
  }
  return m_IsValid;
} // isValid

void Group::setApplicationType(ApplicationType applicationType) {
  // check if we already have proper application type set
  if (m_ApplicationType == applicationType) {
    return;
  }

  // disallow changing for default groups
  if ((isDefaultGroup(getID()) || (isGlobalAppDsGroup(getID()))) && (m_ApplicationType != ApplicationType::none)) {
    DS_ERROR("Blocked change of application type.", m_ApplicationType, applicationType, *this);
    return;
  }

  m_ApplicationType = applicationType;
  [&] {
    switch (applicationType) {
      case ApplicationType::ventilation:
      case ApplicationType::apartmentVentilation:
      case ApplicationType::recirculation:
      case ApplicationType::apartmentRecirculation:
        m_pApplicationBehavior.reset(new VentilationBehavior(m_pPropertyNode, getLastCalledScene()));
        return;
      case ApplicationType::none:
      case ApplicationType::lights:
      case ApplicationType::blinds:
      case ApplicationType::awnings:
      case ApplicationType::heating:
      case ApplicationType::audio:
      case ApplicationType::video:
      case ApplicationType::security:
      case ApplicationType::access:
      case ApplicationType::joker:
      case ApplicationType::cooling:
      case ApplicationType::window:
      case ApplicationType::temperature:
        m_pApplicationBehavior.reset(new DefaultBehavior(m_pPropertyNode, getLastCalledScene()));
        return;
    }
    DS_ERROR("Failed to set unknown application type.", applicationType, *this);
    m_ApplicationType = ApplicationType::none;
    m_pApplicationBehavior.reset(new DefaultBehavior(m_pPropertyNode, getLastCalledScene()));
  }();

  // Status is supported only for apartment ventilation groups for now.
  if (getApartment().getDss()) {
    // TODO(someday): Status object requires IoService instance from DSS instance.
    // But most tests do not provice DSS instance now.
    // It is arguably bad idea to require full DSS instance to get IoService instance.
    // ModelMaintenance::initialize has similar workaround.
    if (!m_status && isApplicationWithStatus(m_ApplicationType, getZoneID())) {
      // Lazy initialization. Once the status is created, it can never be deleted.
      // Device objects capture references to it.
      m_status.reset(new Status(*this));
    }
  }

  if (getZoneID() == 0) {
    return;
  }

  if (m_GroupID == GroupIDYellow) {
    // zone.123.light
    std::shared_ptr<Group> me = sharedFromThis();
    std::shared_ptr<State> state = m_pApartment->allocateGroupState(me, State::makeGroupName(*this));
  } else if (m_GroupID == GroupIDHeating) {
    // zone.123.heating
    std::shared_ptr<Group> me = sharedFromThis();
    std::shared_ptr<State> state = m_pApartment->allocateGroupState(me, State::makeGroupName(*this));
  }
}

int Group::getColor() const {
  return getApplicationTypeColor(m_ApplicationType);
}

void Group::serializeApplicationConfiguration(uint32_t configuration, JSONWriter& writer) const {
  m_pApplicationBehavior->serializeConfiguration(configuration, writer);
}

std::string Group::serializeApplicationConfiguration(uint32_t configuration) const {
  JSONWriter writer(JSONWriter::jsonNoneResult);
  m_pApplicationBehavior->serializeConfiguration(configuration, writer);
  return writer.successJSON();
}

void Group::setFromSpec(const GroupSpec_t& spec) {
  setName(spec.Name);
  setApplicationType(spec.applicationType);
  setApplicationConfiguration(spec.applicationConfiguration);
}

bool Group::isConfigEqual(const GroupSpec_t& spec) {
  return ((getName() == spec.Name) && (getApplicationType() == spec.applicationType) &&
          (getApplicationConfiguration() == spec.applicationConfiguration));
}

Set Group::getDevices() const {
  return m_pApartment->getZone(m_ZoneID)->getDevices().getByGroup(m_GroupID);
} // getDevices

dsModel::ZoneId Group::zoneId() const {
  return dsModel::ZoneId{ds::str(m_ZoneID)};
}

Group& Group::operator=(const Group& _other) {
  m_GroupID = _other.m_GroupID;
  m_ZoneID = _other.m_ZoneID;
  return *this;
} // operator=

void Group::setLastCalledScene(const int _value) {
  if (m_GroupID == GroupIDControlTemperature) {
    if (_value >= 16) {
      return;
    }
  }
  if (_value != m_LastCalledScene) {
    m_LastButOneCalledScene = m_LastCalledScene;
    m_LastCalledScene = _value;
    m_pApplicationBehavior->setCurrentScene(_value);
  }
}

// Legacy awnings are at power line and dsm-api level in shades application (design mistake).
// But they behave inconsistently. We model legacy awnings in new awnings application at higher level.
// this->applicationType == shades && isAwnings -> scene call is intended for awnings
// this->applicationType == blinds && !isAwnings -> scene call is intended for blinds (shades)
//
// The installation must not have shades and legacy awnings devices
// in one zone for correct operation.
// Otherwise the behavior is not well defined.
void Group::callSceneBlinds(const CallOrigin _origin, const SceneAccessCategory _category, const int _sceneNr,
                            const std::string _token, const bool _force, const bool isAwnings) {
  DS_REQUIRE(this->getApplicationType() == ApplicationType::blinds, "Blinds application needs to be used",
             this->getApplicationType());
  if (this->getZoneID() == ZONE_ID_BROADCAST && !isCluster(m_GroupID)) {
    if (m_pApartment->getZone(ZONE_ID_BROADCAST)
            ->getGroup(static_cast<int>(ApplicationType::blinds))
            ->hasLegacyAwnings()) {
      // Split the apartment wide shades call into per zone call.
      // This allows us to call the scene on shades only / legacy awnings only zones correctly.
      const auto& zones = m_pApartment->getZones();
      for (auto& zone : zones) {
        const auto& zoneApp = zone->getGroup(static_cast<int>(ApplicationType::blinds));
        if (zoneApp == nullptr || zone->getID() == ZONE_ID_BROADCAST) {
          continue;
        }
        if (isAwnings) {
          if (zoneApp->hasShades()) {
            continue;
          } else if (zoneApp->hasLegacyAwnings()) {
            zoneApp->callScene(_origin, _category, _sceneNr, _token, _force);
            usleep(500 * 1000); // 500ms
          }
        } else if (zoneApp->hasShades()) {
          zoneApp->callScene(_origin, _category, _sceneNr, _token, _force);
          usleep(500 * 1000); // 500ms
        }
      }
    } else {
      callScene(_origin, _category, _sceneNr, _token, _force);
    }
  } else {
    if (isAwnings && this->hasShades()) {
      DS_FAIL_REQUIRE("There are both shades and awnings devices in the zone. Calling action failed.");
    } else if (!isAwnings && !this->hasShades() && this->hasLegacyAwnings()) {
      DS_FAIL_REQUIRE("There are no shades devices in the zone, but awning device is present. Calling action failed.");
    } else {
      callScene(_origin, _category, _sceneNr, _token, _force);
    }
  }
}

void Group::callScene(const CallOrigin _origin, const SceneAccessCategory _category, const int _sceneNr,
                      const std::string _token, const bool _force) {
  // this might be redundant, but since a set could be
  // optimized if it contains only one device its safer like that...
  if (SceneHelper::rememberScene(_sceneNr & 0x00ff)) {
    setLastCalledScene(_sceneNr & 0x00ff);
  }

  AddressableModelItem::callScene(_origin, _category, _sceneNr, _token, _force);

  // TODO: checkAccess might have blocked the actual scene call
  switch (m_GroupID) {
    case GroupIDControlTemperature: {
      std::shared_ptr<Zone> pZone = m_pApartment->getZone(m_ZoneID);
      if (_sceneNr <= HeatingOperationModeIDMax) {
        pZone->setHeatingOperationMode(_sceneNr);
      }
    } break;
    default: setOnState(_origin, _sceneNr); break;
  }
} // callScene

unsigned long Group::getPowerConsumption() {
  return getDevices().getPowerConsumption();
} // getPowerConsumption

void Group::nextScene(const CallOrigin _origin, const SceneAccessCategory _category) {
  callScene(_origin, _category, m_pApplicationBehavior->getNextScene(), "", false);
} // nextScene

void Group::previousScene(const CallOrigin _origin, const SceneAccessCategory _category) {
  callScene(_origin, _category, m_pApplicationBehavior->getPreviousScene(), "", false);
} // previousScene

void Group::setSceneName(int _sceneNumber, const std::string& _name) {
  boost::recursive_mutex::scoped_lock lock(PropertyNode::m_GlobalMutex);
  m_SceneNames[_sceneNumber] = _name;
  if (m_pPropertyNode != NULL) {
    PropertyNodePtr scene = m_pPropertyNode->createProperty("scenes/scene" + intToString(_sceneNumber));

    scene->createProperty("scene")->setIntegerValue(_sceneNumber);
    scene->createProperty("name")->setStringValue(_name);
  }
} // setSceneName

std::string Group::getSceneName(int _sceneNumber) {
  boost::recursive_mutex::scoped_lock lock(PropertyNode::m_GlobalMutex);
  m_SceneNames_t::iterator it = m_SceneNames.find(_sceneNumber);
  if (it == m_SceneNames.end()) {
    return std::string();
  } else {
    return m_SceneNames[_sceneNumber];
  }
} // getSceneName

void Group::setOnState(const CallOrigin _origin, bool _on) {
  if (getZoneID() == 0) {
    return;
  }
  if (m_GroupID == GroupIDYellow) {
    try {
      DS_ASSERT(m_ApplicationType == ApplicationType::lights, "");
      auto state = m_pApartment->getNonScriptState(State::makeGroupName(*this));
      state->setState(_origin, _on == true ? State_Active : State_Inactive);
    } catch (ItemNotFoundException& ex) {
    } // should never happen
  } else if (m_GroupID == GroupIDHeating) {
    try {
      DS_ASSERT(m_ApplicationType == ApplicationType::heating, "");
      auto state = m_pApartment->getNonScriptState(State::makeGroupName(*this));
      state->setState(_origin, _on == true ? State_Active : State_Inactive);
    } catch (ItemNotFoundException& ex) {
    } // should never happen
  }
}

std::vector<int> Group::namedScenes() {
  auto groupAvailableScenes = getAvailableScenes();
  groupAvailableScenes.erase(std::remove_if(groupAvailableScenes.begin(), groupAvailableScenes.end(),
                                            [&](int i) { return getSceneName(i).empty(); }),
                             groupAvailableScenes.end());
  return groupAvailableScenes;
}

bool Group::isNamedScene(int sceneNr) {
  auto renamedScenes = namedScenes();
  return std::find(renamedScenes.begin(), renamedScenes.end(), sceneNr) != renamedScenes.end();
}

int Group::availablePreset() {
  std::vector<int> presets = {Scene2,  Scene3,  Scene4,  Scene12, Scene13, Scene14, Scene22, Scene23,
                              Scene24, Scene32, Scene33, Scene34, Scene42, Scene43, Scene44};
  const auto devices = getDevices();
  for (int i = 0; i < devices.length(); i++) {
    switch (devices[i].getDevice()->getButtonID()) {
      case ButtonId_Zone_Extended4:
      case ButtonId_Area4_Extended:
        presets.erase(
            std::remove_if(presets.begin(), presets.end(),
                           [&](int scene) { return ((scene == Scene42) || (scene == Scene43) || (scene == Scene44)); }),
            presets.end());
        break;
      case ButtonId_Zone_Extended3:
      case ButtonId_Area3_Extended:
        presets.erase(
            std::remove_if(presets.begin(), presets.end(),
                           [&](int scene) { return (scene == Scene32 || scene == Scene33 || scene == Scene34); }),
            presets.end());
        break;
      case ButtonId_Zone_Extended2:
      case ButtonId_Area2_Extended:
        presets.erase(
            std::remove_if(presets.begin(), presets.end(),
                           [&](int scene) { return (scene == Scene22 || scene == Scene23 || scene == Scene24); }),
            presets.end());
        break;
      case ButtonId_Zone_Extended1:
      case ButtonId_Area1_Extended:
        presets.erase(
            std::remove_if(presets.begin(), presets.end(),
                           [&](int scene) { return (scene == Scene12 || scene == Scene13 || scene == Scene14); }),
            presets.end());
        break;
      case ButtonId_Area1:
      case ButtonId_Area2:
      case ButtonId_Area3:
      case ButtonId_Area4:
      case ButtonId_Zone:
        presets.erase(
            std::remove_if(presets.begin(), presets.end(),
                           [&](int scene) { return (scene == Scene2 || scene == Scene3 || scene == Scene4); }),
            presets.end());
        break;
    }
  }
  std::reverse(presets.begin(), presets.end());
  for (auto preset : presets) {
    if (!isNamedScene(preset)) {
      return preset;
    }
  }
  DS_FAIL_REQUIRE("Could not assign scenario to preset in group", getID());
}

void Group::setOnState(const CallOrigin _origin, const int _sceneId) {
  if (getZoneID() == 0) {
    return;
  }
  if (m_GroupID == GroupIDYellow) {
    SceneHelper::SceneOnState isOn = SceneHelper::isOnScene(m_GroupID, _sceneId);
    if (isOn != SceneHelper::DontCare) {
      try {
        DS_ASSERT(m_ApplicationType == ApplicationType::lights, "");
        auto state = m_pApartment->getNonScriptState(State::makeGroupName(*this));
        state->setState(_origin, isOn == SceneHelper::True ? State_Active : State_Inactive);
      } catch (ItemNotFoundException& ex) {
      } // should never happen
    }
  } else if (m_GroupID == GroupIDHeating) {
    SceneHelper::SceneOnState isOn = SceneHelper::isOnScene(m_GroupID, _sceneId);
    if (isOn != SceneHelper::DontCare) {
      try {
        DS_ASSERT(m_ApplicationType == ApplicationType::heating, "");
        auto state = m_pApartment->getNonScriptState(State::makeGroupName(*this));
        state->setState(_origin, isOn == SceneHelper::True ? State_Active : State_Inactive);
      } catch (ItemNotFoundException& ex) {
      } // should never happen
    }
  }
}

eState Group::getState() {
  if (getZoneID() == 0) {
    return State_Unknown;
  }

  // only publish states for light
  if (m_GroupID == GroupIDYellow) {
    try {
      DS_ASSERT(m_ApplicationType == ApplicationType::lights, "");
      auto state = m_pApartment->getNonScriptState(State::makeGroupName(*this));
      return static_cast<eState>(state->getState());
    } catch (ItemNotFoundException& ex) {
    } // should never happen
  } else if (m_GroupID == GroupIDHeating) {
    try {
      DS_ASSERT(m_ApplicationType == ApplicationType::heating, "");
      auto state = m_pApartment->getNonScriptState(State::makeGroupName(*this));
      return static_cast<eState>(state->getState());
    } catch (ItemNotFoundException& ex) {
    } // should never happen
  }

  return State_Unknown;
}

void Group::publishToPropertyTree() {
  if (m_pPropertyNode == NULL) {
    if (m_pApartment->getPropertyNode() != NULL) {
      m_pPropertyNode = m_pApartment->getPropertyNode()->createProperty("zones/zone" + intToString(m_ZoneID) +
                                                                        "/groups/group" + intToString(m_GroupID));
      m_pPropertyNode->createProperty("group")->setIntegerValue(m_GroupID);
      m_pPropertyNode->createProperty("color")->linkToProxy(
          PropertyProxyMemberFunction<Group, int>(*this, &Group::getColor));
      m_pPropertyNode->createProperty("applicationType")
          ->linkToProxy(PropertyProxyMemberFunction<Group, int>(*this, &Group::getApplicationTypeInt,
                                                                &Group::setApplicationTypeInt));
      m_pPropertyNode->createProperty("name")->linkToProxy(
          PropertyProxyMemberFunction<Group, std::string>(*this, &Group::getName, &Group::setName));
      m_pPropertyNode->createProperty("lastCalledScene")
          ->linkToProxy(PropertyProxyMemberFunction<Group, int>(*this, &Group::getLastCalledScene));
      m_pPropertyNode->createProperty("connectedDevices")
          ->linkToProxy(PropertyProxyReference<int>(m_connectedDevices, false));

      m_pApplicationBehavior->publishToPropertyTree();
    }
  }
} // publishToPropertyTree

void Group::sensorPush(const dsuid_t& _sourceID, SensorType _type, double _value) {
  DateTime now;
  std::shared_ptr<Zone> pZone = m_pApartment->getZone(m_ZoneID);

  if (m_ZoneID == 0) {
    switch (_type) {
      case SensorType::TemperatureOutdoors: m_pApartment->setTemperature(_value, now); break;
      case SensorType::HumidityOutdoors: m_pApartment->setHumidityValue(_value, now); break;
      case SensorType::BrightnessOutdoors: m_pApartment->setBrightnessValue(_value, now); break;
      case SensorType::WindSpeed: m_pApartment->setWindSpeed(_value, now); break;
      case SensorType::WindDirection: m_pApartment->setWindDirection(_value, now); break;
      case SensorType::GustSpeed:
        m_pApartment->setWindGust(_value, now);
        m_pApartment->updateWeatherProtectedClusters(_value);
        break;
      case SensorType::GustDirection: // ignored by system
        break;
      case SensorType::Precipitation: m_pApartment->setPrecipitation(_value, now); break;
      case SensorType::AirPressure: m_pApartment->setAirPressure(_value, now); break;
      case SensorType::SunAzimuth: m_pApartment->setSunAzimuth(_value, now); break;
      case SensorType::SunElevation: m_pApartment->setSunElevation(_value, now); break;
      case SensorType::ActivePower:                    // no break
      case SensorType::OutputCurrent:                  // no break
      case SensorType::ElectricMeter:                  // no break
      case SensorType::TemperatureIndoors:             // no break
      case SensorType::BrightnessIndoors:              // no break
      case SensorType::HumidityIndoors:                // no break
      case SensorType::CO2Concentration:               // no break
      case SensorType::COConcentration:                // no break
      case SensorType::SoundPressureLevel:             // no break
      case SensorType::RoomTemperatureSetpoint:        // no break
      case SensorType::RoomTemperatureControlVariable: // no break
      case SensorType::Status:           ///< Status sensor value. Value bits are defined by \re StatusFieldType
      case SensorType::Reserved1:        // no break
      case SensorType::Reserved2:        // no break
      case SensorType::OutputCurrent16A: // no break
      case SensorType::ActivePowerVA:    // no break
      case SensorType::Temperature:      // no break
      case SensorType::Brightness:       // no break
      case SensorType::Humidity:         // no break
      case SensorType::WaterQuantity:    // no break
      case SensorType::WaterFlowRate:    // no break
      case SensorType::NotUsed:          // no break
      case SensorType::UnknownType:      // no break
      default: break;
    }
  } else {
    switch (_type) {
      case SensorType::TemperatureIndoors: pZone->setTemperature(_value, now); break;
      case SensorType::RoomTemperatureSetpoint: pZone->setNominalValue(_value, now); break;
      case SensorType::RoomTemperatureControlVariable: pZone->setControlValue(_value, now); break;
      case SensorType::HumidityIndoors: pZone->setHumidityValue(_value, now); break;
      case SensorType::BrightnessIndoors: pZone->setBrightnessValue(_value, now); break;
      case SensorType::CO2Concentration: pZone->setCO2ConcentrationValue(_value, now); break;
      case SensorType::ActivePower:         // no break
      case SensorType::OutputCurrent:       // no break
      case SensorType::ElectricMeter:       // no break
      case SensorType::TemperatureOutdoors: // no break
      case SensorType::BrightnessOutdoors:  // no break
      case SensorType::HumidityOutdoors:    // no break
      case SensorType::AirPressure:         // no break
      case SensorType::GustSpeed:           // no break
      case SensorType::GustDirection:       // no break
      case SensorType::WindSpeed:           // no break
      case SensorType::WindDirection:       // no break
      case SensorType::Precipitation:       // no break
      case SensorType::COConcentration:     // no break
      case SensorType::SoundPressureLevel:  // no break
      case SensorType::Status:              ///< Status sensor value. Value bits are defined by \re StatusFieldType
      case SensorType::Reserved1:           // no break
      case SensorType::Reserved2:           // no break
      case SensorType::OutputCurrent16A:    // no break
      case SensorType::ActivePowerVA:       // no break
      case SensorType::Temperature:         // no break
      case SensorType::Brightness:          // no break
      case SensorType::Humidity:            // no break
      case SensorType::WaterQuantity:       // no break
      case SensorType::WaterFlowRate:       // no break
      case SensorType::NotUsed:             // no break
      case SensorType::UnknownType:         // no break
      case SensorType::SunAzimuth:
      case SensorType::SunElevation:
      default: break;
    }
  }

  if (m_pPropertyNode != NULL) {
    PropertyNodePtr node = m_pPropertyNode->createProperty("sensor/type" + intToString(static_cast<int>(_type)));
    node->createProperty("type")->setIntegerValue(static_cast<int>(_type));
    node->createProperty("value")->setFloatingValue(_value);
    node->createProperty("sourcedsuid")->setStringValue(dsuid2str(_sourceID));
    node->createProperty("time")->setIntegerValue(now.secondsSinceEpoch());
    node->createProperty("timestamp")->setStringValue(now.toString());
  }
} // sensorPush

void Group::sensorInvalid(SensorType _type) {
  if (m_pPropertyNode != NULL) {
    PropertyNodePtr node = m_pPropertyNode->getProperty("sensor/type" + intToString(static_cast<int>(_type)));
    if (node) {
      node->getParentNode()->removeChild(node);
    }
  }
}

Status& Group::getStatus() {
  if (auto status = tryGetStatus()) {
    return *status;
  }
  DS_FAIL_REQUIRE("Group does not support status.", *this);
}

void Group::setStatusField(const std::string& fieldName, const std::string& valueName) {
  auto&& fieldType = statusFieldTypeFromName(fieldName);
  DS_REQUIRE(fieldType, "Unknown", fieldName);

  auto&& value = statusFieldValueFromName(valueName);
  DS_REQUIRE(value, "Unknown", valueName);

  auto&& field = getStatus().getField(*fieldType);
  field.setValueAndPush(*value);
}

void Group::addConnectedDevice() {
  ++m_connectedDevices;
}

void Group::removeConnectedDevice() {
  if (m_connectedDevices > 0) {
    --m_connectedDevices;
  }
}

std::shared_ptr<Group> Group::make(const GroupSpec_t& _groupSpec, std::shared_ptr<Zone> _pZone) {
  std::shared_ptr<Group> pGroup(new Group(_groupSpec.GroupID, _pZone));
  pGroup->setFromSpec(_groupSpec);

  return pGroup;
}

std::ostream& operator<<(std::ostream& stream, const Group& x) {
  return stream << "zone." << x.getZoneID() << ".group." << x.getID();
}

std::vector<GroupAction> Group::actions() const {
  std::vector<GroupAction> result;
  for (const auto& action : groupActions) {
    if (action.type == m_ApplicationType) {
      result.push_back(action);
    }
  }
  return result;
}

void Group::callAction(CallOrigin origin, SceneAccessCategory category, const std::string& actionId, bool isAwnings) {
  const auto action =
      std::find_if(groupActions.begin(), groupActions.end(), [&](const GroupAction a) { return a.id == actionId; });
  if (action == groupActions.end()) {
    throw std::runtime_error("Failed to call unknown action '" + actionId + "'");
  }
  action->invoke(*this, origin, category, isAwnings);
}

bool Group::hasLegacyAwnings() const {
  const auto devices = this->getDevices();
  for (auto i = 0; i < devices.length(); i++) {
    if (devices[i].getDevice()->isLegacyAwning()) {
      return true;
    }
  }
  return false;
}

bool Group::hasShades() const {
  const auto devices = this->getDevices();
  for (auto i = 0; i < devices.length(); i++) {
    const auto& device = devices[i].getDevice();
    if (device->isShadeDevice() && !device->isBlindsButton()) {
      return true;
    }
  }
  return false;
}

} // namespace dss
