/*
    Copyright (c) 2018 digitalSTROM AG, Zurich, Switzerland

    This file is part of dSS.

    vdSM is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 2 of the License.

    vdSM is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with dSS. If not, see <http://www.gnu.org/licenses/>.
*/

#include "modelmaintenance.h"

#include <ds/asio/test.h>
#include <ds/catch/catch.h>
#include <ds/log.h>

#include <boost/asio/steady_timer.hpp>
#include <trompeloeil.hpp>

#include "apartment-mock.h"
#include "businterface-mock.h"
#include "device-mock.h"
#include "dss.h"
#include "event-mock.h"
#include "modelmaintenance-mock.h"
#include "propertysystem-mock.h"
#include "webservice_connection-mock.h"

static ds::LogChannel dsLogChannel("modelMaintenanceTest", ds::LogSeverity::DEBUG);
static const char* TAGS = "[dss][modelMaintenance]";

using namespace dss;
using namespace std::literals;
using namespace trompeloeil;
using trompeloeil::_;

static const Dsuid dsuid1{"0000000000000000000000000000000001"};

namespace {

struct ModelMaintenanceHooksMock {
public:
  MAKE_MOCK0(linkPresent, void());
  MAKE_MOCK0(linkLost, void());
  MAKE_MOCK1(ds485MemberLost, void(dsuid_t));
  MAKE_MOCK1(ds485MemberJoin, void(dsuid_t));

  MAKE_MOCK0(minimalTransactionTimeExpired, void());
  MAKE_MOCK0(transactionDone, void());
  MAKE_MOCK0(snapshotDone, void());

  ModelMaintenanceHooks modelMaintenanceHooks() {
    return {
        .linkPresent = [&] { return linkPresent(); },
        .linkLost = [&] { return linkLost(); },
        .ds485MemberLost = [&](dsuid_t arg) { return ds485MemberLost(arg); },
        .ds485MemberJoin = [&](dsuid_t arg) { return ds485MemberJoin(arg); },

        .minimalTransactionTimeExpired = [&] { return minimalTransactionTimeExpired(); },
        .transactionDone = [&] { return transactionDone(); },
        .snapshotDone = [&] { return snapshotDone(); },
    };
  }
};

SCENARIO("ModelMaintenance", TAGS) {
  PropertySystemMock propertySystem{};
  EventSinkMock eventSink{};

  ds::asio::TestLoopResource loop; // RAII, ensure destructor runs last

  BusInterfaceMock busInterface{};
  ModelMaintenance modelMaintenance(nullptr, propertySystem, eventSink);
  modelMaintenance.setBusInterface(busInterface);

  ApartmentMock apartment{};
  apartment.setBusInterface(&busInterface); // StructureManipulator pulls from Apartment
  modelMaintenance.setApartment(&apartment);

  using Timer = boost::asio::basic_waitable_timer<ds::Clock, ds::ClockWaitTraits>;
  Timer timer(loop.io);

  GIVEN("work queue") {
    THEN("added work is done") {
      bool test{false};
      modelMaintenance.addWork([&test] { test = true; });
      modelMaintenance.handleWork();
      CHECK(test == true);
    }
  }

  GIVEN("PIMPL initialized") {
    ModelMaintenanceHooksMock hooks{};
    ModelMaintenanceConfig config{.minTransactionTime = 2ms, .maxAggregationWindow = 5ms};

    loop.post(
        [&] { return modelMaintenance.makeImpl(loop, ModelMaintenanceConfig(config), hooks.modelMaintenanceHooks()); });
    loop.postStopRun();

    ds::Slot<std::shared_ptr<ModelSnapshot>> DS_UNIQUE_NAME(slot){
        loop, apartment.modelStructureChanged(), [&](auto arg) { DS_NOTICE("modelStructureChanged signal"); }};

    auto expireMinTransactionTime = [&] {
      REQUIRE_CALL(hooks, minimalTransactionTimeExpired()).LR_SIDE_EFFECT(loop.stop());
      loop.run();
    };

    auto expireMaxAggregationTime = [&] {
      timer.expires_from_now(config.maxAggregationWindow);
      timer.async_wait([&](boost::system::error_code e) { loop.stop(); });
      loop.run();
    };

    auto completeTransaction = [&]() {
      REQUIRE_CALL(hooks, transactionDone()).LR_SIDE_EFFECT(loop.stop());
      modelMaintenance.handleWork();
      loop.run();
    };

    WHEN("event occurs") {
      REQUIRE_CALL(hooks, linkPresent());
      busInterface.linkPresent.emit();
      expireMinTransactionTime();

      THEN("work is scheduled") {
        REQUIRE_CALL(busInterface.structureQuery, getBusMembers()).RETURN(std::vector<DSMeterSpec>{});
        modelMaintenance.handleWork();

        THEN("transactionDone") {
          REQUIRE_CALL(hooks, transactionDone());
          loop.postStopRun();

          THEN("snapshotDone in separate cycle") {
            REQUIRE_CALL(hooks, snapshotDone());
            loop.postPostStopRun();
          }
        }
      }

      WHEN("2nd event arrives") {
        REQUIRE_CALL(hooks, linkPresent());
        busInterface.linkPresent.emit();
        loop.postStopRun();

        THEN("is combined in same transcation") {
          REQUIRE_CALL(busInterface.structureQuery, getBusMembers()).RETURN(std::vector<DSMeterSpec>{});
          REQUIRE_CALL(busInterface.structureQuery, getBusMembers()).RETURN(std::vector<DSMeterSpec>{});
          modelMaintenance.handleWork();
        }
      }

      WHEN("2nd event arrives after maxAggregationWindow") {
        expireMaxAggregationTime();

        REQUIRE_CALL(hooks, linkPresent());
        busInterface.linkPresent.emit();
        loop.postStopRun();

        THEN("transaction only contains first action") {
          REQUIRE_CALL(busInterface.structureQuery, getBusMembers()).RETURN(std::vector<DSMeterSpec>{});
          modelMaintenance.handleWork();

          WHEN("transaction/snaphost completes") {
            REQUIRE_CALL(hooks, transactionDone());
            REQUIRE_CALL(hooks, snapshotDone());
            loop.postPostStopRun(); // 2 mainloop iterations

            THEN("2nd action is handled in new transaction") {
              REQUIRE_CALL(busInterface.structureQuery, getBusMembers()).RETURN(std::vector<DSMeterSpec>{});
              modelMaintenance.handleWork();
            }
          }

          WHEN("only transaction completes") {
            REQUIRE_CALL(hooks, transactionDone());
            loop.postStopRun();

            THEN("2nd action not yet scheduled") { // needs snapshotDone
              modelMaintenance.handleWork();
            }

            WHEN("new events arrive while snapshotDone pending") {
              REQUIRE_CALL(hooks, linkPresent());
              busInterface.linkPresent.emit();
              loop.postStopRun();

              THEN("next transaction is created with both actions") {
                // snapshotDone not required, mind that order should be checked
                REQUIRE_CALL(busInterface.structureQuery, getBusMembers()).RETURN(std::vector<DSMeterSpec>{});
                REQUIRE_CALL(busInterface.structureQuery, getBusMembers()).RETURN(std::vector<DSMeterSpec>{});
                modelMaintenance.handleWork();
              }
            }
          }
        }
      }
    }

    WHEN("link lost") {
      REQUIRE_CALL(hooks, linkLost());
      busInterface.linkLost.emit();
      expireMinTransactionTime();
      completeTransaction();
    }

    WHEN("asyncDiscoverDS485Devices") {
      modelMaintenance.asyncDiscoverDS485Devices();
      loop.postStopRun();

      THEN("executed via work queue") {
        REQUIRE_CALL(busInterface.structureQuery, getBusMembers()).RETURN(std::vector<DSMeterSpec>{});
        modelMaintenance.handleWork();

        THEN("confirmed") {
          expireMinTransactionTime();
          REQUIRE_CALL(hooks, transactionDone());
          loop.postStopRun();
        }
      }
    }

    GIVEN("dsm with device") {
      DSUID_DEFINE(meter1DSID, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 10);
      auto& meter1 = *apartment.allocateDSMeter(meter1DSID);

      auto& dev1 = *createDevice(apartment, {.hardwareType = HardwareType::GE_TKM210});
      dev1.setDSMeter(meter1.shared_from_this());

      CHECK(meter1.isPresent() == false);
      CHECK(dev1.isPresent() == false);

      WHEN("ds485 member joins") {
        REQUIRE_CALL(hooks, ds485MemberJoin(_));
        busInterface.ds485MemberJoin.emit(meter1.getDSID());
        loop.postStopRun();
        modelMaintenance.handleWork();

        THEN("meter and device are present") {
          CHECK(meter1.isPresent() == true);
          markOnline(dev1); // TODO requires MeterMaintencance::triggerMeterSynchronization
          CHECK(dev1.isPresent() == true);
        }

        WHEN("ds485 member is lost") {
          REQUIRE_CALL(hooks, ds485MemberLost(_));
          busInterface.ds485MemberLost.emit(meter1.getDSID());
          loop.postStopRun();
          modelMaintenance.handleWork();

          THEN("meter and device is marked absent") {
            CHECK(meter1.isPresent() == false);
            CHECK(dev1.isPresent() == false);
          }
        }
      }
    }
  }
}

} // namespace
