/*
    Copyright (c) 2019 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation," " either version 3 of the License," " or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful," "
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not," " see <http://www.gnu.org/licenses/>.
*/

#include "model/shademanipulator.h"

#include <ds/catch/catch.h>
#include <ds/log.h>

#include "businterface-mock.h"
#include "model/apartment-mock.h"
#include "model/device-mock.h"
#include "structuremanipulator.h"

static ds::LogChannel dsLogChannel("dssModelShadeManipulatorTest");
static const char* TAGS = "[model][device][manipulator]";

using namespace dss;
using namespace dss::model;
using namespace trompeloeil;
using trompeloeil::_;

namespace {

SCENARIO("init", TAGS) {
  auto busInterfaceMock = BusInterfaceMock{};
  ApartmentMock apartment(nullptr, nullptr, &busInterfaceMock);

  auto shadeManipulator = makeShadeManipulator(apartment, busInterfaceMock.device);

  GIVEN("various devices") {
    struct X {
      HardwareType hardwareType;
      std::optional<dsModel::ShadeType> expectedType;
    };
    const auto& x = GENERATE(values(std::initializer_list<X>{
        {HardwareType::GR_KL200, dsModel::ShadeType::rollerShutter},
        {HardwareType::GR_KL210, dsModel::ShadeType::awningTerrace},
        {HardwareType::GR_KL220, dsModel::ShadeType::windowBlind},
        {HardwareType::GR_HKL230, dsModel::ShadeType::windowBlind},
        {HardwareType::GR_KL300, dsModel::ShadeType::none},  // multi device
        {HardwareType::GR_HKL330, dsModel::ShadeType::none}, // multi device
        {HardwareType::GE_KM200, std::nullopt},
    }));
    THEN("static shade type is correct") {
      INFO(ds::str(x.hardwareType));
      auto& device = *createDevice(apartment, {.hardwareType = x.hardwareType});
      CHECK(tryDefaultShadeType(device) == x.expectedType);
      CHECK(defaultShadeType(device) == x.expectedType.value_or(dsModel::ShadeType::none));

      ALLOW_CALL(busInterfaceMock.device, getDeviceConfig(_, 1, 0x5a)).RETURN(0); // KL-300
      CHECK(shadeManipulator->shadeType(device) == x.expectedType.value_or(dsModel::ShadeType::none));
    }
  }
  GIVEN("device which supports multiple shadeType") {
    struct X {
      uint8_t bankValue;
      dsModel::ShadeType expectedType;
    };
    auto& device = *createDevice(apartment, {.hardwareType = HardwareType::GR_KL300});
    device.setIsPresent(true);
    device.setIsConnected(true);
    const auto& x = GENERATE(values(std::initializer_list<X>{{0, dsModel::ShadeType::none},
                                                             {1, dsModel::ShadeType::windowBlind},
                                                             {2, dsModel::ShadeType::rollerShutter},
                                                             {3, dsModel::ShadeType::awningFacade},
                                                             {4, dsModel::ShadeType::awningTerrace},
                                                             {5, dsModel::ShadeType::awningFacadeHorizontal},
                                                             {7, dsModel::ShadeType::curtain},
                                                             {8, dsModel::ShadeType::internalBlind}}));
    THEN("bank values are read and interpreted") {
      INFO(ds::str(x.expectedType));
      REQUIRE_CALL(busInterfaceMock.device, getDeviceConfig(_, 1, 0x5a)).RETURN(x.bankValue);
      CHECK(shadeManipulator->shadeType(device) == x.expectedType);
      THEN("shade type is cached") { CHECK(shadeManipulator->shadeType(device) == x.expectedType); }
    }
    THEN("bank values are set") {
      INFO(ds::str(x.expectedType));
      REQUIRE_CALL(busInterfaceMock.device, setDeviceConfig(_, 1, 0x5a, x.bankValue));
      shadeManipulator->setShadeType(device, x.expectedType);
      THEN("shade type is cached") { CHECK(shadeManipulator->shadeType(device) == x.expectedType); }
    }
  }
}

}; // namespace
