/*
    Copyright (c) 2022 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/

#include "patch-diff.h"

#include <ds/catch/catch.h>
#include <ds/json.h>
#include <ds/log.h>

using namespace std::literals;
static ds::LogChannel dsLogChannel("dssStringDiffTest");
static const char* TAGS = "[dss][dssStringDiff]";

using namespace trompeloeil;
using namespace dss;

namespace {

SCENARIO("dssStringDiff", TAGS) {
  SECTION("splitBuffer") {
    WHEN("buffer not terminated with delim") {
      auto buffer = "aa\nbb"s;
      auto lines = splitBuffer(buffer);
      CHECK(lines[0] == "aa"sv);
      CHECK(lines[1] == "bb"s);
    }
    WHEN("empty buffer") {
      auto buffer = ""s;
      auto lines = splitBuffer(buffer);
      CHECK(lines.empty());
    }
    WHEN("buffer start with empty line") {
      auto buffer = "\naa\n"s;
      auto lines = splitBuffer(buffer);
      CHECK(lines[0] == ""sv);
      CHECK(lines[1] == "aa"sv);
    }
  }

  SECTION("shortestPath") {
    WHEN("same length") {
      auto left = std::vector{"aa"sv, "bb"sv};
      auto right = std::vector{"aa"sv, "cc"sv};
      CHECK(shortestPath(left, right) == std::vector<std::pair<int, int>>{{1, 1}, {2, 1}, {2, 2}});
    }
    WHEN("M != N") {
      WHEN("left shorter") {
        auto left = std::vector{"aa"sv};
        auto right = std::vector{"aa"sv, "cc"sv};
        CHECK(shortestPath(left, right) == std::vector<std::pair<int, int>>{{1, 1}, {1, 2}});
      }
      WHEN("left empty") {
        auto left = std::vector<std::string_view>{};
        auto right = std::vector{"aa"sv, "bb"sv};
        CHECK(shortestPath(left, right) == std::vector<std::pair<int, int>>{{0, 0}, {0, 1}, {0, 2}});
      }
      WHEN("right buffer short") {
        auto left = std::vector{"aa"sv, "bb"sv};
        auto right = std::vector{"aa"sv};
        CHECK(shortestPath(left, right) == std::vector<std::pair<int, int>>{{1, 1}, {2, 1}});
      }
      WHEN("right empty") {
        auto left = std::vector{"aa"sv, "bb"sv};
        auto right = std::vector<std::string_view>{};
        CHECK(shortestPath(left, right) == std::vector<std::pair<int, int>>{{0, 0}, {1, 0}, {2, 0}});
      }
    }
    WHEN("differs in first line") {
      auto left = std::vector{"aa"sv};
      auto right = std::vector{"cc"sv};
      CHECK(shortestPath(left, right) == std::vector<std::pair<int, int>>{{0, 0}, {1, 0}, {1, 1}});
    }
    WHEN("2 differences") {
      auto left = std::vector{"aa"sv, "bb"sv, "dd"sv, "ee"sv, "gg"sv};
      auto right = std::vector{"aa"sv, "cc"sv, "dd"sv, "ff"sv, "gg"sv};
      CHECK(shortestPath(left, right) == std::vector<std::pair<int, int>>{{1, 1}, {2, 1}, {3, 3}, {4, 3}, {5, 5}});
    }
    WHEN("common end") {
      auto left = std::vector{"aa"sv, "cc"sv};
      auto right = std::vector{"bb"sv, "cc"sv};
      CHECK(shortestPath(left, right) == std::vector<std::pair<int, int>>{{0, 0}, {1, 0}, {2, 2}});
    }
  }

  SECTION("printPatch") {
    WHEN("common first line") {
      auto new_ = std::vector{"aa"sv, "bb"sv};
      auto old = std::vector{"aa"sv, "cc"sv};
      std::vector<std::pair<int, int>> shortestPath{{1, 1}, {2, 1}, {2, 2}};
      auto buffer = createPatch(new_, old, shortestPath);
      CHECK(buffer == std::vector{" aa"s, "+bb"s, "-cc"s});
    }
    WHEN("first line differs") {
      auto new_ = std::vector{"aa"sv, "cc"sv};
      auto old = std::vector{"bb"sv, "cc"sv};
      std::vector<std::pair<int, int>> shortestPath{{0, 0}, {1, 0}, {2, 2}};
      auto buffer = createPatch(new_, old, shortestPath);
      CHECK(buffer == std::vector{"+aa"s, "-bb"s, " cc"s});
    }
    WHEN("trim trailing context") {
      auto new_ = std::vector{"xx"sv, "bb"sv, "cc"sv, "dd"sv};
      auto old = std::vector{"yy"sv, "bb"sv, "cc"sv, "dd"sv};
      std::vector<std::pair<int, int>> shortestPath{{0, 0}, {1, 0}, {4, 4}};
      CHECK(createPatch(new_, old, shortestPath, 0) == std::vector{"+xx"s, "-yy"s});
      CHECK(createPatch(new_, old, shortestPath, 1) == std::vector{"+xx"s, "-yy"s, " bb"s});
      CHECK(createPatch(new_, old, shortestPath, 2) == std::vector{"+xx"s, "-yy"s, " bb"s, " cc"s});
      CHECK(createPatch(new_, old, shortestPath, -1) == std::vector{"+xx"s, "-yy"s, " bb"s, " cc"s, " dd"s});
    }
    WHEN("trim leading context") {
      auto new_ = std::vector{"aa"sv, "bb"sv, "cc"sv, "xx"sv};
      auto old = std::vector{"aa"sv, "bb"sv, "cc"sv, "yy"sv};
      std::vector<std::pair<int, int>> shortestPath{{3, 3}, {4, 3}, {4, 4}};
      CHECK(createPatch(new_, old, shortestPath, 0) == std::vector{"+xx"s, "-yy"s});
      CHECK(createPatch(new_, old, shortestPath, 1) == std::vector{" cc"s, "+xx"s, "-yy"s});
      CHECK(createPatch(new_, old, shortestPath, -1) == std::vector{" aa"s, " bb"s, " cc"s, "+xx"s, "-yy"s});
    }
  }

  SECTION("Patch class") {
    GIVEN("patch with single hook") {
      auto new_ = "aa\nbb\ndd\nff\ngg"s;
      auto old = "aa\ncc\ndd\nff\ngg"s;
      Patch patch{new_, old};

      THEN("shortest path matches") {
        // editPath:[(1,1),(2,1),(5,5)] string-diff.cpp:
        SUCCEED();
      }

      THEN("hooks are found") { CHECK(patch.diff(-1) == std::vector{" aa"s, "+bb"s, "-cc"s, " dd"s, " ff"s, " gg"s}); }
    }
    GIVEN("patch with multiple hook") {
      auto new_ = "aa\nbb\ndd\ndd\nee\ngg\ngg\nhh\ngg"s;
      auto old = "aa\ncc\ndd\ndd\nff\ngg\ngg\nii\ngg"s;
      Patch patch{new_, old};

      THEN("both hooks are returned") {
        CHECK(patch.diff() == std::vector<std::string>{" aa", //
                                                       "+bb", //
                                                       "-cc", //
                                                       " dd", //
                                                       " dd", //
                                                       "+ee", //
                                                       "-ff", //
                                                       " gg", //
                                                       " gg", //
                                                       "+hh", //
                                                       "-ii", //
                                                       " gg"});
      }
      THEN("context of hooks can be trimmed") {
        CHECK(patch.diff(0) == std::vector<std::string>{"+bb", //
                                                        "-cc", //
                                                        "+ee", //
                                                        "-ff", //
                                                        "+hh", //
                                                        "-ii"});
      }
    }
  }
}

} // namespace
