/*
    Copyright (c) 2022 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/

#include "patch-diff.h"

#include <ds/log.h>

#include <iterator>
#include <limits>
#include <sstream>

using namespace std::literals;
static ds::LogChannel dsLogChannel("dssStringDiff");

namespace dss {

std::vector<std::string_view> splitBuffer(std::string_view buffer) {
  std::vector<std::string_view> result{};

  std::string::size_type pos = 0;
  std::string::size_type prev = 0;

  while ((pos = buffer.find('\n', prev)) != std::string::npos) {
    result.push_back(buffer.substr(prev, pos - prev));
    prev = pos + 1;
  }

  DS_DEBUG(prev, buffer.size());
  if (prev < buffer.size()) {
    result.push_back(buffer.substr(prev));
  }
  return result;
};

std::vector<std::pair<int, int>> shortestPath(const std::vector<std::string_view>& new_,
                                              const std::vector<std::string_view>& old) {
  int M = new_.size(); // -> x
  int N = old.size();  // -> y
  auto half = M + N;
  DS_DEBUG(M, N, half);

  // M != N, and either M, N == 0 -> patch is just the other vector

  // VD uses a lot of memory
  DS_ASSERT(M + N < std::numeric_limits<uint16_t>::max());

  using V_t = std::vector<uint16_t>; // x of endpoint for for diagonal with offset k
  std::vector<std::vector<uint16_t>> VD;

  auto shortestPath = [&VD](int D, int k, int x) -> std::vector<std::pair<int, int>> {
    std::vector<std::pair<int, int>> result{};
    result.emplace_back(std::pair{x, x - k});

    while (D > 0) {
      // we go from D backward -> D - 1
      auto Dend = std::exchange(D, D - 1);

      const auto& V = VD.back();
      DS_ASSERT(VD.size() == static_cast<size_t>(D) + 1, VD.size(), D);
      DS_ASSERT(V.size() >= static_cast<size_t>(2 * D + 1), D, V.size());

      // -D .. 0 .. D
      auto getV = [V_ = VD.back(), D](int index) {
        auto shifted = index + D;
        DS_ASSERT(shifted >= 0);
        DS_ASSERT(static_cast<size_t>(shifted) < V_.size(), D, index, V_.size(), shifted);
        return V_[shifted];
      };

      if ((k == -Dend) || (k != Dend && getV(k - 1) < getV(k + 1))) {
        k = k + 1;
      } else {
        k = k - 1;
      }
      x = getV(k);
      result.emplace_back(std::pair{x, x - k});
      VD.pop_back();
    }

    std::reverse(result.begin(), result.end());
    return result;
  };

  // 2 * 1 + 1 -> 3; k = 0 + 2
  V_t V_(2 * (M + N) + 1, 0); // intialize only once, updated by each iteration
  auto V = [&V_, half](int index) -> uint16_t& { return V_[index + half]; };

  // EUGENE W. MYERS: An O(ND) Difference Algorithm and Its Variations
  for (int D = 0; D <= M + N; D++) {
    DS_ASSERT(VD.size() == static_cast<size_t>(D), D, VD.size());
    for (int k = -D; k <= D; k += 2) {
      int x, y;
      if ((k == -D) || (k != D && V(k - 1) < V(k + 1))) {
        x = V(k + 1); // longest diagonal with x - y == k + 1 -> increment y
      } else {
        x = V(k - 1) + 1; // increment x
      }
      y = x - k;

      while (x < M && y < N && new_[x] == old[y]) {
        x++, y++; // longest diagonal, lines match
      }
      DS_DEBUG(D, k, x, y);
      DS_ASSERT(x <= std::numeric_limits<uint16_t>::max());
      V(k) = x;

      if (x >= M && y >= N) {
        DS_NOTICE("found path ", D);
        auto path = shortestPath(D, k, x);
        DS_DEBUG("found path ", D, path);
        return path;
      }
    }

    {
      // not all M + N indices are used, in step D, only [ -D, .. 0, .. D] -> 2 * D + 1
      // auto V = [&V_, half](int index) -> uint16_t& { return V_[index + half]; };
      // -> shift  (half - D) to the left
      V_t shrinked{};
      std::copy(std::next(V_.begin(), half - D), std::next(V_.begin(), half + D + 1), std::back_inserter(shrinked));
      DS_ASSERT(shrinked.size() >= static_cast<size_t>(2 * D + 1), D, shrinked.size());
      VD.push_back(std::move(shrinked)); // keep a copy of endpoints for paths with length == D
      DS_ASSERT(VD.size() == static_cast<size_t>(D) + 1, VD.size(), D); // 0-based
    }
  }

  DS_FAIL_REQUIRE("no shortest path found");
}

std::vector<std::string> formatHook(const std::vector<std::string_view>& new_, const std::vector<std::string_view>& old,
                                    const std::vector<std::pair<int, int>>& path,
                                    std::vector<std::pair<int, int>>::const_iterator& pos, int context) {
  if (pos == path.cend()) {
    DS_NOTICE("empty");
    return {};
  }

  if (context == -1) {
    context = std::numeric_limits<int>::max();
  }
  DS_DEBUG(context); // print number of lines pre/post hook

  auto commonLines = [](auto pos, int D) -> int {
    const auto& [endX, endY] = *pos;

    // single change per snake, either only common lines in other direction
    if (D == 0) {
      return std::min(endX, endY); // 0, 0 not part path -> segfault
    } else {
      const auto& [startX, startY] = *std::prev(pos);
      return std::min(endX - startX, endY - startY);
    }
  };

  auto calcPre = [context, &commonLines](auto pos, int D) {
    int common = commonLines(std::prev(pos), D - 1);

    if (D == 1) {
      return std::min(common, context); // D == 0 is not printed
    } else {
      int unprinted = std::max(common - context, 0); // skip post lines
      return std::min(context, unprinted);
    }
  };

  if (pos == path.cbegin()) {
    pos++; // D == 0 has only common lines
  }

  int D = std::distance(path.cbegin(), pos);
  int x = std::prev(pos)->first;
  int y = std::prev(pos)->second;
  int k = x - y;

  std::vector<std::string> result{};

  // start of snake
  for (; pos != path.cend(); pos++, D++) {
    const auto& [nextX, nextY] = *pos;
    DS_DEBUG(D, x, y, nextX, nextY);

    auto pre = calcPre(pos, D);
    DS_DEBUG(D, pre, x);
    for (int p = x - pre; p < x; p++) {
      DS_DEBUG(x, y);
      DS_ASSERT(static_cast<unsigned>(p) < new_.size(), p, new_.size());
      result.push_back(" "s + std::string{new_[p]}); // common lines
    }

    int oldK = std::exchange(k, nextX - nextY);
    DS_DEBUG(oldK, k);
    if (oldK < k) {
      DS_ASSERT(static_cast<unsigned>(x) < new_.size(), x, new_.size());
      result.push_back("+"s + std::string{new_[x++]}); // move right
    } else {
      DS_ASSERT(static_cast<unsigned>(y) < old.size(), y, old.size());
      result.push_back("-"s + std::string{old[y++]}); // move down
    }

    for (int post = 0; x < nextX; x++, y++, post++) {
      DS_DEBUG(D, post, x, y);
      if (post < context) {
        DS_ASSERT(static_cast<unsigned>(x) < new_.size(), x, new_.size());
        result.push_back(" "s + std::string{new_[x]}); // common lines
      }
    }
    auto tmp = nextY;
    DS_ASSERT(y == tmp);
  }
  return result;
}

std::vector<std::string> createPatch(const std::vector<std::string_view>& new_,
                                     const std::vector<std::string_view>& old,
                                     const std::vector<std::pair<int, int>>& path, int context) {
  int M = new_.size(); // -> x
  int N = old.size();  // -> y

  DS_DEBUG(M, new_);
  DS_DEBUG(N, old);
  DS_DEBUG(path);

  std::vector<std::string> result{};

  for (auto pos = path.cbegin(); pos != path.cend();) {
    auto buffer = formatHook(new_, old, path, pos, context);
    std::move(buffer.cbegin(), buffer.cend(), back_inserter(result));
  }
  return result;
}

Patch::Patch(const std::string& new_, const std::string& old)
    : m_new{splitBuffer(new_)}, m_old{splitBuffer(old)}, editPath(shortestPath(m_new, m_old)) {}

Patch::Patch(std::vector<std::string>&& new_, std::vector<std::string>&& old)
    : m_newOwn(std::move(new_)), m_oldOwn(std::move(old)) {
  auto convert = [](const std::vector<std::string>& in) -> std::vector<std::string_view> {
    return std::vector<std::string_view>{in.cbegin(), in.cend()};
  };

  m_new = convert(m_newOwn);
  m_old = convert(m_oldOwn);
  editPath = shortestPath(m_new, m_old);
}

std::vector<std::string> Patch::diff(int context) const {
  // pos -> iterate hook by hook,
  auto pos = editPath.cbegin();
  return formatHook(m_new, m_old, editPath, pos, context);
};

} // namespace dss
