/*
    Copyright (c) 2010,2011,2022 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/

#include "metermaintenance.h"

#include <ds/log.h>

#include "businterface.h"
#include "comm-channel.h"
#include "dss.h"
#include "event.h"
#include "model/apartment.h"
#include "model/autoclustermaintenance.h"
#include "model/busscanner.h"
#include "model/dsmeter.h"
#include "model/modelmaintenance.h"
#include "model/set.h"
#include "model/state.h"
#include "security/security.h"
#include "structuremanipulator.h"
#include "util.h" // synchronizeGroups(m_pApartment, m_pModifyingBusInterface);
#include "http_client.h"
#include "ds/json.h"

static ds::LogChannel dsLogChannel("scannerMeterMaintenance");

DS_JSON_DEFINE_STRUCT((dss), BlockListAttributes);
DS_JSON_DEFINE_STRUCT((dss), BlockListData);
DS_JSON_DEFINE_STRUCT((dss), BlockList);

namespace dss {

BlockListAttributes::~BlockListAttributes() = default;
BlockListData::~BlockListData() = default;
BlockList::~BlockList() = default;

static const int MAX_RETRY_COUNT = 30;

MeterMaintenance::MeterMaintenance(DSS* _pDSS, const std::string& _name)
    : Thread(_name),
      m_pDSS(_pDSS),
      m_pApartment(NULL), // lazy initialization
      m_pModifyingBusInterface(NULL),
      m_pQueryBusInterface(NULL),
      m_IsInitializing(true),
      m_triggerSynchronize(false),
      m_retryCount(0),
      m_totalCount(0) {}

void MeterMaintenance::triggerMeterSynchronization() {
  std::unique_lock<std::mutex> lock(m_syncMutex);
  m_triggerSynchronize = true;
}

void MeterMaintenance::execute() {
  DS_CONTEXT("meterMaintenance");
  static const int METER_DELAY_TIME = 1000;

  if (m_pDSS == NULL) {
    DS_FATAL("DSS Instance is NULL. Can not continue");
    return;
  }
  m_pApartment = &m_pDSS->getApartment();
  m_pModifyingBusInterface = m_pDSS->getBusInterface().getStructureModifyingBusInterface();
  m_pQueryBusInterface = m_pDSS->getBusInterface().getStructureQueryBusInterface();

  DS_ASSERT(m_pApartment, "");
  DS_ASSERT(m_pModifyingBusInterface, "");
  DS_ASSERT(m_pQueryBusInterface, "");

  if (DSS::hasInstance()) {
    DSS::getInstance()->getSecurity().loginAsSystemUser("ModelMaintenance needs system-rights");
  }

  while (!m_Terminated) {
    readOutPendingMeter();
    synchronizeMeters();
    sleepMS(METER_DELAY_TIME);
    if (m_totalCount % 600 == 0) {
      dSM11HouseKeeping();
    }
    m_totalCount++;
  }
}

void MeterMaintenance::synchronizeMeters() {
  if (m_triggerSynchronize) {
    {
      std::unique_lock<std::mutex> lock(m_syncMutex);
      m_triggerSynchronize = false;
    }
    StructureManipulator manipulator(*m_pApartment);

    for (auto&& zone : DSS::getInstance()->getApartment().getZones()) {
      manipulator.autoAssignZoneSensors(*zone);
      manipulator.synchronizeZoneSensorAssignment(*zone);
    }
  }
}

int MeterMaintenance::getNumValiddSMeters() const {
  int cntReadOutMeters = 0;
  for (std::shared_ptr<DSMeter> pDSMeter : m_pApartment->getDSMeters()) {
    if (busMemberIsDsmApiDSM(pDSMeter->getBusMemberType()) && pDSMeter->isPresent() && pDSMeter->isValid()) {
      ++cntReadOutMeters;
    }
  }
  return cntReadOutMeters;
}

void MeterMaintenance::readOutPendingMeter() {
  bool hadToUpdate = false;
  for (std::shared_ptr<DSMeter> pDSMeter : m_pApartment->getDSMeters()) {
    if (pDSMeter->isPresent() && (!pDSMeter->isValid())) {
      // call for all bus participants (vdc, dsm,..)
      dsMeterReady(pDSMeter->getDSID());
      // only for non virtual devices
      if (busMemberIsDsmApiDSM(pDSMeter->getBusMemberType())) {
        hadToUpdate = true;
        break;
      }
    }
  }

  // If dSMeter configuration has changed we need to synchronize user-groups
  if (!m_IsInitializing && hadToUpdate) {
    synchronizeGroups(m_pApartment, m_pModifyingBusInterface);
  }

  // If we didn't have to update for one cycle, assume that we're done
  if (m_IsInitializing && !hadToUpdate) {
    int cntReadOutMeters = getNumValiddSMeters();
    int busMemberCount = getdSMBusMemberCount();
    DS_DEBUG("Initializing: ReadOutMeters:", cntReadOutMeters, "BusMemberCounts:", busMemberCount);
    if ((busMemberCount > 0) && (cntReadOutMeters == busMemberCount)) {
      setupInitializedState();
    } else {
      monitorInitialization();
    }
  }
} // readOutPendingMeter

void MeterMaintenance::setupInitializedState() {
  synchronizeGroups(m_pApartment, m_pModifyingBusInterface);
  DS_INFO("******** Finished loading model from dSM(s)...");
  m_IsInitializing = false;
  m_pDSS->getModelMaintenance().addModelEvent(std::make_shared<ModelEvent>(ModelEvent::etModelDirty));
  setApartmentState();
  triggerMeterSynchronization();
  {
    m_pDSS->getModelMaintenance().addModelEvent(std::make_shared<ModelEvent>(ModelEvent::etMeterReady));
    std::shared_ptr<Event> readyEvent = std::make_shared<Event>(EventName::ModelReady);
    raiseEvent(readyEvent);
    try {
      CommChannel::getInstance()->resumeUpdateTask();
      CommChannel::getInstance()->requestLockedScenes();
    } catch (std::runtime_error& err) {
      DS_ERROR(err.what());
    }
  }
  AutoClusterMaintenance maintenance(m_pApartment);
  maintenance.joinIdenticalClusters();
} // setupInitializedState

void MeterMaintenance::monitorInitialization() {
  ++m_retryCount;
  if (m_retryCount >= MAX_RETRY_COUNT) {
    if (getdSMBusMemberCount() == 0) {
      setupInitializedState();
    } else {
      m_retryCount = 0;
      DS_INFO("");
      m_pDSS->getModelMaintenance().asyncDiscoverDS485Devices();
    }
  }
} // monitorInitialization

int MeterMaintenance::getdSMBusMemberCount() {
  if (m_pQueryBusInterface == NULL) {
    return -1;
  }
  int busMemberCount = 0;
  try {
    std::vector<DSMeterSpec_t> vecMeterSpec = m_pQueryBusInterface->getBusMembers();
    for (DSMeterSpec_t spec : vecMeterSpec) {
      DS_DEBUG("getdSMBusMemberCount Device:", spec.DSID, spec.DeviceType);
      if (busMemberIsDsmApiDSM(spec.DeviceType)) {
        // ignore dSMx with older api version
        if (spec.APIVersion >= 0x300) {
          ++busMemberCount;
          DS_DEBUG("getdSMBusMemberCount Device:", spec.DSID, "counting.");
        }
      }
    }
    return busMemberCount;
  } catch (BusApiError& e) {
    DS_ERROR("Bus error getting dSM bus member count.", e.what());
  }
  return -1;
}

void MeterMaintenance::dSM11HouseKeeping() {
  dss::HttpClient client;

  if (m_totalCount % 86400 == 600) {
    dss::HttpRequest request{.url = "http://localhost:8089/backend/v1/apartment/dsdevices/blocklist",
                            .type = dss::HttpMethod::GET};

    std::string result;
    long resultCode = client.get(request.url, &result);

    if (resultCode == 200) {
      auto blockList = ds::fromJson<BlockList>(ds::parseJsonFragment(result.c_str()));
      DS_DEBUG("Blocklist", blockList.data.attributes.global, blockList.data.attributes.local);

      for (auto& item : blockList.data.attributes.local) {
        m_dsuidsToBlock.push_back(item);
      }

      for (auto& item : blockList.data.attributes.global) {
        m_prefixesToBlock.push_back(item);
      }
    } else {
      DS_WARNING("Failed to get blocklist", result, resultCode);
    }
  }

  for (std::shared_ptr<DSMeter> pDSMeter : m_pApartment->getDSMeters()) {
    if (pDSMeter->getBusMemberType() == BusMember_dSM11) {
      dsid_t dsid;
      if (dsuid_to_dsid(pDSMeter->getDSID(), &dsid)) {
        auto dsidString = dsid2str(dsid);

        auto dsuidString = dsuid2str(pDSMeter->getDSID());

        auto blockToggle = [&](const bool value) {
          if (value) {
            pDSMeter->setIsPresent(!value);
            pDSMeter->setIsConnected(!value);
          }

          std::string valueString = "false";
          if (value) {
            valueString = "true";
          }

          dss::HttpRequest request{.url = "http://localhost:8087/dsa/json/dsm/toggleBlockDsm?dsuid=" + dsuidString + "&value=" + valueString,
                            .type = dss::HttpMethod::GET};
          std::string result;
          long resultCode = client.get(request.url, &result);
          if (resultCode != 200) {
            DS_WARNING("Failure to disable dSM11 in dSA.", resultCode, result);
          }
        };

        bool mustBlock = false;

        for (auto& blockId : m_dsuidsToBlock) {
          if (blockId == dsuidString) {
            DS_NOTICE("Device found that is overheating", dsidString);
            mustBlock = true;
          }
        }

        for (auto& blockPrefix : m_prefixesToBlock) {
          if (dsidString.find(blockPrefix) == 0) {
            DS_NOTICE("Device found that is overheating", dsidString);
            mustBlock = true;
          }
        }

        blockToggle(mustBlock);

      } else {
        DS_WARNING("Could not convert dsuid");
      }
    }
  }
}

void MeterMaintenance::dsMeterReady(const dsuid_t& _dsMeterBusID) {
  DS_INFO("Scanning dS485 bus device:", _dsMeterBusID);
  try {
    std::shared_ptr<DSMeter> mod;
    try {
      mod = m_pApartment->getDSMeterByDSID(_dsMeterBusID);
    } catch (ItemNotFoundException& e) {
      DS_ERROR("Error scanning dS485 bus device:", _dsMeterBusID, "not found in data model");
      return; // nothing we could do here ...
    }

    try {
      BusScanner scanner(*m_pQueryBusInterface, *m_pApartment, *m_pApartment->getModelMaintenance());
      if (!scanner.scanDSMeter(mod)) {
        DS_ERROR("Error scanning dS485 device:", _dsMeterBusID, ", data model incomplete");
        return;
      }

      if (mod->getCapability_HasDevices()) {
        Set devices = mod->getDevices();
        for (int i = 0; i < devices.length(); i++) {
          devices[i].getDevice()->setIsConnected(true);
        }

        // synchronize devices with binary inputs
        if (mod->hasPendingEvents()) {
          scanner.syncBinaryInputStates(mod, std::shared_ptr<Device>());
        } else {
          DS_DEBUG("Event counter match on dSM", _dsMeterBusID);
        }

        // additionally set all previously connected device to valid now
        devices = m_pApartment->getDevices().getByLastKnownDSMeter(_dsMeterBusID);
        for (int i = 0; i < devices.length(); i++) {
          devices[i].getDevice()->setIsValid(true);
        }
      }

      std::shared_ptr<Event> dsMeterReadyEvent = std::make_shared<Event>(EventName::DSMeterReady);
      dsMeterReadyEvent->setProperty("dsMeter", dsuid2str(mod->getDSID()));
      raiseEvent(dsMeterReadyEvent);

    } catch (BusApiError& e) {
      DS_FATAL("Bus error scanning dSM", _dsMeterBusID, ":", e.what());
      auto pEvent = std::make_shared<ModelEventWithDSID>(ModelEvent::etDSMeterReady, _dsMeterBusID);
      m_pApartment->getModelMaintenance()->addModelEvent(pEvent);
    }
  } catch (ItemNotFoundException& e) {
    DS_ERROR("dsMeterReady", _dsMeterBusID, ": item not found:", e.what());
  }
} // dsMeterReady

void MeterMaintenance::setApartmentState() {
  if (!DSS::hasInstance()) {
    return;
  }

  // logic discussed with mtr:
  //
  // if dSMs do not return the same state, then go with whatever is in
  // the dSS
  //
  // dSMs that report "unknown" do not fall into consideration
  //
  // if all dSMs have the same state set (either present or absent but
  // not unknown), then this state shall be set in the dSS
  //
  // if neither the dSMs nor the dSS have a valid known state, we will use
  // "present" as default as defined by our PO. this case is extremely rare.
  //
  // "as defined above, the dSS keeps the last known state persistent and will
  // set this state when starting up. So the dSS can always provide a state,
  // even after a 3h break down.  We neither need another default state
  // (except for the very first dSS startup where it should be present) nor a
  // timeout."

  // start with whatever state was saved in the dSS
  uint8_t dssaptstate = DSM_APARTMENT_STATE_UNKNOWN;
  std::shared_ptr<State> state;
  try {
    state = m_pApartment->getState(StateType_Service, "presence");
  } catch (ItemNotFoundException& ex) {
    DS_ERROR("setApartmentState: error accessing apartment presence state");
    return;
  }
  if (state->getState() == State_Active) {
    dssaptstate = DSM_APARTMENT_STATE_PRESENT;
  } else if (state->getState() == State_Inactive) {
    dssaptstate = DSM_APARTMENT_STATE_ABSENT;
  }

  uint8_t lastdsmaptstate = DSM_APARTMENT_STATE_UNKNOWN;
  std::shared_ptr<DSMeter> pDSMeter;
  for (size_t i = 0; i < m_pApartment->getDSMeters().size(); i++) {
    pDSMeter = m_pApartment->getDSMeters().at(i);
    if (pDSMeter->isPresent() && pDSMeter->isValid()) {
      uint8_t dsmaptstate = pDSMeter->getApartmentState();

      // dSMs that do not know their state will be ignored
      if (dsmaptstate == DSM_APARTMENT_STATE_UNKNOWN) {
        continue;
      }

      // remember last state in order to be able to compare the dSMs
      if (lastdsmaptstate == DSM_APARTMENT_STATE_UNKNOWN) {
        lastdsmaptstate = dsmaptstate;
      }

      // if dSMs disagree go with the dSS state
      if (lastdsmaptstate != dsmaptstate) {
        lastdsmaptstate = DSM_APARTMENT_STATE_UNKNOWN;
        break;
      }
    }
  }

  // result from the dSMs was conclusive
  if (lastdsmaptstate != DSM_APARTMENT_STATE_UNKNOWN) {
    dssaptstate = lastdsmaptstate;
  }

  // dSS state is still not known and dSMs were inconclusive:
  // use present as default as defined by the PO
  if (dssaptstate == DSM_APARTMENT_STATE_UNKNOWN) {
    dssaptstate = DSM_APARTMENT_STATE_PRESENT;
  }

  std::string strstate;
  if (dssaptstate == DSM_APARTMENT_STATE_PRESENT) {
    strstate = "present";
  } else if (dssaptstate == DSM_APARTMENT_STATE_ABSENT) {
    strstate = "absent";
  }

  DS_DEBUG("setApartmentState: apartment state set to", strstate);
  state->setState(CallOrigin::system, strstate);
}

void MeterMaintenance::raiseEvent(const std::shared_ptr<Event>& event) {
  if (DSS::hasInstance()) {
    DSS::getInstance()->raiseEvent(event);
  }
}

} // namespace dss
