/*
    Copyright (c) 2010,2022 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/

#include "websocketevent.h"

#include "ds485types.h"
#include "dss.h"
#include "includerapidjson.h"
#include "model/apartment.h" // ItemNotFoundException
#include "model/device.h"
#include "model/dsmeter.h"
#include "model/group.h"
#include "model/state.h"
#include "web/webserver.h"

namespace dss {

WebSocketEvent::WebSocketEvent(std::shared_ptr<Event> _event) : Task(), m_event(_event) {}

void WebSocketEvent::run() {
  if (m_event == NULL) {
    return;
  }

  DSS::getInstance()->getWebServer().sendToWebSockets(*m_event);
}

std::string WebSocketEvent::prepareJson(const Event& event) {
  rapidjson::StringBuffer s;
  rapidjson::Writer<rapidjson::StringBuffer> json(s);

  json.StartObject();

  json.String("name");
  json.String(event.getName().c_str());

  json.String("properties");
  json.StartObject();

  for (auto&& param : event.getProperties().getContainer()) {
    json.String((param.first).c_str());
    json.String((param.second).c_str());
  }
  json.EndObject();

  json.String("source");
  json.StartObject();

  EventRaiseLocation raiseLocation = event.getRaiseLocation();
  if ((raiseLocation == erlGroup) || (raiseLocation == erlApartment)) {
    if (DSS::hasInstance()) {
      std::shared_ptr<const Group> group = event.getRaisedAtGroup(DSS::getInstance()->getApartment());
      json.String("set");
      json.String(
          (".zone(" + intToString(group->getZoneID()) + ").group(" + intToString(group->getID()) + ")").c_str());

      json.String("groupID");
      json.Int(group->getID());

      json.String("zoneID");
      json.Int(group->getZoneID());

      json.String("isApartment");
      json.Bool(raiseLocation == erlApartment);

      json.String("isGroup");
      json.Bool(raiseLocation == erlGroup);

      json.String("isDevice");
      json.Bool(false);
    }
  } else if (raiseLocation == erlDevice) {
    std::shared_ptr<const DeviceReference> device = event.getRaisedAtDevice();
    try {
      json.String("set");
      json.String(("dsid(" + dsuid2str(device->getDSID()) + ")").c_str());

      json.String("dsid");
      json.String(dsuid2str(device->getDSID()).c_str());

      json.String("zoneID");
      json.Int(device->getDevice()->getZoneID());
    } catch (ItemNotFoundException& e) {
    }
    json.String("isApartment");
    json.Bool(false);

    json.String("isGroup");
    json.Bool(false);

    json.String("isDevice");
    json.Bool(true);
  } else if (raiseLocation == erlState) {
    std::shared_ptr<const State> state = event.getRaisedAtState();
    if (state->getType() == StateType_Device) {
      if (std::shared_ptr<Device> device = state->getProviderDevice().lock()) {
        dsid_t dsid;
        if (dsuid_to_dsid(device->getDSID(), &dsid)) {
          json.String("dsid");
          json.String(dsid2str(dsid).c_str());
        } else {
          json.String("dsid");
          json.String("");
        }
        json.String("dSUID");
        json.String(dsuid2str(device->getDSID()).c_str());

        json.String("zoneID");
        json.Int(device->getZoneID());

        json.String("isApartment");
        json.Bool(false);

        json.String("isGroup");
        json.Bool(false);

        json.String("isDevice");
        json.Bool(true);
      }
    } else if (state->getType() == StateType_Apartment) {
      json.String("isApartment");
      json.Bool(true);

      json.String("isGroup");
      json.Bool(false);

      json.String("isDevice");
      json.Bool(false);
    } else if (state->getType() == StateType_Group) {
      if (std::shared_ptr<Group> group = state->getProviderGroup().lock()) {
        json.String("isApartment");
        json.Bool(false);

        json.String("isGroup");
        json.Bool(true);

        json.String("isDevice");
        json.Bool(false);

        json.String("groupID");
        json.Int(group->getID());

        json.String("zoneID");
        json.Bool(group->getZoneID());
      }
    } else if (state->getType() == StateType_Service) {
      json.String("isService");
      json.Bool(true);

      json.String("isGroup");
      json.Bool(false);

      json.String("isDevice");
      json.Bool(false);

      json.String("serviceName");
      json.String(state->getProviderService().c_str());
    } else if (state->getType() == StateType_Script) {
      json.String("isScript");
      json.Bool(true);

      json.String("isGroup");
      json.Bool(false);

      json.String("isDevice");
      json.Bool(false);

      json.String("serviceName");
      json.String(state->getProviderService().c_str());
    } else if (state->getType() == StateType_Circuit) {
      if (std::shared_ptr<DSMeter> meter = state->getProviderDsm().lock()) {
        dsid_t dsid;
        if (dsuid_to_dsid(meter->getDSID(), &dsid)) {
          json.String("dsid");
          json.String(dsid2str(dsid).c_str());
        } else {
          json.String("dsid");
          json.String("");
        }
        json.String("dSUID");
        json.String(dsuid2str(meter->getDSID()).c_str());

        json.String("isApartment");
        json.Bool(false);

        json.String("isGroup");
        json.Bool(false);

        json.String("isDevice");
        json.Bool(true);
      }
    }
  }

  json.EndObject();
  json.EndObject();

  return s.GetString();
}
} // namespace dss
