/*
    Copyright (c) 2010 digitalSTROM.org, Zurich, Switzerland

    Author: Patrick Staehlin, futureLAB AG <pstaehlin@futurelab.ch>

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "deviceinterfacerequesthandler.h"

#include <ds/log.h>
#include <limits.h>

#include <stdexcept>

#include "base.h"
#include "ds485types.h"
#include "model/deviceinterface.h"
#include "model/group.h"
#include "model/modelconst.h"
#include "model/scenehelper.h"
#include "sceneaccess.h"
#include "security/user.h"
#include "session.h"
#include "structuremanipulator.h"
#include "util.h"
#include "web/restful.h"

DS_JSON_DEFINE_STRUCT((dss::web::handler), ConsumptionReply);

namespace dss {

static ds::LogChannel dsLogChannel("jsonApiDeviceInterface");

//=========================================== DeviceInterfaceRequestHandler

DeviceInterfaceRequestHandler::DeviceInterfaceRequestHandler(Apartment& apartment) : m_apartment(apartment) {}

// compatibility helper which will be removed in R1.8
SceneAccessCategory web::handler::getCategory(const RestfulRequest& request) {
  return SceneAccess::stringToCategory(request.tryGetParameter("category").value_or("manual"));
}

std::string DeviceInterfaceRequestHandler::turnOn(const RestfulRequest& request, DeviceInterfacePtr& interface,
    SceneAccessCategory sceneCategory, const std::string& sessionToken) {
  DS_NOTICE("Turn on.");
  interface->turnOn(CallOrigin::json, sceneCategory);
  return JSONWriter::success();
}
std::string DeviceInterfaceRequestHandler::turnOff(const RestfulRequest& request, DeviceInterfacePtr& interface,
    SceneAccessCategory sceneCategory, const std::string& sessionToken) {
  DS_NOTICE("Turn off.");
  interface->turnOff(CallOrigin::json, sceneCategory);
  return JSONWriter::success();
}
std::string DeviceInterfaceRequestHandler::increaseValue(const RestfulRequest& request, DeviceInterfacePtr& interface,
    SceneAccessCategory sceneCategory, const std::string& sessionToken) {
  DS_NOTICE("Increase value.");
  interface->increaseValue(CallOrigin::json, sceneCategory);
  return JSONWriter::success();
}
std::string DeviceInterfaceRequestHandler::decreaseValue(const RestfulRequest& request, DeviceInterfacePtr& interface,
    SceneAccessCategory sceneCategory, const std::string& sessionToken) {
  DS_NOTICE("Decrease value.");
  interface->decreaseValue(CallOrigin::json, sceneCategory);
  return JSONWriter::success();
}
std::string DeviceInterfaceRequestHandler::setValue(const RestfulRequest& request, DeviceInterfacePtr& interface,
    SceneAccessCategory sceneCategory, const std::string& sessionToken) {
  std::string valueStr = request.getParameter("value");
  int value = strToIntDef(valueStr, -1);
  if ((value < 0) || (value > UCHAR_MAX)) {
    return JSONWriter::failure("Invalid or missing parameter value: '" + valueStr + "'");
  } else {
    DS_NOTICE("Set value.", value);
    interface->setValue(CallOrigin::json, sceneCategory, value, sessionToken);
  }
  return JSONWriter::success();
}
std::string DeviceInterfaceRequestHandler::callScene(const RestfulRequest& request, DeviceInterfacePtr& interface,
    SceneAccessCategory sceneCategory, const std::string& sessionToken) {
  std::string sceneStr = request.getParameter("sceneNumber");
  int sceneID = strToIntDef(sceneStr, -1);
  bool force = request.getParameter("force") == "true";
  if (sceneID != -1) {
    if (SceneHelper::isInRange(sceneID, 0)) {
      DS_NOTICE("Call scene.", sceneID, force);
      interface->callScene(CallOrigin::json, sceneCategory, sceneID, sessionToken, force);
    } else {
      return JSONWriter::failure("Parameter 'sceneNumber' out of bounds ('" + sceneStr + "')");
    }
  } else {
    return JSONWriter::failure("Invalid sceneNumber: '" + sceneStr + "'");
  }
  return JSONWriter::success();
}
std::string DeviceInterfaceRequestHandler::callSceneMin(const RestfulRequest& request, DeviceInterfacePtr& interface,
    SceneAccessCategory sceneCategory, const std::string& sessionToken) {
  std::string sceneStr = request.getParameter("sceneNumber");
  int sceneID = strToIntDef(sceneStr, -1);
  if (sceneID != -1) {
    if (SceneHelper::isInRange(sceneID, 0)) {
      DS_NOTICE("Call scene min.", sceneID);
      interface->callSceneMin(CallOrigin::json, sceneCategory, sceneID, sessionToken);
    } else {
      return JSONWriter::failure("Parameter 'sceneNumber' out of bounds ('" + sceneStr + "')");
    }
  } else {
    return JSONWriter::failure("Invalid sceneNumber: '" + sceneStr + "'");
  }
  return JSONWriter::success();
}
std::string DeviceInterfaceRequestHandler::saveScene(const RestfulRequest& request, DeviceInterfacePtr& interface,
    SceneAccessCategory sceneCategory, const std::string& sessionToken) {
  std::string sceneStr = request.getParameter("sceneNumber");
  int sceneID = strToIntDef(sceneStr, -1);
  if (sceneID != -1) {
    if (SceneHelper::isInRange(sceneID, 0)) {
      DS_NOTICE("Save scene.", sceneID);
      StructureManipulator(m_apartment).saveScene(*interface, CallOrigin::json, sceneID, sessionToken);
    } else {
      return JSONWriter::failure("Parameter 'sceneNumber' out of bounds ('" + sceneStr + "')");
    }
  } else {
    return JSONWriter::failure("Invalid sceneNumber: '" + sceneStr + "'");
  }
  return JSONWriter::success();
}
std::string DeviceInterfaceRequestHandler::undoScene(const RestfulRequest& request, DeviceInterfacePtr& interface,
    SceneAccessCategory sceneCategory, const std::string& sessionToken) {
  std::string sceneStr = request.getParameter("sceneNumber");
  int sceneID = strToIntDef(sceneStr, -1);
  if (sceneID == -1) {
    DS_NOTICE("Undo last scene.");
    interface->undoSceneLast(CallOrigin::json, sceneCategory, sessionToken);
  } else {
    DS_NOTICE("Undo scene.", sceneID);
    interface->undoScene(CallOrigin::json, sceneCategory, sceneID, sessionToken);
  }
  return JSONWriter::success();
}
std::string DeviceInterfaceRequestHandler::getConsumption(const RestfulRequest& request, DeviceInterfacePtr& interface,
    SceneAccessCategory sceneCategory, const std::string& sessionToken) {
  JSONWriter json;
  json.add("consumption", static_cast<unsigned long long>(interface->getPowerConsumption()));
  return json.successJSON();
}
std::string DeviceInterfaceRequestHandler::nextScene(const RestfulRequest& request, DeviceInterfacePtr& interface,
    SceneAccessCategory sceneCategory, const std::string& sessionToken) {
  auto groupPtr = std::dynamic_pointer_cast<Group>(interface);
  if ((groupPtr != NULL) && ((groupPtr->getApplicationType() == ApplicationType::ventilation) ||
                             (groupPtr->getApplicationType() == ApplicationType::apartmentVentilation))) {
    DS_NOTICE("Next scene.");
    groupPtr->nextScene(CallOrigin::json, sceneCategory);
    return JSONWriter::success();
  } else {
    return JSONWriter::failure("Operation allowed only on Ventilation groups.");
  }
}
std::string DeviceInterfaceRequestHandler::previousScene(const RestfulRequest& request, DeviceInterfacePtr& interface,
    SceneAccessCategory sceneCategory, const std::string& sessionToken) {
  auto groupPtr = std::dynamic_pointer_cast<Group>(interface);
  if ((groupPtr != NULL) && ((groupPtr->getApplicationType() == ApplicationType::ventilation) ||
                             (groupPtr->getApplicationType() == ApplicationType::apartmentVentilation))) {
    DS_NOTICE("Previous scene.");
    groupPtr->previousScene(CallOrigin::json, sceneCategory);
    return JSONWriter::success();
  } else {
    return JSONWriter::failure("Operation allowed only on Ventilation groups.");
  }
}
std::string DeviceInterfaceRequestHandler::blink(const RestfulRequest& request, DeviceInterfacePtr& interface,
    SceneAccessCategory sceneCategory, const std::string& sessionToken) {
  DS_NOTICE("Blink.");
  interface->blink(CallOrigin::json, sceneCategory, sessionToken);
  return JSONWriter::success();
}
std::string DeviceInterfaceRequestHandler::increaseOutputChannelValue(const RestfulRequest& request,
    DeviceInterfacePtr& interface, SceneAccessCategory sceneCategory, const std::string& sessionToken) {
  if (request.getParameter("channel").empty()) {
    return JSONWriter::failure("Missing parameter 'channel'");
  }
  auto channel = getOutputChannelIdAndSize(request.getParameter("channel"));
  DS_NOTICE("Increase output value.", channel);
  interface->increaseOutputChannelValue(CallOrigin::json, sceneCategory, channel.first, sessionToken);
  return JSONWriter::success();
}
std::string DeviceInterfaceRequestHandler::decreaseOutputChannelValue(const RestfulRequest& request,
    DeviceInterfacePtr& interface, SceneAccessCategory sceneCategory, const std::string& sessionToken) {
  if (request.getParameter("channel").empty()) {
    return JSONWriter::failure("Missing parameter 'channel'");
  }
  auto channel = getOutputChannelIdAndSize(request.getParameter("channel"));
  DS_NOTICE("Decrease output value.", channel);
  interface->decreaseOutputChannelValue(CallOrigin::json, sceneCategory, channel.first, sessionToken);
  return JSONWriter::success();
}
std::string DeviceInterfaceRequestHandler::stopOutputChannelValue(const RestfulRequest& request,
    DeviceInterfacePtr& interface, SceneAccessCategory sceneCategory, const std::string& sessionToken) {
  if (request.getParameter("channel").empty()) {
    return JSONWriter::failure("Missing parameter 'channel'");
  }
  auto channel = getOutputChannelIdAndSize(request.getParameter("channel"));
  DS_NOTICE("Stop output channel value.", channel);
  interface->stopOutputChannelValue(CallOrigin::json, sceneCategory, channel.first, sessionToken);
  return JSONWriter::success();
}
std::string DeviceInterfaceRequestHandler::pushSensorValue(const RestfulRequest& request, DeviceInterfacePtr& interface,
    SceneAccessCategory sceneCategory, const std::string& sessionToken) {
  dsuid_t sourceID;
  std::string deviceIDStr = request.getParameter("sourceDSID");
  std::string dsuidStr = request.getParameter("sourceDSUID");
  if (deviceIDStr.empty() && dsuidStr.empty()) {
    sourceID = DSUID_NULL;
  } else {
    sourceID = dsidOrDsuid2dsuid(deviceIDStr, dsuidStr);
  }
  int sensorType = strToIntDef(request.getParameter("sensorType"), -1);
  std::string sensorValueString = request.getParameter("sensorValue");
  if (sensorType == -1 || sensorValueString.length() == 0) {
    return JSONWriter::failure("Need valid parameter 'sensorType' and 'sensorValue'");
  }

  SensorType _sensorType = static_cast<SensorType>(sensorType);
  double floatSensorValue = ::strtod(sensorValueString.c_str(), 0);
  m_apartment.aggregateSensorValue(_sensorType, sourceID, floatSensorValue);

  DS_NOTICE("Push sensor.", _sensorType, floatSensorValue);
  interface->pushSensor(CallOrigin::json, sceneCategory, sourceID, _sensorType, floatSensorValue, sessionToken);
  return JSONWriter::success();
}

std::string DeviceInterfaceRequestHandler::handleDeviceInterfaceRequest(
    const RestfulRequest& request, DeviceInterfacePtr interface, std::shared_ptr<Session> session) {
  DS_ASSERT(interface != NULL, "");
  DS_ASSERT(web::handler::isDeviceInterfaceCall(request), "");
  SceneAccessCategory sceneCategory = web::handler::getCategory(request);
  std::string sessionToken = "";
  if (session && session->getUser() != NULL) {
    sessionToken = session->getUser()->getToken();
  }

  using dss::web::handler::DeviceInterfaceRpc;
  auto rpcCall = ds::tryToEnum<DeviceInterfaceRpc>(request.getMethod());
  if (!rpcCall) {
    throw std::runtime_error("Unknown function");
  }

  switch (*rpcCall) {
    case DeviceInterfaceRpc::blink: return blink(request, interface, sceneCategory, sessionToken);
    case DeviceInterfaceRpc::callScene: return callScene(request, interface, sceneCategory, sessionToken);
    case DeviceInterfaceRpc::callSceneMin: return callSceneMin(request, interface, sceneCategory, sessionToken);
    case DeviceInterfaceRpc::decreaseOutputChannelValue:
      return decreaseOutputChannelValue(request, interface, sceneCategory, sessionToken);
    case DeviceInterfaceRpc::decreaseValue: return decreaseValue(request, interface, sceneCategory, sessionToken);
    case DeviceInterfaceRpc::getConsumption: return getConsumption(request, interface, sceneCategory, sessionToken);
    case DeviceInterfaceRpc::increaseOutputChannelValue:
      return increaseOutputChannelValue(request, interface, sceneCategory, sessionToken);
    case DeviceInterfaceRpc::increaseValue: return increaseValue(request, interface, sceneCategory, sessionToken);
    case DeviceInterfaceRpc::nextScene: return nextScene(request, interface, sceneCategory, sessionToken);
    case DeviceInterfaceRpc::previousScene: return previousScene(request, interface, sceneCategory, sessionToken);
    case DeviceInterfaceRpc::pushSensorValue: return pushSensorValue(request, interface, sceneCategory, sessionToken);
    case DeviceInterfaceRpc::saveScene: return saveScene(request, interface, sceneCategory, sessionToken);
    case DeviceInterfaceRpc::setValue: return setValue(request, interface, sceneCategory, sessionToken);
    case DeviceInterfaceRpc::stopOutputChannelValue:
      return stopOutputChannelValue(request, interface, sceneCategory, sessionToken);
    case DeviceInterfaceRpc::turnOff: return turnOff(request, interface, sceneCategory, sessionToken);
    case DeviceInterfaceRpc::turnOn: return turnOn(request, interface, sceneCategory, sessionToken);
    case DeviceInterfaceRpc::undoScene: return undoScene(request, interface, sceneCategory, sessionToken);
  };

  DS_FAIL_REQUIRE("Unhandled RPC call", *rpcCall);
} // handleRequest

} // namespace dss

namespace dss::web::handler {
bool isDeviceInterfaceCall(const RestfulRequest& _request) {
  auto rpcCall = ds::tryToEnum<DeviceInterfaceRpc>(_request.getMethod());
  return rpcCall.has_value();
} // isDeviceInterfaceCall

} // namespace dss::web::handler

