#ifndef __WEBSERVICE_API__
#define __WEBSERVICE_API__

#include <stdexcept>

#include "logger.h"
#include "http_client.h"
#include "webservice_connection.h" // URLRequestCallback

namespace dss {

class JSONWriter;

// --------------------------------------------------------------------
// helper

namespace MsHub {
  std::vector<std::string> uploadEvents(void);
  void toJson(const std::shared_ptr<Event> &event, JSONWriter& json);

  enum {
    OK = 0,
    RC_NOT_ENABLED = 7,
  };
}

struct WebserviceReply {
  int code;
  std::string desc;
};


struct WeatherInfoIncludesWeather {
  int id;
  std::string main;
  std::string description;
  std::string icon;
};

struct WeatherInfoIncludesSun {
  ds::DateTime sunrise;
  ds::DateTime sunset;
};

struct WeatherInfoIncludesMain {
  float temperature;
  int pressure;
  int humidity;
};

struct WeatherInfoIncludes {
  WeatherInfoIncludesMain main;
  WeatherInfoIncludesSun sun;
  WeatherInfoIncludesWeather weather;
};

struct WeatherGeoInfo {
  std::string city;
  std::string country;
};

struct WeatherInfoData {
  std::string type;
  ds::DateTime lastChanged;
  WeatherGeoInfo attributes;
  WeatherInfoIncludes includes;
};

struct WeatherInfo {
  WeatherInfoData data;
};

typedef enum {
  REST_OK = 0,
  NETWORK_ERROR,
  JSON_ERROR,
} RestTransferStatus_t;

class ParseError : public std::runtime_error {
public:
    ParseError(const std::string& _message);
};

class WebserviceCallDone {
public:
  virtual ~WebserviceCallDone(){}
  /**
   * @RestTransferStatus -- connection established, valid json received
   * @WebserviceReply -- webserver reply
   */
  virtual void done(RestTransferStatus_t status, WebserviceReply reply) = 0;
};

typedef std::shared_ptr<WebserviceCallDone> WebserviceCallDone_t;

/// WebserviceCallDone storing std::function
class WebserviceCallDoneFunction : public WebserviceCallDone {
public:
  typedef std::function<void (RestTransferStatus_t, WebserviceReply)> Function;
  WebserviceCallDoneFunction(Function &&f) : m_f(std::move(f)) {}
  void done(RestTransferStatus_t status, WebserviceReply reply) /* override */ {
    m_f(status, reply);
  }
private:
  Function m_f;
};

// --------------------------------------------------------------------
// interface calls

class WebserviceMsHub {
  __DECL_LOG_CHANNEL__
public:
  typedef enum {
    ApartmentChange = 1,
    TimedEventChange = 2,
    UDAChange = 3,
  } ChangeType;

  /**
   * Did the user authorize cloud service
   * @return
   */
  static bool isAuthorized();

  /**
   * doModelChanged - schedules asynchronous http request
   * @type: what part of the model changed
   * @callback: provides network and server status
   */
  static void doModelChanged(ChangeType type, WebserviceCallDone_t callback);

  static void doDssBackAgain(WebserviceCallDone_t callback);
  /**
   * Asynchronously notifies the cloud about removed application token
   * if you want to retry the request in case of network failure you can pass
   * request_done callback argument
   * @token: token to be deleted
   * @callback: provides network and server status
   */
  static void doNotifyTokenDeleted(const std::string &token,
                                   WebserviceCallDone_t callback);

  /**
   * Pull online weather information.
   * @callback: provides network and server status
   */
  static void doGetWeatherInformation();

  template <class iterator>
  static bool doUploadSensorData(iterator begin, iterator end,
                                 WebserviceCallDone_t callback);

  static void doVdcStoreVdcToken(const std::string &vdcToken, WebserviceCallDone_t callback);

  /**
   * Extract return code and message
   * @param json -- json encoded reply
   * @return decodod struct or throw ParseError if failed
   */
  static WebserviceReply parseReply(const char* buf);
};
}

DS_REFLECT_STRUCT((dss), WeatherInfoIncludesWeather,
    (id)
    (main)
    (description)
    (icon)
);

DS_REFLECT_STRUCT((dss), WeatherInfoIncludesSun,
    (sunrise)
    (sunset)
);

DS_REFLECT_STRUCT((dss), WeatherInfoIncludesMain,
    (temperature)
    (pressure)
    (humidity)
);

DS_REFLECT_STRUCT((dss), WeatherInfoIncludes,
    (main)
    (sun)
    (weather)
);

DS_REFLECT_STRUCT((dss), WeatherGeoInfo,
    (city)
    (country)
);

DS_REFLECT_STRUCT((dss), WeatherInfoData,
    (type)
    (lastChanged)
    (attributes)
    (includes)
);

DS_REFLECT_STRUCT((dss), WeatherInfo,
    (data)
);

#endif
