#include <ds/conf.h>
#include <ds/log.h>
#include <ds/system.h>

#include <boost/asio.hpp>
#include <boost/asio/local/stream_protocol.hpp>

#include "metering/config.h"
#include "metering/readings/open-metering-system.h"

static ds::LogChannel dsLogChannel("toolsMeteringOpenMeteringSystemDataGenerator");

using dss::metering::readings::OpenMeteringSystem;
using namespace dss::metering;
using namespace std::literals;
using boost::asio::local::stream_protocol;

// copy & paste from p1 reader, -> move to ds/asio/single-socket-writer
class SocketClient : public std::enable_shared_from_this<SocketClient> {
public:
  SocketClient(boost::asio::io_context &io, std::string path)
      : m_endpoint(stream_protocol::endpoint(path)), m_socket(stream_protocol::socket(io)) {}

  void send(std::string &&message) {
    m_buffer = std::move(message);
    doConnect();
  }

  void doConnect() {
    m_socket.async_connect(m_endpoint, [self = shared_from_this()](boost::system::error_code ec) {
      if (ec) {
        DS_WARNING("Could not connect", ec, self->m_endpoint.path());
        return;
      }
      self->doWrite();
    });
  }

  void disconnect() {
    // m_socket.close(boost::system::error_code));
  }

  void doWrite() {
    m_socket.async_write_some(boost::asio::buffer(m_buffer),
                              [self = shared_from_this()](boost::system::error_code ec, std::size_t size) {
                                if (ec) {
                                  DS_ERROR(ec);
                                  return;
                                }
                                // don't schedule another task should end the client
                              });
  }

private:
  stream_protocol::endpoint m_endpoint;
  stream_protocol::socket m_socket;
  std::string m_buffer;
};

class Timer {
  using Callback =
      std::function<void(const boost::system::error_code &error_code)>;

public:
  Timer(boost::asio::io_service &io, std::chrono::seconds seconds,
        Callback &&callback)
      : m_timer(boost::asio::steady_timer(io, seconds)), m_seconds(seconds),
        m_callback(callback) {
    start();
  }

  void start() {
    m_timer.expires_at(m_timer.expiry() + m_seconds);
    m_timer.async_wait([this](const auto &ec) -> void { (*this)(ec); });
  }

  void operator()(const boost::system::error_code &error_code) {
    DS_NOTICE("retry timer", m_timer.expiry(), error_code);
    m_callback(error_code);
    start();
  }

  boost::asio::steady_timer m_timer;
  std::chrono::seconds m_seconds;
  Callback m_callback;
};

class Triangle {
public:
  Triangle(int high, int low, int gradient = 10)
      : m_high(high), m_low(low), m_gradient(gradient) {
    DS_ASSERT(m_low < m_high);
  }

  int generate() {
    if (m_upwards) {
      m_signal += m_gradient;
    } else {
      m_signal -= m_gradient;
    }

    if (m_signal >= m_high) {
      m_upwards = false;
    } else if (m_signal <= m_low) {
      m_upwards = true;
    }

    return m_signal;
  }

  int m_high;
  int m_low;
  int m_gradient;
  bool m_upwards{true};
  int m_signal{};
};

auto readElectricityMeteringConfig() {
  auto &system = ds::system();
  return system.conf().readOr<ElectricityMeteringConfig>("electricityMetering", {});
}

int main() {
  boost::asio::io_service io{};
  auto meteringConfig = readElectricityMeteringConfig();

  Triangle triangle(1'000, 0, 10);
  auto reading = OpenMeteringSystem{.id = readings::MeterId{"emulator-1"s}};

  auto callback = [&triangle, &reading, &io,
                   socketPath = meteringConfig.socketPath](const boost::system::error_code &error_code) {
    auto signal = triangle.generate();
    reading.lastChanged = std::chrono::system_clock::now();
    reading.powerConsumed = signal;
    reading.powerProduced = 1'000 - signal;

    reading.t1EnergyConsumed += signal;
    reading.t1EnergyProduced += 1'000 - signal;
    DS_NOTICE(reading);
    std::make_shared<SocketClient>(io, socketPath)->send(ds::toJson(reading));
  };

  Timer timer(io, 1s, std::move(callback));

  io.run();
}
